/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <stdio.h>

#include "grid.h"
#include "geometry.h"
#include "points.h"
#include "grid_unstruct.h"
#include "dep_list.h"
#include "read_cube_csv_grid.h"
#include "utils.h"

static FILE * open_cube_csv_file    ( const char *filename);

static void close_cube_csv_file   ( FILE * file );

static void check_status(int status, const char * filename, int ref, int line) {

  if (status == ref) return;

  fprintf(stderr, "error while reading file %s (%s:%d)\n", filename, __FILE__,
          line);
  exit(EXIT_FAILURE);
}
 
void read_cube_csv_grid_information(const char * filename, int * nbr_vertices,
                                    int * nbr_cells, int ** cell_to_vertex,
                                    double ** x_vertices, double ** y_vertices) {

  // open file

  FILE * file = open_cube_csv_file ( filename );

  // read number of vertices and cells

  check_status(fscanf(file, "%d,%d\n", nbr_vertices, nbr_cells), filename, 2,
               __LINE__);

  // read coordinates of vertices

  *x_vertices = xmalloc (*nbr_vertices * sizeof(**x_vertices));
  *y_vertices = xmalloc (*nbr_vertices * sizeof(**y_vertices));

  for (int i = 0, dummy; i < *nbr_vertices; ++i)
    check_status(fscanf(file, "%d,%lf,%lf\n", &dummy, *x_vertices+i,
                        *y_vertices+i), filename, 3, __LINE__);

  // read indices of cell vertices

  *cell_to_vertex = xmalloc(4 * *nbr_cells * sizeof(**cell_to_vertex));

  for (int i = 0, dummy; i < *nbr_cells; ++i) {
    check_status(fscanf(file, "%d,%d,%d,%d,%d\n", &dummy, *cell_to_vertex+4*i+0,
                        *cell_to_vertex+4*i+1, *cell_to_vertex+4*i+2,
                        *cell_to_vertex+4*i+3), filename, 5, __LINE__);

    for (unsigned j = 0; j < 4; ++j)
      (*cell_to_vertex+4*i)[j]--;
  }

  close_cube_csv_file ( file );
}

struct grid * read_cube_csv_grid(char * filename) {

   int nbr_vertices;
   int nbr_cells;
   int * cell_to_vertex;

   double * x_vertices;
   double * y_vertices;

   read_cube_csv_grid_information(filename, &nbr_vertices, &nbr_cells,
                                  &cell_to_vertex, &x_vertices, &y_vertices);

   size_t * num_vertices_per_cell =
    xmalloc(nbr_cells * sizeof(*num_vertices_per_cell));

   for (int i = 0; i < nbr_cells; ++i)
     num_vertices_per_cell[i] = 4;

   struct dep_list * cell_to_vertex_dep_list =
      yac_dep_list_new((size_t)nbr_cells, num_vertices_per_cell,
                       (unsigned*)cell_to_vertex);

   struct grid * tmp_grid =
      yac_unstruct_grid_new(
         x_vertices, y_vertices, nbr_vertices, cell_to_vertex_dep_list);
   struct grid * grid_copy = yac_copy_grid(tmp_grid);
   yac_delete_grid(tmp_grid);
   yac_dep_list_delete(cell_to_vertex_dep_list);
   free(x_vertices);
   free(y_vertices);

   return grid_copy;
}

static FILE * open_cube_csv_file (const char *filename) {

  FILE * file = fopen(filename, "r");

  if (file == NULL) {
    perror("Error while opening the file.\n");
    exit(EXIT_FAILURE);
  }

  return file;
}

/* ---------------------------------------------------------------- */

static void close_cube_csv_file (FILE * file) {

   fclose(file);
}

