/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#ifdef YAC_MPI_ENABLED
#include <mpi.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "yac_interface.h"
#include "points.h"
#include "utils.h"

#define NBR_CELLS 2
#define NBR_VERTICES 4
#define NBR_SUBDOMAIN_IDS 1

int main () {

   int no_of_fields = 8;

   const char * fieldName[] = { "atmos_out1",
                                "atmos_out2",
                                "atmos_out3",
                                "atmos_out4",
                                "ocean_out1",
                                "ocean_out2",
                                "ocean_out3",
                                "ocean_out4"};

  char * xml_filename = "dummy_coupling.xml";
  char * xsd_filename = "coupling.xsd";

  char * comp_name = "dummy_io";

  char * grid_name[] = { "ocean_grid",
                         "atmos_grid"};

  const int nbr_subdomain_ids = NBR_SUBDOMAIN_IDS;

  const int nbr_cells = NBR_CELLS;
  const int nbr_vertices = NBR_VERTICES;
  int nbr_vertices_per_cell[NBR_CELLS];

  int i, ierror;
  int comp_id;

  int cell_point_ids[2];
  int cell_mask_ids[2];

  int domain_id[2];
  int subdomain_id;
  int subdomain_ids[NBR_SUBDOMAIN_IDS];

  int global_index[NBR_CELLS];
  int owner_local[NBR_CELLS];

  double * buffer_lon;
  double * buffer_lat;
  int * cell_to_vertex;

  int * cell_mask;
  int * field_id;

#ifdef YAC_MPI_ENABLED
  MPI_Comm local_comm;
#else
  int local_comm;
#endif

  int size, rank;

#ifdef YAC_MPI_ENABLED
  MPI_Init (0, NULL);
#endif

  // Initialise the coupler
  yac_cinit ( xml_filename, xsd_filename );

  // Inform the coupler about what we are
  yac_cdef_comp ( comp_name, &comp_id );

  printf ( "YAC Version %s\n", yac_cget_version() );

  yac_cget_localcomm ( &local_comm, comp_id );

#ifdef YAC_MPI_ENABLED
  MPI_Comm_rank ( local_comm, &rank );
  MPI_Comm_size ( local_comm, &size );
#else
  rank = 0;
  size = 1;
#endif

  printf (" %s rank %d : local size %d \n", comp_name, rank, size );

  /* Here we define two grids (although numrically identical) to mimick
     two different Output server groups */

  for ( int igrid = 0; igrid < 2; igrid++ ) {

    // Announce one subdomain (patch) to the coupler

    yac_cdef_subdomain ( comp_id, grid_name[igrid], &subdomain_id );

    subdomain_ids[0] = subdomain_id;

    buffer_lon = xmalloc ( nbr_vertices * sizeof(*buffer_lon) );
    buffer_lat = xmalloc ( nbr_vertices * sizeof(*buffer_lat) );
    cell_to_vertex = xmalloc ( 3 * nbr_cells * sizeof(*cell_to_vertex) );

    /* Define vertices

                0
               / \
              / o \
             /     \
            1-------2   Eq.
             \     /
              \ o /
               \ /
                3

    */

    buffer_lon[0] =  0.0; buffer_lat[0] =  1.1;
    buffer_lon[1] = -1.0; buffer_lat[1] =  0.0;
    buffer_lon[2] =  1.0; buffer_lat[2] =  0.0;
    buffer_lon[3] =  0.0; buffer_lat[3] = -1.0;

    // Connectivity

    cell_to_vertex[0] = 0; cell_to_vertex[1] = 1; cell_to_vertex[2] = 2; // cell 1
    cell_to_vertex[3] = 1; cell_to_vertex[4] = 3; cell_to_vertex[5] = 2; // cell 2

    for ( i = 0; i < nbr_cells; ++i ) nbr_vertices_per_cell[i] = 3;

    // Description of elements, here as unstructured grid

    yac_cdef_elements_unstruct ( subdomain_id,
                                 nbr_vertices,
                                 nbr_cells,
                                 nbr_vertices_per_cell,
                                 buffer_lon,
                                 buffer_lat,
                                 cell_to_vertex );

    // Center points in cells (needed e.g. for patch recovery and nearest neighbour)

    buffer_lon[0] = 0.0; buffer_lat[0] =  0.5;
    buffer_lon[1] = 0.0; buffer_lat[1] = -0.5;

    yac_cdef_points_unstruct ( subdomain_id,
                               nbr_cells,
                               CELL,
                               buffer_lon,
                               buffer_lat,
                               &cell_point_ids[igrid] );

    free ( buffer_lon );
    free ( buffer_lat );

    // Decomposition information

    for ( i = 0; i < nbr_cells; ++i ) {
      global_index[i] = i;
      owner_local[i] = -1;
    }

    yac_cdef_index_location ( subdomain_id,
                              nbr_cells,
                              CELL,
                              global_index,
                              owner_local );

    // Connect subdomains
    yac_cconnect_subdomains ( comp_id,
                              nbr_subdomain_ids,
                              subdomain_ids,
                              &domain_id[igrid] );

    // Mask generation

    cell_mask = xmalloc (nbr_cells * sizeof(*cell_mask));

    for ( i = 0; i < nbr_cells; ++i )
      cell_mask[i] = 1;

    yac_cdef_mask ( cell_mask,
                    cell_point_ids[igrid],
                    &cell_mask_ids[igrid] );

    free (cell_mask);
  }

  field_id = xmalloc ( no_of_fields * sizeof(*field_id));

  // atmosphere output

  for ( i = 0; i < no_of_fields-4; i++ )
    yac_cdef_field ( fieldName[i],
                     comp_id,
                     domain_id[1],
                     &cell_point_ids[1],
                     cell_mask_ids,
                     1,
                     &field_id[i] );

  // ocean output

  for ( i = no_of_fields-4; i < no_of_fields; i++ )
    yac_cdef_field ( fieldName[i],
                     comp_id,
                     domain_id[0],
                     &cell_point_ids[0],
                     cell_mask_ids,
                     1,
                     &field_id[i] );

  yac_csearch ( 1, &comp_id, no_of_fields, field_id, &ierror );

  yac_cfinalize ();

#ifdef YAC_MPI_ENABLED
  MPI_Finalize ();
#endif

  return 0;
}

