/**
 * 
 */
package de.mpg.mpimet.yacgui.junit;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import de.mpg.mpimet.yacgui.utils.XmlDomHandler;
import de.mpg.mpimet.yacgui.utils.XmlDomHandlerException;
import de.mpg.mpimet.yacgui.xml.Component;
import de.mpg.mpimet.yacgui.xml.Couple;
import de.mpg.mpimet.yacgui.xml.Coupling;
import de.mpg.mpimet.yacgui.xml.CouplingOperation;
import de.mpg.mpimet.yacgui.xml.Dates;
import de.mpg.mpimet.yacgui.xml.DebugMode;
import de.mpg.mpimet.yacgui.xml.Grid;
import de.mpg.mpimet.yacgui.xml.TimestepUnit;
import de.mpg.mpimet.yacgui.xml.Transient;
import de.mpg.mpimet.yacgui.xml.TransientCouple;
import de.mpg.mpimet.yacgui.xml.TransientGridRef;
import de.mpg.mpimet.yacgui.xml.interpolation.AverageInterpolation;
import de.mpg.mpimet.yacgui.xml.interpolation.ConservativeInterpolation;
import de.mpg.mpimet.yacgui.xml.interpolation.Interpolation;
import de.mpg.mpimet.yacgui.xml.interpolation.NNearestNeighborInterpolation;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @author teresa
 * 
 */
public class XmlDomHandlerTest extends XmlDomHandler {

	private final static String FILE_COUPLING = "res/xml/coupling.xml";
	private final static String FILE_COMPONENT1 = "res/xml/component1.xml";
	private final static String FILE_INTERPOLATION = "res/xml/interpolation.xml";
	private final static String COUPLING = "coupling";
	private final static String COMPONENT = "component";
	
	@Rule
	public ExpectedException thrown = ExpectedException.none();
	
	public XmlDomHandlerTest() throws XmlDomHandlerException {
		super(FILE_COUPLING, XmlType.COUPLING);
	}

	/**
	 * Test method for {@link XmlDomHandler}.
	 * 
	 * @throws Error XmlDomHandlerException
	 */
	@Test
	public void testXmlDomHandler() {
		try {
			// domHandler creation in general
			XmlDomHandler domHandler = new XmlDomHandler(FILE_COUPLING, XmlType.COUPLING);
			assertNotNull("domHandler is null", domHandler);
			
			// this: new XmlDomHandler(FILE_COUPLING) has been called
			assertNotNull(document);
		} catch (XmlDomHandlerException e) {
			throw new Error(e);
		}
	}

	/**
	 * Test method for {@link XmlDomHandler#parseXml(String, XmlType)}.
	 * 
	 * @throws Error XmlDomHandlerException
	 */
	@Test
	public void testParseXml() {
		try {
			// this: coupling.xml has been parsed
			assertTrue(document.getDocumentElement().getNodeName().equals(COUPLING));
			
			// parse component xml
			parseXml(FILE_COMPONENT1, XmlType.COMPONENT);
			assertTrue(document.getDocumentElement().getNodeName()
					.equals(COMPONENT));
		} catch (XmlDomHandlerException e) {
			throw new Error(e);
		}
	}
	
	/**
	 * Test method for {@link de.mpg.mpimet.yacgui.utils.XmlDomHandler#parseXml(String, XmlType)}.
	 * 
	 * @throws XmlDomHandlerException XmlDomHandlerException
	 */
	@Test
	public void testParseXmlException_FileNotFound() throws XmlDomHandlerException {
		thrown.expect(XmlDomHandlerException.class);
		parseXml("not-existent.xml", null);
	}
	
	/**
	 * Test method for {@link de.mpg.mpimet.yacgui.utils.XmlDomHandler#parseXml(String, XmlType)}.
	 *
	 * @throws XmlDomHandlerException XmlDomHandlerException
	 */
	@Test
	public void testParseXmlException_notValidating1() throws XmlDomHandlerException {
		// neither COUPLING nor COMPONENT
		thrown.expect(XmlDomHandlerException.class);
		parseXml(FILE_INTERPOLATION, XmlType.COMPONENT);
	}
	
	/**
	 * Test method for {@link de.mpg.mpimet.yacgui.utils.XmlDomHandler#buildComponent()}.
	 */
	@Test
	public void testBuildComponent() {
		try {
			// We create an XmlDomHandler that reads the file coming with this
			// application. We then assume that the created Component object is
			// carrying all child objects as members according to the specified
			// component xml file.
			XmlDomHandler domHandler = new XmlDomHandler(FILE_COMPONENT1, XmlType.COMPONENT);
			Component component = domHandler.buildComponent();
			assertNotNull("Component is null", component);

			// <id>
			int id = component.getId();
			assertTrue(id == 1);

			// <name>
			String name = component.getName();
			assertEquals(name, "ICON-ocean");

			// <model>
			String model = component.getModel();
			assertEquals(model, "some_model...");

			// <simulated>
			String simulated = component.getSimulated();
			assertEquals(simulated, "ocean");

			// <transient_grid_refs>
			Map<Integer, TransientGridRef> transientGridRefs = new HashMap<Integer, TransientGridRef>();
			transientGridRefs.put(1, new TransientGridRef(1, 1, 1, 1));
			transientGridRefs.put(2, new TransientGridRef(2, 2, 1, 2));
			transientGridRefs.put(3, new TransientGridRef(3, 3, 2, 1));
			transientGridRefs.put(4, new TransientGridRef(4, 4, 2, 2));
			transientGridRefs.put(5, new TransientGridRef(5, 4, 1, 1));
			assertEquals(component.getTransientGridRefs(), transientGridRefs);

			// <transients>
			Map<Integer, Transient> transients = 
					new HashMap<Integer, Transient>();
			transients.put(1, new Transient(1, "sea_surface_temperature"));
			transients.put(2, new Transient(2, "age_of_sea_ice"));
			transients.put(3, new Transient(3, "air_density"));
			transients.put(4, new Transient(4, "air_pressure_at_sea_level"));
			assertEquals(component.getTransients(), transients);

			// <grids>
			Map<Integer, Grid> grids = new HashMap<Integer, Grid>();
			grids.put(1, new Grid(1, "R2B06"));
			grids.put(2, new Grid(2, "R2B06_pacific"));
			assertEquals(component.getGrids(), grids);
		} catch (XmlDomHandlerException e) {
			throw new Error(e);
		}
	}

	/**
	 * Test method for {@link de.mpg.mpimet.yacgui.utils.XmlDomHandler#buildCoupling()}.
	 */
	@Test
	public void testBuildCoupling() {
		try {
			// We create an XmlDomHandler that reads the file coming with this
			// application. We then assume that the created Coupling object is
			// carrying all child objects as members according to the specified
			// coupling xml file.
			XmlDomHandler domHandler = new XmlDomHandler(FILE_COUPLING, XmlType.COUPLING);
			Coupling coupling = domHandler.buildCoupling();
			assertNotNull("Coupling is null", coupling);

			// <components>
			Map<Integer, Component> components = coupling.getComponentsMap();
			assertNotNull("components map is null", components);
			assertTrue("components map size is expected to be 3", components.size() == 3);
			Component component1 = components.get(1);
			assertNotNull(component1);
			Component component2 = components.get(2);
			assertNotNull(component2);
			Component component3 = components.get(3);
			assertNotNull(component3);

			// <component> 1
			assertTrue(component1.getId() == 1);
			assertEquals(component1.getName(), "ICON-ocean");
			assertEquals(component1.getModel(), "some_model...");
			assertEquals(component1.getSimulated(), "ocean");
			Map<Integer, TransientGridRef> transientGridRefs = new HashMap<Integer, TransientGridRef>();
			transientGridRefs.put(1, new TransientGridRef(1, 1, 1, 1));
			transientGridRefs.put(2, new TransientGridRef(2, 1, 2, 1));
			transientGridRefs.put(3, new TransientGridRef(3, 3, 3, 1));
			transientGridRefs.put(4, new TransientGridRef(4, 4, 4, 1));
			transientGridRefs.put(5, new TransientGridRef(5, 5, 5, 1));
			assertEquals(component1.getTransientGridRefs(), transientGridRefs);
			assertNull(component1.getTransients());
			assertNull(component1.getGrids());

			// <component> 2
			assertTrue(component2.getId() == 2);
			assertTrue(component2.getName().equals("ICON-atmosphere"));
			assertTrue(component2.getModel().equals("..."));
			assertTrue(component2.getSimulated().equals("atmosphere"));
			transientGridRefs.clear();
			transientGridRefs.put(1, new TransientGridRef(1, 1, 2, 1));
			transientGridRefs.put(2, new TransientGridRef(2, 2, 2, 1));
			transientGridRefs.put(3, new TransientGridRef(3, 3, 2, 1));
			transientGridRefs.put(4, new TransientGridRef(4, 4, 6, 1));
			transientGridRefs.put(5, new TransientGridRef(5, 5, 6, 1));
			assertEquals(component2.getTransientGridRefs(), transientGridRefs);
			assertNull(component2.getTransients());
			assertNull(component2.getGrids());

			// <component> 3
			assertTrue(component3.getId() == 3);
			assertTrue(component3.getName().equals("..."));
			assertTrue(component3.getModel().equals("..."));
			assertTrue(component3.getSimulated().equals("atmosphere"));
			transientGridRefs.clear();
			transientGridRefs.put(1, new TransientGridRef(1, 6, 2, 1));
			transientGridRefs.put(2, new TransientGridRef(2, 7, 2, 1));
			transientGridRefs.put(3, new TransientGridRef(3, 8, 2, 1));
			transientGridRefs.put(4, new TransientGridRef(4, 9, 6, 1));
			assertEquals(component3.getTransientGridRefs(), transientGridRefs);
			assertNull(component3.getTransients());
			assertNull(component3.getGrids());

			// <transients>
			Map<Integer, Transient> transients = 
					new HashMap<Integer, Transient>();
			transients.put(1, new Transient(1, "sea_surface_temperature"));
			transients.put(2, new Transient(2, "wind_speed"));
			transients.put(3, new Transient(3, "water_flux_into_sea_water"));
			transients.put(4, new Transient(4, "grid_eastward_wind"));
			transients.put(5, new Transient(5, "grid_northward_wind"));
			transients.put(6, new Transient(6, "std_name_1"));
			transients.put(7, new Transient(7, "std_name_2"));
			transients.put(8, new Transient(8, "std_name_3"));
			transients.put(9, new Transient(9, "std_name_4"));
			assertEquals(coupling.getTransients(), transients);

			// <grids>
			Map<Integer, Grid> grids = new HashMap<Integer, Grid>();
			grids.put(1, new Grid(1, "grid1"));
			grids.put(2, new Grid(2, "grid2"));
			grids.put(3, new Grid(3, "grid3"));
			grids.put(4, new Grid(4, "grid4"));
			grids.put(5, new Grid(5, "grid5"));
			grids.put(6, new Grid(6, "grid6"));
			assertEquals(coupling.getGrids(), grids);

			// <dates>
			Dates dates = coupling.getDates();
			assertEquals(dates.getStartDate(), "2008-03-09 16:05:07Z");
			assertEquals(dates.getEndDate(), "2008-03-10 16:05:07Z");

			// <timestep_unit>
			assertSame(coupling.getTimestepUnit(), TimestepUnit.SECOND);

			// <couples>
			Map<String, Couple> couples = coupling.getCouples();
			assertNotNull(couples);
			assertFalse(couples.isEmpty());
			assertTrue(couples.size() == 1);

			// <couple>
			Couple couple = couples.get("1-2");
			assertNotNull(couple);
			assertTrue(couple.getComponent1Id() == 1);
			assertTrue(couple.getComponent2Id() == 2);

			// <transient_couple>
			List<TransientCouple> transientCouples = couple
					.getTransientCouples();
			assertTrue(transientCouples.size() == 2);
			TransientCouple tc1 = transientCouples.get(0);
			assertNotNull(tc1);
			TransientCouple tc2 = transientCouples.get(1);
			assertNotNull(tc2);

			// <transient_couple> 1
			assertTrue(tc1.getTransientId() == 1);
			assertTrue(tc1.getSourceComponentRef() == 1);
			assertTrue(tc1.getSourceTransientGridRefId() == 1);
			assertTrue(tc1.getTargetTransientGridRefIf() == 1);
			assertTrue(tc1.getTimestep().getSource().equals("10"));
			assertTrue(tc1.getTimestep().getTarget().equals("20"));
			assertTrue(tc1.getTimestep().getCouplingPeriod().equals("60"));
			assertSame(tc1.getTimestep().getCouplingOperation(), CouplingOperation.ACCUMULATE);
			assertTrue(tc1.getTimestep().getSourceTimeLag() == "0");
			assertTrue(tc1.getTimestep().getTargetTimeLag() == "0");
			List<Interpolation> interpolations = tc1.getInterpolations();
			assertTrue(interpolations.size() == 2);
			assertEquals(tc1.getMaskValue(), Double.valueOf(999.0));
			DebugMode debugMode = tc1.getDebugMode();
			assertTrue(debugMode.isAtSourceBeforeInterpolation());
			assertTrue(debugMode.isAtSourceAfterInterpolation());
			assertFalse(debugMode.isAtTarget());
			assertTrue(tc1.isEnforceWriteRestart());

			// <transient_couple> 2
			assertTrue(tc2.getTransientId() == 3);
			assertTrue(tc2.getSourceComponentRef() == 2);
			assertTrue(tc2.getSourceTransientGridRefId() == 3);
			assertTrue(tc2.getTargetTransientGridRefIf() == 3);
			assertTrue(tc2.getTimestep().getSource().equals("10"));
			assertTrue(tc2.getTimestep().getTarget().equals("20"));
			assertTrue(tc2.getTimestep().getCouplingPeriod().equals("60"));
			assertSame(tc2.getTimestep().getCouplingOperation(), CouplingOperation.ACCUMULATE);
			assertTrue(tc2.getTimestep().getSourceTimeLag() == "0");
			assertTrue(tc2.getTimestep().getTargetTimeLag() == "0");
			interpolations = tc2.getInterpolations();
			assertTrue(interpolations.size() == 3);
			assertTrue(interpolations.get(0) instanceof AverageInterpolation);
			
			AverageInterpolation average = (AverageInterpolation) interpolations.get(0);
			assertTrue(average.getAttributesNumber() == 2);
			assertEquals(average.getAttributeName(0), "partial_coverage");
			assertEquals(average.getAttributeValue(0), "false");
			assertEquals(average.getAttributeName(1), "weighted");
			assertEquals(average.getAttributeValue(1), "DISTANCE_WEIGHTED");
			
			assertTrue(interpolations.get(1) instanceof NNearestNeighborInterpolation);
			NNearestNeighborInterpolation nNearestNeighbor = 
					(NNearestNeighborInterpolation) interpolations.get(1);
			assertTrue(nNearestNeighbor.getAttributesNumber() == 3);
			assertEquals(nNearestNeighbor.getAttributeName(1), "n");
			assertEquals(nNearestNeighbor.getAttributeValue(1), "3");
			
			assertTrue(interpolations.get(2) instanceof ConservativeInterpolation);
			ConservativeInterpolation conservative = (ConservativeInterpolation) interpolations.get(2);
			assertTrue(conservative.getAttributesNumber() == 4);
			assertEquals(conservative.getAttributeName(1), "enforced_conservation");
			assertEquals(conservative.getAttributeValue(1), "false");
			assertEquals(conservative.getAttributeName(3), "normalisation");
			assertEquals(conservative.getAttributeValue(3), "FRACAREA");
			debugMode = tc2.getDebugMode();
			assertTrue(debugMode.isAtSourceBeforeInterpolation());
			assertFalse(debugMode.isAtSourceAfterInterpolation());
			assertFalse(debugMode.isAtTarget());
			assertFalse(tc2.isEnforceWriteRestart());
		} catch (XmlDomHandlerException e) {
			throw new Error(e);
		}
	}

	@Test
	public void testWriteCouplingToFile() {
		try {
			File destination = createNewFile("test_write_coupling");
			XmlDomHandler domHandler = new XmlDomHandler(FILE_COUPLING, XmlType.COUPLING);
			Coupling coupling = domHandler.buildCoupling();
			writeCouplingToFile(destination, coupling);
			XmlDomHandler newDomHandler = new XmlDomHandler(destination.getPath(), XmlType.COUPLING);
			Coupling newCoupling = newDomHandler.buildCoupling();
			assertEquals(coupling.getComponentsMap(), newCoupling.getComponentsMap());
			assertEquals(coupling.getTransients(), newCoupling.getTransients());
			assertEquals(coupling.getGrids(), newCoupling.getGrids());
			assertEquals(coupling.getDates(), newCoupling.getDates());
			assertEquals(coupling.getTimestepUnit(),
					newCoupling.getTimestepUnit());
			assertEquals(coupling, newCoupling);
		} catch (XmlDomHandlerException e) {
			throw new Error(e);
		}
	}

	private static File createNewFile(String filename) {
		File file = new File("res/test/" + filename + ".txt");
		// if file doesnt exists, then create it
		if (!file.exists()) {
			try {
				file.createNewFile();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return file;
	}
}
