package de.mpg.mpimet.yacgui.xml.interpolation;

import de.mpg.mpimet.yacgui.xml.Dom;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Definite interpolation method smoothed patch recovery
 * 
 * @author Junpei Xu
 */

public class SmoothedPatchRecoveryInterpolation extends Interpolation {

	private boolean allowSmoothedExtrapolation;
	private int extendSmoothedSourcePatch;
	private int polynomialSmoothedOrder;
	private int gaussSmoothedOrder;
	private boolean allowCornerWeight;
	private float cornerWeight;
	
	public SmoothedPatchRecoveryInterpolation(boolean allowSmoothedExtrapolation, int extendSmoothedSourcePatch,
			int polynomialSmoothedOrder, int gaussSmoothedOrder, boolean allowCornerWeight, float cornerWeight) throws IllegalArgumentException {
		
		if (extendSmoothedSourcePatch < 1) {
			throw new IllegalArgumentException("ExtendSourcePatch must be bigger than 0");
		}
		if (polynomialSmoothedOrder < 1) {
			throw new IllegalArgumentException("PolynomialOrder must be bigger than 0");
		}
		if (gaussSmoothedOrder < 1) {
			throw new IllegalArgumentException("GaussOrder must be bigger than 0");
		}
		if (cornerWeight < 0 || cornerWeight > 1) {
			throw new IllegalArgumentException("GaussOrder must be float between 0 and 1");
		}
		
		attributesNumber = 6;
		method = InterpolationMethod.SMOOTHED_PATCH_RECOVERY;
		this.allowSmoothedExtrapolation = allowSmoothedExtrapolation;
		this.extendSmoothedSourcePatch = extendSmoothedSourcePatch;
		this.polynomialSmoothedOrder = polynomialSmoothedOrder;
		this.gaussSmoothedOrder = gaussSmoothedOrder;
		this.allowCornerWeight = allowCornerWeight;
		this.cornerWeight = cornerWeight;
	}
	
	/**
	 * 
	 * @return true if extrapolation is allowed in smoothed patch recovery  
	 */
	public boolean isAllowExtrapolation() {
		return allowSmoothedExtrapolation;
	}

	/**
	 * 
	 * @return integer value of extend source patch
	 */
	public int getExtendSourcePatch() {
		return extendSmoothedSourcePatch;
	}
	
	/**
	 * 
	 * @return integer value of polynomial order
	 */
	public int getPolynomialOrder() {
		return polynomialSmoothedOrder;
	}
	
	/**
	 * 
	 * @return integer value of gauss order
	 */
	public int getGaussOrder() {
		return gaussSmoothedOrder;
	}
	
	/**
	 * 
	 * @return true if corner weight is allowed in smoothed patch recovery  
	 */
	public boolean isAllowCornerWeight() {
		return allowCornerWeight;
	}

	/**
	 * 
	 * @return float value of corner weight
	 */
	public float getCornerWeight() {
		return cornerWeight;
	}
	
	/**
	 * get attribute name by the order of index
	 * @param index the order of attributes on the panel
	 * @return  the name of attribute
	 * 				if index = 0 return ALLOW_EXTRAPOLATION, if 1 return EXTEND_SOURCE_PATCH, if 2 return POLYNOMIAL_ORDER, 
	 * if 3 return GAUSS_ORDER, if 4 return ALLOW_CORNER_WEIGHT, if 5 return CORNER_WEIGHT
	 */
	public String getAttributeName(int index) {
		if (index == 0) {
			return Dom.Attribute.ALLOW_EXTRAPOLATION;
		} else if (index == 1) {
			return Dom.Attribute.EXTEND_SOURCE_PATCH;
		} else if (index == 2) {
			return Dom.Attribute.POLYNOMIAL_ORDER;
		} else if (index == 3) {
			return Dom.Attribute.GAUSS_ORDER;
		} else if (index == 4) {
			return Dom.Attribute.ALLOW_CORNER_WEIGHT;
		} else if (index == 5) {
			return Dom.Attribute.CORNER_WEIGHT;
		}
		return null;
	}
	
	/**
	 * get attribute value by the order of index
	 * @param index the order of attributes on the panel
	 * @return the value of attribute as String
	 *				if 0 return allowSmoothedExtrapolation, if 1 return extendSmoothedSourcePatch, if 2 return polynomialSmoothedOrder, 
	 * if 3 return gaussSmoothedOrder, if 4 return allowCornerWeight, if 5 return cornerWeight
	 */
	public String getAttributeValue(int index) {
		if (index == 0) {
			return String.valueOf(allowSmoothedExtrapolation);
		} else if (index == 1) {
			return String.valueOf(extendSmoothedSourcePatch);
		} else if (index == 2) {
			return String.valueOf(polynomialSmoothedOrder);
		} else if (index == 3) {
			return String.valueOf(gaussSmoothedOrder);
		} else if (index == 4) {
			return String.valueOf(allowCornerWeight);
		} else if (index == 5) {
			return String.valueOf(cornerWeight);
		}
		return null;
	}
}
