/**
 * @file bucket_search.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <math.h>

#include "grid_search.h"
#include "bucket_search.h"
#include "search.h"
#include "grid_reg2d.h"
#include "utils.h"
#include "ensure_array_size.h"
#include "grid_search_utils.h"

static double const tol = 1.0e-12;

static struct dep_list *
bucket_search_do_cell_search (struct grid_search * search,
                              struct grid * grid_data);
static void
bucket_search_do_cell_search_single (struct grid_search * search,
                                     struct grid_cell grid_cell,
                                     size_t * n_cells,
                                     size_t * cells_size,
                                     size_t ** cells);
static struct dep_list *
bucket_search_do_point_search_c (struct grid_search * search,
                                 struct grid * grid_data);
static struct dep_list *
bucket_search_do_point_search_c2 (struct grid_search * search,
                                  double (*coordinates_xyz)[3],
                                  size_t num_points);
static struct dep_list *
bucket_search_do_point_search_c3(struct grid_search * search,
                                 double (*coordinates_xyz)[3],
                                 size_t num_points,
                                 struct points * points);
static struct dep_list *
bucket_search_do_point_search_p (struct grid_search * search,
                                 struct grid * grid_data);
static struct dep_list *
bucket_search_do_point_search_p2 (struct grid_search * search,
                                  double (*coordinates_xyz)[3],
                                  size_t num_points);
static struct dep_list *
bucket_search_do_point_search_p3 (struct grid_search * search,
                                  double (*coordinates_xyz)[3],
                                  size_t num_points,
                                  struct points * points);
static void
bucket_search_do_point_search_p4 (struct grid_search * search,
                                  double coordinate_xyz[3],
                                  size_t * n_points, size_t * points_size,
                                  size_t ** points);

static void delete_bucket_search(struct grid_search * search);

struct bucket_search {
   struct grid_search_vtable * vtable;
   struct grid * bucket_grid;
   double * bucket_grid_x_coords;
   double * bucket_grid_y_coords;
   struct grid * grid_data;
   size_t num_buckets[2];
   struct dep_list * bucket_to_cell;
};

static struct grid_search_vtable bucket_search_vtable =
{
   .do_cell_search        = bucket_search_do_cell_search,
   .do_cell_search_single = bucket_search_do_cell_search_single,
   .do_point_search_c     = bucket_search_do_point_search_c,
   .do_point_search_c2    = bucket_search_do_point_search_c2,
   .do_point_search_c3    = bucket_search_do_point_search_c3,
   .do_point_search_p     = bucket_search_do_point_search_p,
   .do_point_search_p2    = bucket_search_do_point_search_p2,
   .do_point_search_p3    = bucket_search_do_point_search_p3,
   .do_point_search_p4    = bucket_search_do_point_search_p4,
   .do_bnd_circle_search  = NULL,
   .delete_grid_search    = delete_bucket_search
};

static void delete_bucket_search(struct grid_search * search) {

   struct bucket_search * bucket_search = (struct bucket_search *)search;

   free(bucket_search->bucket_grid_x_coords);
   free(bucket_search->bucket_grid_y_coords);
   yac_delete_grid(bucket_search->bucket_grid);
   yac_dep_list_delete(bucket_search->bucket_to_cell);
   free(search);
}

// ==============================================================================
// Creation of the bucket grid
static
void init_buckets(struct bucket_search * bucket_search, double grid_extent[2][2],
                  unsigned num_cells) {

   double dx, dy;                            // grid spacing of the bucket grid
   double * coordinates_x, * coordinates_y;  // coordinates of the bucket grid
   unsigned bucket_cells[2];                 // number of bucket cells in x and y direction
   unsigned cyclic[2] = {0,0};               // marker to indicate cyclic coordinates
   unsigned i;

   bucket_cells[0] =
    MAX((unsigned)(sqrt((double)num_cells *
                        (grid_extent[0][1] - grid_extent[0][0]) /
                        (grid_extent[1][1] - grid_extent[1][0]))),1);

   bucket_cells[1] = num_cells / bucket_cells[0];


   // if the grid is cyclic
   if (fabs((2*M_PI) - (grid_extent[0][1] - grid_extent[0][0])) < 1e-10)
      cyclic[0] = 1;

   // even though we do not need "bucket_cells[0]+1" coordinates in case of a
   // cyclic grid we still allocate them for the bisection search done later
   // MoHa: I do not like this
   coordinates_x = xmalloc((bucket_cells[0]+1)*sizeof(double));
   coordinates_y = xmalloc((bucket_cells[1]+1)*sizeof(double));

   dx = (grid_extent[0][1] - grid_extent[0][0]) / (double)(bucket_cells[0]);
   dy = (grid_extent[1][1] - grid_extent[1][0]) / (double)(bucket_cells[1]);

   for (i = 0; i < bucket_cells[0]; ++i) {
      coordinates_x[i] = grid_extent[0][0] + dx * i;
   }

   coordinates_x[bucket_cells[0]] = grid_extent[0][1];

   for (i = 0; i < bucket_cells[1]; ++i) {
      coordinates_y[i] = grid_extent[1][0] + dy * i;
   }
   coordinates_y[bucket_cells[1]] = grid_extent[1][1];

   bucket_search->num_buckets[0] = (size_t)bucket_cells[0];
   bucket_search->num_buckets[1] = (size_t)bucket_cells[1];

   bucket_search->bucket_grid_x_coords = coordinates_x;
   bucket_search->bucket_grid_y_coords = coordinates_y;

   // register the bucket grid as a regular grid in lon and lat

   bucket_search->bucket_grid = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                                   bucket_cells, cyclic);
}

// ==============================================================================
// Search for the grid cells on the bucket grid
//------------------------------------------------------------------
// Because of the properties of the bucket grid this search is
// being optimised by usage of the function bisection_search.
//------------------------------------------------------------------
static struct dep_list *
search_on_bucket_grid(
   struct bucket_search * bucket_search, struct grid * grid_data) {

   unsigned num_x_coords, num_y_coords;
   int * found_x, * position_x, * found_y, * position_y;
   double const * x_coords;
   double const * y_coords;

   x_coords = yac_get_x_coords(grid_data);
   y_coords = yac_get_y_coords(grid_data);

   num_x_coords = yac_get_size_x_coords(grid_data);
   num_y_coords = yac_get_size_y_coords(grid_data);

   found_x    = xmalloc (num_x_coords * sizeof(*found_x));
   found_y    = xmalloc (num_y_coords * sizeof(*found_y));
   position_x = xmalloc (num_x_coords * sizeof(*position_x));
   position_y = xmalloc (num_y_coords * sizeof(*position_y));

   yac_bisection_search(x_coords, num_x_coords, 
                        bucket_search->bucket_grid_x_coords,
                        (unsigned)(bucket_search->num_buckets[0])+1,
                        position_x, found_x, 2*M_PI);

   yac_bisection_search(y_coords, num_y_coords, 
                        bucket_search->bucket_grid_y_coords,
                        (unsigned)(bucket_search->num_buckets[1])+1,
                        position_y, found_y, 0);

   //----------------------------------------------
   // generate initial cell->bucket dependency list
   //----------------------------------------------

   size_t num_cells = (size_t)yac_get_num_grid_cells(grid_data);
   size_t cell_to_bucket_dependencies_size = num_cells * 4;
   unsigned * cell_to_bucket_dependencies =
      xmalloc(cell_to_bucket_dependencies_size *
              sizeof(*cell_to_bucket_dependencies));
   size_t * num_buckets_per_cell =
      xcalloc (num_cells, sizeof(*num_buckets_per_cell));

   size_t num_total_dependencies = 0;

   // for all cells
   for (size_t i = 0; i < num_cells; ++i) {

      size_t num_corners = (size_t)yac_get_num_cell_corners(grid_data, (unsigned)i);

      unsigned const * curr_x_coords =
         yac_get_cell_x_coord_indices(grid_data, (unsigned)i);
      unsigned const * curr_y_coords =
         yac_get_cell_y_coord_indices(grid_data, (unsigned)i);

      for (size_t j = 0; j < num_corners; ++j) {

         unsigned curr_cell_x_corner = curr_x_coords[j];
         unsigned curr_cell_y_corner = curr_y_coords[j];

         if (!found_x[curr_cell_x_corner] || !found_y[curr_cell_y_corner])
            continue;

         size_t num_matching_buckets = 1;
         unsigned bucket_flags = 0;
         unsigned matching_buckets[4] = {(unsigned)position_x[curr_cell_x_corner] +
                                         (unsigned)position_y[curr_cell_y_corner] *
                                         (unsigned)(bucket_search->num_buckets[0])};

         // if the current point falls onto the left edge of the bucket
         if ((position_x[curr_cell_x_corner] != 0) &&
             (fabs(x_coords[curr_cell_x_corner] -
                   bucket_search->bucket_grid_x_coords[position_x[curr_cell_x_corner]]) < tol)) {

            matching_buckets[num_matching_buckets++] =
               (unsigned)position_x[curr_cell_x_corner] - 1 +
               (unsigned)position_y[curr_cell_y_corner] *
               (unsigned)(bucket_search->num_buckets[0]);

            bucket_flags |= 1;

         // if the current point falls onto the right edge of the bucket
         } else if ((position_x[curr_cell_x_corner] != (int)(bucket_search->num_buckets[0]) - 1) &&
             (fabs(x_coords[curr_cell_x_corner] -
                   bucket_search->bucket_grid_x_coords[position_x[curr_cell_x_corner] + 1]) < tol)) {

            matching_buckets[num_matching_buckets++] =
               (unsigned)position_x[curr_cell_x_corner] + 1 +
               (unsigned)position_y[curr_cell_y_corner] *
               (unsigned)(bucket_search->num_buckets[0]);

            bucket_flags |= 2;
         }

         // if the current point falls onto the lower edge of the bucket
         if ((position_y[curr_cell_y_corner] != 0) &&
             (fabs(y_coords[curr_cell_y_corner] -
                   bucket_search->bucket_grid_y_coords[position_y[curr_cell_y_corner]]) < tol)) {

            matching_buckets[num_matching_buckets++] =
               (unsigned)position_x[curr_cell_x_corner] +
               ((unsigned)position_y[curr_cell_y_corner] - 1) *
               (unsigned)(bucket_search->num_buckets[0]);

            bucket_flags |= 4;

         // if the current point falls onto the upper edge of the bucket
         } else if ((position_y[curr_cell_y_corner] != (int)(bucket_search->num_buckets[1]) - 1) &&
             (fabs(y_coords[curr_cell_y_corner] -
                   bucket_search->bucket_grid_y_coords[position_y[curr_cell_y_corner]+1]) < tol)) {

            matching_buckets[num_matching_buckets++] =
               (unsigned)position_x[curr_cell_x_corner] +
               ((unsigned)position_y[curr_cell_y_corner] + 1) *
               (unsigned)(bucket_search->num_buckets[0]);

            bucket_flags |= 8;
         }

         if (bucket_flags == 5)
            matching_buckets[num_matching_buckets++] =
               (unsigned)position_x[curr_cell_x_corner] - 1 +
               ((unsigned)position_y[curr_cell_y_corner] - 1) *
               (unsigned)(bucket_search->num_buckets[0]);
         else if (bucket_flags == 9)
            matching_buckets[num_matching_buckets++] =
               (unsigned)position_x[curr_cell_x_corner] - 1 +
               ((unsigned)position_y[curr_cell_y_corner] + 1) *
               (unsigned)(bucket_search->num_buckets[0]);
         else if (bucket_flags == 6)
            matching_buckets[num_matching_buckets++] =
               (unsigned)position_x[curr_cell_x_corner] + 1 +
               ((unsigned)position_y[curr_cell_y_corner] - 1) *
               (unsigned)(bucket_search->num_buckets[0]);
         else if (bucket_flags == 10)
            matching_buckets[num_matching_buckets++] =
               (unsigned)position_x[curr_cell_x_corner] + 1 +
               ((unsigned)position_y[curr_cell_y_corner] + 1) *
               (unsigned)(bucket_search->num_buckets[0]);

         if (num_matching_buckets > 4)
            yac_internal_abort_message("ERROR: too many matching buckets\n", __FILE__, __LINE__);

         ENSURE_ARRAY_SIZE(cell_to_bucket_dependencies,
                           cell_to_bucket_dependencies_size,
                           num_total_dependencies+num_matching_buckets);

         for (size_t k = 0; k < num_matching_buckets; ++k) {

            cell_to_bucket_dependencies[num_total_dependencies++] =
               matching_buckets[k];
            num_buckets_per_cell[i]++;
         }
      }
   }

   if (cell_to_bucket_dependencies_size != num_total_dependencies)
      cell_to_bucket_dependencies = xrealloc(cell_to_bucket_dependencies,
                                             num_total_dependencies *
                                             sizeof(*cell_to_bucket_dependencies));

   struct dep_list * initial_cell_to_bucket =
      yac_dep_list_new(
         num_cells, num_buckets_per_cell, cell_to_bucket_dependencies);

   //----------------------------------------------
   // generate cell->bucket dependency list
   //----------------------------------------------

   struct dep_list * cell_to_bucket =
      yac_find_overlapping_cells(
         grid_data, bucket_search->bucket_grid, initial_cell_to_bucket);

   //-------------
   // free memory
   //-------------

   yac_dep_list_delete(initial_cell_to_bucket);
   free(found_x); free(found_y); free(position_x); free(position_y);

   return cell_to_bucket;
}

static
void search_on_bucket_grid_single (struct bucket_search * bucket_search,
                                   struct grid_cell grid_cell,
                                   size_t ** deps, size_t * deps_size,
                                   size_t * num_deps, unsigned src_index,
                                   unsigned * tgts_already_touched,
                                   unsigned ** stack, unsigned * stack_size) {

   int found_x[grid_cell.num_corners];
   int position_x[grid_cell.num_corners];
   int found_y[grid_cell.num_corners];
   int position_y[grid_cell.num_corners];

   yac_bisection_search(grid_cell.coordinates_x,
                        (unsigned)(grid_cell.num_corners),
                        bucket_search->bucket_grid_x_coords,
                        (unsigned)(bucket_search->num_buckets[0])+1,
                        position_x, found_x, 2*M_PI);

   yac_bisection_search(grid_cell.coordinates_y,
                        (unsigned)(grid_cell.num_corners),
                        bucket_search->bucket_grid_y_coords,
                        (unsigned)(bucket_search->num_buckets[1])+1,
                        position_y, found_y, 0);

   //----------------------------------------------
   // generate initial cell->bucket dependency list
   //----------------------------------------------

   unsigned k;
   unsigned * buckets = NULL;
   unsigned buckets_size = 0;
   unsigned num_buckets = 0;

   for (size_t j = 0; j < grid_cell.num_corners; ++j) {

      if (!found_x[j] || !found_y[j])
         continue;

      unsigned num_matching_buckets = 1;
      unsigned bucket_flags = 0;
      unsigned matching_buckets[4] = {(unsigned)position_x[j] +
                                      (unsigned)position_y[j] *
                                      (unsigned)(bucket_search->num_buckets[0])};

      // if the current point falls onto the left edge of the bucket
      if ((position_x[j] != 0) &&
          (fabs(grid_cell.coordinates_x[j] -
                bucket_search->bucket_grid_x_coords[position_x[j]]) < tol)) {

         matching_buckets[num_matching_buckets++] =
            (unsigned)position_x[j] - 1 +
            (unsigned)position_y[j] *
            (unsigned)(bucket_search->num_buckets[0]);

         bucket_flags |= 1;

      // if the current point falls onto the right edge of the bucket
      } else if ((position_x[j] != (int)(bucket_search->num_buckets[0]) - 1) &&
          (fabs(grid_cell.coordinates_x[j] -
                bucket_search->bucket_grid_x_coords[position_x[j] + 1]) < tol)) {

         matching_buckets[num_matching_buckets++] =
            (unsigned)position_x[j] + 1 +
            (unsigned)position_y[j] *
            (unsigned)(bucket_search->num_buckets[0]);

         bucket_flags |= 2;
      }

      // if the current point falls onto the lower edge of the bucket
      if ((position_y[j] != 0) &&
          (fabs(grid_cell.coordinates_y[j] -
                bucket_search->bucket_grid_y_coords[position_y[j]]) < tol)) {

         matching_buckets[num_matching_buckets++] =
            (unsigned)position_x[j] +
            ((unsigned)position_y[j] - 1) *
            (unsigned)(bucket_search->num_buckets[0]);

         bucket_flags |= 4;

      // if the current point falls onto the upper edge of the bucket
      } else if ((position_y[j] != (int)(bucket_search->num_buckets[1]) - 1) &&
          (fabs(grid_cell.coordinates_y[j] -
                bucket_search->bucket_grid_y_coords[position_y[j]+1]) < tol)) {

         matching_buckets[num_matching_buckets++] =
            (unsigned)position_x[j] +
            ((unsigned)position_y[j] + 1) *
            (unsigned)(bucket_search->num_buckets[0]);

         bucket_flags |= 8;
      }

      if (bucket_flags == 5)
         matching_buckets[num_matching_buckets++] =
            (unsigned)position_x[j] - 1 +
            ((unsigned)position_y[j] - 1) *
            (unsigned)(bucket_search->num_buckets[0]);
      else if (bucket_flags == 9)
         matching_buckets[num_matching_buckets++] =
            (unsigned)position_x[j] - 1 +
            ((unsigned)position_y[j] + 1) *
            (unsigned)(bucket_search->num_buckets[0]);
      else if (bucket_flags == 6)
         matching_buckets[num_matching_buckets++] =
            (unsigned)position_x[j] + 1 +
            ((unsigned)position_y[j] - 1) *
            (unsigned)(bucket_search->num_buckets[0]);
      else if (bucket_flags == 10)
         matching_buckets[num_matching_buckets++] =
            (unsigned)position_x[j] + 1 +
            ((unsigned)position_y[j] + 1) *
            (unsigned)(bucket_search->num_buckets[0]);

      if (num_matching_buckets > 4)
         yac_internal_abort_message("ERROR: too many matching buckets\n", __FILE__, __LINE__);

      ENSURE_ARRAY_SIZE(buckets, buckets_size,
                        num_buckets+num_matching_buckets);

      for (k = 0; k < num_matching_buckets; ++k)
         buckets[num_buckets++] = matching_buckets[k];
   }

   //----------------------------------------------
   // generate cell->bucket dependency list
   //----------------------------------------------

   struct bounding_circle bnd_circle;

   yac_get_cell_bounding_circle(grid_cell, &bnd_circle);
   yac_find_overlapping_cells_s (grid_cell, bnd_circle, bucket_search->bucket_grid,
                                 buckets, num_buckets, deps, deps_size, num_deps,
                                 src_index, tgts_already_touched, stack, stack_size);

   free(buckets);
}


// ==============================================================================
// Generation of the grid_search object
struct grid_search * yac_bucket_search_new (struct grid * grid_data) {

   struct bucket_search * search;

   search = xmalloc(1 * sizeof(*search));

   search->vtable = &bucket_search_vtable;

   double grid_extent[2][2];

   if (yac_get_num_grid_corners(grid_data) == 0)
      yac_internal_abort_message("ERROR: bucket_search_new empty input grid\n",
                                 __FILE__, __LINE__);

   //-------------------------
   // initiate the search data
   //-------------------------

   yac_get_2d_grid_extent(grid_data, grid_extent);

   search->grid_data = grid_data;

   init_buckets (search, grid_extent, yac_get_num_grid_cells(grid_data));

   //----------------------------------------------
   // generate cell->bucket dependency list
   //----------------------------------------------

   struct dep_list * cell_to_bucket = search_on_bucket_grid(search, grid_data);

   //----------------------------------------------
   // generate bucket->cell dependency list
   //----------------------------------------------

   search->bucket_to_cell = yac_invert_dep_list(cell_to_bucket);

   //-------------
   // free memory
   //-------------
   
   yac_dep_list_delete(cell_to_bucket);

   return (struct grid_search *)search;
}

// ==============================================================================
// This routine searches the given grid data in the data associated with
//  the search handle and returns a list of dependencies between the grids.
static struct dep_list *
bucket_search_do_cell_search (struct grid_search * search,
                              struct grid * grid_data) {

   struct bucket_search * bucket_search = (struct bucket_search *)search;

   //--------------------------------
   // find grid cells on bucket grid
   //--------------------------------

   struct dep_list * tgt_to_bucket =
      search_on_bucket_grid(bucket_search, grid_data);

   //----------------------------------------------------------
   // use initial cell-to-bucket-results to fill initial result
   // for src to tgt grid dependencies
   //----------------------------------------------------------

   unsigned * initial_tgt_to_src_dependencies = NULL;
   size_t curr_size_tgt_to_src_dependencies = 0;
   size_t num_total_initial_tgt_to_src_deps = 0;

   size_t num_tgt_grid_cells = (size_t)yac_get_num_grid_cells(grid_data);
   size_t * initial_num_srcs_per_tgt =
      xcalloc(num_tgt_grid_cells, sizeof(*initial_num_srcs_per_tgt));
   size_t num_src_grid_cells =
      (size_t)yac_get_num_grid_cells(bucket_search->grid_data);
   size_t * src_cell_mask =
      xcalloc(num_src_grid_cells, sizeof(*src_cell_mask)); // array that helps
                                                           // double entries of
                                                           // the same tgt-to-src
                                                           // dependency in
                                                           // initial_tgt_to_src_dependencies

   size_t bucket_to_cell_num_elements =
      yac_get_num_elements(bucket_search->bucket_to_cell);

   // for all target grid cells
   for (size_t i = 0; i < num_tgt_grid_cells; ++i) {

      unsigned const * curr_buckets =
         yac_get_dependencies_of_element(tgt_to_bucket, i);
      size_t curr_num_buckets =
         yac_get_num_dependencies_of_element(tgt_to_bucket, i);

      // for all matching bucket grid cells of the current tgt grid cell
      for (size_t j = 0; j < curr_num_buckets; ++j) {

         size_t curr_bucket = (size_t)(curr_buckets[j]);

         if (curr_bucket >= bucket_to_cell_num_elements) continue;

         unsigned const * curr_src_cells =
            yac_get_dependencies_of_element(
               bucket_search->bucket_to_cell, curr_bucket);
         size_t curr_num_cells =
            yac_get_num_dependencies_of_element(
               bucket_search->bucket_to_cell, curr_bucket);

         // for all src grid cells that overlap with the current bucket cell
         for (size_t k = 0; k < curr_num_cells; ++k) {

            // check for double src entries
            if (src_cell_mask[curr_src_cells[k]] != i + 1) {

               ENSURE_ARRAY_SIZE(initial_tgt_to_src_dependencies,
                                 curr_size_tgt_to_src_dependencies,
                                 num_total_initial_tgt_to_src_deps+1);
               initial_tgt_to_src_dependencies[num_total_initial_tgt_to_src_deps++] =
                  curr_src_cells[k];
               initial_num_srcs_per_tgt[i]++;
               src_cell_mask[curr_src_cells[k]] = i + 1;
            }
         } // (k = 0; k < bucket_search->bucket_to_cell.num_deps_per_element[curr_bucket], ++k)
      } // (j = 0; j < tgt_to_bucket.num_deps_per_element[i]; ++j)
   } // (i = 0; i < num_tgt_grid_cells; ++i)

   struct dep_list * initial_tgt_to_src =
      yac_dep_list_new(num_tgt_grid_cells, initial_num_srcs_per_tgt,
                       initial_tgt_to_src_dependencies);

   //----------------------------------------------
   // generate actual src to tgt grid dependencies
   //----------------------------------------------

   struct dep_list * tgt_to_src_cells =
      yac_find_overlapping_cells(
         grid_data, bucket_search->grid_data, initial_tgt_to_src);
   //-------------
   // free memory
   //-------------

   free(src_cell_mask);

   yac_dep_list_delete(tgt_to_bucket);
   yac_dep_list_delete(initial_tgt_to_src);

   return tgt_to_src_cells;
}

static
void bucket_search_do_cell_search_single (struct grid_search * search,
                                          struct grid_cell cell,
                                          size_t * n_cells,
                                          size_t * cells_size,
                                          size_t ** cells) {

   struct bucket_search * bucket_search = (struct bucket_search *)search;

   //--------------------------------
   // find grid cells on bucket grid
   //--------------------------------

   size_t * cell_to_bucket = NULL;
   size_t cell_to_bucket_size = 0, num_buckets = 0;
   size_t num_grid_cells_bucket =
      (size_t)yac_get_num_grid_cells(bucket_search->bucket_grid);
   size_t num_grid_cells_base =
      (size_t)yac_get_num_grid_cells(bucket_search->grid_data);
   unsigned * helper_array =
     xcalloc(
       MAX(num_grid_cells_bucket, num_grid_cells_base), sizeof(*helper_array));
   unsigned * stack = NULL;
   unsigned stack_size = 0;

   search_on_bucket_grid_single (bucket_search, cell, &cell_to_bucket,
                                 &cell_to_bucket_size, &num_buckets, 0,
                                 helper_array, &stack, &stack_size);

   //----------------------------------------------------------
   // use initial cell-to-bucket-results to fill initial result
   // for src to tgt grid dependencies
   //----------------------------------------------------------

   unsigned * initial_tgt_to_src_cells = NULL;
   unsigned initial_tgt_to_src_cells_size = 0;
   unsigned initial_num_cells = 0;

   size_t bucket_to_cell_num_elements =
      yac_get_num_elements(bucket_search->bucket_to_cell);

   // for all matching bucket grid cells of the current tgt grid cell
   for (size_t j = 0; j < num_buckets; ++j) {

      size_t curr_bucket = cell_to_bucket[j];

      if (curr_bucket >= bucket_to_cell_num_elements) continue;

      unsigned const * curr_src_cells =
         yac_get_dependencies_of_element(
            bucket_search->bucket_to_cell, curr_bucket);
      size_t curr_num_src_cells =
         yac_get_num_dependencies_of_element(
            bucket_search->bucket_to_cell, curr_bucket);

      // for all src grid cells that overlap with the current bucket cell
      for (size_t k = 0; k < curr_num_src_cells; ++k) {

         // check for double src entries
         if (helper_array[curr_src_cells[k]] != 2) {

            ENSURE_ARRAY_SIZE(initial_tgt_to_src_cells,
                              initial_tgt_to_src_cells_size,
                              initial_num_cells+1);
            initial_tgt_to_src_cells[initial_num_cells++] = curr_src_cells[k];
            helper_array[curr_src_cells[k]] = 2;
         }
      } // (k = 0; k < bucket_search->bucket_to_cell.num_deps_per_element[curr_bucket], ++k)
   } // (j = 0; j < tgt_to_bucket.num_deps_per_element[i]; ++j)

   //----------------------------------------------
   // generate actual src to tgt grid dependencies
   //----------------------------------------------

   struct bounding_circle bnd_circle;

   yac_get_cell_bounding_circle(cell, &bnd_circle);
   yac_find_overlapping_cells_s (cell, bnd_circle, bucket_search->grid_data,
                                 initial_tgt_to_src_cells, initial_num_cells,
                                 cells, cells_size, n_cells, 3, helper_array,
                                 &stack, &stack_size);
                             
   //-------------
   // free memory
   //-------------

   free(initial_tgt_to_src_cells);
   free(helper_array);
   free(stack);
   free(cell_to_bucket);
}

// ==============================================================================
// This routine searches the given grid data in the data associated with
// the search handle and returns a list of dependencies between the grids.
static struct dep_list *
bucket_search_do_point_search_c (struct grid_search * search,
                                 struct grid * grid_data) {

   struct bucket_search * bucket_search = (struct bucket_search *)search;

   //--------------------------------
   // find grid cells on bucket grid
   //--------------------------------

   unsigned num_x_coords, num_y_coords;
   int * found_x, * position_x, * found_y, * position_y;
   double const * x_coords;
   double const * y_coords;

   x_coords = yac_get_x_coords(grid_data);
   y_coords = yac_get_y_coords(grid_data);

   num_x_coords = yac_get_size_x_coords(grid_data);
   num_y_coords = yac_get_size_y_coords(grid_data);

   found_x    = xmalloc (num_x_coords * sizeof(*found_x));
   found_y    = xmalloc (num_y_coords * sizeof(*found_y));
   position_x = xmalloc (num_x_coords * sizeof(*position_x));
   position_y = xmalloc (num_y_coords * sizeof(*position_y));

   yac_bisection_search(x_coords, num_x_coords, 
                        bucket_search->bucket_grid_x_coords,
                        (unsigned)(bucket_search->num_buckets[0])+1,
                        position_x, found_x, 2*M_PI);

   yac_bisection_search(y_coords, num_y_coords, 
                        bucket_search->bucket_grid_y_coords,
                        (unsigned)(bucket_search->num_buckets[1])+1,
                        position_y, found_y, 0);

   //----------------------------------------------------------
   // use inital cell-to-bucket-results to find the find the
   // actual tgt point to src cell dependency
   //----------------------------------------------------------

   struct grid_cell src_cell;

   size_t num_tgt_corners = (size_t)yac_get_num_grid_corners(grid_data);
   size_t num_total_dependencies = 0;
   size_t * num_src_per_tgt =
      xcalloc(num_tgt_corners, sizeof(*num_src_per_tgt));
   unsigned * tgt_to_src_cell =
      xmalloc(num_tgt_corners * sizeof(*tgt_to_src_cell));

   yac_init_grid_cell(&src_cell);

   size_t bucket_to_cell_num_elements =
      yac_get_num_elements(bucket_search->bucket_to_cell);

   // for all corners of the target grid
   for (size_t i = 0; i < num_tgt_corners; ++i) {

      unsigned curr_corner_index_x =
         yac_get_corner_x_coord_index(grid_data, (unsigned)i);
      unsigned curr_corner_index_y =
         yac_get_corner_y_coord_index(grid_data, (unsigned)i);

      // if the current target corner was not found on the source grid
      if (!found_x[curr_corner_index_x] || !found_y[curr_corner_index_y]) continue; // continue with next corner

      // compute the current bucket cell index
      size_t curr_bucket = (size_t)position_x[curr_corner_index_x] +
         (size_t)position_y[curr_corner_index_y] *
         bucket_search->num_buckets[0];

      if (curr_bucket >= bucket_to_cell_num_elements) continue;

      unsigned const * curr_src_cells =
         yac_get_dependencies_of_element(
            bucket_search->bucket_to_cell, curr_bucket);
      size_t curr_num_src_cells =
         yac_get_num_dependencies_of_element(
            bucket_search->bucket_to_cell, curr_bucket);
      // for all source cells associated with the current bucket cell
      for (size_t j = 0; j < curr_num_src_cells; ++j) {

         yac_get_grid_cell(bucket_search->grid_data, curr_src_cells[j],
                       &src_cell);

         double tgt_coords[3];
         LLtoXYZ(x_coords[curr_corner_index_x], y_coords[curr_corner_index_y],
                 tgt_coords);

         // if the target point is within the current source cell
         if (yac_point_in_cell(tgt_coords, src_cell)) {

            num_src_per_tgt[i] = 1;
            tgt_to_src_cell[num_total_dependencies++] = curr_src_cells[j];
            break; // continue with next target point
         }
      } // for j
   } // for i

   tgt_to_src_cell = xrealloc(tgt_to_src_cell, num_total_dependencies *
                              sizeof(*tgt_to_src_cell));

   struct dep_list * tgt_to_src_cells =
      yac_dep_list_new(num_tgt_corners, num_src_per_tgt, tgt_to_src_cell);

   //-------------
   // free memory
   //-------------

   yac_free_grid_cell(&src_cell);
   free(found_x); free(found_y); free(position_x); free(position_y);

   return tgt_to_src_cells;
}

// this routine searches the given point data in the data associated with the search
// handle and returns a dependency list containing the source cell for each point
static struct dep_list *
bucket_search_do_point_search_c2 (struct grid_search * search,
                                  double (*coordinates_xyz)[3],
                                  size_t num_points) {

   struct bucket_search * bucket_search = (struct bucket_search *)search;

   double * x_coordinates = xmalloc(num_points * sizeof(*x_coordinates));
   double * y_coordinates = xmalloc(num_points * sizeof(*y_coordinates));

   // Intel compilers starting version 16.0.2 and ending (at least) version
   // 17.0.6 fail to vectorize this loop correctly. Versions 16.0.1 and before
   // do not even try to do so. Versions 18.0.0 and above do it correctly. The
   // negative effect we want to avoid is that some elements of the array
   // 'y_coordinates' become significantly less than zero, whereas are expected
   // to be greater than or equal to zero (e.g. M_PI_2 - acos(0.0) < -1e-16).
   // This case is covered by test 'tests/test_point_search.c'.
   #if defined(__INTEL_COMPILER) && (__INTEL_COMPILER < 1800)
     #pragma novector
   #endif
   for (size_t i = 0; i < num_points; ++i)
     XYZtoLL(coordinates_xyz[i], x_coordinates + i, y_coordinates + i);

   //--------------------------------
   // find grid cells on bucket grid
   //--------------------------------

   int * found_x, * position_x, * found_y, * position_y;

   found_x    = xmalloc (num_points * sizeof(*found_x));
   found_y    = xmalloc (num_points * sizeof(*found_y));
   position_x = xmalloc (num_points * sizeof(*position_x));
   position_y = xmalloc (num_points * sizeof(*position_y));

   yac_bisection_search(x_coordinates, (unsigned)num_points,
                        bucket_search->bucket_grid_x_coords,
                        (unsigned)(bucket_search->num_buckets[0])+1,
                        position_x, found_x, 2.0*M_PI);

   yac_bisection_search(y_coordinates, (unsigned)num_points,
                        bucket_search->bucket_grid_y_coords,
                        (unsigned)(bucket_search->num_buckets[1])+1,
                        position_y, found_y, 0);

   //----------------------------------------------------------
   // use initial cell-to-bucket-results to find the find the
   // actual tgt point to src cell dependency
   //----------------------------------------------------------

   struct grid_cell src_cell;
   size_t num_total_dependencies = 0;
   size_t * num_src_per_tgt = xcalloc(num_points, sizeof(*num_src_per_tgt));
   unsigned * tgt_to_src_cell = xmalloc(num_points * sizeof(*tgt_to_src_cell));

   yac_init_grid_cell(&src_cell);

   size_t bucket_to_cell_num_elements =
      yac_get_num_elements(bucket_search->bucket_to_cell);

   // for all corners of the target grid
   for (size_t i = 0; i < num_points; ++i) {

      // if the current target corner was not found on the source grid
      if (!found_x[i] || !found_y[i]) continue; // continue with next corner

      // compute the current bucket cell index
      size_t curr_bucket = (size_t)position_x[i] + (size_t)position_y[i] *
         bucket_search->num_buckets[0];

      if (curr_bucket >= bucket_to_cell_num_elements) continue;

      unsigned const * curr_src_cells =
         yac_get_dependencies_of_element(
            bucket_search->bucket_to_cell, curr_bucket);
      size_t curr_num_src_cells =
         yac_get_num_dependencies_of_element(
           bucket_search->bucket_to_cell, curr_bucket);
      // for all source cells associated with the current bucket cell
      for (size_t j = 0; j < curr_num_src_cells; ++j) {

         yac_get_grid_cell(bucket_search->grid_data, curr_src_cells[j],
                       &src_cell);

         double tgt_coords[3];
         LLtoXYZ(x_coordinates[i], y_coordinates[i], tgt_coords);

         // if the target point is within the current source cell
         if (yac_point_in_cell(tgt_coords, src_cell)) {

            num_src_per_tgt[i] = 1;
            tgt_to_src_cell[num_total_dependencies++] = curr_src_cells[j];
            break; // continue with next target point
         }
      } // for j
   } // for i

   free(x_coordinates);
   free(y_coordinates);

   tgt_to_src_cell = xrealloc (tgt_to_src_cell, num_total_dependencies *
                               sizeof(tgt_to_src_cell[0]));

   struct dep_list * tgt_to_src_cells =
      yac_dep_list_new(num_points, num_src_per_tgt, tgt_to_src_cell);
   //-------------
   // free memory
   //-------------

   yac_free_grid_cell(&src_cell);
   free(found_x); free(found_y); free(position_x); free(position_y);

   return tgt_to_src_cells;
}

static struct dep_list *
bucket_search_do_point_search_c3(struct grid_search * search,
                                 double (*coordinates_xyz)[3],
                                 size_t num_points,
                                 struct points * points) {

  return
    yac_grid_search_utils_do_point_search_c3(
      search, coordinates_xyz, num_points, points);
}

static struct dep_list *
bucket_search_do_point_search_p (struct grid_search * search,
                                 struct grid * grid_data) {

   struct bucket_search * bucket_search = (struct bucket_search *)search;

   return
      yac_grid_search_utils_do_point_search_p(
         search, bucket_search->grid_data, grid_data);
}

static struct dep_list *
bucket_search_do_point_search_p2 (struct grid_search * search,
                                  double (*coordinates_xyz)[3],
                                  size_t num_points) {

   struct bucket_search * bucket_search = (struct bucket_search *)search;

   return
      yac_grid_search_utils_do_point_search_p2(
         search, bucket_search->grid_data, coordinates_xyz, num_points);
}

static struct dep_list *
bucket_search_do_point_search_p3 (struct grid_search * search,
                                  double (*coordinates_xyz)[3],
                                  size_t num_points,
                                  struct points * points) {

   return
      yac_grid_search_utils_do_point_search_p3(
         search, coordinates_xyz, num_points, points);
}

static
void bucket_search_do_point_search_p4 (struct grid_search * search,
                                       double coordinate_xyz[3],
                                       size_t * n_points,
                                       size_t * points_size,
                                       size_t ** points) {

   struct bucket_search * bucket_search = (struct bucket_search *)search;

   yac_grid_search_utils_do_point_search_p4(search, bucket_search->grid_data,
                                            coordinate_xyz, n_points,
                                            points_size, points);
}

