/**
 * @file field_interpolation_gen.h
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef FIELD_INTERPOLATION_GEN_H
#define FIELD_INTERPOLATION_GEN_H

#include "fields.h"
#include "communicator.h"
#include "interpolation_method.h"
#include "interpolation_plan.h"
#include "global_search.h"
#include "field_interpolation_gen_tgt.h"

/** \example test_field_interpolation.c
 * This shows how to use the field interpolation routines.
 */

/** \example test_field_interpolation_parallel.c
 * This contains a parallel example for the usage of field interpolation. The
 * example uses two source and two target processes.
 */

/**
 * \example test_field_interpolation_parallel2.c
 * This contains a parallel example for the usage of field interpolation. The
 * example uses two source and two target processes. In addition to
 * \ref test_field_interpolation_parallel.c the data distribution is more
 * complex and it makes use of masks.
 */

/**
 * \example test_field_interpolation_parallel3.c
 * This contains a parallel example for the usage of field interpolation. The
 * example uses two source and two target processes. In addition to
 * \ref test_field_interpolation_parallel.c the data distribution is more
 * complex, it makes use of masks and has multiple points per subdomain.
 */

/**
 * \example test_field_interpolation_parallel4.c
 * This contains a parallel example for the usage of field interpolation. The
 * example uses two source and two target processes. In addition to
 * \ref test_field_interpolation_parallel.c the data distribution is more
 * complex, it makes use of masks and has multiple points per subdomain.
 * However, for the points no coordinates are provided.
 */

/**
 * \example test_field_interpolation_parallel5.c
 * This contains a parallel example for the usage of field interpolation and
 * how to write an interpolation stack to a file. The example uses four source
 * and four target processes. For the interpolation a previously generated
 * weight file is used in combination with an average interpolation. The
 * combination of both interpolations is written to a weight file.
 */
 
typedef struct field_interpolation_gen * field_interpolation_gen;
typedef struct field_interpolation * field_interpolation;

struct field_interpolation_tags {
   unsigned results_exchange;
   unsigned point_data_exchange;
   unsigned point_data_header_exchange;
   unsigned tgt_point_location_exchange;
   unsigned interp_data_msg_count;
   unsigned interp_data_header;
   unsigned interp_data;
   unsigned exchange;
};

struct field_interpolation_gen_vtable {

   void (*enddef)(field_interpolation_gen interpolation_gen);
   field_interpolation (*get_field_interpolation)
     (field_interpolation_gen interpolation_gen);
   void (*add_weight_file_name)
     (field_interpolation_gen interpolation_gen,
      char const * weight_file_name);
   void (*delete)(field_interpolation_gen interpolation_gen);
};

struct field_interpolation_vtable {
   void (*put)(field_interpolation interpolation, double **** data, int collection_size);
   void (*get)(field_interpolation interpolation, double *** data, int collection_size);
   void (*delete)(field_interpolation interpolation);
};

struct field_interpolation_gen {

   struct field_interpolation_gen_vtable * vtable;
};

struct field_interpolation {

   struct field_interpolation_vtable * vtable;
};

struct field_interpolation_tags
yac_field_interpolation_generate_tag(struct communicator * comm,
                                     char const * source_comp_name,
                                     char const * target_comp_name,
                                     char const * source_domain_name,
                                     char const * target_domain_name,
                                     char const * source_field_name,
                                     char const * target_field_name);

/**
 * generates a field interpolation generator
 * @param[in] points                  point sets
 * @param[in] num_pointsets           number of point sets in points
 * @param[in] field_config_name       name if the target field configuration
 * @param[in] search                  global search data that is to be used for
 *                                    the interpolation
 * @param[in] other_field_config_name name of partner field
 * @param[in] is_source               is the local process the source for the
 *                                    interpolation
 * @param[in] is_target               is the local process the target for the
 *                                    interpolation
 * @param[in] methods                 interpolation methods that are to be used
 *                                    for the interpolation
 * @param[in] points_idx              which points set of the domain is to be
 *                                    interpolated
 * @param[in] weight_file_name        if != NULL the result of the interpolation
 *                                    will be written to the file
 * @param[in] io_proc_step            stride between ranks of io processes
 *                                    (e.g. for writing weight files)
 * @param[in] mapping_on_source       is the mapping in the put-get-operation to
 *                                    be executed on the source processes (on
 *                                    the target otherwise)
 * @return field interpolation generator or NULL if ((!is_source) && (!is_target))
 *
 * \remarks - a process can be source and target
 * \remarks - methods argument is ignored on processes that are no source
 * \remarks - points_idx and mask_handling arguments are ignored on processes that are no target
 * \remarks - not all interpolations support the writing of the result to file
 *            (e.g. interpolation method fixed)
 */
struct field_interpolation_gen *
yac_new_field_interpolation_gen(struct points *** points,
                                unsigned num_pointsets,
                                char const * field_config_name,
                                struct global_search * search,
                                const char * other_field_config_name,
                                unsigned is_source, unsigned is_target,
                                struct interpolation_method ** methods,
                                unsigned points_idx,
                                char const * weight_file_name,
                                unsigned io_proc_step,
                                int mapping_on_source);

/**
 * generates a field interpolation generator
 * @param[in] points                  point sets
 * @param[in] masks                   masks
 * @param[in] num_pointsets           number of point sets in points
 * @param[in] field_config_name       name if the target field configuration
 * @param[in] search                  global search data that is to be used for
 *                                    the interpolation
 * @param[in] other_field_config_name name of partner field
 * @param[in] is_source               is the local process the source for the interpolation
 * @param[in] is_target               is the local process the target for the interpolation
 * @param[in] methods                 interpolation methods that are to be used for the
 *                                    interpolation
 * @param[in] points_idx              which points set is to of the domain is to interpolated
 * @param[in] mask_handling           how to handle masked target points
 * @param[in] weight_file_name        if != NULL the result of the interpolation
 *                                    will be written to the file
 * @param[in] io_proc_step            stride between ranks of io processes
 *                                    (e.g. for writing weight files)
 * @param[in] mapping_on_source       is the mapping in the put-get-operation to
 *                                    be executed on the source processes (on
 *                                    the target otherwise)
 * @return field interpolation generator or NULL if ((!is_source) && (!is_target))
 *
 * \remarks - a process can be source and target
 * \remarks - methods argument is ignored on processes that are no source
 * \remarks - points_idx and mask_handling arguments are ignored on processes that are no target
 * \remarks - not all interpolations support the writing of the result to file
 *            (e.g. interpolation method fixed)
 */
struct field_interpolation_gen *
yac_new_field_interpolation_gen_mask(struct points *** points,
                                 struct mask *** masks,
                                 unsigned num_pointsets,
                                 char const * field_config_name,
                                 struct global_search * search,
                                 const char * other_field_config_name,
                                 unsigned is_source, unsigned is_target,
                                 struct interpolation_method ** methods,
                                 unsigned points_idx,
                                 enum yac_field_interpolation_tgt_mask_handling
                                   mask_handling,
                                 char const * weight_file_name,
                                 unsigned io_proc_step,
                                 int mapping_on_source);

void yac_enddef_field_interpolation_gen(struct field_interpolation_gen * interpolation_gen);

struct field_interpolation *
yac_get_field_interpolation(field_interpolation_gen interpolation_gen);

void yac_field_interpolation_gen_add_weight_file_name(
  field_interpolation_gen interpolation_gen, char const * weight_file_name);

void yac_delete_field_interpolation_gen(struct field_interpolation_gen * interpolation_gen);

// dimensions of data: [subdomain idx][points set idx]
void yac_put_field(struct field_interpolation * interpolation, double **** data, int collection_size);
void yac_get_field(struct field_interpolation * interpolation, double *** data, int collection_size);

void yac_delete_field_interpolation(struct field_interpolation * interpolation);

#endif // FIELD_INTERPOLATION_GEN_H

