/**
 * @file field_interpolation_gen_src_tgt.h
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef FIELD_INTERPOLATION_GEN_SRC_TGT_H
#define FIELD_INTERPOLATION_GEN_SRC_TGT_H

#include "field_interpolation_gen.h"
#include "field_interpolation_gen_tgt.h"

/**
 * constructor for a field_interpolation_gen where both sides are source and target
 * @param[in] points                  point sets
 * @param[in] num_pointsets           number of point sets in points
 * @param[in] field_config_name       name of field configuration
 * @param[in] search                  global search used as the basis for the
 *                                    interpolation
 * @param[in] other_field_config_name name of the exchange partner field
 * @param[in] methods                 interpolation method stack that is to be
 *                                    used for the interpolation
 * @param[in] points_idx              which points set is to of the domain is to
 *                                    interpolated
 * @param[in] weight_file_name        if != NULL the result of the interpolation
 *                                    will be written to the file
 * @param[in] io_proc_step            stride between ranks of io processes
 *                                    (e.g. for writing weight files)
 * @param[in] mapping_on_source       is the mapping in the put-get-operation to
 *                                    be executed on the source processes (on
 *                                    the target otherwise)
 * @return field interpolation generator
 *
 * \remarks - not all interpolations support the writing of the result to file
 *            (e.g. interpolation method fixed)
 */
struct field_interpolation_gen *
yac_new_field_interpolation_gen_src_tgt(struct points *** points,
                                        unsigned num_pointsets,
                                        char const * field_config_name,
                                        struct global_search * search,
                                        const char * other_field_config_name,
                                        struct interpolation_method ** methods,
                                        unsigned points_idx,
                                        char const * weight_file_name,
                                        unsigned io_proc_step,
                                        int mapping_on_source);

/**
 * constructor for a field_interpolation_gen where source and target field are
 * both available locally
 * @param[in] points_src            source point sets
 * @param[in] num_pointsets_src     number of source point sets in points
 * @param[in] field_config_name_src name of the local field configuration
 * @param[in] search_src            global search used as the basis for the
 *                                  interpolation
 * @param[in] methods_src           interpolation method stack that is to be
 *                                  used for the interpolation
 * @param[in] weight_file_name_src  if != NULL the result of the interpolation
 *                                  will be written to the file
 * @param[in] io_proc_step_src      stride between ranks of io processes
 *                                  (e.g. for writing weight files)
 * @param[in] points_tgt            target point sets
 * @param[in] num_pointsets_tgt     number of target point sets in points
 * @param[in] field_config_name_tgt name if the target field configuration
 * @param[in] search_tgt            global search used as the basis for the
 *                                  interpolation
 * @param[in] points_idx_tgt        which points set is to of the domain is to
 *                                  interpolated
 * @param[out] interp_gen_src       source field interpolation generator
 * @param[out] interp_gen_tgt       target field interpolation generator
 * @param[in] mapping_on_source     is the mapping in the put-get-operation to
 *                                  be executed on the source processes (on
 *                                  the target otherwise)
 *
 * \remarks - not all interpolations support the writing of the result to file
 */
void yac_new_field_interpolation_gen_src_tgt_2(
  struct points *** points_src, unsigned num_pointsets_src,
  char const * field_config_name_src, struct global_search * search_src,
  struct interpolation_method ** methods_src, char const * weight_file_name_src,
  unsigned io_proc_step_src, struct points *** points_tgt,
  unsigned num_pointsets_tgt, char const * field_config_name_tgt,
  struct global_search * search_tgt, unsigned points_idx_tgt,
  struct field_interpolation_gen ** interp_gen_src,
  struct field_interpolation_gen ** interp_gen_tgt, int mapping_on_source);

/**
 * constructor for a field_interpolation_gen where both sides are source and target
 * @param[in] points                  point sets
 * @param[in] masks                   masks
 * @param[in] num_pointsets           number of point sets in points
 * @param[in] field_config_name       name of field configuration
 * @param[in] search                  global search used as the basis for the
 *                                    interpolation
 * @param[in] other_field_config_name name of the exchange partner field
 *                                    configuration
 * @param[in] methods                 interpolation method stack that is to be
 *                                    used for the interpolation
 * @param[in] points_idx              which points set of the domain is to
 *                                    interpolated
 * @param[in] mask_handling           how to handle masked target points
 * @param[in] weight_file_name        if != NULL the result of the interpolation
 *                                    will be written to the file
 * @param[in] io_proc_step            stride between ranks of io processes
 *                                    (e.g. for writing weight files)
 * @param[in] mapping_on_source       is the mapping in the put-get-operation to
 *                                    be executed on the source processes (on
 *                                    the target otherwise)
 * @return field interpolation generator
 *
 * \remarks - not all interpolations support the writing of the result to file
 *            (e.g. interpolation method fixed)
 */
struct field_interpolation_gen *
yac_new_field_interpolation_gen_src_tgt_mask(struct points *** points,
                                         struct mask *** masks,
                                         unsigned num_pointsets,
                                         char const * field_config_name,
                                         struct global_search * search,
                                         const char * other_field_config_name,
                                         struct interpolation_method ** methods,
                                         unsigned points_idx,
                                         enum yac_field_interpolation_tgt_mask_handling
                                           mask_handling,
                                         char const * weight_file_name,
                                         unsigned io_proc_step,
                                         int mapping_on_source);

/**
 * constructor for a field_interpolation_gen where source and target field are
 * both available locally
 * @param[in] points_src            source point sets
 * @param[in] num_pointsets_src     number of source point sets in points
 * @param[in] field_config_name_src name of the local field configuration
 * @param[in] search_src            global search used as the basis for the
 *                                  interpolation
 * @param[in] methods_src           interpolation method stack that is to be
 *                                  used for the interpolation
 * @param[in] weight_file_name_src  if != NULL the result of the interpolation
 *                                  will be written to the file
 * @param[in] io_proc_step_src      stride between ranks of io processes
 *                                  (e.g. for writing weight files)
 * @param[in] points_tgt            target point sets
 * @param[in] masks_tgt             target masks
 * @param[in] num_pointsets_tgt     number of target point sets in points
 * @param[in] field_config_name_tgt name if the target field configuration
 * @param[in] search_tgt            global search used as the basis for the
 *                                  interpolation
 * @param[in] points_idx_tgt        which points set is to of the domain is to
 *                                  interpolated
 * @param[in] mask_handling_tgt     how to handle masked target points
 * @param[out] interp_gen_src       source field interpolation generator
 * @param[out] interp_gen_tgt       target field interpolation generator
 * @param[in] mapping_on_source     is the mapping in the put-get-operation to
 *                                  be executed on the source processes (on
 *                                  the target otherwise)
 *
 * \remarks - not all interpolations support the writing of the result to file
 */
void yac_new_field_interpolation_gen_src_tgt_mask_2(
  struct points *** points_src, unsigned num_pointsets_src,
  char const * field_config_name_src, struct global_search * search_src,
  struct interpolation_method ** methods_src, char const * weight_file_name_src,
  unsigned io_proc_step_src, struct points *** points_tgt,
  struct mask *** masks_tgt, unsigned num_pointsets_tgt,
  char const * field_config_name_tgt, struct global_search * search_tgt,
  unsigned points_idx_tgt,
  enum yac_field_interpolation_tgt_mask_handling mask_handling_tgt,
  struct field_interpolation_gen ** interp_gen_src,
  struct field_interpolation_gen ** interp_gen_tgt, int mapping_on_source);

#endif // FIELD_INTERPOLATION_GEN_SRC_TGT_H

