/**
 * @file interpolation_method_conserv.h
 * @brief conservative remapping interpolation method
 *
 * cell-based 1st order conservative remapping
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INTERPOLATION_METHOD_CONSERV_H
#define INTERPOLATION_METHOD_CONSERV_H

 /** \example test_interpolation_method_conserv.c
 * This shows how the conservative remapping interpolation works
 */

 /** \example test_interpolation_method_conserv_parallel.c
 * This shows how the parallel conservative remapping interpolation works
 */

#include "interpolation_method.h"

/** \file interpolation_method_conserv.h
 *  \brief 1st order conservative interpolation method
 *
 * cell interpolation method that searches for each target cell the overlapping
 * source cells. Polygon intersections are determined and partial areas are 
 * calculated. The target value is determined by adding up the partial contributions
 * from each overlapping source cell. Weights are provided as partial areas normalised
 * by the target cell area.
 *
 * Currently it is assumed that the first struct points of each subdomain is of type
 * CELL. This is used for the interpolation.
 */

/**
 * normalisation options for conservative interpolation\n
 * (see SCRIP user manual for a more detailed description of the options)
 */
enum yac_interpolation_method_conserv_normalisation {
   CONSERV_DESTAREA, //!< interpolation values are normalised by the area of
                     //!< the respective target cell
                     //!< (this is the default option)
   CONSERV_FRACAREA, //!< interpolation values will be normalised by the area
                     //!< of the respective target cell that is covered by
                     //!< non-masked target cells
};

/**
 * constructor for a interpolation method of type interpolation_method_conserv
 * @param[in] search           global search object that is to be used for the
 *                             interpolation
 * @param[in] domain           %domain of the local process
 * @param[in] stack_name       name of the interpolation stack
 * @param[in] points           points to be used for the weight generation
 * @param[in] masks            mask to be used for the weight generation
 * @param[in] num_pointsets    number of point sets
 * @param[in] points_idx       which points set of the domain is to be used for
 *                             the interpolation
 * @param[in] order            1st or 2nd order remapping
 * @param[in] enforced_conserv enforce conservation by correcting truncation errors
 * @param[in] partial_coverage if == 0, target cells that are not completely
 *                             covered by non-masked source cells will not be
 *                             interpolated (this is the default option)\n
 *                             if != 0, for target cells that are only
 *                             partially covered by non-mask source cells the
 *                             interpolation value will be determined by the
 *                             contributions of the respective source cells
 * @param[in] normalisation    specifies how to do the normalisation
 */
struct interpolation_method *
yac_new_interpolation_method_conserv ( struct global_search * search,
                                       struct domain * domain,
                                       char const * stack_name,
                                       struct points *** points,
                                       struct mask *** masks,
                                       unsigned num_pointsets,
                                       unsigned points_idx,
                                       int order,
                                       int enforced_conserv,
                                       int partial_coverage,
                                       enum yac_interpolation_method_conserv_normalisation normalisation );

/**
 * constructor for a interpolation method of type interpolation_method_conserv
 * @param[in] search           global search object that is to be used for the
 *                             interpolation
 * @param[in] domain           %domain of the local process
 * @param[in] stack_name       name of the interpolation stack
 * @param[in] points           points to be used for the weight generation
 * @param[in] masks            mask to be used for the weight generation
 * @param[in] num_pointsets    number of point sets
 * @param[in] points_idx       which points set of the domain is to be used for
 *                             the interpolation
 * @param[in] order            1st or 2nd order remapping
 * @param[in] enforced_conserv enforce conservation by correcting truncation errors
 * @param[in] partial_coverage if == 0, target cells that are not completely
 *                             covered by non-masked source cells will not be
 *                             interpolated (this is the default option)\n
 *                             if != 0, for target cells that are only
 *                             partially covered by non-mask source cells the
 *                             interpolation value will be determined by the
 *                             contributions of the respective source cells
 * @param[in] normalisation    specifies how to do the normalisation
 * @param[in] use_source_mask  specifies whether masked source cells shall
 *                             be considered (true) or not (false)
 *
 */
struct interpolation_method *
yac_new_interpolation_method_conserv_mask (struct global_search * search,
                                           struct domain * domain,
                                           char const * stack_name,
                                           struct points *** points,
                                           struct mask *** masks,
                                           unsigned num_pointsets,
                                           unsigned points_idx,
                                           int order,
                                           int enforced_conserv,
                                           int partial_coverage,
                                           enum yac_interpolation_method_conserv_normalisation normalisation,
                                           unsigned use_source_mask);

#endif // INTERPOLATION_METHOD_CONSERV_H

