/**
 * @file interpolation_method_fixed.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "interpolation_method.h"
#include "interpolation_method_fixed.h"
#include "interpolation_plan.h"
#include "communicator.h"
#include "utils.h"

static void
do_search_interpolation_method_fixed(struct interpolation_method ** method,
                                     struct interpolation_plan_element ** plan_stack,
                                     int * global_ids, double (*coordinates_xyz)[3],
                                     enum yac_location tgt_location,
                                     unsigned num_points,
                                     unsigned * num_interpolated_points,
                                     struct interpolation_method_data_request ** data_request);
static void 
enddef_interpolation_method_fixed(struct interpolation_method ** method);
static void 
get_source_exchange_data_interpolation_method_fixed(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request);
static void delete_interpolation_method_fixed(struct interpolation_method * method);
static struct global_search *
get_global_search(struct interpolation_method * method);
static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data);

static struct interpolation_method_vtable
   interpolation_method_fixed_vtable = {do_search_interpolation_method_fixed,
                                        enddef_interpolation_method_fixed,
                                        get_source_exchange_data_interpolation_method_fixed,
                                        delete_interpolation_method_fixed,
                                        get_global_search,
                                        set_finish_callback};

struct interpolation_method_fixed_tags {

   unsigned global_id_exchange[2];
};

struct interpolation_method_fixed {

   struct interpolation_method_vtable * vtable;
   void (*finish_callback)(void * user_data);
   void * finish_user_data;
   struct global_search * search;
   double fixed_value;
   struct communicator * comm;
   struct interpolation_method_data_request ** data_request;
   struct interpolation_method_fixed_tags tags;
};

static void
generate_tag_string(char * buf, size_t buffer_size,
                    char const * tag_base_str,
                    char const * source_comp_name,
                    char const * target_comp_name,
                    char const * source_domain_name,
                    char const * target_domain_name,
                    int unique_interpolation_method_id,
                    char const * tag_str) {

   int ret_val;
   ret_val = snprintf(buf, buffer_size, "%s%s%s%s%s%d%s", tag_base_str,
                      source_comp_name, target_comp_name,
                      source_domain_name, target_domain_name,
                      unique_interpolation_method_id, tag_str);

   if (ret_val < 0 || (size_t)ret_val >= buffer_size)
      yac_internal_abort_message("ERROR(generate_tag_string): truncation", __FILE__, __LINE__);
}

static void
generate_interpolation_method_fixed_tag(struct interpolation_method_fixed * method) {

   char const * str_base_str             = "interpolation method fixed";
   char const * str_generator_string     = "generator string";
   char const * str_global_id_exchange_a = "global id exchange a";
   char const * str_global_id_exchange_b = "global id exchange b";

   struct global_search * search;

   search = get_global_search((struct interpolation_method *)method);

   struct communicator * comm;

   comm = yac_get_domain_comm(yac_get_local_domain(search));

   char const * source_comp_name = yac_get_local_component_name(search);
   char const * target_comp_name = yac_get_remote_component_name(search);

   char const * source_domain_name = yac_get_local_domain_name(search);
   char const * target_domain_name = yac_get_remote_domain_name(search);

   unsigned max_tag_strlen = strlen(str_global_id_exchange_a);

   unsigned max_str_buf_size = max_tag_strlen;
   max_str_buf_size += strlen(str_base_str);
   max_str_buf_size += strlen(source_comp_name);
   max_str_buf_size += strlen(target_comp_name);
   max_str_buf_size += strlen(source_domain_name);
   max_str_buf_size += strlen(target_domain_name) + 10 + 1; // (+ 10) for unique_method_id
                                                            // (+ 1)  for '\0'

   char * unique_generator_string = xmalloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name,
                       source_domain_name, target_domain_name,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_generator_string);

   char * tag_strings[2];
   tag_strings[0] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name,
                       source_domain_name, target_domain_name,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_a);

   tag_strings[1] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], max_str_buf_size, str_base_str,
                       source_comp_name, target_comp_name,
                       source_domain_name, target_domain_name,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_b);

   unsigned tags[2];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 2, tags);

   free(unique_generator_string);

   method->tags.global_id_exchange[0] = tags[0];
   method->tags.global_id_exchange[1] = tags[1];

   for (unsigned i = 0; i < 2; ++i) free(tag_strings[i]);
}

struct interpolation_method *
yac_new_interpolation_method_fixed(double fixed_value,
                                   struct global_search * search) {

   struct interpolation_method_fixed * method;

   method = xmalloc(1 * sizeof(*method));

   method->fixed_value = fixed_value;
   method->search = search;
   method->comm = yac_get_domain_comm(yac_get_local_domain(search));
   method->vtable = &interpolation_method_fixed_vtable;
   method->finish_callback = NULL;
   method->finish_user_data = NULL;

   yac_init_interpolation_method((struct interpolation_method *)method, search);
   generate_interpolation_method_fixed_tag(method);

   return (struct interpolation_method *)method;
}

struct search_data_fixed {

   struct interpolation_method ** method_stack;
   struct interpolation_plan_element ** plan_stack;

   int * global_ids;
   double (*coordinates_xyz)[3];
   enum yac_location tgt_location;
   unsigned num_points;
   unsigned * num_interpolated_points;
};

static void
do_search_continue_fixed(void * user_data,
                         int * global_ids_to_remove,
                         unsigned num_global_ids_to_remove) {

   struct search_data_fixed * data = user_data;

   if (num_global_ids_to_remove > 0) {

      unsigned new_num_points = data->num_points;
      int * remaining_points;
      yac_remove_points(data->global_ids, &new_num_points, global_ids_to_remove,
                        num_global_ids_to_remove, &remaining_points);

      if (data->coordinates_xyz != NULL) {
        for (unsigned i = 0; i < new_num_points; ++i) {
          int idx = remaining_points[i];
          data->coordinates_xyz[i][0] = data->coordinates_xyz[idx][0];
          data->coordinates_xyz[i][1] = data->coordinates_xyz[idx][1];
          data->coordinates_xyz[i][2] = data->coordinates_xyz[idx][2];
        }
      }

      *(data->num_interpolated_points) -= data->num_points - new_num_points;
      data->num_points = new_num_points;

      free(remaining_points);
   }

   struct interpolation_method_fixed * method_fixed;

   method_fixed = (struct interpolation_method_fixed*)data->method_stack[0];

   // generate interpolation plan
   if (*(data->num_interpolated_points) > 0) {

      // generate a plan to execute the interpolation
      struct interpolation_plan_element * plan;

      plan = xmalloc(1 * sizeof(*plan));

      yac_init_interpolation_plan_element_fixed(plan, *(data->num_interpolated_points),
                                                method_fixed->fixed_value);

      data->plan_stack[0] = plan;
      data->plan_stack++;
   }

   // call the do search interpolation_method for the next method
   // however, there are no points left
   if (data->method_stack[1] != NULL)
      data->method_stack[1]->vtable->do_search(
         data->method_stack+1, data->plan_stack, NULL, NULL,
         data->tgt_location, 0, data->num_interpolated_points+1,
         method_fixed->data_request+1);

   free(data);

   if (method_fixed->finish_callback != NULL)
      method_fixed->finish_callback(method_fixed->finish_user_data);
}

static void
do_search_interpolation_method_fixed(struct interpolation_method ** method,
                                     struct interpolation_plan_element ** plan_stack,
                                     int * global_ids, double (*coordinates_xyz)[3],
                                     enum yac_location tgt_location,
                                     unsigned num_points,
                                     unsigned * num_interpolated_points,
                                     struct interpolation_method_data_request ** data_request) {

   struct interpolation_method_fixed * method_fixed;

   method_fixed = (struct interpolation_method_fixed *)(method[0]);
   method_fixed->data_request = data_request;

   // we can process all points
   *num_interpolated_points = num_points;

   struct search_data_fixed * data;

   data = xmalloc(1 * sizeof(*data));

   data->method_stack = method;
   data->plan_stack = plan_stack;
   data->global_ids = global_ids;
   data->coordinates_xyz = coordinates_xyz;
   data->tgt_location = tgt_location;
   data->num_points = num_points;
   data->num_interpolated_points = num_interpolated_points;

   // exchange results with other source processes
   yac_exchange_results(global_ids, *num_interpolated_points,
                        num_points, method_fixed->tags.global_id_exchange,
                        method_fixed->comm, do_search_continue_fixed, data);

   return;
}

static void
enddef_interpolation_method_fixed(struct interpolation_method ** method) {

   struct interpolation_method_fixed * method_fixed;

   method_fixed = (struct interpolation_method_fixed *)(method[0]);

   while(yac_comm_wait(method_fixed->comm,
                       method_fixed->tags.global_id_exchange[0]) ||
         yac_comm_wait(method_fixed->comm,
                       method_fixed->tags.global_id_exchange[1]));

   if (method[1] != NULL)
      method[1]->vtable->enddef(method + 1);
}

static void
get_source_exchange_data_interpolation_method_fixed(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request) {

   struct interpolation_method_fixed * method_fixed;

   method_fixed = (struct interpolation_method_fixed *)(method[0]);

   *data_request = method_fixed->data_request;
   method_fixed->data_request = NULL;
}

void delete_interpolation_method_fixed(struct interpolation_method * method) {

   free(method);
}

static struct global_search *
get_global_search(struct interpolation_method * method) {

   return ((struct interpolation_method_fixed*)(method))->search;
}

 static void
 set_finish_callback(struct interpolation_method * method,
                     void (*callback)(void * user_data),
                     void * user_data) {

   struct interpolation_method_fixed * method_fixed;

   method_fixed = (struct interpolation_method_fixed*)method;

   method_fixed->finish_callback = callback;
   method_fixed->finish_user_data = user_data;
}

