/**
 * @file interpolation_method_hcsbb.h
 * @brief Hybrid cubic spherical Bernstein-Bezier patch interpolation method
 *
 * Point interpolation method that uses hybrid cubic spherical Bernstein-Bezier
 * patches [1].
 *
 * [1]: Xiaoyu Liu, Larry L. Schumaker, Hybrid Béer patches on sphere-like
 *      surfaces, In Journal of Computational and Applied Mathematics,
 *      Volume 73, Issues 1.2, 1996, Pages 157-172, ISSN 0377-0427,
 *      https://doi.org/10.1016/0377-0427(96)00041-6.
 *
 * @copyright Copyright  (C)  2017 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INTERPOLATION_METHOD_HCSBB_H
#define INTERPOLATION_METHOD_HCSBB_H

 /** \example test_interpolation_method_hcsbb.c
 * This shows how the hcsbb remapping interpolation works
 */

#include "interpolation_method.h"

/**
 * constructor for a interpolation method of type interpolation_method_hcsbb
 * @param[in] search             global search object that is to be used for the
 *                               interpolation
 * @param[in] domain             %domain of the local process
 * @param[in] stack_name         name of the interpolation stack
 * @param[in] points             points to be used for the weight generation
 * @param[in] num_pointsets      number of point sets
 * @param[in] points_idx         which points set of the domain is to be used
 *                               for the interpolation
 *
 * \remarks all subdomains need to have a halo of at least one cell in order to
 *          be able to do interpolation across process boundaries
 */
struct interpolation_method *
yac_new_interpolation_method_hcsbb(struct global_search * search,
                                   struct domain * domain,
                                   char const * stack_name,
                                   struct points *** points,
                                   unsigned num_pointsets,
                                   unsigned points_idx);

/**
 * constructor for a interpolation method of type interpolation_method_hcsbb
 * @param[in] search            global search object that is to be used for the
 *                              interpolation
 * @param[in] domain            %domain of the local process
 * @param[in] stack_name        name of the interpolation stack
 * @param[in] points            points to be used for the weight generation
 * @param[in] masks             mask to be used for the weight generation
 * @param[in] num_pointsets     number of point sets
 * @param[in] points_idx        which points set of the domain is to be used for
 *                              the interpolation
 * @param[in] use_source_mask   specifies how to handle masked source points
 *
 * \remarks all subdomains need to have a halo of at least one cell in order to
 *          be able to do interpolation across process boundaries
 * \remarks halo points need to have valid mask values
 */
struct interpolation_method *
yac_new_interpolation_method_hcsbb_mask(struct global_search * search,
                                        struct domain * domain,
                                        char const * stack_name,
                                        struct points *** points,
                                        struct mask *** masks,
                                        unsigned num_pointsets,
                                        unsigned points_idx,
                                        unsigned use_source_mask);

#endif // INTERPOLATION_METHOD_HCSBB_H

