/**
 * @file interpolation_method_nnn.c
 * @brief implementation of the n-Nearest Neighbour interpolation method
 *
 * @copyright Copyright  (C)  2017 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "yac_lapack_interface.h"
#include "interpolation_method.h"
#include "interpolation_method_nnn.h"
#include "interpolation_plan.h"
#include "grid_search.h"
#include "global_search.h"
#include "ensure_array_size.h"
#include "utils.h"
#include "grid.h"
#include "nnn_search.h"

static void
do_search_interpolation_method_nnn(struct interpolation_method ** method,
                                   struct interpolation_plan_element ** plan_stack,
                                   int * global_ids, double (*coordinates_xyz)[3],
                                   enum yac_location tgt_location,
                                   unsigned num_points,
                                   unsigned * num_interpolated_points,
                                   struct interpolation_method_data_request ** data_request);
static void
enddef_interpolation_method_nnn(struct interpolation_method ** method);
static void
get_source_exchange_data_interpolation_method_nnn(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request);
static void
delete_interpolation_method_nnn(struct interpolation_method * method);
static struct global_search *
get_global_search(struct interpolation_method * method);
static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data);

static struct interpolation_method_vtable
   interpolation_method_nnn_vtable = {
      .do_search                = do_search_interpolation_method_nnn,
      .enddef                   = enddef_interpolation_method_nnn,
      .get_source_exchange_data = get_source_exchange_data_interpolation_method_nnn,
      .delete                   = delete_interpolation_method_nnn,
      .get_global_search        = get_global_search,
      .set_finish_callback      = set_finish_callback};

struct interpolation_method_nnn_tags {

   unsigned global_id_exchange_1[2];  //!< tags for the first global id exchange in \ref do_search_interpolation_method_nnn
   unsigned global_id_exchange_2[2];  //!< tags for the second global id exchange in \ref do_search_continue_nnn
   unsigned exchange_results[2];      //!< tags for exchanging which target points are actually interpolated
};

struct interpolation_method_nnn {

   struct interpolation_method_vtable * vtable;
   void (*finish_callback)(void * user_data);  //!< callback routine that is called once the interpolation is finished
   void * finish_user_data;                    //!< user data that is passed to finish_callback
   struct global_search * search;              //!< global search object used a the basis for the interpolation
   struct domain * domain;
   struct points *** points;
   struct mask *** masks;
   unsigned num_pointsets;
   enum yac_nnn_reduction_type reduction_type; //!< reduction type used to compute a target point based on n source points
   unsigned n;                                 //!< number of nearest neighbours per target point
   double rbf_scale;                           //!< scaling factor in case reduction_type == NNN_RBF
   double gauss_scale;                         //!< scaling factor in case reduction_type == NNN_GAUSS_WEIGHTED
   unsigned points_idx;                        //!< index of the points set of field that is to be used for the interpolation
   struct interpolation_method_data_request ** data_request;
   struct interpolation_method_nnn_tags tags;
   unsigned use_source_mask;
   struct nnn_search * nnn_search;
};

enum interp_flag {
  LOCAL = 0,  //!< interpolated on the local process
  NOT = 1,    //!< not assigned to any process yet
  REMOTE = 2, //!< interpolated by a remote process
};

/**
 * contains all temporary search specific data for one target point
 */
struct tgt_point_search_data {

   int global_id;                 //!< global id of the target point
   enum interp_flag flag;
   double coordinate_xyz[3];

   double min_cos_angle; //!< minimum cosine angle of the results

   struct result_point * results; //!< the source points found for the target point
   unsigned num_found_points;          //!< number of found source points
};

struct search_data {

   struct tgt_point_search_data * tgt_point_data;
   size_t num_tgt_points;                //!< total number of entries in tgt_point_data
   size_t num_tgt_points_to_interpolate; //!< number of tgt_point_data entries
                                         //!< that are marked as LOCAL
};

//! collection of data that is required through out the nearest neighbourhood search
struct intermediate_method_data {

   struct search_data search_data; //!< search data for the target points that
                                   //!< are assigned to the process

   struct interpolation_method ** method_stack;
   struct interpolation_plan_element ** plan_stack;
   unsigned * num_interpolated_points;
   double (*coordinates_xyz)[3];
   int * global_ids;
   enum yac_location tgt_location;
};

struct data_request_point {
  int global_id;
  unsigned rank;
};

static void
generate_tag_string(char * buf,  size_t buffer_size,
                    char const * tag_base_str,
                    char const * source_comp_name,
                    char const * target_comp_name,
                    char const * source_domain_name,
                    char const * target_domain_name,
                    char const * stack_name,
                    unsigned points_idx,
                    enum yac_nnn_reduction_type reduction_type,
                    unsigned n,
                    unsigned use_source_mask,
                    double rbf_scale,
                    double gauss_scale,
                    int unique_interpolation_method_id,
                    char const * tag_str) {

   int ret_val;
   ret_val = snprintf(buf, buffer_size, "%s%s%s%s%s%s%u%d%u%u%e%e%d%s", tag_base_str,
                      source_comp_name, target_comp_name, source_domain_name,
                      target_domain_name, stack_name, points_idx,
                      (int)reduction_type, n, use_source_mask, rbf_scale, gauss_scale,
                      unique_interpolation_method_id, tag_str);

   if (ret_val < 0 || (size_t)ret_val >= buffer_size)
      yac_internal_abort_message(
         "ERROR(generate_tag_string): truncation", __FILE__, __LINE__);
}

static void
generate_interpolation_method_nnn_tag(struct interpolation_method_nnn * method,
                                      char const * stack_name) {

   char const * str_base_str               = "interpolation method nnn";
   char const * str_generator_string       = "generator string";
   char const * str_global_id_exchange_1_a = "global id exchange 1 a";
   char const * str_global_id_exchange_1_b = "global id exchange 1 b";
   char const * str_global_id_exchange_2_a = "global id exchange 2 a";
   char const * str_global_id_exchange_2_b = "global id exchange 2 b";
   char const * str_exchange_results_a     = "exchange results a";
   char const * str_exchange_results_b     = "exchange results b";

   struct global_search * search =
      get_global_search((struct interpolation_method *)method);

   struct communicator * comm = yac_get_domain_comm(method->domain);

   char const * source_comp_name = yac_get_local_component_name(search);
   char const * target_comp_name = yac_get_remote_component_name(search);

   char const * source_domain_name = yac_get_local_domain_name(search);
   char const * target_domain_name = yac_get_remote_domain_name(search);

   unsigned max_tag_strlen = strlen(str_global_id_exchange_1_a);

   unsigned max_str_buf_size = max_tag_strlen;
   max_str_buf_size += strlen(str_base_str);
   max_str_buf_size += strlen(source_comp_name);
   max_str_buf_size += strlen(target_comp_name);
   max_str_buf_size += strlen(source_domain_name);
   max_str_buf_size += strlen(target_domain_name);
   max_str_buf_size += strlen(stack_name) + 10 + 10 + 10 + 2 + 10 + 10 + 10 + 1;
                                      // (+ 10) for points_idx
                                      // (+ 10) for reduction_type
                                      // (+ 10) for n
                                      // (+ 2)  for mask handling
                                      // (+ 10) for unique_method_id
                                      // (+ 10) for rbf_scale
                                      // (+ 10) for gauss_scale
                                      // (+ 1)  for '\0'

   char * unique_generator_string = xmalloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->reduction_type,
                       method->n, method->use_source_mask, method->rbf_scale, method->gauss_scale,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_generator_string);

   char * tag_strings[6];

   tag_strings[0] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->reduction_type,
                       method->n, method->use_source_mask, method->rbf_scale, method->gauss_scale,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_1_a);

   tag_strings[1] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->reduction_type,
                       method->n, method->use_source_mask, method->rbf_scale, method->gauss_scale,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_1_b);

   tag_strings[2] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[2], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->reduction_type,
                       method->n, method->use_source_mask, method->rbf_scale, method->gauss_scale,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_2_a);

   tag_strings[3] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[3], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->reduction_type,
                       method->n, method->use_source_mask, method->rbf_scale, method->gauss_scale,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_2_b);

   tag_strings[4] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[4], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->reduction_type,
                       method->n, method->use_source_mask, method->rbf_scale, method->gauss_scale,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_exchange_results_a);

   tag_strings[5] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[5], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->reduction_type,
                       method->n, method->use_source_mask, method->rbf_scale, method->gauss_scale,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_exchange_results_b);

   unsigned tags[6];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 6, tags);

   free(unique_generator_string);

   method->tags.global_id_exchange_1[0] = tags[0];
   method->tags.global_id_exchange_1[1] = tags[1];
   method->tags.global_id_exchange_2[0] = tags[2];
   method->tags.global_id_exchange_2[1] = tags[3];
   method->tags.exchange_results[0] = tags[4];
   method->tags.exchange_results[1] = tags[5];

   for (unsigned i = 0; i < 6; ++i) free(tag_strings[i]);
}

static int compare_int(const void * a, const void * b) {

  return ((*(int*)a) - (*(int*)b));
}

static void
init_tgt_search_data(struct tgt_point_search_data * data, double coordinate_xyz[3],
                     int global_id, double min_cos_angle) {

   data->global_id = global_id;
   data->flag = NOT;
   data->coordinate_xyz[0] = coordinate_xyz[0];
   data->coordinate_xyz[1] = coordinate_xyz[1];
   data->coordinate_xyz[2] = coordinate_xyz[2];

   data->min_cos_angle = min_cos_angle;

   data->results = NULL;
   data->num_found_points = 0;
}

static struct tgt_point_search_data *
tgt_point_search_data_new(unsigned num_points,
                          double (*coordinates_xyz)[3],
                          int * global_ids) {

   struct tgt_point_search_data * data = xmalloc(num_points * sizeof(*data));

   for (unsigned i = 0; i < num_points; ++i)
      init_tgt_search_data(
        data+i, coordinates_xyz[i], global_ids[i], -1.0);

   return data;
}

/**
 * search for result source points that are marked as remote, but are actually
 * available locally
 */
static void detect_local_points(struct intermediate_method_data * data) {

   struct search_data * search_data = &(data->search_data);
   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);

   struct domain * domain = method->domain;

   int comm_rank = yac_get_comm_rank(yac_get_domain_comm(domain));

   int * global_ids = NULL;
   unsigned global_ids_array_size = 0;
   unsigned num_global_ids = 0;

   // search for remote points that have the same rank as the local process
   for (unsigned i = 0; i < data->num_interpolated_points[0]; ++i) {

      for (unsigned j = 0;
           j < search_data->tgt_point_data[i].num_found_points; ++j) {

         if (search_data->tgt_point_data[i].results[j].is_local == 0 &&
             search_data->tgt_point_data[i].results[j].data.remote.rank ==
                comm_rank) {

            ENSURE_ARRAY_SIZE(global_ids, global_ids_array_size,
                              num_global_ids + 1);
            global_ids[num_global_ids++] =
               search_data->tgt_point_data[i].results[j].global_id;
         }
      }
   }

   // if some points were found
   if (num_global_ids > 0) {

      enum yac_location location = method->points[0][method->points_idx]->location;

      int * matched_global_ids;
      unsigned * local_ids;
      unsigned * subdomain_ids;
      unsigned num_matched_global_ids = 0;
      unsigned curr_num_matched_global_ids = 0;
      unsigned num_subdomains;

      matched_global_ids = xmalloc(num_global_ids * sizeof(*matched_global_ids));
      local_ids = xmalloc(num_global_ids * sizeof(*local_ids));
      subdomain_ids = xmalloc(num_global_ids * sizeof(*subdomain_ids));

      num_subdomains = yac_get_num_subdomains(domain);

      // sort global_ids (required by match_global_ids)
      qsort(global_ids, num_global_ids, sizeof(*global_ids), compare_int);

      // search for the global ids within the local data
      for (unsigned i = 0; i < num_subdomains; ++i) {

         yac_match_global_ids(domain, i, location, global_ids, num_global_ids,
                              matched_global_ids + num_matched_global_ids, NULL,
                              &curr_num_matched_global_ids, local_ids +
                              num_matched_global_ids, 0);


         for (unsigned j = 0; j < curr_num_matched_global_ids; ++j)
            subdomain_ids[num_matched_global_ids+j] = i;

         num_matched_global_ids += curr_num_matched_global_ids;
      }

      if (num_matched_global_ids != num_global_ids)
         yac_internal_abort_message(
            "ERROR(detect_local_points): is the definition of the global"
            " ids correct?", __FILE__, __LINE__);

      // convert the "remote" points to local ones
      for (unsigned i = 0; i < data->num_interpolated_points[0]; ++i) {

         for (unsigned j = 0;
              j < search_data->tgt_point_data[i].num_found_points; ++j) {

            if (search_data->tgt_point_data[i].results[j].is_local == 0 &&
                search_data->tgt_point_data[i].results[j].data.remote.rank ==
                   comm_rank) {

               for (unsigned k = 0; k < num_global_ids; ++k) {

                  if (search_data->tgt_point_data[i].results[j].global_id ==
                      matched_global_ids[k]) {

                     search_data->tgt_point_data[i].results[j].is_local = 1;
                     search_data->tgt_point_data[i].results[j].data.local.
                        local_id = local_ids[k];
                     search_data->tgt_point_data[i].results[j].data.local.
                        subdomain_id = subdomain_ids[k];
                  }
               }
            }
         }
      }

      free(matched_global_ids);
      free(local_ids);
      free(subdomain_ids);
   }

   free(global_ids);
}

static void insertion_sort_data_request_point(
  struct data_request_point * a, size_t * curr_a_length,
  struct data_request_point * b, size_t b_length) {

  size_t tmp_a_length = *curr_a_length;
  size_t num_elements_to_copy = 0;

  // count the number of unique elements that need to be copied from b to a
  for (size_t i = 0, j = 0; i < b_length;) {

    struct data_request_point curr_b = b[i++];

    // try to find curr_b in a
    while ((j < tmp_a_length) && (a[j].global_id < curr_b.global_id)) ++j;
    // skip all b that are equal to curr_b
    while ((i < b_length) && (b[i].global_id == curr_b.global_id)) ++i;

    if ((j >= tmp_a_length) || (a[j].global_id != curr_b.global_id)) {
      b[num_elements_to_copy++] = curr_b;
    }
  }

  *curr_a_length += num_elements_to_copy;

  // copy elements from back to front
  size_t i = num_elements_to_copy - 1,
         j = tmp_a_length - 1,
         k = tmp_a_length + num_elements_to_copy - 1;

  while ((i < num_elements_to_copy) && (j < tmp_a_length))
    a[k--] = (a[j].global_id > b[i].global_id)?a[j--]:b[i--];

  if (i < num_elements_to_copy) memcpy(a, b, (i+1)*sizeof(*a));
}

static int compare_data_request_points(const void * a, const void * b) {

  return ((struct data_request_point const *)a)->global_id -
         ((struct data_request_point const *)b)->global_id;
}

/**
 * analyses the results of the search and generates a data_request object, which
 * is required by the field_interpolation in order to be able to provide the
 * remote points needed for the local interpolation
 */
static void generate_data_request(struct intermediate_method_data * data) {

   struct search_data * search_data = &(data->search_data);
   struct tgt_point_search_data * tgt_point_data = search_data->tgt_point_data;
   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);
   unsigned num_pointsets = method->num_pointsets;
   method->data_request[0] = xmalloc(num_pointsets *
                                     sizeof(*(method->data_request[0])));
   struct interpolation_method_data_request * data_request =
      method->data_request[0];
   size_t num_interpolated_points = (size_t)(data->num_interpolated_points[0]);

   for (unsigned i = 0; i < num_pointsets; ++i) {
      data_request[i].num_values = 0;
      data_request[i].global_ids = NULL;
      data_request[i].ranks = NULL;
   }

   // select the data request that is actually required
   data_request += method->points_idx;

   struct data_request_point * data_request_points = NULL;
   size_t data_request_points_array_size = 0;
   size_t num_requested_points = 0;

   struct data_request_point * tmp_data_request_points =
      xmalloc(MAX(1024, 256 * method->n) * sizeof(*tmp_data_request_points));

   for (size_t i = 0; i < (num_interpolated_points + 255)/256; ++i) {

      size_t curr_num_tgt_points = MIN(256, num_interpolated_points - i * 256);
      size_t tmp_num_requested_points = 0;

      for (size_t j = 0; j < curr_num_tgt_points; ++j) {

         struct tgt_point_search_data * curr_tgt_point_data =
            tgt_point_data + j;
         size_t num_found_points =
            (size_t)(curr_tgt_point_data->num_found_points);

         for (size_t k = 0; k < num_found_points; ++k) {

            if (!curr_tgt_point_data->results[k].is_local) {

               tmp_data_request_points[tmp_num_requested_points].global_id =
                  curr_tgt_point_data->results[k].global_id;
               tmp_data_request_points[tmp_num_requested_points].rank =
                  curr_tgt_point_data->results[k].data.remote.rank;
               tmp_num_requested_points++;
            }
         }
      }
      tgt_point_data += 256;

      if (tmp_num_requested_points > 0) {
         ENSURE_ARRAY_SIZE(
            data_request_points, data_request_points_array_size,
            num_requested_points + tmp_num_requested_points);

         qsort(tmp_data_request_points, tmp_num_requested_points,
               sizeof(*tmp_data_request_points), compare_data_request_points);
         insertion_sort_data_request_point(
            data_request_points, &num_requested_points,
            &(tmp_data_request_points[0]), tmp_num_requested_points);
      }
   }

   free(tmp_data_request_points);

   data_request->num_values = (unsigned)num_requested_points;
   int * global_ids =
      xmalloc(num_requested_points * sizeof(*(data_request->global_ids)));
  int * ranks =
      xmalloc(num_requested_points * sizeof(*(data_request->ranks)));
   data_request->global_ids = global_ids;
   data_request->ranks = ranks;

   for (size_t i = 0; i < num_requested_points; ++i) {

      global_ids[i] = data_request_points[i].global_id;
      ranks[i] = (int)(data_request_points[i].rank);
   }

   free(data_request_points);
}

static void free_search_data(struct search_data * search_data) {

   for (size_t i = 0; i < search_data->num_tgt_points; ++i)
      free(search_data->tgt_point_data[i].results);

   free(search_data->tgt_point_data);
}

static void
free_intermediate_method_data(struct intermediate_method_data * data) {

   free_search_data(&(data->search_data));
   free(data);
}

/**
 * generates the interpolation plan based on the results \n
 * uses a point-type plan for the interpolation plan
 */
static struct interpolation_plan_element *
generate_interpolation_plan_1nn(struct intermediate_method_data * data) {

   if (*(data->num_interpolated_points) == 0) return NULL;

   unsigned points_idx =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->points_idx;
   unsigned num_interpolated_points = data->num_interpolated_points[0];
   struct tgt_point_search_data * tgt_results =
      data->search_data.tgt_point_data;
   struct interpolation_method_data_request * data_request =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->
         data_request[0] + points_idx;
   unsigned num_pointsets =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->
         num_pointsets;

   unsigned * src_indices =
      xmalloc(3  * num_interpolated_points * sizeof(*src_indices));

   for (unsigned i = 0; i < num_interpolated_points; ++i) {

      if (tgt_results[i].results[0].is_local) {

         src_indices[3*i+0] = tgt_results[i].results[0].data.local.subdomain_id+1;
         src_indices[3*i+1] = points_idx;
         src_indices[3*i+2] = tgt_results[i].results[0].data.local.local_id;

      } else {

         unsigned k;
         for (k = 0; k < data_request->num_values; ++k)
            if (data_request->global_ids[k] == tgt_results[i].results[0].global_id)
               break;

         src_indices[3*i+0] = 0;
         src_indices[3*i+1] = points_idx;
         src_indices[3*i+2] = k;
      }
   }

   unsigned * num_remote_points =
      xcalloc(num_pointsets, sizeof(*num_remote_points));
   num_remote_points[points_idx] = data_request->num_values;

   // generate a plan to execute the interpolation
   struct interpolation_plan_element * plan = xmalloc(1 * sizeof(*plan));

   yac_init_interpolation_plan_element_point(
      plan, num_interpolated_points, src_indices, num_remote_points,
      num_pointsets);

   free(num_remote_points);
   free(src_indices);

   return plan;
}

static void inverse(double * A, int n) {

// LAPACKE_dsytrf_work and LAPACKE_dsytri might not be available (see yac_lapack_interface.h).
#ifdef YAC_LAPACK_NO_DSYTR
  lapack_int ipiv[n+1];
  double work[n*n];

  for (size_t i = 0; i < n+1; ++i) ipiv[i] = 0;
  for (size_t i = 0; i < n*n; ++i)
    work[i] = 0;

  if (LAPACKE_dgetrf(
        LAPACK_COL_MAJOR, (lapack_int) n, (lapack_int) n,
        A, (lapack_int) n, ipiv) != 0 )
    yac_internal_abort_message("internal ERROR: dgetrf", __FILE__, __LINE__);

  if (LAPACKE_dgetri_work(
        LAPACK_COL_MAJOR, (lapack_int) n, A, (lapack_int) n,
        ipiv, work, (lapack_int) (n * n)) != 0 )
    yac_internal_abort_message("internal ERROR: dgetri", __FILE__, __LINE__);
#else
  lapack_int ipiv[n];
  double work[n];

  if (LAPACKE_dsytrf_work(
        LAPACK_COL_MAJOR, 'L', (lapack_int) n , A,
        (lapack_int) n, ipiv, work, (lapack_int) n) != 0)
    yac_internal_abort_message("internal ERROR: dsytrf", __FILE__, __LINE__);

  if (LAPACKE_dsytri_work(
        LAPACK_COL_MAJOR, 'L', (lapack_int) n , A,
        (lapack_int) n, ipiv, work) != 0)
    yac_internal_abort_message("internal ERROR: dsytri", __FILE__, __LINE__);

  for (size_t i = 0; i < n; ++i)
    for (size_t j = i + 1; j < n; ++j)
      A[j*n+i] = A[i*n+j];
#endif
}

/**
 * generates the interpolation plan based on the results \n
 * uses a wpointsum as for the interpolation plan
 */
static struct interpolation_plan_element *
generate_interpolation_plan_nnn(struct intermediate_method_data * data) {

   if (*(data->num_interpolated_points) == 0) return NULL;

   unsigned points_idx =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->points_idx;
   enum yac_nnn_reduction_type reduction_type =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->
         reduction_type;
   unsigned num_src_per_tgt =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->n;
   double rbf_scale =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->rbf_scale;
   double gauss_scale =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->gauss_scale;
   unsigned num_interpolated_points = data->num_interpolated_points[0];
   struct tgt_point_search_data * tgt_results =
      data->search_data.tgt_point_data;
   struct interpolation_method_data_request * data_request =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->
         data_request[0] + points_idx;
   unsigned num_pointsets =
      ((struct interpolation_method_nnn*)(data->method_stack[0]))->
         num_pointsets;

   unsigned * src_indices =
      xmalloc(
         3 * num_src_per_tgt * num_interpolated_points * sizeof(*src_indices));

   unsigned * curr_src_indices = src_indices;

   for (unsigned i = 0; i < num_interpolated_points; ++i) {

      for (unsigned j = 0; j < num_src_per_tgt; ++j) {

         if (tgt_results[i].results[j].is_local) {

            curr_src_indices[0] = tgt_results[i].results[j].data.local.subdomain_id+1;
            curr_src_indices[1] = points_idx;
            curr_src_indices[2] = tgt_results[i].results[j].data.local.local_id;

         } else {

            unsigned k;
            for (k = 0; k < data_request->num_values; ++k)
               if (data_request->global_ids[k] == tgt_results[i].results[j].global_id)
                  break;

            curr_src_indices[0] = 0;
            curr_src_indices[1] = points_idx;
            curr_src_indices[2] = k;
         }
         curr_src_indices += 3;
      }
   }

   double * weights =
      xmalloc(num_interpolated_points * num_src_per_tgt * sizeof(*weights));
   unsigned * num_src_per_tgt_array =
      xmalloc(num_interpolated_points * sizeof(*num_src_per_tgt_array));
   unsigned * num_remote_points =
      xcalloc(num_pointsets, sizeof(*num_remote_points));
   double * distances = xmalloc(num_src_per_tgt * sizeof(*distances));

   switch (reduction_type) {

      case (NNN_GAUSS_WEIGHTED):

         for (unsigned i = 0; i < num_interpolated_points; ++i) {

            unsigned j;

            for (j = 0; j < num_src_per_tgt; ++j) {
               distances[j] = acos(tgt_results[i].results[j].cos_angle);
               if (distances[j] <= yac_angle_tol) break;
               distances[j] *= distances[j];
            }

            if (j == num_src_per_tgt) {

               // if no source point directly overlaps with the target point

               // a) compute average distance between source points

               double sum_d = 0.0;

               for (int j = 0; j < num_src_per_tgt-1; ++j) {
                  for (int k = j + 1; k < num_src_per_tgt; ++k) {
                     double d =
                        get_vector_angle(tgt_results[i].results[j].coordinate_xyz,
                                         tgt_results[i].results[k].coordinate_xyz);
                     sum_d += d;
                  }
               }

               sum_d = sum_d * sum_d / (double)num_src_per_tgt;;

               // b) calculate weights from Gauss weighted inverse distances

               double sum_weights = 0.0;

               for (unsigned k = 0; k < num_src_per_tgt; ++k) {
                  weights[i*num_src_per_tgt+k] = exp ( -1.0 * distances[k] / ( gauss_scale * sum_d ) );
                  sum_weights += weights[i*num_src_per_tgt+k];
               }
               for (unsigned k = 0; k < num_src_per_tgt; ++k)
                  weights[i*num_src_per_tgt+k] = weights[i*num_src_per_tgt+k] / sum_weights;

            } else {
              // special case: source and target at same location
              for ( unsigned k = 0; k < num_src_per_tgt; ++k )
                 weights[i*num_src_per_tgt+k] = 0.0;
              weights[i*num_src_per_tgt+j] = 1.0;
            }

         }
         break;

      case (NNN_DISTANCE_WEIGHTED): 
         for (unsigned i = 0; i < num_interpolated_points; ++i) {

            double distance_sum = 0.0;
            unsigned j;

            for (j = 0; j < num_src_per_tgt; ++j) {
               distances[j] = acos(tgt_results[i].results[j].cos_angle);
               if (distances[j] <= yac_angle_tol) break;
               distances[j] *= distances[j];
               distance_sum += 1.0 / distances[j];
            }

            // if no source point directly overlaps with the target point
            if (j == num_src_per_tgt) {

              distance_sum = 1.0 / distance_sum;
              for (unsigned k = 0; k < num_src_per_tgt; ++k)
                 weights[i*num_src_per_tgt+k] = 1.0 / distances[k] * distance_sum;

            } else {

              // special case: source and target at same location
              for ( unsigned k = 0; k < num_src_per_tgt; ++k )
                 weights[i*num_src_per_tgt+k] = 0.0;
              weights[i*num_src_per_tgt+j] = 1.0;
            }
         }
         break;

      case (NNN_AVERAGE):

         // arithmetic average
         for (unsigned i = 0; i < num_interpolated_points * num_src_per_tgt; ++i)
            weights[i] = 1.0 / (double)num_src_per_tgt;
         break;

      case (NNN_RBF): {

         double * A =
           xmalloc(num_src_per_tgt * num_src_per_tgt * sizeof(*A));
         double * distance_matrix = A;
         double * A_inv = A;
         double * a = xmalloc(num_src_per_tgt * sizeof(*a));

         for (unsigned i = 0; i < num_interpolated_points; ++i) {

           double sum_d = 0.0, scale_d = 1.0;

           // compute distance matrix for all found source points
           for (int j = 0; j < num_src_per_tgt; ++j)
             distance_matrix[j + j * num_src_per_tgt] = 0.0;
           for (int j = 0; j < num_src_per_tgt-1; ++j) {
             for (int k = j + 1; k < num_src_per_tgt; ++k) {
               double d =
                 get_vector_angle(tgt_results[i].results[j].coordinate_xyz,
                                  tgt_results[i].results[k].coordinate_xyz);
               distance_matrix[k + j * num_src_per_tgt] = d;
               distance_matrix[j + k * num_src_per_tgt] = d;
               sum_d += d;
             }
           }

           // compute and apply scale factor for distance matrix
           if (sum_d > 0.0)
             scale_d =
               ((double)((num_src_per_tgt - 1) * num_src_per_tgt)) / (2.0 * sum_d);
           scale_d /= rbf_scale;

           double sq_scale_d = scale_d * scale_d;

           // compute matrix A[num_src_per_tgt][num_src_per_tgt]
           // with A = rbf(distance_matrix)
           //      rbf(a) => Radial basis function
           for (int j = 0; j < num_src_per_tgt * num_src_per_tgt; ++j) {
             double d = distance_matrix[j];
             // gauß rbf kernel
             A[j] = exp(-1.0 * d * d * sq_scale_d);
           }

           // compute inverse of A
           inverse(A_inv, num_src_per_tgt);

           // compute a[NUM_NEIGH]
           // with a_i = rbf(d(x_i, y))
           //      x => vector containing the coordinates of the
           //           num_src_per_tgt nearest neighbours of the current
           //           target point
           //      y => coordinates of target point
           //      d(a, b) => great circle distance between point a and b
           //      rbf(a) => Radial basis function
           for (int j = 0; j < num_src_per_tgt; ++j) {
             double d = acos(tgt_results[i].results[j].cos_angle);
             // gauß rbf kernel
             a[j] = exp(-1.0 * d * d * sq_scale_d);
           }

           // compute weights
           // with w_i = SUM(A_inv[i][j]*a[j]) // j = 0..num_src_per_tgt-1
           for (int j = 0; j < num_src_per_tgt; ++j) {
             weights[i*num_src_per_tgt+j] = 0.0;
             for (int k = 0; k < num_src_per_tgt; ++k)
               weights[i*num_src_per_tgt+j] +=
                 A_inv[j*num_src_per_tgt+k] * a[k];
           }
        }

        free(a);
        free(A);
        break;
      }
      default:
        yac_internal_abort_message(
           "ERROR(generate_interpolation_plan_nnn): invalid reduction_type",
           __FILE__, __LINE__);
   }

   free(distances);

   for (unsigned i = 0; i < num_interpolated_points; ++i)
      num_src_per_tgt_array[i] = num_src_per_tgt;
   num_remote_points[points_idx] = data_request->num_values;

   // generate a plan to execute the interpolation
   struct interpolation_plan_element * plan = xmalloc(1 * sizeof(*plan));

   yac_init_interpolation_plan_element_wpointsum(
      plan, num_interpolated_points, src_indices, num_src_per_tgt_array,
      weights, num_remote_points, num_pointsets);

   free(num_remote_points);
   free(num_src_per_tgt_array);
   free(src_indices);
   free(weights);

   return plan;
}

static void
retrieve_tgt_point_data(struct tgt_point_search_data * tgt_point_data,
                        size_t num_tgt_points, double (*coordinates_xyz)[3],
                        int * global_ids) {

  for (size_t i = 0; i < num_tgt_points; ++i) {
    global_ids[i] = tgt_point_data[i].global_id;
    coordinates_xyz[i][0] = tgt_point_data[i].coordinate_xyz[0];
    coordinates_xyz[i][1] = tgt_point_data[i].coordinate_xyz[1];
    coordinates_xyz[i][2] = tgt_point_data[i].coordinate_xyz[2];
  }
}

/**
 * Marks points from the list of target points (even if they could
 * be interpolated by the local process). This is done because they are going
 * to be interpolated by another source process, which might later
 * request some data from the local process to do the interpolation.
 * @param[in] num_points_to_be_marked number of points that are interpolated
 *                                    by another source process
 * @param[in] global_ids_to_be_marked array with global ids of target points
 *                                    that are to be marked from the local list
 * @param[in,out] data                basic search data
 *                                    (see \ref intermediate_method_data)
 */
static void
mark_tgt_points(size_t num_points_to_be_marked, int * global_ids_to_be_marked,
                struct intermediate_method_data * data) {

  if (num_points_to_be_marked == 0) return;

  // first sort all global ids that need to be marked
  yac_quicksort_index(global_ids_to_be_marked, num_points_to_be_marked, NULL);

  // for all target points that need to be marked
  for (size_t i = 0, j = 0; i < num_points_to_be_marked; ++i) {

    // find the matching entry in the local target points
    while ((j < data->search_data.num_tgt_points) &&
           (data->search_data.tgt_point_data[j].global_id <
            global_ids_to_be_marked[i])) ++j;

    // if no matching entry could be found
    if ((j == data->search_data.num_tgt_points) ||
        (data->search_data.tgt_point_data[j].global_id !=
        global_ids_to_be_marked[i]))
      yac_internal_abort_message(
        "ERROR(mark_tgt_points): invalid global target id\n",
        __FILE__, __LINE__);

    // mark the target point
    data->search_data.tgt_point_data[j].flag = REMOTE;
  }
}

static inline int compare_tgt_point_data(const void * a,const void * b) {

   return ((struct tgt_point_search_data*)a)->global_id -
          ((struct tgt_point_search_data*)b)->global_id;
}

static inline int compare_tgt_point_data_flag(const void * a,const void * b) {

   if (((struct tgt_point_search_data*)a)->flag !=
       ((struct tgt_point_search_data*)b)->flag)
      return ((struct tgt_point_search_data*)a)->flag -
             ((struct tgt_point_search_data*)b)->flag;
   else
      return (((struct tgt_point_search_data*)a)->global_id -
              ((struct tgt_point_search_data*)b)->global_id);
}

/**
 * this routine receive all global ids
 * once the search is finished, this routine generates the data_request object for
 * the field_interpolation and the interpolation_plan \n
 * afterwards it informs all other source processes that the nnn search on this
 * process is finished
 * finally it call the next interpolation method on the interpolation method
 * stack for the remaining target points that could not be interpolated
 */
static void
finish_nnn_interpolation_cont(void * user_data, int * global_ids_to_mark,
                              unsigned num_global_ids) {

   struct intermediate_method_data * data =
      (struct intermediate_method_data *)user_data;
   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);

   //--------------------------------------------------------------
   // reset all REMOTE flags to NOT
   //--------------------------------------------------------------
   for (size_t i = 0; i < data->search_data.num_tgt_points; ++i)
      if (data->search_data.tgt_point_data[i].flag == REMOTE)
         data->search_data.tgt_point_data[i].flag = NOT;

   //--------------------------------------------------------------
   // mark all target points that are interpolated by remote
   // source processes
   //--------------------------------------------------------------
   mark_tgt_points((size_t)num_global_ids, global_ids_to_mark, data);

   //--------------------------------------------------------------
   // sort tgt_point_data by the global ids and put the ones
   // flagged as LOCAL in front
   //--------------------------------------------------------------
   qsort(data->search_data.tgt_point_data, data->search_data.num_tgt_points,
         sizeof(*data->search_data.tgt_point_data),
         compare_tgt_point_data_flag);

   // count the number of interpolated points
   unsigned num_interpolated_points = 0;
   unsigned num_not_interpolated_points = 0;
   for (size_t i = 0; i < data->search_data.num_tgt_points; ++i) {
      if (data->search_data.tgt_point_data[i].flag == LOCAL)
         num_interpolated_points++;
      else if (data->search_data.tgt_point_data[i].flag == NOT)
         num_not_interpolated_points++;
   }
   data->search_data.num_tgt_points_to_interpolate = num_interpolated_points;
   *(data->num_interpolated_points) = num_interpolated_points;

   // detect local result points that are marked as remote ones
   detect_local_points(data);

   // collect all remote points required for the interpolation
   generate_data_request(data);

   // generate final results
   if (method->n == 1)
      data->plan_stack[0] = generate_interpolation_plan_1nn(data);
   else
      data->plan_stack[0] = generate_interpolation_plan_nnn(data);
   if (data->plan_stack[0] != NULL) data->plan_stack++;

   // retrieve coordinates_xyz and global_ids from tgt_point_data
   retrieve_tgt_point_data(data->search_data.tgt_point_data,
                           num_interpolated_points +
                           num_not_interpolated_points,
                           data->coordinates_xyz,
                           data->global_ids);

   struct interpolation_method ** method_stack     = data->method_stack+1;
   struct interpolation_plan_element ** plan_stack = data->plan_stack;
   unsigned * p_num_interpolated_points            = data->num_interpolated_points+1;
   double (*coordinates_xyz)[3]                    = data->coordinates_xyz;
   int * global_ids                                = data->global_ids+*(data->num_interpolated_points);
   enum yac_location tgt_location                  = data->tgt_location;

   coordinates_xyz += *(data->num_interpolated_points);

   free_intermediate_method_data(data);

   // call next interpolation method on stack
   if (*method_stack != NULL)
      (*method_stack)->vtable->do_search(
         method_stack, plan_stack, global_ids, coordinates_xyz, tgt_location,
         num_not_interpolated_points, p_num_interpolated_points,
         method->data_request+1);

   if (method->finish_callback != NULL)
      method->finish_callback(method->finish_user_data);
}

/**
 * once the search is finished, this routine exchanges the points that can be
 * interpolated by the local process with the other source processes. This
 * allows then to remove these points from their target point lists, which are
 * then passed to the next interpolation method in the stack.
 */
static void
finish_nnn_interpolation(struct intermediate_method_data * data){

   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);
   struct communicator * comm = yac_get_domain_comm(method->domain);

   // search for target points for which the number of found source
   // points is insufficient
   for (size_t i = 0; i < data->search_data.num_tgt_points_to_interpolate;
        ++i)
      if (data->search_data.tgt_point_data[i].num_found_points < method->n)
         data->search_data.tgt_point_data[i].flag = NOT;

   // get all points that can be interpolated
   unsigned num_interpolated_points = 0;
   for (size_t i = 0; i < data->search_data.num_tgt_points_to_interpolate;
        ++i)
      if (data->search_data.tgt_point_data[i].flag == LOCAL)
         data->global_ids[num_interpolated_points++] =
            data->search_data.tgt_point_data[i].global_id;
   // get the remaining points
   for (unsigned i = 0, j = num_interpolated_points;
        i < data->search_data.num_tgt_points; ++i)
      if (data->search_data.tgt_point_data[i].flag != LOCAL)
         data->global_ids[j++] =
            data->search_data.tgt_point_data[i].global_id;

   // sort the target point data by global ids
   qsort(data->search_data.tgt_point_data, data->search_data.num_tgt_points,
         sizeof(*data->search_data.tgt_point_data), compare_tgt_point_data);

   // exchange final results with other source processes
   yac_exchange_results(data->global_ids, num_interpolated_points,
                        data->search_data.num_tgt_points,
                        method->tags.exchange_results,
                        comm, finish_nnn_interpolation_cont, data);
}

//! constructor for a nearest neighbour interpolation method
struct interpolation_method *
yac_new_interpolation_method_nnn_mask(struct global_search * search,
                                      struct domain * domain,
                                      char const * stack_name,
                                      struct points *** points,
                                      struct mask *** masks,
                                      unsigned num_pointsets,
                                      enum yac_nnn_reduction_type reduction_type,
                                      unsigned n, double rbf_scale, double gauss_scale,
                                      unsigned points_idx,
                                      unsigned use_source_mask) {

   // ensure that the right point data is available
   if (num_pointsets == 0)
      yac_internal_abort_message ("ERROR(new_interpolation_method_nnn): missing "
                                  "data points.", __FILE__, __LINE__);

   if (points_idx >= num_pointsets)
      yac_internal_abort_message ("ERROR(new_interpolation_method_nnn): invalid "
                                  "points_idx.", __FILE__, __LINE__);

   if (points == NULL || points[0] == NULL || points[0][points_idx] == NULL)
      yac_internal_abort_message ("ERROR(new_interpolation_method_nnn): no point set "
                                  "data available", __FILE__, __LINE__);

   struct interpolation_method_nnn * method = xmalloc(1 * sizeof(*method));

   method->vtable = &interpolation_method_nnn_vtable;
   method->search = search;
   method->domain = domain;
   method->points = points;
   method->masks = (use_source_mask)?masks:NULL;
   method->num_pointsets = num_pointsets;
   method->finish_callback = NULL;
   method->finish_user_data = NULL;
   method->reduction_type = reduction_type;
   method->n = n;
   method->rbf_scale = rbf_scale;
   method->gauss_scale = gauss_scale;
   method->points_idx = points_idx;
   method->use_source_mask = use_source_mask;

   yac_init_interpolation_method((struct interpolation_method *)method, search);

   generate_interpolation_method_nnn_tag(method, stack_name);

   char * unique_nnn_string;
   {
      char const * str_base_str = "interpolation method nnn";
      char const * str_nnn_search = "nnn search";
      char const * source_comp_name = yac_get_local_component_name(search);
      char const * target_comp_name = yac_get_remote_component_name(search);
      char const * source_domain_name = yac_get_local_domain_name(search);
      size_t str_buf_size = strlen(str_nnn_search);
      char const * target_domain_name = yac_get_remote_domain_name(search);
      str_buf_size += strlen(str_base_str);
      str_buf_size += strlen(source_comp_name);
      str_buf_size += strlen(target_comp_name);
      str_buf_size += strlen(source_domain_name);
      str_buf_size += strlen(target_domain_name);
      str_buf_size += strlen(stack_name) + 10 + 10 + 10 + 2 + 16 + 16 + 10 + 1;
         // (+ 10) for points_idx
         // (+ 10) for reduction_type
         // (+ 10) for n
         // (+ 2)  for mask handling
         // (+ 16) for rbf scale
         // (+ 16) for gauss scale
         // (+ 10) for unique_method_id
         // (+ 1)  for '\0'

      unique_nnn_string = xmalloc(str_buf_size);
      generate_tag_string(
         unique_nnn_string, str_buf_size, str_base_str, source_comp_name,
         target_comp_name, source_domain_name, target_domain_name, stack_name,
         points_idx, reduction_type, n, use_source_mask, rbf_scale, gauss_scale,
         yac_get_unique_id_interpolation_method(
            (struct interpolation_method *)method), str_nnn_search);
   }
   method->nnn_search =
    yac_nnn_search_new(
      domain, unique_nnn_string, points, masks, num_pointsets, points_idx);
   free(unique_nnn_string);

   return (struct interpolation_method *)method;
}

//! constructor for a nearest neighbour interpolation method
struct interpolation_method *
yac_new_interpolation_method_nnn(struct global_search * search,
                                 struct domain * domain,
                                 char const * stack_name,
                                 struct points *** points,
                                 unsigned num_pointsets,
                                 enum yac_nnn_reduction_type reduction_type,
                                 unsigned n, double rbf_scale, double gauss_scale,
                                 unsigned points_idx) {

   unsigned use_source_mask = 0;

   return (struct interpolation_method *)yac_new_interpolation_method_nnn_mask(
      search, domain, stack_name, points, NULL, num_pointsets, reduction_type,
      n, rbf_scale, gauss_scale, points_idx, use_source_mask);
}

static void finish_nnn_callback(
  void * user_data, struct request_point * tgt_points) {

  struct intermediate_method_data * data =
     (struct intermediate_method_data *)user_data;

  size_t num_points = data->search_data.num_tgt_points_to_interpolate;
  for (size_t i = 0; i < num_points; ++i) {

    struct request_point * curr_tgt_point = tgt_points + i;
    struct tgt_point_search_data * curr_tgt_point_data =
      data->search_data.tgt_point_data + i;

    curr_tgt_point_data->min_cos_angle = curr_tgt_point->cos_angle;
    curr_tgt_point_data->num_found_points = curr_tgt_point->num_found_points;

    curr_tgt_point_data->results = curr_tgt_point->results;
    curr_tgt_point->results = NULL;
  }

  yac_nnn_search_free_request_points(tgt_points, num_points);
  free(tgt_points);
  finish_nnn_interpolation(data);
}

/**
 * callback routine passed to exchange_results_init by
 * do_search_continue_nnn
 */
static void
do_search_continue_nnn_2(void * user_data,
                         int * global_ids_to_mark,
                         unsigned num_global_ids) {

   struct intermediate_method_data * data =
      (struct intermediate_method_data *)user_data;
   struct interpolation_method_nnn * method_nnn =
      (struct interpolation_method_nnn*)(data->method_stack[0]);

   //-------------------------------------------------------------------
   // mark target points that are interpolated by another source process
   //-------------------------------------------------------------------
   mark_tgt_points((size_t)num_global_ids, global_ids_to_mark, data);

   //------------------------------------------------------------------------
   // all target points not marked as LOCAL or REMOTE have to be interpolated
   // by the local process
   //------------------------------------------------------------------------
   for (size_t i = 0; i < data->search_data.num_tgt_points; ++i)
     if (data->search_data.tgt_point_data[i].flag == NOT)
       data->search_data.tgt_point_data[i].flag = LOCAL;

   //--------------------------------------------------------------
   // sort tgt_point_data by the global ids and put the ones
   // flagged as LOCAL in front
   //--------------------------------------------------------------
   qsort(data->search_data.tgt_point_data, data->search_data.num_tgt_points,
         sizeof(*data->search_data.tgt_point_data),
         compare_tgt_point_data_flag);

   //-------------------------------------------------------------
   // count the number of target point that are to be interpolated
   // by the local process
   //-------------------------------------------------------------
   size_t num_tgt_points_to_interpolate = 0;
   for (size_t i = 0; i < data->search_data.num_tgt_points; ++i)
      if (data->search_data.tgt_point_data[i].flag == LOCAL)
         ++num_tgt_points_to_interpolate;
   data->search_data.num_tgt_points_to_interpolate =
      num_tgt_points_to_interpolate;

   double (*tgt_coordinates_xyz)[3] =
      xmalloc(num_tgt_points_to_interpolate * sizeof(*tgt_coordinates_xyz));
   int * global_tgt_ids = data->global_ids;
   struct tgt_point_search_data * tgt_point_data =
      data->search_data.tgt_point_data;
   for (size_t i = 0; i < num_tgt_points_to_interpolate; ++i) {
      tgt_coordinates_xyz[i][0] = tgt_point_data[i].coordinate_xyz[0];
      tgt_coordinates_xyz[i][1] = tgt_point_data[i].coordinate_xyz[1];
      tgt_coordinates_xyz[i][2] = tgt_point_data[i].coordinate_xyz[2];
      global_tgt_ids[i] = data->search_data.tgt_point_data[i].global_id;
   }

   //-------------------------------------------------------------
   // get the n nearest neighbour for target point
   //-------------------------------------------------------------
   yac_nnn_search_do_search(
      method_nnn->nnn_search, method_nnn->n, num_tgt_points_to_interpolate,
      tgt_coordinates_xyz, global_tgt_ids, data, finish_nnn_callback);

   free(tgt_coordinates_xyz);
}

/**
 * callback routine passed to yac_exchange_results by
 * do_search_interpolation_method_nnn
 */
static void
do_search_continue_nnn(void * user_data,
                       int * global_ids_to_mark,
                       unsigned num_global_ids) {

   struct intermediate_method_data * data =
      (struct intermediate_method_data *)user_data;
   struct interpolation_method_nnn * method_nnn =
      (struct interpolation_method_nnn*)(data->method_stack[0]);

   //-------------------------------------------------------------------
   // mark target points that are interpolated by another source process
   //-------------------------------------------------------------------
   mark_tgt_points((size_t)num_global_ids, global_ids_to_mark, data);

   //----------------------------------------------------------------
   // distribute the target points that did not match any source cell
   //----------------------------------------------------------------
   unsigned num_uninterpolated_points = 0;
    for (size_t i = 0; i < data->search_data.num_tgt_points; ++i)
      if (data->search_data.tgt_point_data[i].flag == NOT)
        data->global_ids[num_uninterpolated_points++] =
          data->search_data.tgt_point_data[i].global_id;

   yac_exchange_results(data->global_ids, num_uninterpolated_points,
                        num_uninterpolated_points,
                        method_nnn->tags.global_id_exchange_2,
                        yac_get_domain_comm(method_nnn->domain),
                        do_search_continue_nnn_2, data);
}

static unsigned
contains_local_cell(struct domain * domain, unsigned subdomain_id,
                    unsigned const * local_cell_ids, size_t num_cells) {

  for (size_t i = 0; i < num_cells; ++i)
    if (!yac_is_halo_value(domain, subdomain_id, CELL, local_cell_ids[i], NULL))
      return 1;
  return 0;
}

static void
do_search_interpolation_method_nnn(struct interpolation_method ** method,
                                   struct interpolation_plan_element ** plan_stack,
                                   int * global_ids, double (*coordinates_xyz)[3],
                                   enum yac_location tgt_location,
                                   unsigned num_points,
                                   unsigned * num_interpolated_points,
                                   struct interpolation_method_data_request ** data_request) {

  if ( ( coordinates_xyz == NULL ) && num_points > 0 )
    yac_internal_abort_message("ERROR: missing target point coordinates "
                  "(coordinates_xyz == NULL)", __FILE__, __LINE__);

  struct interpolation_method_nnn * method_nnn;
  method_nnn = (struct interpolation_method_nnn*)(method[0]);

  method_nnn->data_request = data_request;

  struct intermediate_method_data * data;
  struct search_data * search_data;

  struct domain * domain = method_nnn->domain;

  struct communicator * domain_comm = yac_get_domain_comm(domain);

  // initialise data that is required for the nearest neighbourhood search

  data = xmalloc(1 * sizeof(*data));
  data->method_stack = method;
  data->plan_stack = plan_stack;
  data->num_interpolated_points = num_interpolated_points;
  *num_interpolated_points = 0;
  data->coordinates_xyz = coordinates_xyz;
  data->global_ids = global_ids;
  data->tgt_location = tgt_location;
  search_data = &(data->search_data);
  search_data->tgt_point_data =
    tgt_point_search_data_new(
      num_points, coordinates_xyz, global_ids);
  search_data->num_tgt_points = (size_t)num_points;
  search_data->num_tgt_points_to_interpolate = (size_t)num_points;

  unsigned num_subdomains = yac_get_num_subdomains(domain);

  // do an initial search that tries to find for each target point a matching
  // local cell
  for (unsigned subdomain_id = 0; subdomain_id < num_subdomains;
       ++subdomain_id) {

    struct dep_list * tgt_to_src_cell = yac_do_point_search_c2(
      yac_get_subdomain_grid_search(method_nnn->search, subdomain_id),
      coordinates_xyz, (size_t)num_points);

    for (size_t i = 0; i < num_points; ++i) {
      unsigned const * curr_src_cells =
        yac_get_dependencies_of_element(tgt_to_src_cell, i);
      size_t curr_num_src_cells =
        yac_get_num_dependencies_of_element(tgt_to_src_cell, i);
      if (contains_local_cell(
            domain, subdomain_id, curr_src_cells, curr_num_src_cells))
       search_data->tgt_point_data[i].flag = LOCAL;
    }

    yac_dep_list_delete(tgt_to_src_cell);

    // give the communicator layer the chance to process messages
    yac_comm_check(domain_comm);
  }

  // sort the target point data by global ids
  qsort(data->search_data.tgt_point_data, data->search_data.num_tgt_points,
        sizeof(*data->search_data.tgt_point_data), compare_tgt_point_data);

  // count the number of target points for which at least one matching local
  // source cell was found
  unsigned num_matching_tgt_points = 0;
  for (unsigned i = 0; i < num_points; ++i)
    if (search_data->tgt_point_data[i].flag == LOCAL)
      data->global_ids[num_matching_tgt_points++] =
        search_data->tgt_point_data[i].global_id;
  for (unsigned i = 0, j = num_matching_tgt_points; i < num_points; ++i)
    if (search_data->tgt_point_data[i].flag != LOCAL)
      data->global_ids[j++] =
        search_data->tgt_point_data[i].global_id;

  // exchange the cells results of the initial search with the other source
  // processes and determine points for which multiple source processes have
  // found matching cells once this exchange is finished the algorithms
  // continues in the routine do_search_continue_nnn
  yac_exchange_results(data->global_ids, num_matching_tgt_points, num_points,
                       method_nnn->tags.global_id_exchange_1,
                       domain_comm, do_search_continue_nnn, data);
}

/**
 * blocking routine that ensures that the search is finished and everything
 * is cleaned up
 */
static void
enddef_interpolation_method_nnn(struct interpolation_method ** method) {

   struct interpolation_method_nnn * method_nnn =
      (struct interpolation_method_nnn*)(method[0]);

   struct communicator * domain_comm = yac_get_domain_comm(method_nnn->domain);

   while(yac_comm_wait(domain_comm, method_nnn->tags.global_id_exchange_1[0]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.global_id_exchange_1[1]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.global_id_exchange_2[0]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.global_id_exchange_2[1]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.exchange_results[0]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.exchange_results[1]) ||
         yac_nnn_search_wait(method_nnn->nnn_search));

   if (method[1] != NULL) method[1]->vtable->enddef(method + 1);
}

static void get_source_exchange_data_interpolation_method_nnn (
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request) {

   struct interpolation_method_nnn * method_nnn =
      (struct interpolation_method_nnn*)(method[0]);

   *data_request = method_nnn->data_request;
   method_nnn->data_request = NULL;
}

static void
delete_interpolation_method_nnn(struct interpolation_method * method) {

   struct interpolation_method_nnn * method_nnn =
      (struct interpolation_method_nnn*)method;

   yac_nnn_search_delete(method_nnn->nnn_search);
   free(method);
}

static struct global_search *
get_global_search(struct interpolation_method * method) {

   return ((struct interpolation_method_nnn*)(method))->search;
}

static void
set_finish_callback(struct interpolation_method * method,
                     void (*callback)(void * user_data),
                     void * user_data) {

   struct interpolation_method_nnn * method_nnn =
      (struct interpolation_method_nnn*)method;

   method_nnn->finish_callback = callback;
   method_nnn->finish_user_data = user_data;
}
