/**
 * @file interpolation_method_patch_smooth.h
 * @brief patch recovery interpolation method
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INTERPOLATION_METHOD_PATCH_SMOOTH_H
#define INTERPOLATION_METHOD_PATCH_SMOOTH_H

/** \example test_interpolation_method_patch_smooth.c
 * This shows how the patch recovery interpolation works
 */

/** \example test_interpolation_method_patch_smooth_parallel.c
 * This contains a parallel example for the patch recovery interpolation
 * method. The example uses four source and four target processes.
 */

#include "interpolation_method.h"

/**
 * @file interpolation_method_patch_smooth.h
 * @brief patch recovery interpolation method
 *
 * cell interpolation method that searches for each target cell the overlapping
 * source cells which make up the patch. A polynomial fit is applied over the
 * which then is used to provide function values at any point within this patch.
 *
 */


/**
 * constructor for a interpolation method of type interpolation_method_patch_smooth
 * @param[in] search              global search object that is to be used for the interpolation
 * @param[in] domain              %domain of the local process
 * @param[in] stack_name          name of the interpolation stack
 * @param[in] points              points to be used for the weight generation
 * @param[in] masks               mask to be used for the weight generation
 * @param[in] num_pointsets       number of point sets
 * @param[in] polynom_order       order of the polynomial to be fitted
 * @param[in] gauss_order         gauss_order, is transferred into the number of gauss points per cell
 * @param[in] corner_weight       weight of corner values compared to the target point value
 * @param[in] extent_source_patch_smooth allow extension of the overlapping patch by one row of neighbouring cells
 * @param[in] allow_extrapolation allow interpolation for target points that are outside the patch
 * @param[in] points_idx          which points set of the domain is to be used for the interpolation
 *
 */
struct interpolation_method *
yac_new_interpolation_method_patch_smooth ( struct global_search * search,
                                     struct domain * domain,
                                     char const * stack_name,
                                     struct points *** points,
                                     struct mask *** masks,
                                     unsigned num_pointsets,
                                     unsigned polynom_order,
                                     unsigned gauss_order,
                                     double corner_weight,
                                     unsigned extent_source_patch_smooth,
                                     unsigned allow_extrapolation,
                                     unsigned points_idx );

/**
 * constructor for a interpolation method of type interpolation_method_patch_smooth
 * @param[in] search              global search object that is to be used for the interpolation
 * @param[in] domain              %domain of the local process
 * @param[in] stack_name          name of the interpolation stack
 * @param[in] points              points to be used for the weight generation
 * @param[in] masks               mask to be used for the weight generation
 * @param[in] num_pointsets       number of point sets
 * @param[in] polynom_order       order of the polynomial to be fitted
 * @param[in] corner_weight       weight of corner values compared to the target point value
 * @param[in] gauss_order         gauss_order, is transferred into the number of gauss points per cell
 * @param[in] extent_source_patch_smooth allow extension of the overlapping patch by one row of neighbouring cells
 * @param[in] allow_extrapolation allow interpolation for target points that are outside the patch
 * @param[in] points_idx          which points set of the domain is to be used for the interpolation
 * @param[in] use_source_mask     specifies whether masked source cells shall be considered (true) or not (false)
 *
 */
struct interpolation_method *
yac_new_interpolation_method_patch_smooth_mask ( struct global_search * search,
                                          struct domain * domain,
                                          char const * stack_name,
                                          struct points *** points,
                                          struct mask *** masks,
                                          unsigned num_pointsets,
                                          unsigned polynom_order,
                                          unsigned gauss_order,
                                          double corner_weight,
                                          unsigned extent_source_patch_smooth,
                                          unsigned allow_extrapolation,
                                          unsigned points_idx,
                                          unsigned use_source_mask );

#endif // INTERPOLATION_METHOD_PATCH_SMOOTH_H

