/**
 * @file interpolation_method_spmap.c
 * @brief implementation source point mapping interpolation method
 *
 * @copyright Copyright  (C)  2016 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <string.h>
#include <math.h>
#include <sys/time.h>

#include "interpolation_method.h"
#include "interpolation_method_spmap.h"
#include "ensure_array_size.h"
#include "geometry.h"
#include "grid.h"
#include "sphere_part.h"
#include "utils.h"

static void
do_search_interpolation_method_spmap(struct interpolation_method ** method,
                                     struct interpolation_plan_element ** plan_stack,
                                     int * global_ids, double (*coordinates_xyz)[3],
                                     enum yac_location tgt_location,
                                     unsigned num_points,
                                     unsigned * num_interpolated_points,
                                     struct interpolation_method_data_request ** data_request);

static void
enddef_interpolation_method_spmap(struct interpolation_method ** method);

static void
get_source_exchange_data_interpolation_method_spmap(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request);

static void
delete_interpolation_method_spmap(struct interpolation_method * method);

static struct global_search *
get_global_search(struct interpolation_method * method);

static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data);

static struct interpolation_method_vtable
  interpolation_method_spmap_vtable = {
    .do_search                = do_search_interpolation_method_spmap,
    .enddef                   = enddef_interpolation_method_spmap,
    .get_source_exchange_data = get_source_exchange_data_interpolation_method_spmap,
    .delete                   = delete_interpolation_method_spmap,
    .get_global_search        = get_global_search,
    .set_finish_callback      = set_finish_callback
};

struct interpolation_method_spmap_tags {
  unsigned request_header;
  unsigned request_data;
  unsigned request_answer_header;
  unsigned request_answer_data;
  unsigned continue_barrier;
  unsigned create_dist_dir;
  unsigned distribute_results;
};

struct interpolation_method_spmap {
  struct interpolation_method_vtable * vtable;
  void (*finish_callback)(void * user_data); //!< callback routine that is called once the interpolation is finished
  void * finish_user_data;                   //!< user data that is passed to finish_callback
  struct global_search * search;             //!< global search object used a the basis for the interpolation
  struct domain * domain;                    //!< domain of the local process
  struct points *** points;                  //!< source points to be used for the weight generation
  struct mask *** masks;                     //!< mask to be used for the weight generation (can be NULL -> no mask)
  unsigned num_pointsets;                    //!< number of point sets
  unsigned points_idx;                       //!< index of the points set of field that is to be used for the interpolation
  struct interpolation_method_data_request ** data_request; //!< contains points required from other
  struct interpolation_method_spmap_tags tags; //!< unique tags to handle the communication for spmapping
  enum yac_location src_location;
};

/**
 * contains all temporary search specific data for one source point
 */
struct src_point_search_data {

  int global_id;                 //!< global id of the source point
  unsigned local_id;             //!< local id of the source point
  unsigned subdomain_idx;        //!< subdomain index
  struct bounding_circle circle; //!< current bounding circle used to identify
                                 //!< potential target points (the centre of the
                                 //!< bounding circle is the source point)

  struct {
    int global_id; //!< global id of the target point
    unsigned rank; //!< rank of the source process that has the target point
  } tgt_point; //!< the target point found for the source point

  unsigned * checked_ranks; //!< ranks of all source processes that have been
                            //!< checked
  unsigned num_checked_ranks;

  unsigned * to_be_checked_ranks; //!< ranks of all source processes that need
                                  //!< to be checked
  unsigned to_be_checked_ranks_array_size;
  unsigned num_to_be_checked_ranks;
};

struct search_data {

  struct src_point_search_data * src_point_data;
  unsigned num_src_points;
  double (*coordinates_xyz)[3];
};

struct tgt_point_data {
  double (*coordinates_xyz)[3];
  int * global_ids;
  unsigned num_tgt_points;
  struct point_sphere_part_search * point_sphere_part;
};

struct halo_cells {
  unsigned rank;
  struct bounding_circle bnd_circle;
};

//! collection of data that is required through out the search
struct intermediate_method_data {

  struct search_data search_data; //!< search data for the local source points
  unsigned num_open_remote_requests; //!< number for search requests that have
                                     //!< not yet been answered
  struct interpolation_method ** method_stack;
  struct interpolation_plan_element ** plan_stack;
  unsigned * num_interpolated_points;
  struct tgt_point_data tgt_points;
  enum yac_location tgt_location;
  struct interpolation_method_spmap_tags tags;
};

struct search_request_user_data {

  struct tgt_point_data tgt_points;
  struct halo_cells * halo_cells;
  unsigned num_halo_cells;
  struct interpolation_method_spmap_tags tags;
};

struct dist_dir_data {

  int tgt_global_id;
  unsigned master_rank;
  int * src_global_ids;
  unsigned * src_owner;
  unsigned num_src;
};

struct search_results {

  int tgt_global_id;
  unsigned num_src; // -1 -> interpolated by other source process
                    // 0  -> not interpolated
                    // n  -> n source points used for the interpolation
  int * src_global_ids;
  unsigned * owner;
};

// Generation of tags
// -----------------------------------------------------------------------------

static void
generate_tag_string(char * buf,  size_t buffer_size,
                    char const * tag_base_str,
                    char const * source_comp_name,
                    char const * target_comp_name,
                    char const * source_domain_name,
                    char const * target_domain_name,
                    char const * stack_name,
                    unsigned points_idx,
                    unsigned use_source_mask,
                    int unique_interpolation_method_id,
                    char const * tag_str) {

   int ret_val;
   ret_val = snprintf(buf, buffer_size, "%s%s%s%s%s%s%u%u%d%s", tag_base_str,
                      source_comp_name, target_comp_name,
                      source_domain_name, target_domain_name, stack_name,
                      points_idx, use_source_mask,
                      unique_interpolation_method_id, tag_str);

   if (ret_val < 0 || (size_t)ret_val >= buffer_size)
      yac_internal_abort_message("ERROR(generate_tag_string): truncation",
      __FILE__, __LINE__);
}

static void
generate_interpolation_method_spmap_tag(
  struct interpolation_method_spmap * method_spmap, char const * stack_name) {

   char const * str_base_str              = "interpolation method spmap 2";
   char const * str_generator_string      = "generator string";
   char const * str_request_header        = "request header";
   char const * str_request_data          = "request data";
   char const * str_request_answer_header = "request answer header";
   char const * str_request_answer_data   = "request answer data";
   char const * str_continue_barrier      = "continue barrier";
   char const * str_create_dist_dir       = "create dist dir";
   char const * str_distribute_results    = "distribute results";

   struct global_search * search;

   search = get_global_search((struct interpolation_method *)method_spmap);

   struct communicator * comm = yac_get_domain_comm(method_spmap->domain);

  char const * source_comp_name = yac_get_local_component_name(search);
  char const * target_comp_name = yac_get_remote_component_name(search);

  char const * source_domain_name = yac_get_local_domain_name(search);
  char const * target_domain_name = yac_get_remote_domain_name(search);

   unsigned max_tag_strlen = strlen(str_request_answer_header);

   unsigned max_str_buf_size = max_tag_strlen;
   max_str_buf_size += strlen(str_base_str);
   max_str_buf_size += strlen(source_comp_name);
   max_str_buf_size += strlen(target_comp_name);
   max_str_buf_size += strlen(source_domain_name);
   max_str_buf_size += strlen(target_domain_name);
   max_str_buf_size += strlen(stack_name) + 10 + 2 + 10 + 1; // (+ 10) for points_idx
                                                            // (+ 2)  for mask handling
                                                            // (+ 10) for unique_method_id
                                                            // (+ 1)  for '\0'

   char * unique_generator_string = xmalloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method_spmap->points_idx, method_spmap->masks != NULL,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method_spmap),
                       str_generator_string);

   char * tag_strings[7];
   tag_strings[0] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method_spmap->points_idx, method_spmap->masks != NULL,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method_spmap),
                       str_request_header);

   tag_strings[1] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method_spmap->points_idx, method_spmap->masks != NULL,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method_spmap),
                       str_request_data);

   tag_strings[2] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[2], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method_spmap->points_idx, method_spmap->masks != NULL,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method_spmap),
                       str_request_answer_header);

   tag_strings[3] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[3], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method_spmap->points_idx, method_spmap->masks != NULL,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method_spmap),
                       str_request_answer_data);

   tag_strings[4] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[4], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method_spmap->points_idx, method_spmap->masks != NULL,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method_spmap),
                       str_continue_barrier);

   tag_strings[5] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[5], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method_spmap->points_idx, method_spmap->masks != NULL,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method_spmap),
                       str_create_dist_dir);

   tag_strings[6] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[6], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method_spmap->points_idx, method_spmap->masks != NULL,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method_spmap),
                       str_distribute_results);

   unsigned tags[7];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 7, tags);

   free(unique_generator_string);

   method_spmap->tags.request_header = tags[0];
   method_spmap->tags.request_data = tags[1];
   method_spmap->tags.request_answer_header = tags[2];
   method_spmap->tags.request_answer_data = tags[3];
   method_spmap->tags.continue_barrier = tags[4];
   method_spmap->tags.create_dist_dir = tags[5];
   method_spmap->tags.distribute_results = tags[6];

   for (unsigned i = 0; i < 7; ++i) free(tag_strings[i]);
}

struct interpolation_method *
yac_new_interpolation_method_spmap_mask(struct global_search * search,
                                        struct domain * domain,
                                        char const * stack_name,
                                        struct points *** points,
                                        struct mask *** masks,
                                        unsigned num_pointsets,
                                        unsigned points_idx,
                                        unsigned use_source_mask) {

  struct interpolation_method_spmap * method_spmap =
    xmalloc (1 * sizeof(*method_spmap));

  method_spmap->vtable = &interpolation_method_spmap_vtable;
  method_spmap->finish_callback = NULL;
  method_spmap->finish_user_data = NULL;
  method_spmap->search = search;
  method_spmap->domain = domain;
  method_spmap->points = points;
  method_spmap->masks = (use_source_mask)?masks:NULL;
  method_spmap->num_pointsets = num_pointsets;
  method_spmap->points_idx = points_idx;
  method_spmap->data_request = NULL;

  yac_init_interpolation_method((struct interpolation_method *)method_spmap,
                                search);

  generate_interpolation_method_spmap_tag(method_spmap, stack_name);

  // ensure that the right point data is available

  if (num_pointsets == 0)
    yac_internal_abort_message(
      "ERROR(yac_new_interpolation_method_spmap_mask): missing data points.",
      __FILE__, __LINE__);

  if (points_idx >= num_pointsets)
    yac_internal_abort_message(
      "ERROR(yac_new_interpolation_method_spmap_mask): invalid points_idx.",
      __FILE__, __LINE__);

  unsigned num_subdomains = yac_get_num_subdomains(domain);

  if (num_subdomains > 0) {

    unsigned use_mask = masks != NULL;

    if (points == NULL)
      yac_internal_abort_message(
        "ERROR(yac_new_interpolation_method_spmap_mask): no point set data "
        "available", __FILE__, __LINE__);

    for (unsigned subdomain_idx = 0; subdomain_idx < num_subdomains;
         ++subdomain_idx) {

      if (use_mask && (masks[subdomain_idx] == NULL ||
                       masks[subdomain_idx][points_idx] == NULL))
        yac_internal_abort_message(
          "ERROR(yac_new_interpolation_method_spmap_mask): mask data is "
          "missing", __FILE__, __LINE__);

      if (points[subdomain_idx] == NULL ||
          points[subdomain_idx][points_idx] == NULL)
        yac_internal_abort_message(
          "ERROR(yac_new_interpolation_method_spmap_mask): no point set data "
          "available", __FILE__, __LINE__);

      if (points[0][points_idx]->location !=
          points[subdomain_idx][points_idx]->location)
        yac_internal_abort_message(
          "ERROR(yac_new_interpolation_method_spmap_mask): location of "
          "points do not match", __FILE__, __LINE__);
    }

    method_spmap->src_location = points[0][points_idx]->location;

    // if (method_spmap->src_location != CELL)
    //   yac_internal_abort_message(
    //     "ERROR(yac_new_interpolation_method_spmap_mask): data points have "
    //     "wrong location type (needs to be CELL).", __FILE__, __LINE__);
  }

  return (struct interpolation_method *)method_spmap;
}

struct interpolation_method *
yac_new_interpolation_method_spmap(struct global_search * search,
                                   struct domain * domain,
                                   char const * stack_name,
                                   struct points *** points,
                                   unsigned num_pointsets,
                                   unsigned points_idx) {

  return yac_new_interpolation_method_spmap_mask(
    search, domain, stack_name, points, NULL, num_pointsets, points_idx, 0);
}

static int
compare_src_point_data(const void * a, const void * b) {

  return ((((struct src_point_search_data *)a)->global_id) >
          (((struct src_point_search_data *)b)->global_id)) -
         ((((struct src_point_search_data *)a)->global_id) <
          (((struct src_point_search_data *)b)->global_id));
}

// gets all non-masked local source core points
static void
init_local_search_data(struct search_data * search_data, struct domain * domain,
                       struct points *** points, struct mask *** masks,
                       unsigned points_idx, enum yac_location src_location) {

  unsigned domain_rank = yac_get_comm_rank(yac_get_domain_comm(domain));

  unsigned num_subdomains = yac_get_num_subdomains(domain);

  struct src_point_search_data * src_point_data = NULL;
  unsigned src_point_array_size = 0;
  unsigned num_src_points = 0;

  for (unsigned subdomain_idx = 0; subdomain_idx < num_subdomains;
       ++subdomain_idx) {

    unsigned subdomain_size =
      yac_get_subdomain_data_size(domain, subdomain_idx, src_location);
    unsigned num_core_points = subdomain_size -
      yac_get_num_total_halo_points(domain, subdomain_idx, src_location);
    struct mask * mask =
      (masks != NULL)?(masks[subdomain_idx][points_idx]):(NULL);

    ENSURE_ARRAY_SIZE(src_point_data, src_point_array_size,
                      num_src_points + num_core_points);

    int const * global_ids =
      yac_get_global_ids(domain, subdomain_idx, src_location);
    unsigned const * global_ranks =
      yac_get_global_ranks(domain, subdomain_idx, src_location);

    for (unsigned i = 0; i < subdomain_size; ++i) {

      // if the source point is local and not masked
      if (((global_ranks == NULL) ||
           ((global_ranks[i] == -1) || (global_ranks[i] == domain_rank))) &&
          ((mask == NULL) || !(mask->mask[i]))) {

        src_point_data[num_src_points].global_id = global_ids[i];
        src_point_data[num_src_points].local_id = i;
        src_point_data[num_src_points].subdomain_idx = subdomain_idx;
        double base_point[2];
        yac_get_point_coordinates(
          points[subdomain_idx][points_idx], i, base_point);
        LLtoXYZ(base_point[0], base_point[1],
                src_point_data[num_src_points].circle.base_vector);
        src_point_data[num_src_points].circle.inc_angle = SIN_COS_M_PI;
        src_point_data[num_src_points].tgt_point.global_id = -1;
        src_point_data[num_src_points].tgt_point.rank = (unsigned)-2;
        src_point_data[num_src_points].checked_ranks = NULL;
        src_point_data[num_src_points].num_checked_ranks = 0;
        src_point_data[num_src_points].to_be_checked_ranks = NULL;
        src_point_data[num_src_points].to_be_checked_ranks_array_size = 0;
        src_point_data[num_src_points].num_to_be_checked_ranks = 0;
        num_src_points++;
      }
    }
  }

  // sort source points data based on global ids of the source points
  qsort(src_point_data, num_src_points, sizeof(*src_point_data),
        compare_src_point_data);
  search_data->src_point_data =
    xrealloc(src_point_data, num_src_points * sizeof(*src_point_data));
  search_data->num_src_points = num_src_points;

  search_data->coordinates_xyz =
    xmalloc(num_src_points * sizeof(*(search_data->coordinates_xyz)));
  for (unsigned i = 0; i < num_src_points; ++i) {

    search_data->coordinates_xyz[i][0] =
      search_data->src_point_data[i].circle.base_vector[0];
    search_data->coordinates_xyz[i][1] =
      search_data->src_point_data[i].circle.base_vector[1];
    search_data->coordinates_xyz[i][2] =
      search_data->src_point_data[i].circle.base_vector[2];
  }
}

// searches through local target points
static void search_local(struct search_data * search_data,
                         struct tgt_point_data tgt_points,
                         unsigned rank) {

  struct src_point_search_data * src_point_data = search_data->src_point_data;
  size_t num_src_points = search_data->num_src_points;

  size_t * local_point_ids = NULL;
  size_t local_point_ids_array_size = 0;
  size_t * num_local_point_ids =
    xmalloc(num_src_points * sizeof(*num_local_point_ids));

  yac_point_sphere_part_search_NN(
    tgt_points.point_sphere_part, num_src_points, search_data->coordinates_xyz,
    NULL, NULL, NULL, &local_point_ids, &local_point_ids_array_size,
    num_local_point_ids);

  // if more than on local point, use highest; because tgt point are sorted by
  // global_id, it will also be the highest one
  size_t offset = 0;
  for (size_t src_idx = 0; src_idx < num_src_points; ++src_idx) {

    if (num_local_point_ids[src_idx] == 0) continue;

    unsigned tgt_idx = (unsigned)(local_point_ids[offset]);

    struct sin_cos_angle distance =
      get_vector_angle_2(tgt_points.coordinates_xyz[tgt_idx],
                         src_point_data[src_idx].circle.base_vector);

    // if the result is not better than the one already found
    if (compare_angles(distance, src_point_data[src_idx].circle.inc_angle) > 0)
      continue;

    int tgt_global_id = tgt_points.global_ids[tgt_idx];
    // find the target point with the highest global id
    for (unsigned i = 1; i < num_local_point_ids[src_idx]; ++i) {

      unsigned tmp_tgt_idx = (unsigned)(local_point_ids[offset + i]);
      int tmp_tgt_global_id = tgt_points.global_ids[tmp_tgt_idx];
      if (tmp_tgt_global_id > tgt_global_id) {
        tgt_idx = tmp_tgt_idx;
        tgt_global_id = tmp_tgt_global_id;
      }
    }

    src_point_data[src_idx].circle.inc_angle = distance;
    src_point_data[src_idx].tgt_point.global_id = tgt_global_id;
    src_point_data[src_idx].tgt_point.rank = rank;

    offset += num_local_point_ids[src_idx];
  }

  free(local_point_ids);
  free(num_local_point_ids);
}

// gets the bounding circle and the owner rank of all halo cells
static void
init_halo_cells(struct halo_cells ** halo_cells, unsigned * num_halo_cells,
                struct domain * domain) {

  struct halo_cells * tmp_halo_cells;
  unsigned tmp_num_halo_cells = 0;

  unsigned num_subdomains = yac_get_num_subdomains(domain);
  unsigned domain_rank = yac_get_comm_rank(yac_get_domain_comm(domain));

  for (unsigned subdomain_idx = 0; subdomain_idx < num_subdomains;
       ++subdomain_idx) {

    tmp_num_halo_cells +=
      yac_get_num_total_halo_points(domain, subdomain_idx, CELL);
  }

  tmp_halo_cells = xmalloc(tmp_num_halo_cells * sizeof(*tmp_halo_cells));
  tmp_num_halo_cells = 0;

  struct grid_cell src_cell;

  yac_init_grid_cell(&src_cell);

  for (unsigned subdomain_idx = 0; subdomain_idx < num_subdomains;
       ++subdomain_idx) {

    unsigned const * ranks = yac_get_global_ranks(domain, subdomain_idx, CELL);

    if (ranks == NULL) continue;

    struct grid * grid = yac_get_subdomain_grid(*domain, subdomain_idx);

    unsigned subdomain_size =
      yac_get_subdomain_data_size(domain, subdomain_idx, CELL);

    for (unsigned local_cell_id = 0; local_cell_id < subdomain_size;
         ++local_cell_id) {

      if ((ranks[local_cell_id] == -1) || (ranks[local_cell_id] == domain_rank))
        continue;

      tmp_halo_cells[tmp_num_halo_cells].rank = ranks[local_cell_id];
      yac_get_grid_cell2(
        grid, local_cell_id, &src_cell,
        &(tmp_halo_cells[tmp_num_halo_cells].bnd_circle));
      tmp_num_halo_cells++;
    }
  }

  yac_free_grid_cell(&src_cell);

  *halo_cells = tmp_halo_cells;
  *num_halo_cells = tmp_num_halo_cells;
}

static void
insertion_sort_uint(unsigned element, unsigned * a, unsigned * curr_length) {

   unsigned i, j;

   if (*curr_length > 0) {

      for (i = 0; i < *curr_length; ++i)
         if (a[i] >= element) break;

      // if the element is already in the array
      if ((i != *curr_length) && (a[i] == element)) return;

      // copy new element into array and move bigger elements one position up
      for (j = *curr_length; j > i; --j) {

         a[j] = a[j-1];
      }

      a[i] = element;

      // increase array length indicator
      ++(*curr_length);

   } else {

      a[0] = element;
      *curr_length = 1;
   }
}

// check for overlaps with halos
static void
check_for_potential_remote_points(struct search_data * search_data,
                                  struct halo_cells * halo_cells,
                                  unsigned num_halo_cells,
                                  unsigned ignore_rank) {

  for (unsigned src_idx = 0; src_idx < search_data->num_src_points; ++src_idx) {

    struct bounding_circle * src_circle =
      &(search_data->src_point_data[src_idx].circle);

    for (unsigned halo_idx = 0; halo_idx < num_halo_cells; ++halo_idx) {
      
      if (halo_cells[halo_idx].rank == ignore_rank) continue;

      if (yac_extents_overlap(src_circle, &(halo_cells[halo_idx].bnd_circle))) {

        ENSURE_ARRAY_SIZE(
          search_data->src_point_data[src_idx].to_be_checked_ranks,
          search_data->src_point_data[src_idx].to_be_checked_ranks_array_size,
          search_data->src_point_data[src_idx].num_to_be_checked_ranks + 1);

        insertion_sort_uint(
          halo_cells[halo_idx].rank,
          search_data->src_point_data[src_idx].to_be_checked_ranks,
          &(search_data->src_point_data[src_idx].num_to_be_checked_ranks));
      }
    }
  }
}

static void
reorder_target_points(struct tgt_point_data tgt_points,
                      struct search_results * results,
                      unsigned * num_interpolated_points,
                      unsigned * num_points) {

  unsigned i = 0;

  double (*tmp_coordinates_xyz)[3] =
    xmalloc(tgt_points.num_tgt_points * sizeof(*tmp_coordinates_xyz));

  memcpy(tmp_coordinates_xyz, tgt_points.coordinates_xyz,
    tgt_points.num_tgt_points * sizeof(*tmp_coordinates_xyz));
  // put interpolated points to the front
  for (unsigned tgt_idx = 0; tgt_idx < tgt_points.num_tgt_points; ++tgt_idx) {
    if ((results[tgt_idx].num_src > 0) && (results[tgt_idx].num_src != -1)) {
      tgt_points.coordinates_xyz[i][0] = tmp_coordinates_xyz[tgt_idx][0];
      tgt_points.coordinates_xyz[i][1] = tmp_coordinates_xyz[tgt_idx][1];
      tgt_points.coordinates_xyz[i][2] = tmp_coordinates_xyz[tgt_idx][2];
      tgt_points.global_ids[i]    = results[tgt_idx].tgt_global_id;
      ++i;
    }
  }
  *num_interpolated_points = i;
  // put not interpolated points behind
  for (unsigned tgt_idx = 0; tgt_idx < tgt_points.num_tgt_points; ++tgt_idx) {
    if (results[tgt_idx].num_src == 0) {
      tgt_points.coordinates_xyz[i][0] = tmp_coordinates_xyz[tgt_idx][0];
      tgt_points.coordinates_xyz[i][1] = tmp_coordinates_xyz[tgt_idx][1];
      tgt_points.coordinates_xyz[i][2] = tmp_coordinates_xyz[tgt_idx][2];
      tgt_points.global_ids[i]    = results[tgt_idx].tgt_global_id;
      ++i;
    }
  }
  *num_points = i;

  for (unsigned tgt_idx = 0, j = 0; tgt_idx < tgt_points.num_tgt_points;
       ++tgt_idx) {

    if ((results[tgt_idx].num_src == 0) || (results[tgt_idx].num_src == -1))
      continue;

    results[j++] = results[tgt_idx];
  }

  free(tmp_coordinates_xyz);
}

/**
 * Check for source points that are required for the interpolation, which are
 * located on other source processes
 */
static void
check_for_remote_points(unsigned * num_remote_points,
                        int ** remote_points,
                        int ** remote_point_ranks,
                        struct search_results * results,
                        unsigned num_interpolated_points,
                        unsigned comm_rank) {

  size_t tmp_num_remote_points = 0;

  for (unsigned i = 0; i < num_interpolated_points; ++i)
    tmp_num_remote_points += (size_t)(results[i].num_src);

  int * tmp_remote_points =
    xmalloc(tmp_num_remote_points * sizeof(*tmp_remote_points));
  unsigned * tmp_remote_point_ranks =
    xmalloc(tmp_num_remote_points * sizeof(*tmp_remote_points));

  tmp_num_remote_points = 0;

  for (unsigned i = 0; i < num_interpolated_points; ++i) {
    for (unsigned j = 0; j < results[i].num_src; ++j) {

      if (results[i].owner[j] == comm_rank) continue;

      tmp_remote_points[tmp_num_remote_points] = results[i].src_global_ids[j];
      tmp_remote_point_ranks[tmp_num_remote_points] = results[i].owner[j];
      ++tmp_num_remote_points;
    }
  }

  yac_quicksort_index(tmp_remote_points, tmp_num_remote_points,
                      (int*)tmp_remote_point_ranks);

  // remote duplicated entries
  size_t j = tmp_num_remote_points > 0;
  for (size_t i = 1; i < tmp_num_remote_points; ++i) {
    if (tmp_remote_points[i] == tmp_remote_points[i-1]) continue;
    if (i != j) {
      tmp_remote_points[j] = tmp_remote_points[i];
      tmp_remote_point_ranks[j] = tmp_remote_point_ranks[i];
    }
    ++j;
  }

  *num_remote_points = (unsigned)j;
  *remote_points =
    xrealloc(tmp_remote_points,
             tmp_num_remote_points * sizeof(*tmp_remote_points));
  *remote_point_ranks =
    xrealloc(tmp_remote_point_ranks,
             tmp_num_remote_points * sizeof(*tmp_remote_point_ranks));
}

/**
 * Check for source points that are required for the interpolation, which are located on
 * other source processes or on another local subdomain (which the target point is
 * not assigned to).
 *
 * @param[in] len  size of list
 * @param[in] list list of integer numbers
 * @param[in] list_req list element for which the index is requested
 *
 * @return returns the index of list containing list_req
 *
 */

static int
index_location_spmap ( unsigned len, int * list, int list_req ) {

   //
   // simple bisection algorithm, list must be ordered.
   // -------------------------------------------------

   int j1, j2, j3;

   // Initialise lower and upper limits

   int location = -1;

   j1 = 0;

   j3 = len;

   while ( (j3-j1) > 0 ) {

      // Compute midpoint

      j2 = (j3+j1)/2;

      // replace lower or upper limit

      if ( (list[len-1] > list[0]) == (list_req > list[j2]) )
         j1 = j2;
      else
         j3 = j2;

      if ( list_req == list[j2] ) {
         location = j2;
         break;
      }
   }
   return location;
}

static struct interpolation_plan_element *
generate_interpolation_plan_spmap(struct search_results * results,
                                  struct search_data * search_data,
                                  unsigned num_interpolated_points,
                                  int * remote_points,
                                  unsigned num_remote_points, 
                                  unsigned comm_rank, unsigned points_idx,
                                  unsigned num_pointsets) {

  unsigned total_num_mapped_source_cells = 0;
  unsigned * num_src_per_tgt =
    xmalloc(num_interpolated_points * sizeof(*num_src_per_tgt));

  for (unsigned i = 0; i < num_interpolated_points; ++i) {
    total_num_mapped_source_cells += results[i].num_src;
    num_src_per_tgt[i] = results[i].num_src;
  }

  unsigned * src_indices =
    xmalloc(3 * total_num_mapped_source_cells * sizeof(*src_indices));
  unsigned * curr_src_indices = src_indices;
  unsigned src_idx = 0;

  for (unsigned tgt_idx = 0; tgt_idx < num_interpolated_points; ++tgt_idx) {
    for (unsigned i = 0; i < results[tgt_idx].num_src; ++i) {

      // if the current source point is local
      if (results[tgt_idx].owner[i] == comm_rank) {

        int tgt_global_id = results[tgt_idx].tgt_global_id;
        while (search_data->src_point_data[src_idx].tgt_point.global_id !=
               tgt_global_id) ++src_idx;

#if 1
        if ((src_idx >= search_data->num_src_points) ||
            (results[tgt_idx].tgt_global_id !=
             search_data->src_point_data[src_idx].tgt_point.global_id) ||
            (results[tgt_idx].src_global_ids[i] !=
             search_data->src_point_data[src_idx].global_id))
          yac_internal_abort_message(
            "ERROR(generate_interpolation_plan_spmap): internal error\n",
            __FILE__, __LINE__);
#endif

        curr_src_indices[0] =
          search_data->src_point_data[src_idx].subdomain_idx + 1;
        curr_src_indices[1] = points_idx;
        curr_src_indices[2] = search_data->src_point_data[src_idx].local_id;
        ++src_idx;

      } else {

        curr_src_indices[0] = 0;
        curr_src_indices[1] = points_idx;
        curr_src_indices[2] =
          index_location_spmap(num_remote_points, remote_points,
                               results[tgt_idx].src_global_ids[i]);
      }
      curr_src_indices += 3;
    }
  }

  unsigned * num_remote_points_per_point_set =
    xcalloc(num_pointsets, sizeof(*num_remote_points_per_point_set));
  num_remote_points_per_point_set[points_idx] = num_remote_points;

  struct interpolation_plan_element * plan = xmalloc(1 * sizeof(*plan));
   yac_init_interpolation_plan_element_pointsum(
    plan, num_interpolated_points, src_indices, num_src_per_tgt,
    num_remote_points_per_point_set, num_pointsets);

  free(num_remote_points_per_point_set);
  free(src_indices);
  free(num_src_per_tgt);

  return plan;
}

static struct interpolation_method_data_request *
generate_data_request(unsigned num_remote_points, int * remote_points,
                      int * remote_point_ranks, unsigned num_pointsets,
                      unsigned points_idx) {

  struct interpolation_method_data_request * data_request =
    xmalloc(num_pointsets * sizeof(*data_request));

  for (unsigned i = 0; i < num_pointsets; ++i) {
    data_request[i].num_values = 0;
    data_request[i].global_ids = NULL;
    data_request[i].ranks = NULL;
  }

  data_request[points_idx].num_values = num_remote_points;
  data_request[points_idx].global_ids = remote_points;
  data_request[points_idx].ranks = remote_point_ranks;

  return data_request;
}

static void
do_search_final(struct intermediate_method_data * data,
                struct search_results * results,
                struct communicator * comm) {

  struct interpolation_method_spmap * method_spmap =
    (struct interpolation_method_spmap*)(data->method_stack[0]);

  unsigned num_interpolated_points, num_points;

  // reorder target points (put interpolated points to the front and remove
  // target points that are interpolated by other processes)
  reorder_target_points(data->tgt_points, results, &num_interpolated_points,
                        &num_points);

  data->num_interpolated_points[0] = num_interpolated_points;

  // Prepare information about source points required for the mapping that are
  // not locally available
  unsigned num_remote_points;
  int * remote_points;
  int * remote_point_ranks;
  check_for_remote_points(&num_remote_points, &remote_points,
                          &remote_point_ranks, results, num_interpolated_points,
                          yac_get_comm_rank(comm));

  // generate data request used by field interpolation
  method_spmap->data_request[0] =
    generate_data_request(num_remote_points, remote_points, remote_point_ranks,
                          method_spmap->num_pointsets,
                          method_spmap->points_idx);

  // generate interpolation plan
  data->plan_stack[0] =
    generate_interpolation_plan_spmap(results, &(data->search_data),
                                      num_interpolated_points,
                                      remote_points, num_remote_points,
                                      yac_get_comm_rank(comm),
                                      method_spmap->points_idx,
                                      method_spmap->num_pointsets);

  // advance interpolation plan stack if an interpolation plan was generated
  if (data->plan_stack[0] != NULL) data->plan_stack++;

  data->method_stack++;
  data->tgt_points.global_ids += num_interpolated_points;
  data->tgt_points.coordinates_xyz += num_interpolated_points;
  unsigned num_remaining_points = num_points - num_interpolated_points;
  data->num_interpolated_points++;

  // call next interpolation method on stack
  if (data->method_stack[0] != NULL)
     data->method_stack[0]->vtable->do_search(
        data->method_stack, data->plan_stack, data->tgt_points.global_ids,
        data->tgt_points.coordinates_xyz, data->tgt_location,
        num_remaining_points, data->num_interpolated_points,
        method_spmap->data_request + 1);

  // call finish_callback
  if (method_spmap->finish_callback != NULL)
    method_spmap->finish_callback(method_spmap->finish_user_data);

  // clean up
  for (unsigned i = 0; i < data->search_data.num_src_points; ++i) {
    free(data->search_data.src_point_data[i].checked_ranks);
    free(data->search_data.src_point_data[i].to_be_checked_ranks);
  }
  yac_delete_point_sphere_part_search(data->tgt_points.point_sphere_part);
  free(data->search_data.coordinates_xyz);
  free(data->search_data.src_point_data);
  free(data);
}

static int
compare_results(const void * a, const void * b) {

  return ((((struct search_results *)a)->tgt_global_id) >
          (((struct search_results *)b)->tgt_global_id)) -
         ((((struct search_results *)a)->tgt_global_id) <
          (((struct search_results *)b)->tgt_global_id));
}

static void
alltoallv_distribute_results_callback(struct communicator * comm,
                                      struct comm_message msg,
                                      void * user_data) {

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)user_data;

  int * recv_buffer = (int *)(msg.buffer);

  struct search_results * results =
    xmalloc(data->tgt_points.num_tgt_points * sizeof(*results));

  // unpack results
  for (unsigned i = 0; i < data->tgt_points.num_tgt_points; ++i) {

    results[i].tgt_global_id = *(recv_buffer++);
    results[i].num_src = *(recv_buffer++);
    if (results[i].num_src == -1) {
      results[i].src_global_ids = NULL;
      results[i].owner = NULL;
    } else {
      results[i].src_global_ids = recv_buffer;
      results[i].owner = (unsigned*)recv_buffer + results[i].num_src;
      recv_buffer += 2 * results[i].num_src;
    }
  }

  qsort(results, data->tgt_points.num_tgt_points, sizeof(*results),
        compare_results);

  do_search_final(data, results, comm);

  free(results);
}

static void
insertion_sort_int(int element, int * a, unsigned * curr_length) {

   unsigned i, j;

   if (*curr_length > 0) {

      for (i = 0; i < *curr_length; ++i)
         if (a[i] >= element) break;

      // if the element is already in the array
      if ((i != *curr_length) && (a[i] == element)) return;

      // copy new element into array and move bigger elements one position up
      for (j = *curr_length; j > i; --j) {

         a[j] = a[j-1];
      }

      a[i] = element;

      // increase array length indicator
      ++(*curr_length);

   } else {

      a[0] = element;
      *curr_length = 1;
   }
}

static void
remove_duplicated_src_points(struct dist_dir_data * dist_dir_data,
                             unsigned num_dist_dir_data) {

  for (unsigned i = 0; i < num_dist_dir_data; ++i) {

    yac_quicksort_index(dist_dir_data[i].src_global_ids,
                        (size_t)(dist_dir_data[i].num_src),
                        (int*)(dist_dir_data[i].src_owner));
    unsigned tmp_num_src = dist_dir_data[i].num_src > 0;
    for (unsigned j = 1; j < dist_dir_data[i].num_src; ++j) {

      if (dist_dir_data[i].src_global_ids[j] ==
          dist_dir_data[i].src_global_ids[j-1]) continue;

      if (j != tmp_num_src) {
        dist_dir_data[i].src_global_ids[tmp_num_src] =
          dist_dir_data[i].src_global_ids[j];
        dist_dir_data[i].src_owner[tmp_num_src] =
          dist_dir_data[i].src_owner[j];
      }
      tmp_num_src++;
    }
  }
}

static void
alltoallv_dist_dir_data_callback(struct communicator * comm,
                                 struct comm_message msg, void * user_data) {

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)user_data;
  unsigned comm_size = yac_get_comm_size(comm);

  int * recv_buffer = (int*)(msg.buffer);

  // get a list of all target points count total number of source points
  int * tgt_global_ids = NULL;
  unsigned tgt_global_ids_array_size = 0;
  unsigned num_tgt_global_ids = 0;
  unsigned num_src_global_ids = 0;

  for (unsigned rank = 0; rank < comm_size; ++rank) {

    unsigned curr_num_tgt_global_ids = *(recv_buffer++);

    ENSURE_ARRAY_SIZE(tgt_global_ids, tgt_global_ids_array_size,
                      num_tgt_global_ids + curr_num_tgt_global_ids);

    for (unsigned i = 0; i < curr_num_tgt_global_ids; ++i)
      insertion_sort_int(recv_buffer[i], tgt_global_ids, &num_tgt_global_ids);

    recv_buffer += curr_num_tgt_global_ids;

    unsigned curr_num_src_points = *(recv_buffer++);
    num_src_global_ids += curr_num_src_points;

    // skip next part of the message
    recv_buffer += 2 * curr_num_src_points;
  }

  // allocate distributed data
  struct dist_dir_data * dist_dir_data =
    xmalloc(num_tgt_global_ids * sizeof(*dist_dir_data));
  int * src_global_ids_buffer =
    xmalloc(num_src_global_ids * sizeof(*src_global_ids_buffer));
  unsigned * src_owner_buffer =
    xmalloc(num_src_global_ids * sizeof(*src_owner_buffer));

  for (unsigned i = 0; i < num_tgt_global_ids; ++i) {
    dist_dir_data[i].tgt_global_id = tgt_global_ids[i];
    dist_dir_data[i].master_rank = -1;
    dist_dir_data[i].num_src = 0;
  }
  free(tgt_global_ids);

  // count number of sources per target points and determine master ranks for
  // each target point
  recv_buffer = (int*)(msg.buffer);
  for (unsigned rank = 0; rank < comm_size; ++rank) {

    unsigned curr_num_tgt_global_ids = *(recv_buffer++);
    int * curr_tgt_global_ids = recv_buffer;
    recv_buffer += curr_num_tgt_global_ids;
    unsigned dist_dir_data_idx = 0;

    unsigned curr_num_src_points = *(recv_buffer++);

    for (unsigned i = 0, tgt_idx = 0; i < curr_num_src_points; ++i) {

      int curr_tgt_global_id = *(recv_buffer++);
      // required to skip curr_src_global_id in recv_buffer:
      ++recv_buffer;

      while (dist_dir_data[dist_dir_data_idx].tgt_global_id !=
        curr_tgt_global_id) dist_dir_data_idx++;

      // if the current target point does not yet have a master rank
      if (dist_dir_data[dist_dir_data_idx].master_rank == -1) {

        // check whether the current target points is available on the current
        // source rank
        while ((tgt_idx < curr_num_tgt_global_ids) &&
               (curr_tgt_global_ids[tgt_idx] < curr_tgt_global_id)) ++tgt_idx;

        if ((tgt_idx < curr_num_tgt_global_ids) &&
            (curr_tgt_global_ids[tgt_idx] == curr_tgt_global_id))
          dist_dir_data[dist_dir_data_idx].master_rank = rank;
      }

      dist_dir_data[dist_dir_data_idx].num_src++;
    }
  }

  // ensure that each target that is to be interpolated has a valid master rank
  {
    unsigned dist_dir_data_idx;
    for (dist_dir_data_idx = 0; dist_dir_data_idx < num_tgt_global_ids;
         ++dist_dir_data_idx)
      if ((dist_dir_data[dist_dir_data_idx].num_src != 0) &&
          (dist_dir_data[dist_dir_data_idx].master_rank == -1))
        break;

    // if there are target points that have to be interpolated and have not yet
    // a master process assigned to them
    if (dist_dir_data_idx != num_tgt_global_ids) {

      recv_buffer = (int*)(msg.buffer);
      for (unsigned rank = 0; rank < comm_size; ++rank) {

        unsigned curr_num_tgt_global_ids = *(recv_buffer++);
        dist_dir_data_idx = 0;

        for (unsigned i = 0; i < curr_num_tgt_global_ids; ++i) {

          int curr_tgt_global_id = recv_buffer[i];

          while (dist_dir_data[dist_dir_data_idx].tgt_global_id !=
                 curr_tgt_global_id) dist_dir_data_idx++;

          if ((dist_dir_data[dist_dir_data_idx].num_src != 0) &&
              (dist_dir_data[dist_dir_data_idx].master_rank == -1))
            dist_dir_data[dist_dir_data_idx].master_rank = rank;
        }

        recv_buffer += curr_num_tgt_global_ids;

        unsigned curr_num_src_points = *(recv_buffer++);
        recv_buffer += 2 * curr_num_src_points;
      }
    }
  }

  for (unsigned i = 0, offset = 0; i < num_tgt_global_ids; ++i) {
    dist_dir_data[i].src_global_ids = src_global_ids_buffer + offset;
    dist_dir_data[i].src_owner = src_owner_buffer + offset;
    offset += dist_dir_data[i].num_src;
    dist_dir_data[i].num_src = 0;
  }

  // unpack src points
  recv_buffer = (int*)(msg.buffer);
  for (unsigned rank = 0; rank < comm_size; ++rank) {

    // skip first part of the message
    unsigned curr_num_tgt_global_ids = *(recv_buffer++);
    recv_buffer += curr_num_tgt_global_ids;

    unsigned curr_num_src_points = *(recv_buffer++);
    unsigned dist_dir_data_idx = 0;

    for (unsigned i = 0; i < curr_num_src_points; ++i) {

      int curr_tgt_global_id = *(recv_buffer++);
      int curr_src_global_id = *(recv_buffer++);

      while (dist_dir_data[dist_dir_data_idx].tgt_global_id !=
        curr_tgt_global_id) dist_dir_data_idx++;

      unsigned src_idx = dist_dir_data[dist_dir_data_idx].num_src++;

      dist_dir_data[dist_dir_data_idx].src_global_ids[src_idx] =
        curr_src_global_id;
      dist_dir_data[dist_dir_data_idx].src_owner[src_idx] = rank;
    }
  }

  // remove duplicated source points
  remove_duplicated_src_points(dist_dir_data, num_tgt_global_ids);

  // pack and send answer
  // answer per rank:
  // 
  // N * (tgt_global_id
  //      {-1 = remove tgt,
  //       0 = keep tgt (not interpolated)
  //       n = number src used to interpolate this tgt
  //           -> this is followed by the global ids of the respective src's
  //              and the respective owner})
  int * send_buffer = NULL;
  unsigned send_buffer_array_size = 0;
  unsigned prev_send_buffer_size = 0;
  unsigned send_buffer_size = 0;
  unsigned * send_buffer_size_per_rank =
    xmalloc(comm_size * sizeof(*send_buffer_size_per_rank));

  recv_buffer = (int*)(msg.buffer);
  for (unsigned rank = 0; rank < comm_size; ++rank) {

    unsigned dist_dir_data_idx = 0;
    unsigned curr_num_tgt_points = *(recv_buffer++);

    for (unsigned i = 0; i < curr_num_tgt_points; ++i) {

      int tgt_global_id = *(recv_buffer++);

      while ((dist_dir_data_idx < num_tgt_global_ids) &&
             (dist_dir_data[dist_dir_data_idx].tgt_global_id < tgt_global_id))
        ++dist_dir_data_idx;

      ENSURE_ARRAY_SIZE(send_buffer, send_buffer_array_size,
                        send_buffer_size + 2 +
                        2 * dist_dir_data[dist_dir_data_idx].num_src);

      send_buffer[send_buffer_size++] = tgt_global_id;

      // if there is no source for the current target point
      if (dist_dir_data[dist_dir_data_idx].num_src == 0) {

        send_buffer[send_buffer_size++] = 0;

      // if the target point is interpolated by another process
      } else if (dist_dir_data[dist_dir_data_idx].master_rank != rank) {

        send_buffer[send_buffer_size++] = -1;

      } else {

        send_buffer[send_buffer_size++] =
          dist_dir_data[dist_dir_data_idx].num_src;
        for (unsigned j = 0; j < dist_dir_data[dist_dir_data_idx].num_src; ++j)
          send_buffer[send_buffer_size++] =
            dist_dir_data[dist_dir_data_idx].src_global_ids[j];
        for (unsigned j = 0; j < dist_dir_data[dist_dir_data_idx].num_src; ++j)
          send_buffer[send_buffer_size++] =
            dist_dir_data[dist_dir_data_idx].src_owner[j];
      }
    }

    // we now ignore the second part of the message of each process
    unsigned curr_num_src_points = *(recv_buffer++);
    recv_buffer += 2 * curr_num_src_points;

    send_buffer_size_per_rank[rank] = send_buffer_size - prev_send_buffer_size;
    prev_send_buffer_size = send_buffer_size;
  }

  // distribute the results
  yac_comm_alltoallv_callback(send_buffer, send_buffer_size_per_rank, COMM_INT,
                              data->tags.distribute_results, comm, data,
                              alltoallv_distribute_results_callback,
                              MSG_PRIO_MIDDLE);

  free(send_buffer_size_per_rank);
  free(send_buffer);
  free(src_owner_buffer);
  free(src_global_ids_buffer);
  free(dist_dir_data);
}

static void
generate_result_dist_dir(struct intermediate_method_data * data,
                         struct communicator * comm) {

  unsigned comm_size = yac_get_comm_size(comm);

  // pack data for each dist dir process
  // N - num "local" target points
  // N * tgt_global_id
  // M - local source points
  // M * (tgt_global_id, src_global_id)

  unsigned * tmp_buffer = xcalloc(4 * comm_size, sizeof(*tmp_buffer));

  // count number of tgt's per rank
  unsigned * num_tgt_per_rank = tmp_buffer + 0 * comm_size;
  for (unsigned i = 0; i < data->tgt_points.num_tgt_points; ++i)
    num_tgt_per_rank[(data->tgt_points.global_ids[i]/128)%comm_size]++;

  // count number of src's per rank
  unsigned * num_src_per_rank = tmp_buffer + 1 * comm_size;
  for (unsigned i = 0; i < data->search_data.num_src_points; ++i)
    num_src_per_rank[
      (data->search_data.src_point_data[i].tgt_point.global_id/128)%comm_size]++;

  unsigned * send_size_per_rank = tmp_buffer + 3 * comm_size;
  for (unsigned i = 0; i < comm_size; ++i)
    send_size_per_rank[i] = 2 + num_tgt_per_rank[i] + 2 * num_src_per_rank[i];
  unsigned * pack_offsets = tmp_buffer + 2 * comm_size;
  unsigned accum = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    pack_offsets[i] = accum;
    accum += send_size_per_rank[i];
  }

  unsigned buffer_size = 2 * comm_size + data->tgt_points.num_tgt_points +
                         2 * data->search_data.num_src_points;
  int * send_buffer = xmalloc(buffer_size * sizeof(*send_buffer));

  for (unsigned i = 0; i < comm_size; ++i)
    send_buffer[pack_offsets[i]++] = num_tgt_per_rank[i];
  for (unsigned i = 0; i < data->tgt_points.num_tgt_points; ++i) {
    int tgt_global_id = data->tgt_points.global_ids[i];
    send_buffer[pack_offsets[(tgt_global_id/128)%comm_size]++] = tgt_global_id;
  }

  for (unsigned i = 0; i < comm_size; ++i)
    send_buffer[pack_offsets[i]++] = num_src_per_rank[i];

  for (unsigned i = 0; i < data->search_data.num_src_points; ++i) {
    int tgt_global_id = data->search_data.src_point_data[i].tgt_point.global_id;
    unsigned rank = (tgt_global_id/128)%comm_size;
    send_buffer[pack_offsets[rank]++] = tgt_global_id;
    send_buffer[pack_offsets[rank]++] =
      data->search_data.src_point_data[i].global_id;
  }

  yac_comm_alltoallv_callback(send_buffer, send_size_per_rank, COMM_INT,
                              data->tags.create_dist_dir, comm, data,
                              alltoallv_dist_dir_data_callback,
                              MSG_PRIO_MIDDLE);

  free(send_buffer);
  free(tmp_buffer);
}

static void
check_search_results(struct search_data search_data) {

  for (unsigned src_idx = 0; src_idx < search_data.num_src_points; ++src_idx) {

    // if a target point was found for a local source point
    if ((fabs(search_data.src_point_data[
                src_idx].circle.inc_angle.sin) > 0.0) ||
        (fabs(search_data.src_point_data[
                src_idx].circle.inc_angle.cos + 1.0) > 0.0))
      continue;

    char str_err[1024];
    sprintf(str_err, "ERROR(interpolation_method_spmap) for source point %d "
            "(%lf, %lf, %lf) no matching target point was found\n",
            search_data.src_point_data[src_idx].global_id,
            search_data.src_point_data[src_idx].circle.base_vector[0],
            search_data.src_point_data[src_idx].circle.base_vector[1],
            search_data.src_point_data[src_idx].circle.base_vector[2]);
    yac_internal_abort_message(str_err, __FILE__, __LINE__);
  }
}

static int
compare_src_point_data_tgt(const void * a, const void * b) {

  struct src_point_search_data * a_ = (struct src_point_search_data *)a;
  struct src_point_search_data * b_ = (struct src_point_search_data *)b;

  int ret = a_->tgt_point.global_id - b_->tgt_point.global_id;
  if (ret == 0) ret = a_->global_id - b_->global_id;

  return ret;
}

static void do_search_spmap_continue(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  UNUSED(msg);

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)user_data;

  // at this point, we have found for each source point on all participating
  // source processes the closest target point
  check_search_results(data->search_data);

  // we can cancel the search request handler
  yac_comm_cancel_recv(comm, data->tags.request_header);

  // sort source point data based on global ids of their target points
  qsort(data->search_data.src_point_data, data->search_data.num_src_points,
        sizeof(*(data->search_data.src_point_data)),
        compare_src_point_data_tgt);

  // now we redistribute all data into a distributed directory. we use this to
  // I.   gather for each target point all source points that contribute to its
  //      interpolation value
  // II.  to determine which target points can be interpolated
  // III. to determine the source processes that are resposible for the
  //      interpolation of each target point
  generate_result_dist_dir(data, comm);

  // once the distributed directory is created, it will automatically send the
  // required information to each source process
}

static int compare_uint (const void * a, const void * b) {
  return ( *(unsigned*)a - *(unsigned*)b );
}

static void
unpack_search_request_answer(struct search_data search_data, void * buffer,
                             int buffer_size, struct communicator * comm) {

  int position = 0;
  unsigned src_idx = 0;

  while (position < buffer_size) {

    int src_global_id;

    yac_comm_unpack(buffer, buffer_size, &position, &src_global_id, 1, COMM_INT,
                    comm);

    // find local source cell data matching the received src_global_id
    while((src_idx < search_data.num_src_points) &&
          (search_data.src_point_data[src_idx].global_id != src_global_id))
      ++src_idx;

    if (src_idx == search_data.num_src_points)
      yac_internal_abort_message("ERROR(unpack_search_request_answer): "
                                 "internal error", __FILE__, __LINE__);

    unsigned rank;

    yac_comm_unpack(buffer, buffer_size, &position, &rank, 1, COMM_UINT,
                    comm);

    // if the remote process found a better target point
    if (rank != -1) {

      int tgt_global_id;
      struct sin_cos_angle inc_angle;

      yac_comm_unpack(buffer, buffer_size, &position, &tgt_global_id, 1,
                      COMM_INT, comm);
      yac_comm_unpack(buffer, buffer_size, &position, &(inc_angle.sin), 1,
                      COMM_DBLE, comm);
      yac_comm_unpack(buffer, buffer_size, &position, &(inc_angle.cos), 1,
                      COMM_DBLE, comm);

      int is_better;

      // if the new target point has a differnt distance than the current one
      if ((fabs(search_data.src_point_data[src_idx].circle.inc_angle.sin -
                inc_angle.sin) > 0.0) ||
          (fabs(search_data.src_point_data[src_idx].circle.inc_angle.cos -
                inc_angle.cos) > 0.0))
        is_better =
          compare_angles(search_data.src_point_data[src_idx].circle.inc_angle,
                         inc_angle) > 0;
      else
        is_better = tgt_global_id <
                    search_data.src_point_data[src_idx].tgt_point.global_id;

      if (is_better) {

        search_data.src_point_data[src_idx].tgt_point.rank = rank;
        search_data.src_point_data[src_idx].tgt_point.global_id = tgt_global_id;
        search_data.src_point_data[src_idx].circle.inc_angle = inc_angle;
      }
    }

    unsigned num_to_be_checked_ranks;

    yac_comm_unpack(buffer, buffer_size, &position, &num_to_be_checked_ranks,
                    1, COMM_UINT, comm);

    if (num_to_be_checked_ranks > 0) {
      ENSURE_ARRAY_SIZE(
        search_data.src_point_data[src_idx].to_be_checked_ranks,
        search_data.src_point_data[src_idx].to_be_checked_ranks_array_size,
        search_data.src_point_data[src_idx].num_to_be_checked_ranks +
        num_to_be_checked_ranks);
      unsigned * to_be_checked_ranks =
        search_data.src_point_data[src_idx].to_be_checked_ranks +
        search_data.src_point_data[src_idx].num_to_be_checked_ranks;
      yac_comm_unpack(
        buffer, buffer_size, &position, to_be_checked_ranks,
        num_to_be_checked_ranks, COMM_UINT, comm);

      qsort(to_be_checked_ranks, num_to_be_checked_ranks,
            sizeof(*to_be_checked_ranks), compare_uint);
      unsigned new_num_to_be_checked_ranks = 0;
      unsigned * checked_ranks =
        search_data.src_point_data[src_idx].checked_ranks;
      unsigned num_checked_ranks =
        search_data.src_point_data[src_idx].num_checked_ranks;

      for (unsigned i = 0, j = 0; i < num_to_be_checked_ranks; ++i) {

        while ((j < num_checked_ranks) &&
               (checked_ranks[j] < to_be_checked_ranks[i])) ++j;

        if ((j == num_checked_ranks) ||
            (checked_ranks[j] != to_be_checked_ranks[i])) {

          if (i != new_num_to_be_checked_ranks)
            to_be_checked_ranks[new_num_to_be_checked_ranks] =
              to_be_checked_ranks[i];

          new_num_to_be_checked_ranks++;
        }
      }

      search_data.src_point_data[src_idx].num_to_be_checked_ranks +=
        new_num_to_be_checked_ranks;

      if (new_num_to_be_checked_ranks > 0)
        qsort(search_data.src_point_data[src_idx].to_be_checked_ranks,
              search_data.src_point_data[src_idx].num_to_be_checked_ranks,
              sizeof(*to_be_checked_ranks), compare_uint);

      size_t num_to_be_checked_ranks_ =
        (size_t)(search_data.src_point_data[src_idx].num_to_be_checked_ranks);

      yac_remove_duplicates_uint(
        search_data.src_point_data[src_idx].to_be_checked_ranks,
        &num_to_be_checked_ranks_);

      search_data.src_point_data[src_idx].num_to_be_checked_ranks =
        (unsigned)num_to_be_checked_ranks_;
    }
  }
}

static int
pack_and_send_requests(struct intermediate_method_data * data);

static void
recv_request_answer_callback(struct communicator * comm,
                             struct comm_message msg, void * user_data);

static void
recv_request_answer_cont_callback(struct communicator * comm,
                                  struct comm_message msg, void * user_data) {

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)user_data;

  unpack_search_request_answer(data->search_data, msg.buffer, msg.count, comm);

  data->num_open_remote_requests--;

  // if all requests answers have been received
  if (data->num_open_remote_requests == 0) {

    // pack and send new requires if necessary, otherwise continue with the
    // next step of the interpolation
    if (!pack_and_send_requests(data))
      yac_comm_barrier_callback(data->tags.continue_barrier, comm, data,
                                do_search_spmap_continue, MSG_PRIO_MIDDLE);

  } else {
    // set up receive for next answer
    yac_comm_recv_callback(1, COMM_UINT, COMM_ANYSOURCE,
                           data->tags.request_answer_header, comm,
                           (void*)data, recv_request_answer_callback, NULL,
                           MSG_PRIO_MIDDLE);
  }
}

static void
recv_request_answer_callback(struct communicator * comm,
                             struct comm_message msg, void * user_data) {

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)user_data;

  unsigned recv_buffer_size = ((unsigned*)(msg.buffer))[0];

  // receive actual answer
  yac_comm_recv_callback(recv_buffer_size, COMM_PACKED, msg.src_rank,
                         data->tags.request_answer_data, comm, data,
                         recv_request_answer_cont_callback, NULL,
                         MSG_PRIO_MIDDLE);
}

static int
pack_and_send_requests(struct intermediate_method_data * data) {

  struct interpolation_method_spmap * method_spmap =
    (struct interpolation_method_spmap*)(data->method_stack[0]);
  struct domain * domain = method_spmap->domain;
  struct communicator * domain_comm = yac_get_domain_comm(domain);
  unsigned domain_size = yac_get_comm_size(domain_comm);

  struct search_data * search_data = &(data->search_data);
  struct src_point_search_data * src_point_data = search_data->src_point_data;
  unsigned num_src_points = search_data->num_src_points;

  unsigned * num_src_per_rank = xcalloc(domain_size, sizeof(*num_src_per_rank));

  // count the number of source points per ranks
  for (unsigned src_idx = 0; src_idx < num_src_points; ++src_idx) {

    unsigned * to_be_checked_ranks =
      src_point_data[src_idx].to_be_checked_ranks;
    unsigned num_to_be_checked_ranks =
      src_point_data[src_idx].num_to_be_checked_ranks;

    for (unsigned i = 0; i < num_to_be_checked_ranks; ++i)
      ++num_src_per_rank[to_be_checked_ranks[i]];
  }

  unsigned total_num_points_to_send = 0;
  for (unsigned i = 0; i < domain_size; ++i) {
    total_num_points_to_send += num_src_per_rank[i];
    data->num_open_remote_requests += num_src_per_rank[i] > 0;
  }

  if (total_num_points_to_send == 0) {

    free(num_src_per_rank);
    return 0;
  }

  // number of bytes required to pack each source point
  int src_point_pack_size =
    yac_comm_pack_size(2, COMM_INT, domain_comm);  // global_id (src + tgt)
  src_point_pack_size +=
    yac_comm_pack_size(5, COMM_DBLE, domain_comm); // bnd_circle

  int send_buffer_size = src_point_pack_size * total_num_points_to_send;
  char * send_buffer = xmalloc(send_buffer_size);
  int * position = xmalloc(domain_size * sizeof(*position));
  int accum = 0;

  for (unsigned i = 0; i < domain_size; ++i) {
    position[i] = accum;
    accum += num_src_per_rank[i] * src_point_pack_size;
  }

  // pack the data
  for (unsigned src_idx = 0; src_idx < num_src_points; ++src_idx) {

    unsigned * to_be_checked_ranks =
      src_point_data[src_idx].to_be_checked_ranks;
    unsigned num_to_be_checked_ranks =
      src_point_data[src_idx].num_to_be_checked_ranks;
    int src_global_id = src_point_data[src_idx].global_id;
    int tgt_global_id = src_point_data[src_idx].tgt_point.global_id;
    double * base_vector = &(src_point_data[src_idx].circle.base_vector[0]);
    struct sin_cos_angle inc_angle = src_point_data[src_idx].circle.inc_angle;

    for (unsigned i = 0; i < num_to_be_checked_ranks; ++i) {

      yac_comm_pack(&src_global_id, 1, COMM_INT, send_buffer, send_buffer_size,
                    position + to_be_checked_ranks[i], domain_comm);
      yac_comm_pack(&tgt_global_id, 1, COMM_INT, send_buffer, send_buffer_size,
                    position + to_be_checked_ranks[i], domain_comm);
      yac_comm_pack(base_vector, 3, COMM_DBLE, send_buffer, send_buffer_size,
                    position + to_be_checked_ranks[i], domain_comm);
      yac_comm_pack(&(inc_angle.sin), 1, COMM_DBLE, send_buffer,
                    send_buffer_size, position + to_be_checked_ranks[i],
                    domain_comm);
      yac_comm_pack(&(inc_angle.cos), 1, COMM_DBLE, send_buffer,
                    send_buffer_size, position + to_be_checked_ranks[i],
                    domain_comm);
    }

    src_point_data[src_idx].checked_ranks =
      xrealloc(src_point_data[src_idx].checked_ranks,
              (src_point_data[src_idx].num_checked_ranks +
               src_point_data[src_idx].num_to_be_checked_ranks) *
              sizeof(*(src_point_data[src_idx].checked_ranks)));
    memcpy(src_point_data[src_idx].checked_ranks +
           src_point_data[src_idx].num_checked_ranks,
           to_be_checked_ranks, num_to_be_checked_ranks *
           sizeof(*(src_point_data[src_idx].checked_ranks)));
    src_point_data[src_idx].num_checked_ranks += num_to_be_checked_ranks;
    qsort(src_point_data[src_idx].checked_ranks,
          src_point_data[src_idx].num_checked_ranks,
          sizeof(*(src_point_data[src_idx].checked_ranks)),
          compare_uint);

    if (src_point_data[src_idx].num_checked_ranks > domain_size)
      yac_internal_abort_message(
        "ERROR(pack_and_send_requests): internal error", __FILE__, __LINE__);

    src_point_data[src_idx].num_to_be_checked_ranks = 0;
  }

  // send the data
  char * curr_send_buffer = send_buffer;
  for (unsigned rank = 0; rank < domain_size; ++rank) {

    if (num_src_per_rank[rank] == 0) continue;

    unsigned send_size = num_src_per_rank[rank] * src_point_pack_size;
    // send header
    yac_comm_bsend(&send_size, 1, COMM_UINT, rank,
                   data->tags.request_header, domain_comm);

    // send request data
    yac_comm_bsend(curr_send_buffer, send_size, COMM_PACKED, rank,
                   data->tags.request_data, domain_comm);

    curr_send_buffer += send_size;
  }

  free(position);
  free(send_buffer);
  free(num_src_per_rank);

  // set up receive for the answers
  yac_comm_recv_callback(1, COMM_UINT, COMM_ANYSOURCE,
                         data->tags.request_answer_header, domain_comm,
                         (void*)data, recv_request_answer_callback, NULL,
                         MSG_PRIO_MIDDLE);

  return 1;
}

static void
pack_and_send_search_request_answer(
  struct search_data * search_data, unsigned local_rank, unsigned remote_rank,
  struct communicator * comm, unsigned header_tag, unsigned data_tag) {

  unsigned buffer_size = 0;

  struct src_point_search_data * curr_src_point_data =
    search_data->src_point_data;

  for (unsigned src_idx = 0; src_idx < search_data->num_src_points;
       ++src_idx, ++curr_src_point_data) {

    // if we need to send the result of the current source cell
    if ((curr_src_point_data->tgt_point.rank != local_rank) &&
        (curr_src_point_data->num_to_be_checked_ranks == 0)) 
      continue;

    buffer_size += yac_comm_pack_size(1, COMM_INT, comm);  // src global_id

    if (curr_src_point_data->tgt_point.rank != local_rank)
      buffer_size += yac_comm_pack_size(1, COMM_UINT, comm); // -1
    else {
      buffer_size += yac_comm_pack_size(1, COMM_UINT, comm); // rank
      buffer_size += yac_comm_pack_size(1, COMM_INT, comm);  // tgt_global_id
      buffer_size += yac_comm_pack_size(2, COMM_DBLE, comm); // inc_angle
    }

    if (curr_src_point_data->num_to_be_checked_ranks == 0)
      buffer_size += yac_comm_pack_size(1, COMM_UINT, comm); // -1
    else {
      buffer_size +=
        yac_comm_pack_size(1, COMM_UINT, comm); // num_to_be_checked_ranks
      buffer_size +=
        yac_comm_pack_size(
          curr_src_point_data->num_to_be_checked_ranks,
          COMM_UINT, comm);                     // to_be_checked_ranks
    }
  }

  char * buffer = xmalloc(buffer_size);
  int position = 0;

  curr_src_point_data = search_data->src_point_data;
  for (unsigned src_idx = 0; src_idx < search_data->num_src_points;
       ++src_idx, ++curr_src_point_data) {

    // if we need to send the result of the current source cell
    if ((curr_src_point_data->tgt_point.rank != local_rank) &&
        (curr_src_point_data->num_to_be_checked_ranks == 0)) 
      continue;

    yac_comm_pack(&(curr_src_point_data->global_id), 1, COMM_INT, buffer,
                  buffer_size, &position, comm);

    if (curr_src_point_data->tgt_point.rank != local_rank) {
      unsigned dummy = -1;
      yac_comm_pack(&dummy, 1, COMM_UINT, buffer, buffer_size, &position, comm);
    } else {
      yac_comm_pack(&(curr_src_point_data->tgt_point.rank), 1, COMM_UINT,
                    buffer, buffer_size, &position, comm);
      yac_comm_pack(&(curr_src_point_data->tgt_point.global_id), 1, COMM_INT,
                    buffer, buffer_size, &position, comm);
      yac_comm_pack(&(curr_src_point_data->circle.inc_angle.sin), 1, COMM_DBLE,
                    buffer, buffer_size, &position, comm);
      yac_comm_pack(&(curr_src_point_data->circle.inc_angle.cos), 1, COMM_DBLE,
                    buffer, buffer_size, &position, comm);
    }

    yac_comm_pack(&(curr_src_point_data->num_to_be_checked_ranks), 1,
                  COMM_UINT, buffer, buffer_size, &position, comm);
    if (curr_src_point_data->num_to_be_checked_ranks > 0) {
      yac_comm_pack(curr_src_point_data->to_be_checked_ranks,
                    curr_src_point_data->num_to_be_checked_ranks,
                    COMM_UINT, buffer, buffer_size, &position, comm);
    }
    free(curr_src_point_data->to_be_checked_ranks);
  }

  yac_comm_bsend(&buffer_size, 1, COMM_UINT, remote_rank, header_tag, comm);
  yac_comm_bsend(buffer, buffer_size, COMM_PACKED, remote_rank, data_tag, comm);

  free(buffer);
  free(search_data->coordinates_xyz);
  free(search_data->src_point_data);
}

static void
recv_search_request_callback_cancel(struct communicator * comm,
                                    void * user_data) {

  UNUSED(comm);

  struct search_request_user_data * search_request_data =
    (struct search_request_user_data *)user_data;

  free(search_request_data->halo_cells);
  free(search_request_data);
}

static void
recv_search_request_callback(struct communicator * comm,
                             struct comm_message msg, void * user_data);

static void
recv_search_request_cont_callback(struct communicator * comm,
                                  struct comm_message msg, void * user_data) {

  struct search_request_user_data * search_request_data =
    (struct search_request_user_data *)user_data;

  // number of bytes required to pack each source point
  int src_point_pack_size = yac_comm_pack_size(2, COMM_INT, comm); // global_id (src + tgt)
  src_point_pack_size += yac_comm_pack_size(5, COMM_DBLE, comm); // bnd_circle

  struct search_data search_data;

  search_data.num_src_points = msg.count / src_point_pack_size;
  search_data.src_point_data =
    xmalloc(search_data.num_src_points * sizeof(*(search_data.src_point_data)));

  int position = 0;

  search_data.coordinates_xyz =
    xmalloc(search_data.num_src_points * sizeof(*(search_data.coordinates_xyz)));

  // unpack and initialise search_data
  for (unsigned i = 0; i < search_data.num_src_points; ++i) {

    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &(search_data.src_point_data[i].global_id), 1, COMM_INT,
                    comm);
    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &(search_data.src_point_data[i].tgt_point.global_id), 1,
                    COMM_INT, comm);
    yac_comm_unpack(msg.buffer, msg.count, &position,
                    search_data.src_point_data[i].circle.base_vector, 3,
                    COMM_DBLE, comm);
    search_data.coordinates_xyz[i][0] =
      search_data.src_point_data[i].circle.base_vector[0];
    search_data.coordinates_xyz[i][1] =
      search_data.src_point_data[i].circle.base_vector[1];
    search_data.coordinates_xyz[i][2] =
      search_data.src_point_data[i].circle.base_vector[2];
    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &(search_data.src_point_data[i].circle.inc_angle.sin), 1,
                    COMM_DBLE, comm);
    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &(search_data.src_point_data[i].circle.inc_angle.cos), 1,
                    COMM_DBLE, comm);
    search_data.src_point_data[i].tgt_point.rank = msg.src_rank;
    search_data.src_point_data[i].checked_ranks = NULL;
    search_data.src_point_data[i].num_checked_ranks = 0;
    search_data.src_point_data[i].to_be_checked_ranks = NULL;
    search_data.src_point_data[i].to_be_checked_ranks_array_size = 0;
    search_data.src_point_data[i].num_to_be_checked_ranks = 0;
  }

  // search through local target points to see whether there is a better match
  search_local(&search_data, search_request_data->tgt_points, msg.dst_rank);

  // check the halos
  check_for_potential_remote_points(&search_data,
                                    search_request_data->halo_cells,
                                    search_request_data->num_halo_cells,
                                    msg.src_rank);

  // pack and send search request answer (only cells whose target point rank is
  // the local rank or have updated to_be_checked_ranks array)
  pack_and_send_search_request_answer(
    &search_data, msg.dst_rank, msg.src_rank, comm,
    search_request_data->tags.request_answer_header,
    search_request_data->tags.request_answer_data);

  // set up receive for next request
  yac_comm_recv_callback(1, COMM_UINT, COMM_ANYSOURCE,
                         search_request_data->tags.request_header,
                         comm, user_data, recv_search_request_callback,
                         recv_search_request_callback_cancel, MSG_PRIO_HIGH);
}

static void
recv_search_request_callback(struct communicator * comm,
                             struct comm_message msg, void * user_data) {

  struct search_request_user_data * search_request_data =
    (struct search_request_user_data *)user_data;

  unsigned recv_buffer_size = ((unsigned*)(msg.buffer))[0];

  // set up receive for search request data
  yac_comm_recv_callback(recv_buffer_size, COMM_PACKED, msg.src_rank,
                         search_request_data->tags.request_data,
                         comm, user_data, recv_search_request_cont_callback,
                         NULL, MSG_PRIO_HIGH);
}

static void
do_search_interpolation_method_spmap(
  struct interpolation_method ** method,
  struct interpolation_plan_element ** plan_stack,
  int * global_ids, double (*coordinates_xyz)[3],
  enum yac_location tgt_location, unsigned num_points,
  unsigned * num_interpolated_points,
  struct interpolation_method_data_request ** data_request) {

  if ((coordinates_xyz == NULL) && num_points > 0)
    yac_internal_abort_message(
      "ERROR(do_search_interpolation_method_spmap): "
      "missing target point coordinates (coordinates_xyz == NULL)",
      __FILE__, __LINE__);

  struct interpolation_method_spmap * method_spmap =
    (struct interpolation_method_spmap *)(method[0]);

  method_spmap->data_request = data_request;

  // initialise intermediate method data

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)xmalloc(1 * sizeof(*data));

  data->method_stack = method;
  data->plan_stack = plan_stack;
  data->num_interpolated_points = num_interpolated_points;
  *num_interpolated_points = 0;
  data->tgt_points.coordinates_xyz = coordinates_xyz;
  data->tgt_points.global_ids = global_ids;
  data->tgt_points.num_tgt_points = num_points;
  data->tgt_location = tgt_location;
  data->tags = method_spmap->tags;
  data->num_open_remote_requests = 0;

  // sort target points by global_ids
  yac_sort_points(global_ids, coordinates_xyz, (size_t)num_points);

  // generate search tree for the target points
  data->tgt_points.point_sphere_part =
    yac_point_sphere_part_search_new((size_t)num_points, coordinates_xyz);

  // get all non-masked local source core points and sort them by their
  // global ids
  init_local_search_data(&(data->search_data), method_spmap->domain,
                         method_spmap->points, method_spmap->masks,
                         method_spmap->points_idx, method_spmap->src_location);
  // get all halo cells
  struct halo_cells * halo_cells;
  unsigned num_halo_cells;
  init_halo_cells(&halo_cells, &num_halo_cells, method_spmap->domain);

  // set up receive for requests from other processes
  {
    struct search_request_user_data * search_request_data =
      xmalloc(1*sizeof(*search_request_data));
    search_request_data->tgt_points = data->tgt_points;
    search_request_data->halo_cells = halo_cells;
    search_request_data->num_halo_cells = num_halo_cells;
    search_request_data->tags = data->tags;
    yac_comm_recv_callback(1, COMM_UINT, COMM_ANYSOURCE,
                           search_request_data->tags.request_header,
                           yac_get_domain_comm(method_spmap->domain),
                           search_request_data, recv_search_request_callback,
                           recv_search_request_callback_cancel, MSG_PRIO_HIGH);
  }
  // searches for nearest target points for all local source points
  search_local(&(data->search_data), data->tgt_points, -1);

  // checks for each local source point whether it potentially has an overlap
  // with target points on a neighbouring process
  check_for_potential_remote_points(&(data->search_data), halo_cells,
                                    num_halo_cells, -1);

  // pack and send new requires if necessary, otherwise continue with the next
  // step of the interpolation
  if (!pack_and_send_requests(data))
    yac_comm_barrier_callback(data->tags.continue_barrier,
                              yac_get_domain_comm(method_spmap->domain), data,
                              do_search_spmap_continue, MSG_PRIO_MIDDLE);
}

static void
enddef_interpolation_method_spmap(struct interpolation_method ** method) {

  struct interpolation_method_spmap * method_spmap =
   (struct interpolation_method_spmap*)(method[0]);

  struct communicator * domain_comm = yac_get_domain_comm(method_spmap->domain);

   while(yac_comm_wait(domain_comm,
                       method_spmap->tags.request_answer_header) ||
         yac_comm_wait(domain_comm,
                       method_spmap->tags.request_answer_data) ||
         yac_comm_wait(domain_comm,
                       method_spmap->tags.continue_barrier) ||
         yac_comm_wait(domain_comm,
                       method_spmap->tags.create_dist_dir) ||
         yac_comm_wait(domain_comm,
                       method_spmap->tags.distribute_results));

  if (method[1] != NULL)
    method[1]->vtable->enddef(method + 1);
}

static void
get_source_exchange_data_interpolation_method_spmap(
  struct interpolation_method ** method,
  struct interpolation_method_data_request *** data_request) {

  struct interpolation_method_spmap * method_spmap =
    (struct interpolation_method_spmap*)(method[0]);

  *data_request = method_spmap->data_request;
  method_spmap->data_request = NULL;
}

static void
delete_interpolation_method_spmap(struct interpolation_method * method) {

  free(method);
}

static struct global_search *
get_global_search(struct interpolation_method * method) {

   return ((struct interpolation_method_spmap*)(method))->search;
}

static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data) {


  struct interpolation_method_spmap * method_spmap =
    (struct interpolation_method_spmap *)method;

  method_spmap->finish_callback = callback;
  method_spmap->finish_user_data = user_data;
}

