/**
 * @file interpolation_plan.h
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INTERPLOATION_PLAN_H
#define INTERPLOATION_PLAN_H

/** \example test_interpolation_plan.c
 * This shows how to use interpolation plan elements
 */

/**
 * interpolation plan element data structure
 */
struct interpolation_plan_element {

   void (*do_interpolation)(struct interpolation_plan_element ** plan_stack,
                            double *** input_dat, double * output_data);
   void (*get_link_data)(struct interpolation_plan_element ** plan_stack,
                         int *** global_indices,
                         unsigned * num_targets_per_plan,
                         unsigned * num_links_per_plan,
                         unsigned * num_weights_per_link_per_plan,
                         unsigned ** num_links_per_tgt_per_plan,
                         int ** src_indices_per_plan,
                         unsigned ** pointset_indices_per_plan,
                         double ** weights_per_plan);
   unsigned (*get_num_fixed_values)(struct interpolation_plan_element ** plan_stack);
   void (*get_fixed_data)(struct interpolation_plan_element ** plan_stack,
                          int const * global_ids,
                          int ** global_tgt_indices_per_plan,
                          unsigned * num_tgt_indices_per_plan,
                          double * fixed_value_per_plan);
   void (*free_interpolation_plan_element)(struct interpolation_plan_element * plan);

   void * child_data;
};

/**
 * generates an interpolation plan element that assigns a given fixed
 * value to each point
 *
 * @param[in,out] plan        interpolation plan element
 * @param[in]     num_points  number of points to be interpolated
 * @param[in]     fixed_value fixed value to be assigned to each point
 */
void yac_init_interpolation_plan_element_fixed(struct interpolation_plan_element * plan,
                                               unsigned num_points,
                                               double fixed_value);

/**
 * generates an interpolation plan element that directly assigns a point from
 * the input to the output
 *
 * @param[in,out] plan                    interpolation plan element
 * @param[in]     num_interpolated_points number of points to be interpolated
 * @param[in]     src_indices             specifies the source point of each
 *                                        target point
 * @param[in]     num_remote_points_per_points_set number of source points from
 *                                                 other source processes that
 *                                                 are required for the
 *                                                 interpolation
 * @param[in]     num_pointsets           number of struct points per subdomain
 *
 * \remarks src_indices has size 3*num_interpolated_points, each entry contains
 *          three values, which are the indices to the input_data argument of
 *          routine \ref yac_execute_interpolation_plan_element
 */
void yac_init_interpolation_plan_element_point(
  struct interpolation_plan_element * plan, unsigned num_interpolated_points,
  unsigned * src_indices, unsigned * num_remote_points_per_points_set,
  unsigned num_pointsets);

/**
 * generates an interpolation plan element that assigns a sum to each point
 *
 * @param[in,out] plan                    interpolation plan element
 * @param[in]     num_interpolated_points number of points to be interpolated
 * @param[in]     src_indices             specifies the source points that are
 *                                        to be used for the interpolation of each
 *                                        target point
 * @param[in]     num_src_per_tgt         contains for each target point the
 *                                        number of source points that are to be
 *                                        used for its interpolation
 * @param[in]     num_remote_points_per_points_set number of source points from other
 *                                        source processes that are required for
 *                                        the interpolation
 * @param[in]     num_pointsets           number of struct points per subdomain
 *
 * \remarks src_indices has size 3*SUM(num_src_per_tgt[0..num_interpolated_points-1]),
 *          each entry contains three values, which are the indices to the
 *          input_data argument of routine \ref yac_execute_interpolation_plan_element
 */
void yac_init_interpolation_plan_element_pointsum(struct interpolation_plan_element * plan,
                                                  unsigned num_interpolated_points,
                                                  unsigned * src_indices,
                                                  unsigned * num_src_per_tgt,
                                                  unsigned * num_remote_points_per_points_set,
                                                  unsigned num_pointsets);

/**
 * generates an interpolation plan element that assigns a weighted sum
 * to each point
 *
 * @param[in,out] plan                    interpolation plan element
 * @param[in]     num_interpolated_points number of points to be interpolated
 * @param[in]     src_indices             specifies the source points that are
 *                                        to be used for the interpolation of each
 *                                        target point
 * @param[in]     num_src_per_tgt         contains for each target point the
 *                                        number of source points that are to be
 *                                        used for its interpolation
 * @param[in]     weights                 weights for each source-target relation
 * @param[in]     num_remote_points_per_points_set number of source points from other
 *                                        source processes that are required for
 *                                        the interpolation
 * @param[in]     num_pointsets           number of struct points per subdomain
 *
 * \remarks src_indices has size 3*SUM(num_src_per_tgt[0..num_interpolated_points-1]),
 *          each entry contains three values, which are the indices to the
 *          input_data argument of routine \ref yac_execute_interpolation_plan_element
 */
void yac_init_interpolation_plan_element_wpointsum(struct interpolation_plan_element * plan,
                                                   unsigned num_interpolated_points,
                                                   unsigned * src_indices,
                                                   unsigned * num_src_per_tgt, double * weights,
                                                   unsigned * num_remote_points_per_points_set,
                                                   unsigned num_pointsets);

/**
 * generates an interpolation plan that consists of multiple individual
 * interpolation plans
 * @param[in,out] plan      interpolation plan element
 * @param[in]     num_plans number of entries in plans
 * @param[in]     plans     array containing num_plans interpolation plan
 *                          elements that are to be combined into on
 *                          interpolation plan element
 * @remark the entries in plans will be used by the generated interpolation plan
 *         element and will automatically be free when this element is freed
 */
void yac_init_interpolation_plan_element_collection(struct interpolation_plan_element * plan,
                                                    unsigned num_plans,
                                                    struct interpolation_plan_element ** plans);

/**
 * does an interpolation based on interpolation plans on the stack
 *
 * @param[in]  plan_stack  stack containing all interpolation plans to be executed\n
 *                         the last entry has to be NULL
 * @param[in]  input_data  field data of the local process\n
 *                         the first entry in input_data contains an array with data that
 *                         was received from other processes as specified by the
 *                         interpolation methods
 * @param[out] output_data interpolated points/cells, the order is specified by the
 *                         interpolation method that generated the respective plan
 */
void yac_execute_interpolation_plan_element(struct interpolation_plan_element ** plan_stack,
                                            double *** input_data, double * output_data);

/**
 * returns link data for a given interpolation plan stack
 *
 * @param[in] plan_stack               stack containing all interpolation plans to
 *                                     be executed\n the last entry has to be NULL
 * @param[in] global_indices           global indices of the local process\n the
 *                                     first entry in global_indices contains an
 *                                     array with remote global indices requested
 *                                     by the respective interpolation method (see
 *                                     argument input_data of
 *                                     \ref yac_execute_interpolation_plan_element)
 * @param[in]  num_pointsets           number of point set
 * @param[out] num_targets             number of interpolated target points
 * @param[out] num_links_per_point_set number of links per point set
 * @param[out] num_weights_per_link    number of weights stored per link
 * @param[out] num_links_per_tgt       number of links stored per target point
 * @param[out] src_indices             source indices of all links
 * @param[out] pointset_indices        point set index for each source point
 * @param[out] weights                 weights for each link
 *
 * @remark the global target index for the links correspond to the order of the
 *         indices of the array "global_ids" returned by
 *         \ref yac_do_search_interpolation_method
 * @remark the routine allocates all output arrays internally and the user must
 *         ensure that they are freed once they are not needed anymore
 */
void yac_get_link_data(struct interpolation_plan_element ** plan_stack,
                       int *** global_indices,
                       unsigned num_pointsets,
                       unsigned * num_targets,
                       unsigned ** num_links_per_point_set,
                       unsigned * num_weights_per_link,
                       unsigned ** num_links_per_tgt, int ** src_indices,
                       unsigned ** pointset_indices, double ** weights);

/**
 * returns data for a given interpolation plan stack on target points that are
 * supposed to receive a fixed value
 *
 * @param[in]  plan_stack              stack containing all interpolation plans to
 *                                     be executed\n the last entry has to be NULL
 * @param[in]  global_ids              global_ids array as returned by
 *                                     \ref yac_do_search_interpolation_method
 * @param[out] global_tgt_indices      global target indices of all target points
 *                                     that are supposed to receive a fixed value
 * @param[out] num_tgt_indices         number of entries in global_tgt_indices
 * @param[out] num_tgt_per_fixed_value number of global target indices for each
 *                                     fixed value (entries in global_tgt_indices
 *                                     are sorted accordingly)
 * @param[out] fixed_values            array of all fixed values that are to be used
 * @param[out] num_fixed_values        number of entries in fixed_values
 *
 * @remark the routine allocates all output arrays internally and the user must
 *         ensure that they are freed once they are not needed anymore
 */
void yac_get_fixed_data(struct interpolation_plan_element ** plan_stack,
                        int const * global_ids, int ** global_tgt_indices,
                        unsigned * num_tgt_indices,
                        unsigned ** num_tgt_per_fixed_value,
                        double ** fixed_values,
                        unsigned * num_fixed_values);

/**
 * frees all memory that is allocated within the given interpolation plan
 */
void yac_free_interpolation_plan_element(struct interpolation_plan_element * plan);

#endif // INTERPLOATION_PLAN_H

