/**
 * @file nnn_search.c
 *
 * @copyright Copyright  (C)  2017 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Thomas Jahns <jahns@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <string.h>

#include "nnn_search.h"
#include "communicator.h"
#include "domain.h"
#include "points.h"
#include "mask.h"
#include "sphere_part.h"
#include "utils.h"
#include "ensure_array_size.h"

/** \example test_nnn_search_parallel.c
 * This contains a parallel example for the n-nearest-neighbour search.
 * The example uses four source processes and one target process.
 */

enum nnn_search_state {
  INITIALISED,
  ACTIVE,
};

enum nnn_search_msg_types {
  RESULTS_MSG_TYPE = 0,
  SEARCH_REQUEST_MSG_TYPE = 1,
};

struct nnn_search_tags {
  unsigned results_or_search_request_header;
  unsigned search_request;
  unsigned finished;
  unsigned finished_cont;
  unsigned results;
};

struct nnn_search {
  struct domain * domain;
  size_t num_domain_neighbours;      //!< number of neighbour processes
  unsigned * domain_neighbour_ranks; //!< ranks of all neighbour processes

  struct communicator * comm;
  struct nnn_search_tags tags;

  enum nnn_search_state state;

  struct point_sphere_part_search * non_halo_point_search;
  unsigned * non_halo_local_ids;
  size_t * num_non_halo_local_ids_per_subdomain_prefix;
  struct point_sphere_part_search ** halo_point_search;

  enum yac_location location;

};

struct request_record {
  unsigned rank; //!< rank to which (a) request(s) have been sent
  int * global_ids; //!< global ids of request points that have been sent
                    //!< to this rank
  size_t global_ids_array_size;
  size_t num_global_ids; //!< number of entries in sent_global_ids
};

struct nnn_search_data {

  struct nnn_search * search;

  struct request_point * request_points;

  void (*finish_callback)(void *, struct request_point*);
  void * finish_user_data;

  size_t n;          //!< number of nearest neighbours
  size_t num_points; //!< number of request points

  struct request_record * request_records;
  size_t num_request_records;
  size_t num_open_remote_requests; //!< number for forwarded search requests
                                   //!< that have not yet been answered

  int * reorder_idx;
};

struct temp_search_request_data {

  struct nnn_search_data * search_data;
  size_t num_points;
};

static void generate_tag_string(char * buf, size_t buffer_size,
                                char const * tag_base_str,
                                char const * unique_string,
                                char const * tag_str) {

   int ret_val =
    snprintf(buf, buffer_size, "%s%s%s", tag_base_str, unique_string, tag_str);

   if (ret_val < 0 || (size_t)ret_val >= buffer_size)
      yac_internal_abort_message("ERROR(generate_tag_string): truncation",
                                 __FILE__, __LINE__);
}

static void generate_nnn_search_tags(
  struct nnn_search_tags * tags, char const * unique_string,
  struct communicator * comm) {

  char const * str_base_str                         = "nnn search tags";
  char const * str_generator_string                 = "generator string";
  char const * str_results_or_search_request_header = "results or search request header";
  char const * str_search_request                   = "search request";
  char const * str_finished                         = "finished";
  char const * str_finished_cont                    = "finished continue";
  char const * str_results                          = "results";

  unsigned max_tag_strlen = strlen(str_results_or_search_request_header);

  unsigned max_str_buf_size = max_tag_strlen;
  max_str_buf_size += strlen(str_base_str);
  max_str_buf_size += strlen(unique_string) + 1;

  char * unique_generator_string = xmalloc(max_str_buf_size);
  generate_tag_string(unique_generator_string, max_str_buf_size, str_base_str,
                      unique_string, str_generator_string);

  char * tag_strings[5];
  size_t num_tags = 0;
  tag_strings[num_tags] = xmalloc(max_str_buf_size);
  generate_tag_string(tag_strings[num_tags++], max_str_buf_size, str_base_str,
                      unique_string, str_results_or_search_request_header);

  tag_strings[num_tags] = xmalloc(max_str_buf_size);
  generate_tag_string(tag_strings[num_tags++], max_str_buf_size, str_base_str,
                      unique_string, str_search_request);

  tag_strings[num_tags] = xmalloc(max_str_buf_size);
  generate_tag_string(tag_strings[num_tags++], max_str_buf_size, str_base_str,
                      unique_string, str_finished);

  tag_strings[num_tags] = xmalloc(max_str_buf_size);
  generate_tag_string(tag_strings[num_tags++], max_str_buf_size, str_base_str,
                      unique_string, str_finished_cont);

  tag_strings[num_tags] = xmalloc(max_str_buf_size);
  generate_tag_string(tag_strings[num_tags++], max_str_buf_size, str_base_str,
                      unique_string, str_results);

  unsigned tag_indices[sizeof(tag_strings) / sizeof(tag_strings[0])];

  yac_get_unique_comm_tags(
    comm, unique_generator_string, (char const **)tag_strings, num_tags,
    &(tag_indices[0]));

  free(unique_generator_string);

  tags->results_or_search_request_header = tag_indices[0];
  tags->search_request = tag_indices[1];
  tags->finished = tag_indices[2];
  tags->finished_cont = tag_indices[3];
  tags->results = tag_indices[4];

  for (size_t i = 0; i < num_tags; ++i) free(tag_strings[i]);
}

static void generate_point_sphere_part_searches(
  struct nnn_search * search, struct points *** points, struct mask *** masks,
  unsigned points_idx) {

  struct domain * domain = search->domain;
  size_t num_subdomains = (size_t)(yac_get_num_subdomains(domain));
  size_t num_neighbours = search->num_domain_neighbours;
  enum yac_location location = points[0][points_idx]->location;

  // allocate local_ids, coordinates_xyz
  unsigned * local_ids;
  double (*coordinates_xyz)[3];
  size_t total_num_local_ids = 0;
  {
    size_t max_num_local_ids = 0;
    for (size_t subdomain_id = 0; subdomain_id < num_subdomains;
         ++subdomain_id) {
      size_t data_size =
        (size_t)yac_get_subdomain_data_size(
          domain, (unsigned)subdomain_id, location);
      if (data_size > max_num_local_ids) max_num_local_ids = data_size;
      total_num_local_ids += data_size;
    }
    local_ids = xmalloc(max_num_local_ids * sizeof(*local_ids));
    coordinates_xyz = xmalloc(total_num_local_ids * sizeof(*coordinates_xyz));
  }

  // generate sphere part point search for all non halo points
  unsigned * non_halo_local_ids =
    xmalloc(total_num_local_ids * sizeof(*(non_halo_local_ids)));
  size_t * num_non_halo_local_ids_per_subdomain_prefix = NULL;
  if (num_subdomains > 1) {
    num_non_halo_local_ids_per_subdomain_prefix =
      xmalloc((num_subdomains + 1) *
              sizeof(*(num_non_halo_local_ids_per_subdomain_prefix)));
  }

  size_t num_non_masked_points = 0;
  for (size_t subdomain_id = 0; subdomain_id < num_subdomains; ++subdomain_id) {

    struct mask * mask = (masks != NULL)?masks[subdomain_id][points_idx]:NULL;

    size_t num_points =
      (size_t)yac_get_subdomain_data_size(domain, subdomain_id, location);
    num_points -=
      (size_t)yac_get_num_total_halo_points(domain, subdomain_id, location);

    // get the local ids for the non halo points of the current subdomain
    yac_get_non_halo_points(
      domain, subdomain_id, location, NULL, local_ids);

    // get the coordinates for the non halo points
    size_t curr_num_non_masked_points = 0;
    for (size_t i = 0; i < num_points; ++i) {

      // if the point is masked
      if ((mask != NULL) && (mask->mask[local_ids[i]])) continue;

      double lonlat[2];
      yac_get_point_coordinates(
        points[subdomain_id][points_idx], local_ids[i], lonlat);
      LLtoXYZ(lonlat[0], lonlat[1], coordinates_xyz[num_non_masked_points]);
      non_halo_local_ids[num_non_masked_points] = local_ids[i];
      num_non_masked_points++;
      curr_num_non_masked_points++;
    }
    if (num_non_halo_local_ids_per_subdomain_prefix != NULL) {
      num_non_halo_local_ids_per_subdomain_prefix[subdomain_id] =
        curr_num_non_masked_points;
      if (subdomain_id > 0)
        num_non_halo_local_ids_per_subdomain_prefix[subdomain_id] +=
          num_non_halo_local_ids_per_subdomain_prefix[subdomain_id - 1];
    }
  }
  struct point_sphere_part_search * non_halo_point_search =
    yac_point_sphere_part_search_new(
      num_non_masked_points, coordinates_xyz);
  if (num_non_masked_points)
    non_halo_local_ids =
      xrealloc(non_halo_local_ids, num_non_masked_points *
               sizeof(*non_halo_local_ids));

  // generate sphere part point search for all halo points
  struct point_sphere_part_search ** halo_point_search =
    xmalloc(num_neighbours * sizeof(*halo_point_search));
  for (size_t neigh_idx = 0; neigh_idx < num_neighbours; ++neigh_idx) {

    unsigned rank = search->domain_neighbour_ranks[neigh_idx];
    size_t num_halo_points = 0;

    for (size_t subdomain_id = 0; subdomain_id < num_subdomains;
         ++subdomain_id) {

      // get the local ids for the halo points of the current subdomain and
      // current domain neighbour
      unsigned num_points;
      yac_get_halo_points(domain, subdomain_id, location, NULL,
                          local_ids, &num_points, rank);

      // get the coordinates for the non halo points
      for (unsigned i = 0; i < num_points; ++i, ++num_halo_points) {
        double lonlat[2];
        yac_get_point_coordinates(
          points[subdomain_id][points_idx], local_ids[i], lonlat);
        LLtoXYZ(lonlat[0], lonlat[1], coordinates_xyz[num_halo_points]);
      }
    }
    halo_point_search[neigh_idx] =
      yac_point_sphere_part_search_new(num_halo_points, coordinates_xyz);
  }

  free(coordinates_xyz);
  free(local_ids);

  search->non_halo_point_search = non_halo_point_search;
  search->non_halo_local_ids = non_halo_local_ids;
  search->num_non_halo_local_ids_per_subdomain_prefix =
    num_non_halo_local_ids_per_subdomain_prefix;
  search->halo_point_search = halo_point_search;
}

static int compare_uint (const void * a, const void * b) {
  return ( *(unsigned*)a - *(unsigned*)b );
}

struct nnn_search * yac_nnn_search_new(
  struct domain * domain, const char * unique_string, struct points *** points,
  struct mask *** masks, unsigned num_pointsets, unsigned points_idx) {

  // ensure that the right point data is available
  if (num_pointsets == 0)
    yac_internal_abort_message(
      "ERROR(yac_nnn_search_new): missing data points.", __FILE__, __LINE__);

  if (points_idx >= num_pointsets)
    yac_internal_abort_message(
      "ERROR(yac_nnn_search_new): invalid points_idx.", __FILE__, __LINE__);

  if (points == NULL || points[0] == NULL || points[0][points_idx] == NULL)
    yac_internal_abort_message(
      "ERROR(yac_nnn_search_new): no point set data available",
      __FILE__, __LINE__);

  struct nnn_search * search = xmalloc(1 * sizeof(*search));

  search->domain = domain;
  search->comm = yac_get_domain_comm(domain);
  search->state = INITIALISED;
  search->location = points[0][points_idx]->location;

  // generate tags
  generate_nnn_search_tags(&(search->tags), unique_string, search->comm);

  unsigned num_domain_neighbours;
  yac_get_domain_neighbours(
    *domain, CELL, &(search->domain_neighbour_ranks), &num_domain_neighbours);
  search->num_domain_neighbours = (size_t)num_domain_neighbours;
  qsort(search->domain_neighbour_ranks, search->num_domain_neighbours,
        sizeof(*(search->domain_neighbour_ranks)), compare_uint);
  generate_point_sphere_part_searches(search, points, masks, points_idx);

  return search;
}

static void init_request_point(
  struct request_point * request_point, double coordinate_xyz[3],
  int global_id, double cos_angle) {

  request_point->global_id = global_id;
  request_point->coordinate_xyz[0] = coordinate_xyz[0];
  request_point->coordinate_xyz[1] = coordinate_xyz[1];
  request_point->coordinate_xyz[2] = coordinate_xyz[2];
  request_point->cos_angle = cos_angle;
}

static int
compare_request_points(const void * a, const void * b) {

  return ((struct request_point*)a)->global_id -
         ((struct request_point*)b)->global_id;
}

static struct request_point * generate_request_points(
  size_t num_points, double (*coordinates_xyz)[3],
  int * global_ids, int ** reorder_idx) {

  struct request_point * request_points =
    xmalloc(num_points * sizeof(*request_points));

  for (size_t i = 0; i < num_points; ++i)
    init_request_point(
      request_points+i, coordinates_xyz[i], global_ids[i], -1.0);

  int sort_flag = 0;
  for (size_t i = 1; i < num_points; ++i)
    if (global_ids[i-1] > global_ids[i]) sort_flag |= 1;

  if (sort_flag) {

    int * sorted_global_ids = xmalloc(num_points * sizeof(*sorted_global_ids));
    memcpy(
      sorted_global_ids, global_ids, num_points * sizeof(*sorted_global_ids));
    *reorder_idx = xmalloc(num_points * sizeof(**reorder_idx));
    for (size_t i = 0; i < num_points; ++i) (*reorder_idx)[i] = (int)i;
    yac_quicksort_index(sorted_global_ids, num_points, *reorder_idx);
    free(sorted_global_ids);
    qsort(request_points, num_points, sizeof(*request_points),
          compare_request_points);

  } else {
    *reorder_idx = NULL;
  }

  return request_points;
}

static int
compare_results(const void * a, const void * b) {

  int res = (((struct result_point*)b)->cos_angle >
             ((struct result_point*)a)->cos_angle) -
            (((struct result_point*)a)->cos_angle >
             ((struct result_point*)b)->cos_angle);
  if (res == 0)
    res = ((struct result_point*)a)->global_id -
          ((struct result_point*)b)->global_id;
  return res;
}

#define ITERATION_LENGTH (1024)

//! searches in the local data for request points
static void nnn_search_local(struct nnn_search_data * search_data) {

  struct nnn_search * search = search_data->search;
  size_t num_points = search_data->num_points;

  double (*curr_coordinates_xyz)[3] =
    xmalloc(MIN(ITERATION_LENGTH, num_points) * sizeof(*curr_coordinates_xyz));

  struct domain * domain = search->domain;
  size_t num_subdomains = (size_t)(yac_get_num_subdomains(domain));

  double * cos_angles = NULL;
  size_t cos_angles_array_size = 0;
  double (*result_coordinate_xyz)[3] = NULL;
  size_t result_coordinate_xyz_array_size = 0;
  size_t * local_point_ids = NULL;
  size_t local_point_ids_array_size = 0;
  size_t * num_local_point_ids =
    xmalloc(num_points * sizeof(*num_local_point_ids));

  union {
    int const * one_sub;
    int const ** n_sub;
  } global_ids;

  {
    enum yac_location location = search->location;
    if (num_subdomains == 1) {
      global_ids.one_sub = yac_get_global_ids(domain, 0, location);
    } else {
      global_ids.n_sub = xmalloc(num_subdomains * sizeof(*(global_ids.n_sub)));
      for (size_t i = 0; i < num_subdomains; ++i)
        global_ids.n_sub[i] = yac_get_global_ids(domain, (unsigned)i, location);
    }
  }

  struct request_point * curr_request_points = search_data->request_points;
  size_t num_interations =
    (num_points + ITERATION_LENGTH - 1) / ITERATION_LENGTH;

  // The computation of the nearest neighbours for each request point can be
  // very time consuming. To allow the communication layer to process tasks
  // with higher priority the request points are split into work packages of
  // size ITERATION_LENGTH. Before processing each each work package,
  // yac_comm_check is called.
  for (size_t i = 0; i < num_interations;
       ++i, curr_request_points += ITERATION_LENGTH) {

    size_t curr_num_points = MIN(ITERATION_LENGTH, num_points);
    num_points -= ITERATION_LENGTH;

    yac_comm_check(search->comm);

    for (size_t j = 0; j < curr_num_points; ++j) {
      curr_coordinates_xyz[j][0] = curr_request_points[j].coordinate_xyz[0];
      curr_coordinates_xyz[j][1] = curr_request_points[j].coordinate_xyz[1];
      curr_coordinates_xyz[j][2] = curr_request_points[j].coordinate_xyz[2];
    }

    unsigned n = search_data->n;
    yac_point_sphere_part_search_NNN(
      search->non_halo_point_search, curr_num_points, curr_coordinates_xyz,
      (size_t)n, &cos_angles, &cos_angles_array_size, &result_coordinate_xyz,
      &result_coordinate_xyz_array_size, &local_point_ids,
      &local_point_ids_array_size, num_local_point_ids);

    for (size_t j = 0, result_offset = 0; j < curr_num_points; ++j) {

      struct result_point * results =
        xmalloc(MAX(num_local_point_ids[j], n) * sizeof(*results));

      // if there is only one subdomain
      if (num_subdomains == 1) {

        for (size_t k = 0; k < num_local_point_ids[j]; ++k, ++result_offset) {
          unsigned local_id =
            search->non_halo_local_ids[local_point_ids[result_offset]];
          results[k].cos_angle = cos_angles[result_offset];
          results[k].coordinate_xyz[0] =
            result_coordinate_xyz[result_offset][0];
          results[k].coordinate_xyz[1] =
            result_coordinate_xyz[result_offset][1];
          results[k].coordinate_xyz[2] =
            result_coordinate_xyz[result_offset][2];
          results[k].global_id = global_ids.one_sub[local_id];
          results[k].is_local = 1;
          results[k].data.local.local_id = local_id;
          results[k].data.local.subdomain_id = 0;
        }
      } else {

        for (size_t k = 0; k < num_local_point_ids[j]; ++k, ++result_offset) {
          unsigned local_id = (size_t)(local_point_ids[result_offset]);
          size_t subdomain_id;
          for (
            subdomain_id = 0; (subdomain_id < num_subdomains) && (local_id >=
            search->num_non_halo_local_ids_per_subdomain_prefix[subdomain_id]);
            ++subdomain_id);
          local_id = search->non_halo_local_ids[local_id];
          results[k].cos_angle = cos_angles[result_offset];
          results[k].coordinate_xyz[0] =
            result_coordinate_xyz[result_offset][0];
          results[k].coordinate_xyz[1] =
            result_coordinate_xyz[result_offset][1];
          results[k].coordinate_xyz[2] =
            result_coordinate_xyz[result_offset][2];
          results[k].global_id = global_ids.n_sub[subdomain_id][local_id];
          results[k].is_local = 1;
          results[k].data.local.local_id = local_id;
          results[k].data.local.subdomain_id = (unsigned)subdomain_id;
        }
      }

      // sort results first by angle then by global id
      qsort(results, num_local_point_ids[j], sizeof(*results),
            compare_results);

      curr_request_points[j].results = results;
      curr_request_points[j].num_found_points = MIN(num_local_point_ids[j], n);
      if (curr_request_points[j].num_found_points == n)
        curr_request_points[j].cos_angle =
          curr_request_points[j].results[n-1].cos_angle;
    }
  }

#undef ITERATION_LENGTH

  free(local_point_ids);
  free(result_coordinate_xyz);
  free(cos_angles);
  free(num_local_point_ids);
  if (num_subdomains > 1) free(global_ids.n_sub);
  free(curr_coordinates_xyz);
}

static struct request_record *
request_records_new(size_t n, unsigned * ranks) {

  struct request_record * records = xmalloc(n * sizeof(*records));

  for (size_t i = 0; i < n; ++i) {
    records[i].rank = ranks[i];
    records[i].global_ids = NULL;
    records[i].global_ids_array_size = 0;
    records[i].num_global_ids = 0;
  }

  return records;
}

static inline void
add_global_id_to_request_record(struct request_record * record, int global_id) {

  ENSURE_ARRAY_SIZE(record->global_ids, record->global_ids_array_size,
                    record->num_global_ids + 1);

  record->global_ids[record->num_global_ids++] = global_id;
}

/**
 * checks whether there may be better points on neighbouring processes
 */
static int
nnn_search_remote(struct nnn_search_data * search_data,
                  struct request_record * request_records,
                  unsigned orig_rank) {

  struct nnn_search * search = search_data->search;
  struct request_point * request_points = search_data->request_points;
  unsigned n = search_data->n;
  unsigned * domain_neighbour_ranks = search->domain_neighbour_ranks;
  size_t num_domain_neighbours = search->num_domain_neighbours;

  // for each request point
  for (size_t i = 0; i < search_data->num_points; ++i) {

    // if there are not enough results for the current request point
    if (request_points[i].num_found_points < n) {

      for (size_t j = 0; j < num_domain_neighbours; ++j)
        if (domain_neighbour_ranks[j] != orig_rank)
          add_global_id_to_request_record(
            request_records + j, request_points[i].global_id);
    } else {

      // get maximum search distance
      struct bounding_circle circle;
      circle.base_vector[0] = request_points[i].coordinate_xyz[0];
      circle.base_vector[1] = request_points[i].coordinate_xyz[1];
      circle.base_vector[2] = request_points[i].coordinate_xyz[2];
      circle.inc_angle =
        sin_cos_angle_new(
          sqrt(1.0 - request_points[i].cos_angle * request_points[i].cos_angle),
          request_points[i].cos_angle);

      // determine which neighbours have to receive a search request
      for (size_t j = 0; j < num_domain_neighbours; ++j) {

        if (domain_neighbour_ranks[j] == orig_rank) continue;

        // if the current neighbour has halo points in the bounding circle
        if (yac_point_sphere_part_search_bnd_circle_contains_points(
              search->halo_point_search[j], circle))
          add_global_id_to_request_record(
            request_records + j, request_points[i].global_id);
      }
    }
  }

  for (size_t i = 0; i < num_domain_neighbours; ++i)
    if (request_records[i].num_global_ids > 0) return 1;
  return 0;
}

static int
compare_request_records(const void * a, const void * b) {

  return (((struct request_record*)a)->rank -
          ((struct request_record*)b)->rank);
}

/**
 * adds a result point to the results for a request point \n
 * the point is only added if the required number of result points per request
 * points is not exceeded \n
 * if there are already enough result points and if there are results points
 * that have a higher distance, the new point will replace the one with the
 * highest distance
 */
static void add_point_to_results(
  struct result_point result_point, struct request_point * request_point,
  unsigned n) {

  unsigned num_found_points = request_point->num_found_points;
  struct result_point * results = request_point->results;

  // check whether point is already in the list
  for (unsigned i = 0; i < num_found_points; ++i)
    if (results[i].global_id == result_point.global_id) return;

  unsigned i;

  // if we have not yet enought points
  if (num_found_points < n) {

    i = num_found_points;
    num_found_points++;
    request_point->num_found_points = num_found_points;

  // if the new point is not better than the currently worst
  } else if (compare_results(results+n-1, &result_point) < 0) {
    return;

  // we already have enough points and at least the currently worst is to be
  // replaced
  } else {
    i = n - 1;
  }

  for (; i > 0; --i) {
    if (compare_results(results+i-1, &result_point) > 0)
      results[i] = results[i-1];
    else
      break;
  }

  results[i] = result_point;
  request_point->cos_angle = results[num_found_points-1].cos_angle;
}

static int compare_int(const void * a, const void * b) {

  return ((*(int*)a) - (*(int*)b));
}

//! unpacks results received from another process that performed a search
static void unpack_results(
  void * buffer, unsigned buffer_size, struct nnn_search_data * search_data,
  struct request_record ** records, size_t * num_records, int remote_rank,
  struct communicator * comm) {

  int position = 0;

  unsigned num_remote_rank;
  yac_comm_unpack(buffer, buffer_size, &position, &num_remote_rank, 1,
                  COMM_UINT, comm);
  *num_records = (size_t)num_remote_rank;

  *records = xmalloc(num_remote_rank * sizeof(**records));

  for (size_t i = 0; i < *num_records; ++i) {

    struct request_record record;

    // unpack data
    unsigned num_global_ids;
    yac_comm_unpack(
      buffer, buffer_size, &position, &(record.rank), 1, COMM_UINT, comm);
    yac_comm_unpack(
      buffer, buffer_size, &position, &num_global_ids, 1, COMM_UINT, comm);
    record.num_global_ids = (size_t)num_global_ids;
    record.global_ids_array_size = (size_t)num_global_ids;
    if (record.num_global_ids > 0) {
      record.global_ids =
        xmalloc(record.num_global_ids * sizeof(*(record.global_ids)));
      yac_comm_unpack(
        buffer, buffer_size, &position, record.global_ids,
        record.num_global_ids, COMM_INT, comm);
    } else {
      record.global_ids = NULL;
    }

    // check whether we already sent the current data to the respective rank
    size_t j;
    for (j = 0; j < search_data->num_request_records; ++j)
      if (search_data->request_records[j].rank == record.rank) break;

    if (j == search_data->num_request_records) {

      search_data->request_records =
        xrealloc(search_data->request_records,
                 (search_data->num_request_records + 1) *
                 sizeof(*(search_data->request_records)));
      struct request_record * curr_request_record =
        search_data->request_records + search_data->num_request_records;
      search_data->num_request_records++;

      *curr_request_record = record;
      size_t global_ids_size =
        record.num_global_ids * sizeof(*(record.global_ids));
      curr_request_record->global_ids = xmalloc(global_ids_size);
      memcpy(
        curr_request_record->global_ids, record.global_ids, global_ids_size);

    } else {

      struct request_record * curr_request_record =
        search_data->request_records + j;

      ENSURE_ARRAY_SIZE(curr_request_record->global_ids,
                        curr_request_record->global_ids_array_size,
                        curr_request_record->num_global_ids +
                        record.num_global_ids);

      size_t m = 0;
      int sort_flag = 0;
      for (size_t k = 0, l = 0; k < record.num_global_ids; ++k) {

        int curr_global_id = record.global_ids[k];
        while ((l < curr_request_record->num_global_ids) &&
               (curr_request_record->global_ids[l] < curr_global_id))
          ++l;

        if ((l == curr_request_record->num_global_ids) ||
            (curr_request_record->global_ids[l] != curr_global_id)) {
          curr_request_record->global_ids[
            curr_request_record->num_global_ids] = curr_global_id;
          curr_request_record->num_global_ids++;
          sort_flag = 1;
          if (m != k) record.global_ids[m] = curr_global_id;
          m++;
        }
      }
      record.num_global_ids = m;

      if (sort_flag) qsort(curr_request_record->global_ids,
                           curr_request_record->num_global_ids,
                           sizeof(*(curr_request_record->global_ids)),
                           compare_int);
    }

    (*records)[i] = record;
  }

  qsort(*records, *num_records, sizeof(**records), compare_request_records);

  struct result_point result_point;
  result_point.is_local = 0;
  result_point.data.remote.rank = remote_rank;

  size_t j = 0;

  while (position < (int)buffer_size) {

    int global_id;
    unsigned num_found_src_points;

    yac_comm_unpack(
      buffer, buffer_size, &position, &global_id, 1, COMM_INT, comm);
    yac_comm_unpack(
      buffer, buffer_size, &position, &num_found_src_points, 1, COMM_UINT,
      comm);

    while ((j < search_data->num_points) &&
           (search_data->request_points[j].global_id != global_id)) ++j;

    if (j >= search_data->num_points)
      yac_internal_abort_message(
        "ERROR (unpack_results): did not find target point for which results "
        "were received\n", __FILE__, __LINE__);

    for (size_t k = 0; k < num_found_src_points; ++k) {

      yac_comm_unpack(
        buffer, buffer_size, &position, &(result_point.cos_angle), 1, COMM_DBLE,
        comm);
      yac_comm_unpack(
        buffer, buffer_size, &position, &(result_point.coordinate_xyz[0]), 3,
        COMM_DBLE, comm);
      yac_comm_unpack(
        buffer, buffer_size, &position, &(result_point.global_id), 1, COMM_INT,
        comm);

      add_point_to_results(
        result_point, search_data->request_points+j, search_data->n);
    }
  }
}

static void pack_and_send_search_request(
  struct nnn_search_data * search_data, struct request_record * request_records,
  size_t num_request_records) {

  struct nnn_search * search = search_data->search;
  struct request_point * request_points = search_data->request_points;
  struct communicator * comm = search->comm;

  char * send_buffer = NULL;
  size_t send_buffer_array_size = 0;

  for (size_t i = 0; i < num_request_records; ++i)
    if (request_records[i].num_global_ids != 0)
      search_data->num_open_remote_requests++;

  // for all entries in request_records
  for (size_t i = 0; i < num_request_records; ++i) {

    if (request_records[i].num_global_ids == 0) continue;

    size_t num_points = request_records[i].num_global_ids;
    size_t send_buffer_size =
      (size_t)yac_comm_pack_size(
        (unsigned)(4*num_points), COMM_DBLE, comm); // coordinate_xyz +
                                                    // cos_angle
    send_buffer_size +=
      (size_t)yac_comm_pack_size(
        (unsigned)(1*num_points), COMM_INT, comm);  // global_id
    int position = 0;

    ENSURE_ARRAY_SIZE(send_buffer, send_buffer_array_size, send_buffer_size);

    for (size_t j = 0, k = 0; j < num_points; ++j) {

      while (request_points[k].global_id != request_records[i].global_ids[j])
        ++k;

      yac_comm_pack(&(request_points[k].global_id), 1, COMM_INT,
                    send_buffer, send_buffer_size, &position, comm);
      yac_comm_pack(&(request_points[k].coordinate_xyz[0]), 3, COMM_DBLE,
                    send_buffer, send_buffer_size, &position, comm);
      yac_comm_pack(&(request_points[k].cos_angle), 1, COMM_DBLE,
                    send_buffer, send_buffer_size, &position, comm);
    }

    unsigned header[3] =
      {SEARCH_REQUEST_MSG_TYPE, position, (unsigned)num_points};

    // send header (does not contain data, just informs other process about
    //              pending message)
    yac_comm_bsend(header, 3, COMM_UINT, request_records[i].rank,
                   search->tags.results_or_search_request_header, comm);

    // send message
    yac_comm_bsend(send_buffer, position, COMM_PACKED, request_records[i].rank,
                   search->tags.search_request, comm);
  }

  free(send_buffer);
}

static void
request_records_delete(struct request_record * records, size_t n) {

  for (size_t i = 0; i < n; ++i) free(records[i].global_ids);
  free(records);
}

static void free_nnn_search_data(struct nnn_search_data * search_data) {

  request_records_delete(
    search_data->request_records, search_data->num_request_records);
  free(search_data->reorder_idx);
  free(search_data);
}

static void finish_barrier_cont_callback(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  UNUSED(comm);
  UNUSED(msg);

  struct nnn_search_data * search_data = (struct nnn_search_data *)user_data;
  struct nnn_search * search = search_data->search;

  search->state = INITIALISED;

  if (search_data->reorder_idx != NULL) {

    int * global_ids = xmalloc(search_data->num_points * sizeof(*global_ids));
    for (size_t i = 0; i < search_data->num_points; ++i) {
      global_ids[i] = search_data->request_points[i].global_id;
      search_data->request_points[i].global_id = search_data->reorder_idx[i];
    }
    qsort(
      search_data->request_points, search_data->num_points,
      sizeof(*(search_data->request_points)), compare_request_points);
    yac_quicksort_index(
      search_data->reorder_idx, search_data->num_points, global_ids);
    for (size_t i = 0; i < search_data->num_points; ++i)
      search_data->request_points[i].global_id = global_ids[i];
    free(global_ids);
  }

  void (*finish_callback)(void *, struct request_point*) =
    search_data->finish_callback;
  void * finish_user_data = search_data->finish_user_data;
  struct request_point * request_points = search_data->request_points;

  free_nnn_search_data(search_data);

  if (finish_callback != NULL)
    finish_callback(finish_user_data, request_points);
}

static void finish_barrier_callback(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  UNUSED(msg);

  struct nnn_search_data * search_data = (struct nnn_search_data *)user_data;
  struct nnn_search * search = search_data->search;

  yac_comm_cancel_recv(comm, search->tags.results_or_search_request_header);

  // at this point we set up another barrier to ensure that all processes have
  // canceled their requests for header messages
  // otherwise a process might return control to the user prematurely, who in
  // turn could start another search, using this nnn_search object, which then
  // could generate requests that could be served by a receive by this
  // nnn_search object, which is about to be deleted and could therefore create
  // a invalid memory access
  yac_comm_barrier_callback(
    search->tags.finished_cont, search->comm, search_data,
    finish_barrier_cont_callback, MSG_PRIO_LOW);
}

static void
finish_nnn_interpolation(struct nnn_search_data * search_data){

  struct nnn_search * search = search_data->search;

  // set up barrier that waits until all processes are finished
  yac_comm_barrier_callback(search->tags.finished, search->comm, search_data,
                            finish_barrier_callback, MSG_PRIO_LOW);
}

static void set_up_recv_results_or_search_request_callback(
  struct nnn_search_data * search_data);

static void
recv_results_cont_callback(struct communicator * comm, struct comm_message msg,
                           void * user_data) {

  struct nnn_search_data * search_data = (struct nnn_search_data *)user_data;

  void * recv_buffer = msg.buffer;
  unsigned recv_buffer_size = msg.count;

  // unpack the results
  struct request_record * records = NULL;
  size_t num_records = 0;
  unpack_results(recv_buffer, recv_buffer_size, search_data, &records,
                 &num_records, msg.src_rank, comm);

  pack_and_send_search_request(search_data, records, num_records);
  request_records_delete(records, num_records);

  // set up receive, because other processes still might need information from
  // this process
  set_up_recv_results_or_search_request_callback(search_data);

  // if all results have been received
  search_data->num_open_remote_requests--;
  if (search_data->num_open_remote_requests == 0)
    finish_nnn_interpolation(search_data);
}

/**
 * this is a callback routine, which is called once the results from a forwarded
 * have been received
 */
static void
recv_results_callback(struct communicator * comm, struct comm_message msg,
                      void * user_data) {

  struct nnn_search_data * search_data = (struct nnn_search_data *)user_data;
  struct nnn_search * search = search_data->search;

  unsigned recv_buffer_size = ((unsigned *)(msg.buffer))[0];

  yac_comm_recv_callback(recv_buffer_size, COMM_PACKED, msg.src_rank,
                         search->tags.results, comm, user_data,
                         recv_results_cont_callback, NULL, MSG_PRIO_HIGH);
}

static struct request_point * unpack_request_points(
  void * buffer, unsigned buffer_size, int * position, size_t num_points,
  struct communicator * comm) {

  struct request_point * request_points =
    xmalloc(num_points * sizeof(*request_points));

  for (size_t i = 0; i < num_points; ++i) {

    double coordinate_xyz[3], cos_angle;
    int global_tgt_id;

    // unpack the information about the target point
    yac_comm_unpack(
      buffer, buffer_size, position, &global_tgt_id, 1, COMM_INT, comm);
    yac_comm_unpack(
      buffer, buffer_size, position, &(coordinate_xyz[0]), 3, COMM_DBLE, comm);

    // unpack information about the minimum cosine angle of potential results
    yac_comm_unpack(
      buffer, buffer_size, position, &cos_angle, 1, COMM_DBLE, comm);

    // initialise the specific target point information
    init_request_point(
      request_points + i, &(coordinate_xyz[0]), global_tgt_id, cos_angle);

  }

  return request_points;
}

//! returns the result of the search request to the origin of the request
static void
pack_and_send_results(struct request_record * records,
                      struct nnn_search_data * search_data,
                      unsigned orig_rank, struct communicator * comm) {

  struct nnn_search * search = search_data->search;
  size_t num_domain_neighbours = search->num_domain_neighbours;

  char * buffer = NULL;
  size_t buffer_array_size = 0;
  size_t buffer_size = 0;
  int position = 0;

  // count the number of remote processes that may contain better result points
  unsigned num_remote_ranks = 0;
  for (size_t i = 0; i < num_domain_neighbours; ++i)
    if (records[i].num_global_ids > 0) num_remote_ranks++;

  buffer_size += yac_comm_pack_size(1, COMM_UINT, comm);
  ENSURE_ARRAY_SIZE(buffer, buffer_array_size, buffer_size);

  // pack the number of remote ranks
  yac_comm_pack(
    &num_remote_ranks, 1, COMM_UINT, buffer, buffer_size, &position, comm);

  buffer_size = position;

  // pack the global ids of the request points that may have to be sent to the
  // respective remote process
  for (size_t i = 0; i < num_domain_neighbours; ++i) {

    if (records[i].num_global_ids == 0) continue;

    buffer_size += yac_comm_pack_size(2, COMM_UINT, comm);
    buffer_size +=
      yac_comm_pack_size(records[i].num_global_ids, COMM_INT, comm);
    ENSURE_ARRAY_SIZE(buffer, buffer_array_size, buffer_size);

    yac_comm_pack(
      &(records[i].rank), 1, COMM_UINT, buffer, buffer_size, &position, comm);
    yac_comm_pack(
      &(records[i].num_global_ids), 1, COMM_UINT, buffer, buffer_size,
      &position, comm);
    if (records[i].num_global_ids > 0)
      yac_comm_pack(
        records[i].global_ids, records[i].num_global_ids, COMM_INT, buffer,
        buffer_size, &position, comm);
    buffer_size = position;
  }

  // pack locally found result points
  for (size_t i = 0; i < search_data->num_points; ++i) {

    unsigned num_found_points = search_data->request_points[i].num_found_points;

    if (num_found_points == 0) continue;

    struct request_point * curr_request_point = search_data->request_points + i;

    buffer_size += yac_comm_pack_size(1, COMM_INT, comm); // global id
    buffer_size += yac_comm_pack_size(1, COMM_UINT, comm); // num_found_points
    buffer_size += yac_comm_pack_size(num_found_points, COMM_INT, comm); // global id
    buffer_size += yac_comm_pack_size(num_found_points, COMM_DBLE, comm); // cos_angle
    buffer_size += yac_comm_pack_size(3 * num_found_points, COMM_DBLE, comm); // coordinate_xyz
    ENSURE_ARRAY_SIZE(buffer, buffer_array_size, buffer_size);

    yac_comm_pack(
      &(curr_request_point->global_id), 1, COMM_INT, buffer, buffer_size,
      &position, comm);
    yac_comm_pack(
      &(curr_request_point->num_found_points), 1, COMM_UINT, buffer,
      buffer_size, &position, comm);
    for (unsigned j = 0; j < curr_request_point->num_found_points; ++j) {
      yac_comm_pack(
        &(curr_request_point->results[j].cos_angle), 1, COMM_DBLE, buffer,
        buffer_size, &position, comm);
      yac_comm_pack(
        &(curr_request_point->results[j].coordinate_xyz[0]), 3, COMM_DBLE,
        buffer, buffer_size, &position, comm);
      yac_comm_pack(
        &(curr_request_point->results[j].global_id), 1, COMM_INT, buffer,
        buffer_size, &position, comm);
    }
    buffer_size = position;
  }

  unsigned header[2] = {RESULTS_MSG_TYPE, buffer_size};

  // send header
  yac_comm_bsend(&header, 2, COMM_UINT, orig_rank,
                 search->tags.results_or_search_request_header, comm);

  // send data
  yac_comm_bsend(buffer, buffer_size, COMM_PACKED, orig_rank,
                 search->tags.results, comm);

  free(buffer);
}

void yac_nnn_search_free_request_points(
  struct request_point * request_points, size_t num_points) {

  for (size_t i = 0; i < num_points; ++i)
    free(request_points[i].results);
}

static void free_search_data(struct nnn_search_data * search_data) {

  if (search_data->num_open_remote_requests != 0)
    yac_internal_abort_message(
      "ERROR (free_search_data): search_data still has open requests",
      __FILE__, __LINE__);

  for (size_t i = 0; i < search_data->num_points; ++i)
    free(search_data->request_points[i].results);
  free(search_data->request_points);
}

static void receive_search_request_cont_callback(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  struct temp_search_request_data * temp_data =
     (struct temp_search_request_data *)user_data;

  struct nnn_search_data * search_data = temp_data->search_data;
  struct nnn_search * search = search_data->search;
  unsigned num_points = temp_data->num_points;
  unsigned num_domain_neighbours = search->num_domain_neighbours;

  free(temp_data);

  int position = 0;

  struct nnn_search_data temp_search_data;

  temp_search_data.search = search;
  temp_search_data.request_points =
    unpack_request_points(msg.buffer, msg.count, &position, num_points, comm);
  temp_search_data.finish_callback = NULL;
  temp_search_data.finish_user_data = NULL;
  temp_search_data.n = search_data->n;
  temp_search_data.num_points = num_points;
  temp_search_data.request_records = NULL;
  temp_search_data.num_request_records = 0;
  temp_search_data.num_open_remote_requests = 0;

  // local search
  nnn_search_local(&temp_search_data);

  struct request_record * records =
     request_records_new(num_domain_neighbours, search->domain_neighbour_ranks);

  // check whether there may be better source points on neighbouring processes
  nnn_search_remote(&temp_search_data, records, msg.src_rank);

  // set up new request
  set_up_recv_results_or_search_request_callback(search_data);

  // pack and send the results of the local and remote search
  pack_and_send_results(records, &temp_search_data, msg.src_rank, comm);
  free_search_data(&temp_search_data);

  request_records_delete(records, num_domain_neighbours);
}

/**
 * callback routine that receives and processes a forwarded search requests
 */
static void
recv_search_request_callback(struct communicator * comm,
                             struct comm_message msg,
                             void * user_data) {

  struct nnn_search_data * search_data = (struct nnn_search_data *)user_data;
  struct nnn_search * search = search_data->search;

  struct temp_search_request_data * temp_data = xmalloc(1 * sizeof(*temp_data));

  unsigned buffer_size = ((unsigned *)(msg.buffer))[0];
  size_t num_points = (size_t)(((unsigned *)(msg.buffer))[1]);

  temp_data->search_data = search_data;
  temp_data->num_points = num_points;

  yac_comm_recv_callback(buffer_size, COMM_PACKED, msg.src_rank,
                         search->tags.search_request, comm, temp_data,
                         receive_search_request_cont_callback, NULL,
                         MSG_PRIO_HIGH);
}

static void recv_results_or_search_request_callback(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  struct comm_message temp_msg = msg;

  unsigned msg_type = *(unsigned*)(temp_msg.buffer);

  unsigned * temp = (unsigned*)(temp_msg.buffer);
  temp_msg.buffer = ++temp;
  temp_msg.count--;

  switch (msg_type) {

    case (0) :
      recv_results_callback(comm, temp_msg, user_data);
      break;
    case (1) :
      recv_search_request_callback(comm, temp_msg, user_data);
      break;
    default:
      yac_internal_abort_message(
        "ERROR(recv_results_or_search_request_callback): "
        "undefined message type\n", __FILE__, __LINE__);
      break;
  }
}

static void set_up_recv_results_or_search_request_callback(
  struct nnn_search_data * search_data) {

  struct nnn_search * search = search_data->search;

  if (search->state == ACTIVE)
    yac_comm_recv_callback(3, COMM_UINT, COMM_ANYSOURCE,
                           search->tags.results_or_search_request_header,
                           search->comm, search_data,
                           recv_results_or_search_request_callback, NULL,
                           MSG_PRIO_HIGH);
}

void yac_nnn_search_do_search(struct nnn_search * search, unsigned n,
  size_t num_points, double (*coordinates_xyz)[3], int * global_ids,
  void * user_data, void (*callback)(void * user_data, struct request_point*)) {

  if ((coordinates_xyz == NULL) && (num_points > 0))
    yac_internal_abort_message(
      "ERROR(yac_nnn_search_do_search): missing point coordinates",
      __FILE__, __LINE__);

  search->state = ACTIVE;

  struct nnn_search_data * search_data = xmalloc(1 * sizeof(*search_data));
  search_data->search = search;
  search_data->request_points =
    generate_request_points(
      (size_t)num_points, coordinates_xyz, global_ids,
      &(search_data->reorder_idx));
  search_data->n = (size_t)n;
  search_data->num_points = (size_t)num_points;
  search_data->num_open_remote_requests = 0;
  search_data->finish_user_data = user_data;
  search_data->finish_callback = callback;

  // compute for every target point the n nearest local neighbours
  nnn_search_local(search_data);

  struct request_record * request_records;
  size_t num_request_records = search->num_domain_neighbours;
  request_records = request_records_new(search->num_domain_neighbours,
                                        search->domain_neighbour_ranks);
  search_data->request_records = request_records;
  search_data->num_request_records = num_request_records;

  int request_flag = nnn_search_remote(search_data, request_records, -1);

  if (request_flag)
    pack_and_send_search_request(
      search_data, request_records, num_request_records);

   // set up recv_callback for remote search requests
   set_up_recv_results_or_search_request_callback(search_data);

   // if there are no remote requests
   if (!request_flag) finish_nnn_interpolation(search_data);
}

unsigned yac_nnn_search_wait(struct nnn_search * search) {

  struct communicator * comm = search->comm;
  struct nnn_search_tags tags = search->tags;

  unsigned ret_val = 0;

  switch (search->state) {
    case (ACTIVE):
    {
      unsigned tmp_ret_val;
      do {
        tmp_ret_val =
          yac_comm_wait(comm, tags.results_or_search_request_header) ||
          yac_comm_wait(comm, tags.search_request) ||
          yac_comm_wait(comm, tags.results) ||
          yac_comm_wait(comm, tags.finished) ||
          yac_comm_wait(comm, tags.finished_cont);
        ret_val |= tmp_ret_val;
      } while (tmp_ret_val);
    }
    case (INITIALISED):
      break;
    default:
      yac_internal_abort_message(
        "ERROR(yac_nnn_search_wait): invalid state", __FILE__, __LINE__);
  };

  return ret_val;
}

void yac_nnn_search_delete(struct nnn_search * search) {

  switch (search->state) {
    case (ACTIVE):
      yac_internal_abort_message(
        "ERROR(yac_nnn_search_delete): search is active",
        __FILE__, __LINE__);
      break;
    default:
      yac_internal_abort_message(
        "ERROR(yac_nnn_search_delete): invalid state",
        __FILE__, __LINE__);
      break;
    case (INITIALISED):
      break;
  };

  free(search->domain_neighbour_ranks);

  size_t num_neighbours = search->num_domain_neighbours;
  yac_delete_point_sphere_part_search(search->non_halo_point_search);
  free(search->non_halo_local_ids);
  free(search->num_non_halo_local_ids_per_subdomain_prefix);
  for (size_t i = 0; i < num_neighbours; ++i)
     yac_delete_point_sphere_part_search(search->halo_point_search[i]);
  free(search->halo_point_search);

  free(search);
}

