/**
 * @file nnn_search.h
 * @brief Parallel n-nearest-neighbour search
 *
 * Structs and interfaces to do a parallel n-nearest-neighbour search on a
 * domain
 *
 * @copyright Copyright  (C)  2017 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Thomas Jahns <jahns@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef NNN_SEARCH_H
#define NNN_SEARCH_H

#include "communicator.h"
#include "domain.h"
#include "points.h"
#include "mask.h"

/**
 * contain the information for one point that was found for a request point
 */
struct result_point {
  double cos_angle; //!< cosine of angle between the request and result point
  double coordinate_xyz[3];
  int global_id;    //!< global id of the point

  int is_local;    //!< true if the point is available in the local domain

  union {

     struct {
        unsigned local_id;
        unsigned subdomain_id;
     } local;

     struct {
        int rank;
     } remote;
  } data;          //!< specific information on the source point (depends on
                   //!< whether the point is locally available or not)
};

/**
 * contains all temporary search specific data for one request point
 */
struct request_point {

   int global_id;            //!< global id of the request point
   double coordinate_xyz[3]; //!< coordinates of request point

   double cos_angle; //!< cosine of angle between the request point and the
                     //!< furthest result point

   struct result_point * results;   //!< the points found for the request point
   unsigned num_found_points;       //!< number of found points
};

struct nnn_search;

/**
 * constructs a search object that allows a parallel n-nearest-neighbour search
 * within the provided domain
 * @param[in] domain
 * @param[in] unique_string string that is unique for this search object
 * @param[in] points        points containing the locations
 * @param[in] masks         mask for the points (if no mask is to be used, pass
 *                          NULL)
 * @param[in] num_pointsets number of point sets
 * @param[in] points_idx    which points set of the domain is to be used for the
 *                          search
 * @returns nnn_search object
 *
 * \remark the call to this routine is collective
 * \remark unique_string needs to be identical on all processes collectively
 *         calling this routine
 */
struct nnn_search * yac_nnn_search_new(
  struct domain * domain, const char * unique_string, struct points *** points,
  struct mask *** masks, unsigned num_pointsets, unsigned points_idx);

/**
 * initiated a parallel n-nearest-neighbour search
 * @param[in] search
 * @param[in] n               number of nearest neighbours
 * @param[in] num_points      number of points
 * @param[in] coordinates_xyz cartesian coordinates of all points
 * @param[in] global_ids      global_ids for each point
 * @param[in] user_data       a pointer that is passed to callback
 * @param[in] callback        pointer to a routine that is called once the
 *                            n-nearest-neighbour search is completed, it
 *                            receives as an argument the result from the search
 *
 * \remark the user has to ensure that no search is active for this search
 *         object
 * \remark a call \ref yac_nnn_search_wait blocks until the search is complete
 * \remark the user is responsible for freeing the results passed to the
 *         callback routine (see \ref yac_nnn_search_free_request_points)
 */
void yac_nnn_search_do_search(struct nnn_search * search, unsigned n,
  size_t num_points, double (*coordinates_xyz)[3], int * global_ids,
  void * user_data, void (*callback)(void * user_data, struct request_point*));

/**
 * frees all memory associated to an array of request_point's
 * @param[in] request_points request_points to be freed
 * @param[in] num_points     number of request points
 *
 * \remark the user still needs to free the request_points pointer after calling
 * this routine
 */
void yac_nnn_search_free_request_points(
  struct request_point * request_points, size_t num_points);

/**
 * blocks until active search is finished, returns immediately if there was no
 * active search
 * @param[in] search
 * @return 1 if yac_nnn_search_wait actually had to wait for a message
 *         0 if no search was active or the search was already completed
 */
unsigned yac_nnn_search_wait(struct nnn_search * search);

/**
 * frees all memory associated to the n-nearest-neighbour search objext
 * @param[in] search
 *
 * \remark the user has to ensure that there are no actives searches and search
 *         results that are not yet retreived for this search object
 */
void yac_nnn_search_delete(struct nnn_search * search);

#endif // NNN_SEARCH_H

