/**
 * @file patch_recovery.h
 * @brief Routines for the superconvergence patch recovery technique
 *
 * Locations for Gauss Quadrature points
 *
 * @copyright Copyright  (C)  2013 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Hendryk Bockelmann <bockelmann@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef PATCH_RECOVERY_H
#define PATCH_RECOVERY_H

/** \example test_patch_recovery.c
 * This test severall functions for the patch recovery. 
 */

#include "grid.h"

struct xyz_triangle {
   double a[3];
   double b[3];
   double c[3];
};

struct xyz_polygon {
   int num_vertices;
   double *x;
   double *y;
   double *z;
};


/**  Calculation of abscissas for Gauss-Legendre Polynomials
  *  for Linear-Domain Gauss-Legendre Quadratures
  *
  *  Adopted from Numerical Recipes 3rd Ed., Press et. al., 2007.
  *
  *  For results see e.g.
  *
  *  - http://processingjs.nihongoresources.com/bezierinfo/legendre-gauss-values.php
  *  - http://en.wikipedia.org/wiki/Gaussian_quadrature
  *  - http://de.wikipedia.org/wiki/Gau%C3%9F-Quadratur
  *
  *
  *  Given an n this routine returns arrays x and w of lengths n,
  *  containing the abscissas of the Gauss-Legendre
  *  n-point quadrature formula in the interval [-1,1].
  *
  *  @param[in]  n         size of abscissa and weight
  *  @param[out] abscissa  abscissa of the Gauss-Legendre n-point quadrature formula
  *
 **/

void yac_linear_gauss_legendre ( unsigned n, double * abscissa );

/**  Tables of abscissas for Gaussian Quadrature for triangles
  *
  *  Numbers taken from Earth System Modeling Framework
  *  ESMCI_Quadrature.C,v 1.5.4.1 2012/01/06
  *
  *  A program to calculate locations is available from
  *
  *  http://people.sc.fsu.edu/~jburkardt/f_src/dunavant/dunavant.html
  *
  *  D.A. Dunavant, 1985: High Efficient Symmetrical Gaussian Quadrature
  *     Rules for the Triangle, J. Num. Engineering, Vol. 21, 1129 - 1148. 
  *
  *  Note 1:
  *
  *  For each weight a triple of locations is obtained. Because of the
  *  dependency e.g. for order 1
  *
  *               locs[0] + locs[1] + locs[2] - 1 = 0
  *
  *  it is sufficient to provide only locs[0] and locs[1]
  *
  *  Note 2:
  *
  *  The weights given by Dunavant (1985) differ by a factor 2.
  *
  *  @param[in]  n         size of abscissa and weight
  *  @param[out] abscissa  abscissa of the Gauss-Legendre n-point quadrature formula
  *
 **/

void yac_triangle_gauss_legendre ( unsigned n, double * abscissa ) ;

/**  Gaussian Quadrature points (abscissa) for 2D quads a particular order
  *
  *  @param[in]  order     order of the polynomial
  *  @param[out] abscissa  x,y coordinates of the Gauss-Legendre points
  *
 **/

void yac_quad_2d_locations ( unsigned order, double * abscissa );

/**  Number of Gaussian Quadrature points (abscissa) for 2D triangle
  *
  *  @param[in]  order     order of the polynomial
  *
 **/

unsigned yac_get_number_of_quadrature_points ( unsigned order );

/**  Number of coefficients of the polynom
  *
  *  @param[in]  order     order of the polynomial
  *
 **/

unsigned yac_get_num_polynomial_coefficients ( unsigned order );

/**  Conversion of Gaussian Quadrature points in 3D
  *
  *  The  Gaussian Quadrature points r are provided for the reference triangle
  *  and are transformed to points X in the model element (defined by triangle_nodes)
  *  using an affine mapping. Extended to 3D. A unit tetrahedron
  *
  *          a = (1,0,0); b = (0,1,0); c = (0,0,1), d= (0,0,0)
  *
  *  is used to construct the affine mapping. The tetrahedron is rotated such that
  *  the face a-b-d (equal to the the unit triangle in the plane) is mapped onto
  *  the sphere.
  *
  *  x = A x r + c
  *
  *  see e.g.
  *
  *  Mark S. Gockenbach (2006): Understanding and Implementing the Finite Element Method, Chap. 4.6
  *
  *  or
  *
  *  http://people.sc.fsu.edu/~jburkardt/presentations/cg_lab_mapping_triangles.pdf
  *
  *
  *  @param[in]  num_points           number of points to be transformed
  *  @param[in]  triangle             containing the xyz vertex coordinates of the target triangle
  *  @param[in]  reference_location   Gauss quadrature points of the reference triangle
  *  @param[out] transformed_location transformed Gauss quadrature points
  *
 **/

void yac_reference_to_xyz_triangle ( unsigned num_points, struct xyz_triangle triangle,
                                     double * reference_location,
                                     double * transformed_location );

/**  The inverse operation to reference_to_triangle_3D
  *
  *  The  Gaussian Quadrature points r are provided for the reference triangle
  *  and are transformed to points X in the model element (defined by triangle_nodes)
  *  using an affine mapping.
  *
  *   r = A^-1 ( x -c )
  *
  *  see e.g.
  *
  *  Mark S. Gockenbach (2006): Understanding and Implementing the Finite Element Method, Chap. 4.6
  *
  *  or
  *
  *  http://people.sc.fsu.edu/~jburkardt/presentations/cg_lab_mapping_triangles.pdf
  *
  *
  *  @param[in]  num_points           number of points to be transformed
  *  @param[in]  triangle             containing the xyz vertex coordinates of the target triangle
  *  @param[in]  reference_location   Gauss quadrature points of the reference triangle
  *  @param[out] transformed_location transformed Gauss quadrature points
  *
 **/

void yac_xyz_triangle_to_reference ( unsigned num_points,  struct xyz_triangle triangle,
                                     double * reference_location,
                                     double * transformed_location );

/**  Transformation of Gaussian Quadrature points from unit squares into 2d-quadrilaterals
  *
  *  The  Gaussian Quadrature points r are provided for the reference rectangle
  *  and are transformed to points X in the model element (defined by nodes of the quadrilateral).
  *
  *  see http://www.geometrictools.com/Documentation/PerspectiveMappings.pdf
  *      http://math.stackexchange.com/questions/13404/mapping-irregular-quadrilateral-to-a-rectangle
  *      http://www.gamedev.net/topic/596392-uv-coordinate-on-a-2d-quadrilateral/#entry4779201
  *
  *  Here we used the formula as given on page 15 in
  *
  *  Paul S. Heckbert, 1989: Fundamentals of Texture Mapping and Image Warping,
  *  Master Thesis, University of California, Berkley
  *
  *  (first pubished by Michael Faux and John Pratt, 1979: Computational geometry for design and manufacture)

  *  @param[in]  num_points           number of points to be transformed
  *  @param[in]  cell                 containing the vertex coordinates of the target quadrilateral
  *  @param[in]  reference_location   Gauss quadrature points of the reference triangle
  *  @param[out] transformed_location transformed Gauss quadrature points
  *
 **/

void yac_reference_to_xy_quadrilateral ( unsigned num_points, struct grid_cell cell,
                                         double * reference_location,
                                         double * transformed_location );

/**  Transformation of Gaussian Quadrature points from unit squares into 3d-quadrilaterals
  *
  *  The  Gaussian Quadrature points r are provided for the reference rectangle
  *  and are transformed to points X in the model element (defined by nodes of the quadrilateral).
  *  For doing this the quadrilateral is split into an upper and lower triangle. All quadrature
  *  points of the unit square falling into the unit triangle x + y <=1 are transformed into
  *  the lower triangle, next all quadrature points of the unit square falling into the unit
  *  triangle x + y < 1 are transformed into the upper triangle in the quadrilateral. This
  *  is possible as quadrature points are symmetrical with reference to the diagonal.

  *  @param[in]  gauss_order          gauss_order
  *  @param[in]  cell                 containing the vertex coordinates of the target quadrilateral
  *  @param[in]  reference_location   Gauss quadrature points of the reference triangle
  *  @param[out] transformed_location transformed Gauss quadrature points
  *
 **/

void yac_reference_to_xyz_quadrilateral ( unsigned gauss_order, struct grid_cell cell,
                                          double * reference_location,
                                          double * transformed_location );

/**  Transformation of Gaussian Quadrature points from unit triangle into triangles of a 3d-polygon
  *
  *  The  Gaussian Quadrature points r are provided for the reference rectangle
  *  and are transformed to points X in the model element (defined by nodes of the polygon).
  *  For doing this the polygon is split into triangles consisting of two adjacent vertices
  *  of the polygon an the barycenter of the vertices.
  *
  *  @param[in]  num_gauss_points     number of Gauss points in the polygon
  *  @param[in]  polygon              containing the xyz vertex coordinates of the target polygon
  *  @param[in]  reference_location   Gauss quadrature points of the reference triangle
  *  @param[out] transformed_location transformed Gauss quadrature points
  *
 **/

void yac_reference_to_xyz_polygon ( unsigned num_gauss_points,
                                    struct xyz_polygon polygon,
                                    double * reference_location,
                                    double * transformed_location );

/**  Set the polynomial terms x, y, z, xyz, etc.
  *
  *  Triples of locations (x,y,z) are provided on input. Based on the requested order of the
  *  polynomial the linear, quadratic and cubic terms for the matrix P are calculated
  *  to later solve Pa = b.
  *  
  *  @param[in]  order        order of the polynomial ( 0 | 1 | 2 | 3 ), determines the number of columns
  *  @param[in]  num_rows     number of rows of P, i.e. the number of coordinates
  *  @param[in]  coordinates  cartesian coordinates
  *  @param[in]  reference    reference by which the patch will be shifted
  *  @param[out] PT           returned matrix
  *
 **/

void yac_set_polynomial_terms ( unsigned order, unsigned num_rows, double * coordinates, double * reference, double *PT );

/**  Determine the unknown coefficients used a least square fit
  *
  *  Triples of b(x,y,z) are provided on input. Based on the requested order of the
  *  polynomial the linear, quadratic and cubic terms for the matrix P are calculated
  *  to later solve Pa = b.
  *  
  *  @param[in]  num_points  number of sampling (Gauss) points
  *  @param[in]  num_coeff   number of coefficients of the polynomial
  *  @param[in]  PT          polynomial terms x, y, z, xyz, ... for locations of Gauss points
  *  @param[out] a           unknown coefficients
  *  @param[in]  b           value at Gauss points
  *
 **/

void yac_get_polynomial_coefficients ( unsigned num_points, unsigned num_coeff, double *PT, double *a, double *b );

/**  Calculates the num_cells weights for a given target location in the patch
  *
  *  @param[in] num_points       number of patch points for the polynomial
  *  @param[in] num_coeff        number of coefficients of the polynomial
  *  @param[in] num_cells        number of cells in the patch
  *  @param[in] num_gausspoints  number of gauss points per cell
  *  @param[in] Pt               polynomial terms evaluated at target location
  *  @param[in] PT               matrix containing the polynomial terms evaluated at patch points
  *  @param[in] weight           weights for target point point
  *
 **/

void yac_patch_P0weights ( unsigned num_points, unsigned num_coeff, unsigned num_cells, unsigned * num_gausspoints, double *Pt, double *PT, double *weight );

/**  Optimised version of patch_P0weights.
  *
  *  Compared to patch_P0weights this routine was three times faster when
  *  tested on IBM pwr6 using xlc12.1.0.6 (and xlf14.1.0.4) without any
  *  specific optimisation.
  *
  *  Returns the num_cells weights for a given target location in the patch
  *
  *  @param[in] num_points       number of patch points for the polynomial
  *  @param[in] num_coeff        number of coefficients of the polynomial
  *  @param[in] num_cells        number of cells in the patch
  *  @param[in] num_gausspoints  number of gauss points per cell
  *  @param[in] PT               matrix containing the polynomial terms evaluated at patch points
  *  @param[in] weight           weights for target point point
  *
 **/

void yac_patch_P0weights_opt ( unsigned num_points, unsigned num_coeff, unsigned num_cells, unsigned * num_gausspoints, double *PT, double *weight );

/**  Returns the value at a given target location in the patch
  *
  *  takes num_points weights calculated by patch_weights and the num_points
  *  functions values at the patch points.
  *
  *  @param[in] num_cells   number of cells in the patch
  *  @param[in] weight      weights[num_cells] for target point
  *  @param[in] fi          function values at cells
  *
 **/

double yac_patch_P0value ( unsigned num_cells, double *weight, double *fi );

/**  Returns a value at a given location in the patch
  *
  *  Pairs of locations (x,y) are provided on input. Based on the requested order of the
  *  polynomial the linear, quadratic and cubic terms for the matrix P are calculated
  *  to later solve Pa = b.
  *  
  *  @param[in] num_points  number of patch points for the polynomial
  *  @param[in] num_coeff   number of coefficients of the polynomial
  *  @param[in] Pt          polynomial terms evaluated at target location
  *  @param[in] PT          matrix containing the polynomial terms evaluated at patch points
  *  @param[in] fi          function values at patch points
  *
 **/

double yac_patch_value_alt1 ( unsigned num_points, unsigned num_coeff, double *Pt, double *PT, double *fi );

/**  Returns a value at a given location in the patch
  *
  *  Pairs of locations (x,y) are provided on input. Based on the requested order of the
  *  polynomial the linear, quadratic and cubic terms for the matrix P are calculated
  *  to later solve Pa = b.
  *  
  *  @param[in] num_coeff   number of coefficients of the polynomial
  *  @param[in] PT          transposed of matrix P containing the polynomial terms
  *  @param[in] a           coefficients of the polynomial
  *
 **/

double yac_patch_value_alt2 ( unsigned num_coeff, double *PT, double *a );

#endif // PATCH_RECOVERY_H

