/**
 * @file test_compute_overlap_area.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "tests.h"
#include "clipping.h"
#include "geometry.h"
#include "test_common.h"

//#define VERBOSE

int main (void) {

  double ref_area, area;

  static double const area_tolerance = 1.0e-10;

#ifdef VERBOSE
  puts ("----- calculating area -----\n");
#endif

  {
    /* set the test cells concave pentagon with a triangle


           10.0 (lon)   11.0 (lon)
           32.0 (lat)   32.0 (lat)
                  x--------x
                 /          \
                /    10.5    \  
               /     31.0     \  
   9.5 (lon)  /       x        \   11.5
  30.0 (lat) x                  x  30.0 


   with


                    10.5 (lon)
                    31.5 (lat)
                      x
                     / \
                    /   \
           10.0    /     \   11.0
           29.5   x-------x  29.5


    */

    struct grid_cell Pentagon = {.coordinates_x = (double[]) {10.0*YAC_RAD, 11.0*YAC_RAD, 11.5*YAC_RAD, 10.5*YAC_RAD,  9.5*YAC_RAD},
                                 .coordinates_y = (double[]) {32.0*YAC_RAD, 32.0*YAC_RAD, 30.0*YAC_RAD, 31.0*YAC_RAD, 30.0*YAC_RAD},
                                 .coordinates_xyz = (double[5][3]){{-1}},
                                 .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                     GREAT_CIRCLE,
                                                                     GREAT_CIRCLE,
                                                                     GREAT_CIRCLE,
                                                                     GREAT_CIRCLE},
                                 .num_corners = 5};

    struct grid_cell Triangle = {.coordinates_x = (double[]) {10.5*YAC_RAD, 11.0*YAC_RAD, 10.0*YAC_RAD},
                                 .coordinates_y = (double[]) {31.5*YAC_RAD, 29.5*YAC_RAD, 29.5*YAC_RAD},
                                 .coordinates_xyz = (double[3][3]){{-1}},
                                 .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                     GREAT_CIRCLE,
                                                                     GREAT_CIRCLE},
                                 .num_corners = 3};

    double Pentagon_node_xyz[3];
    double Triangle_node_xyz[3];

    generate_XYZ(Pentagon);
    LLtoXYZ(10.5*YAC_RAD, 31.5*YAC_RAD, Pentagon_node_xyz);

    generate_XYZ(Triangle);
    LLtoXYZ(10.5*YAC_RAD, 30.0*YAC_RAD, Triangle_node_xyz);

    size_t N = 1;

    // the reference solution for the area
    yac_compute_overlap_areas( N, &Pentagon, Triangle, &ref_area );

    // this should give a bit identical results as we do not perform any triangulation
    yac_compute_concave_overlap_areas ( N, &Pentagon, Triangle, Triangle_node_xyz, &area );
    if ((ref_area > area) != (ref_area < area))
      PUT_ERR("error compute_concave_overlap_areas, bit identical results expected.\n");

    // with triangulation of the pentagon the result should not differ that much
    yac_compute_concave_overlap_areas ( N, &Triangle, Pentagon, Pentagon_node_xyz, &area );
    if ( fabs(ref_area-area) > area_tolerance )
      PUT_ERR("error compute_concave_overlap_areas, area difference is too large.\n");

  }

  {
    /* set the test cells convex square with a triangle


                -5.0      5.0
                  x--------x  5.0
                  |        |
                  |  10.5  |
                  |  31.0  |
                  x--------x -5.0

   with

                -4.0 0.0 4.0
                      x      6.0
                     / \
                    /   \
                   /     \
                  x-------x -4.0


    */

    struct grid_cell Square = {.coordinates_x = (double[]) {-5.0*YAC_RAD,
                                                             5.0*YAC_RAD,
                                                             5.0*YAC_RAD,
                                                            -5.0*YAC_RAD},
                               .coordinates_y = (double[]) {-5.0*YAC_RAD,
                                                            -5.0*YAC_RAD,
                                                             5.0*YAC_RAD,
                                                             5.0*YAC_RAD},
                               .coordinates_xyz = (double[4][3]){{-1}},
                               .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                   GREAT_CIRCLE,
                                                                   GREAT_CIRCLE,
                                                                   GREAT_CIRCLE},
                               .num_corners = 4};

    struct grid_cell Triangle = {.coordinates_x = (double[]) {-4.0*YAC_RAD, 0.0*YAC_RAD,  4.0*YAC_RAD},
                                 .coordinates_y = (double[]) {-4.0*YAC_RAD, 6.0*YAC_RAD, -4.0*YAC_RAD},
                                 .coordinates_xyz = (double[3][3]){{-1}},
                                 .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                     GREAT_CIRCLE,
                                                                     GREAT_CIRCLE},
                                 .num_corners = 3};

    double Square_node_xyz[3];
    double Triangle_node_xyz[3];

    generate_XYZ(Square);
    LLtoXYZ(0.0*YAC_RAD, 0.0*YAC_RAD, Square_node_xyz);

    generate_XYZ(Triangle);
    LLtoXYZ(0.0*YAC_RAD, 0.0*YAC_RAD, Triangle_node_xyz);

    size_t N = 1;

    // the reference solution for the area
    yac_compute_overlap_areas( N, &Square, Triangle, &ref_area );

    // this should give a bit identical results as we do not perform any triangulation
    yac_compute_concave_overlap_areas ( N, &Square, Triangle, Triangle_node_xyz, &area );
    if ((ref_area > area) != (ref_area < area))
      PUT_ERR("error compute_concave_overlap_areas, bit identical results expected.\n");

    // with triangulation of the square the result should not differ that much
    yac_compute_concave_overlap_areas ( N, &Triangle, Square, Square_node_xyz, &area );
    if ( fabs(ref_area-area) > area_tolerance )
      PUT_ERR("error compute_concave_overlap_areas, area difference is too large.\n");

  }
  return TEST_EXIT_CODE;
}

