/**
 * @file test_field_interpolation_parallel6.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>

#include <mpi.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_file.h"
#include "interpolation_method_avg.h"
#include "fields.h"
#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src.h"
#include "field_interpolation_gen_tgt.h"
#include "weight_file_common.h"

// needs to be a multiple of 2
#define NUM_CELLS_X 128
#define NUM_CELLS_Y 128
#define NUM_CELLS_X_2 64
#define NUM_CELLS_Y_2 64

static void generate_input_weights();

static void generate_ref_weights();

static void submain_src(struct communicator * global_comm,
                        struct communicator * domain_comm);

static void submain_tgt(struct communicator * global_comm,
                        struct communicator * domain_comm);

// link data for input weight file
unsigned num_links_file = 4*NUM_CELLS_X_2*NUM_CELLS_Y_2;
int src_address_file[NUM_CELLS_Y_2][NUM_CELLS_X_2][4];
int tgt_address_file[NUM_CELLS_Y_2][NUM_CELLS_X_2][4];
double weights_file[NUM_CELLS_Y_2][NUM_CELLS_X_2][4];

// reference link data for output weight file
unsigned ref_num_links = 4*NUM_CELLS_X*NUM_CELLS_Y;

int ref_src_address[NUM_CELLS_Y][NUM_CELLS_X][4];
int ref_tgt_address[NUM_CELLS_Y][NUM_CELLS_X][4];
double ref_weights[NUM_CELLS_Y][NUM_CELLS_X][4];

int const * ref_tgt_address_fixed = NULL;
double const * ref_fixed_values = NULL;
int const * ref_num_tgt_per_fixed_value = NULL;
unsigned ref_num_fixed_values = 0;

// grid information (is the same for source and target)
unsigned cyclic[2] = {0,0};
double global_coordinates_x[NUM_CELLS_X + 1];
double global_coordinates_y[NUM_CELLS_X + 1];

int mapping_on_source;

int main(int argc, char *argv[]) {

   if (argc != 2) {
      PUT_ERR("wrong number of arguments\n");
      return TEST_EXIT_CODE;
   }

   mapping_on_source = strcmp(argv[1], "src") == 0;
   if (!mapping_on_source && strcmp(argv[1], "tgt")) {
      PUT_ERR("invalid argument (has to be either \"src\" or \"tgt\")\n");
      return TEST_EXIT_CODE;
   }

  generate_input_weights();
  generate_ref_weights();

  for (unsigned i = 0; i < NUM_CELLS_X + 1; ++i)
    global_coordinates_x[i] = (double)i * YAC_RAD;

  for (unsigned i = 0; i < NUM_CELLS_Y + 1; ++i)
    global_coordinates_y[i] = (double)i * YAC_RAD;

  struct communicator * global_comm;

  global_comm = yac_mpi_comm_new();

  if (yac_get_comm_size(global_comm) == 5) {

    int my_global_rank;

    my_global_rank = yac_get_comm_rank(global_comm);

    struct communicator * comm;

    comm = yac_comm_split(global_comm, my_global_rank < 4, 0);

    if (my_global_rank < 4)
      submain_src(global_comm, comm);
    else
      submain_tgt(global_comm, comm);

    yac_free_communicator(comm);
  } else
    PUT_ERR("invalid number of processes\n");

  yac_free_communicator(global_comm);

  MPI_Finalize();

  return TEST_EXIT_CODE;
}

static void submain_tgt(struct communicator * global_comm,
                        struct communicator * domain_comm) {

   struct grid * grid;
   unsigned num_cells[2] = {NUM_CELLS_Y, NUM_CELLS_X};

   grid = yac_reg2d_grid_new(global_coordinates_x, global_coordinates_y,
                             num_cells, cyclic);

   struct subdomain subdomain;
   int global_corner_ids[NUM_CELLS_Y+1][NUM_CELLS_X+1];
   unsigned corner_is_on_rank[NUM_CELLS_Y+1][NUM_CELLS_X+1];
   int global_cell_ids[NUM_CELLS_Y][NUM_CELLS_X];
   unsigned cell_is_on_rank[NUM_CELLS_Y][NUM_CELLS_X];

   for (unsigned i = 0, id = 0; i < NUM_CELLS_Y + 1; ++i) {
     for (unsigned j = 0; j < NUM_CELLS_X + 1; ++j, ++id) {
       global_corner_ids[i][j] = id;
       corner_is_on_rank[i][j] = -1;
     }
   }

   for (unsigned i = 0, id = 0; i < NUM_CELLS_Y; ++i) {
     for (unsigned j = 0; j < NUM_CELLS_X; ++j, ++id) {
       global_cell_ids[i][j] = id;
       cell_is_on_rank[i][j] = -1;
     }
   }

   yac_init_subdomain_unstruct(&subdomain, grid, domain_comm,
                               &global_corner_ids[0][0], &corner_is_on_rank[0][0],
                               &global_cell_ids[0][0], &cell_is_on_rank[0][0], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains = &subdomain;

   yac_init_domain(&domain, domain_comm, &subdomains, 1, "domain 2");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "domain 1", global_comm, 1, 1,
                                  "component 2", "component 1");
   yac_enddef_global_search(search);


   double cell_coordinates_x[NUM_CELLS_X];
   double cell_coordinates_y[NUM_CELLS_Y];

   for (unsigned i = 0; i < NUM_CELLS_X; ++i)
     cell_coordinates_x[i] = ((double)i + 0.5)*YAC_RAD;
   for (unsigned i = 0; i < NUM_CELLS_Y; ++i)
     cell_coordinates_y[i] = ((double)i + 0.5)*YAC_RAD;

   struct points points_cells;

   yac_init_points(&points_cells, grid, CELL, cell_coordinates_x, cell_coordinates_y);

   struct coupling_field * field;
   struct points * points_per_subdomain = &points_cells;

   struct points ** points_set = &points_per_subdomain;

   struct component * component = NULL;
   unsigned local_comp_id = yac_pointer_to_unique_id((void*)component);

   field = yac_coupling_field_new("field 2", local_comp_id, &domain, &points_set, NULL, 1);

   { // test interpolation

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;

      interpolation_gen =
        yac_new_field_interpolation_gen_tgt_mask(&points_set, NULL, 1, "field 2",
                                                 search, "field 1", 0,
                                                 TGT_NO_INTERP,
                                                 mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation =
        yac_get_field_interpolation(interpolation_gen);

      //---------------------
      // do the interpolation
      //---------------------

      double target_field[NUM_CELLS_X][NUM_CELLS_Y];
      double * target_field_ = &target_field[0][0];
      double ** target_data = &target_field_;

      for (unsigned i = 0; i < NUM_CELLS_Y; ++i)
        for (unsigned j = 0; j < NUM_CELLS_X; ++j)
          target_field[i][j] = -1;

      yac_get_field(interpolation, &target_data, 1);

      //---------------------------
      // check interpolation result
      //---------------------------

      // ensure that the weight file has been written
      MPI_Barrier(MPI_COMM_WORLD);

      double ref_target_field[NUM_CELLS_X*NUM_CELLS_Y];

      for (unsigned i = 0; i < NUM_CELLS_X*NUM_CELLS_Y; ++i)
        ref_target_field[i] = 0;

      for (unsigned i = 0; i < NUM_CELLS_Y; ++i)
        for (unsigned j = 0; j < NUM_CELLS_X; ++j)
          for (unsigned k = 0; k < 4; ++k)
            ref_target_field[ref_tgt_address[i][j][k]] +=
              ((double)ref_src_address[i][j][k]) * ref_weights[i][j][k];

      for (unsigned i = 0; i < NUM_CELLS_Y; ++i)
        for (unsigned j = 0; j < NUM_CELLS_X; ++j)
          if (fabs(target_field[i][j] - ref_target_field[i * NUM_CELLS_X + j]) >
              1e-10)
            PUT_ERR("wrong interpolation result\n")

      //--------
      // cleanup
      //--------

      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_field_interpolation(interpolation);

      yac_reset_interpolation_method();
   }

   //--------
   // cleanup
   //--------

   yac_coupling_field_delete(field);
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   yac_free_points(&points_cells);
   yac_delete_grid(grid);
}

static void submain_src(struct communicator * global_comm,
                        struct communicator * domain_comm) {

   char const filename_in[] = "weight_file_in.nc";
   char const filename_out[] = "weight_file_out.nc";

   int my_rank;

   my_rank = yac_get_comm_rank(domain_comm);

    // create a weight file
   if (my_rank == 0) {
     enum yac_location src_location[1] = {CORNER};
     enum yac_location tgt_location = CELL;
     int * tgt_id_fixed = NULL;
     unsigned num_fixed_tgt = 0;
     double * fixed_values = NULL;
     int * num_tgt_per_fixed_value = NULL;
     unsigned num_fixed_values = 0;
     write_weight_file(filename_in, &src_address_file[0][0][0],
                       &tgt_address_file[0][0][0], &weights_file[0][0][0],
                       num_links_file, src_location, 1, (int*)&num_links_file,
                       tgt_id_fixed, num_fixed_tgt, fixed_values,
                       num_tgt_per_fixed_value, num_fixed_values, tgt_location,
                       "domain 1", "domain 2");
   }

   struct grid * grid;
   unsigned num_cells[2] = {NUM_CELLS_X_2+1,NUM_CELLS_Y_2+1};

   int x_offset, y_offset;

   switch (my_rank) {
     default:
     case(0):
       x_offset = 0;
       y_offset = 0;
       break;
     case(1):
       x_offset = NUM_CELLS_X_2 - 1;
       y_offset = 0;
       break;
     case(2):
       x_offset = NUM_CELLS_X_2 - 1;
       y_offset = NUM_CELLS_Y_2 - 1;
       break;
     case(3):
       x_offset = 0;
       y_offset = NUM_CELLS_Y_2 - 1;
       break;
   };

   grid = yac_reg2d_grid_new(global_coordinates_x + x_offset,
                             global_coordinates_y + y_offset, num_cells, cyclic);

   struct subdomain subdomain;

   int global_corner_ids[NUM_CELLS_Y_2+1+1][NUM_CELLS_X_2+1+1];
   unsigned corner_is_on_rank[NUM_CELLS_Y_2+1+1][NUM_CELLS_X_2+1+1];
   int global_cell_ids[NUM_CELLS_Y_2+1][NUM_CELLS_X_2+1];
   unsigned cell_is_on_rank[NUM_CELLS_Y_2+1][NUM_CELLS_X_2+1];

   for (unsigned i = 0; i < NUM_CELLS_Y_2+1+1; ++i)
     for (unsigned j = 0; j < NUM_CELLS_X_2+1+1; ++j)
       global_corner_ids[i][j] =
         (i + y_offset) * (NUM_CELLS_Y + 1) + (j + x_offset);

   for (unsigned i = 0; i < NUM_CELLS_Y_2+1; ++i)
     for (unsigned j = 0; j < NUM_CELLS_X_2+1; ++j)
       global_cell_ids[i][j] = (i + y_offset) * (NUM_CELLS_Y) +
                               (j + x_offset);

   for (unsigned i = 0; i < NUM_CELLS_Y_2+1+1; ++i)
     for (unsigned j = 0; j < NUM_CELLS_X_2+1+1; ++j)
       corner_is_on_rank[i][j] = -1;
   for (unsigned i = 0; i < NUM_CELLS_Y_2+1; ++i)
     for (unsigned j = 0; j < NUM_CELLS_X_2+1; ++j)
       cell_is_on_rank[i][j] = -1;

   // middle point
   switch (my_rank) {
     default:
     case(0):
       corner_is_on_rank[NUM_CELLS_Y_2+1][NUM_CELLS_X_2+1] = 2;
       cell_is_on_rank[NUM_CELLS_Y_2][NUM_CELLS_X_2] = 2;
       break;
     case(1):
       corner_is_on_rank[NUM_CELLS_Y_2+1][0] = 3;
       cell_is_on_rank[NUM_CELLS_Y_2][0] = 3;
       break;
     case(2):
       corner_is_on_rank[0][0] = 0;
       cell_is_on_rank[0][0] = 0;
       break;
     case(3):
       corner_is_on_rank[0][NUM_CELLS_X_2+1] = 1;
       cell_is_on_rank[0][NUM_CELLS_X_2] = 1;
       break;
   };

   // halo edges
   for (unsigned i = 0; i < NUM_CELLS_X_2+1; ++i) {
     switch(my_rank) {
       default:
       case(0):
         corner_is_on_rank[NUM_CELLS_Y_2+1][i] = 3;
         break;
       case(1):
         corner_is_on_rank[NUM_CELLS_Y_2+1][i+1] = 2;
         break;
       case(2):
         corner_is_on_rank[0][i+1] = 1;
         break;
       case(3):
         corner_is_on_rank[0][i] = 0;
         break;
     };
   }
   for (unsigned i = 0; i < NUM_CELLS_Y_2+1; ++i) {
     switch(my_rank) {
       default:
       case(0):
         corner_is_on_rank[i][NUM_CELLS_X_2+1] = 1;
         break;
       case(1):
         corner_is_on_rank[i][0] = 0;
         break;
       case(2):
         corner_is_on_rank[i+1][0] = 3;
         break;
       case(3):
         corner_is_on_rank[i+1][NUM_CELLS_X_2+1] = 2;
         break;
     };
   }

   // halo edges
   for (unsigned i = 0; i < NUM_CELLS_X_2; ++i) {
     switch(my_rank) {
       default:
       case(0):
         cell_is_on_rank[NUM_CELLS_Y_2][i] = 3;
         break;
       case(1):
         cell_is_on_rank[NUM_CELLS_Y_2][i+1] = 2;
         break;
       case(2):
         cell_is_on_rank[0][i+1] = 1;
         break;
       case(3):
         cell_is_on_rank[0][i] = 0;
         break;
     };
   }
   for (unsigned i = 0; i < NUM_CELLS_Y_2; ++i) {
     switch(my_rank) {
       default:
       case(0):
         cell_is_on_rank[i][NUM_CELLS_X_2] = 1;
         break;
       case(1):
         cell_is_on_rank[i][0] = 0;
         break;
       case(2):
         cell_is_on_rank[i+1][0] = 3;
         break;
       case(3):
         cell_is_on_rank[i+1][NUM_CELLS_X_2] = 2;
         break;
     };
   }

   yac_init_subdomain_unstruct(&subdomain, grid, domain_comm,
                               &global_corner_ids[0][0], &corner_is_on_rank[0][0],
                               &global_cell_ids[0][0], &cell_is_on_rank[0][0], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[1] = {&subdomain};

   yac_init_domain(&domain, domain_comm, subdomains, 1, "domain 1");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "domain 2", global_comm, 1, 1,
                                  "component 1", "component 2");
   yac_enddef_global_search(search);

   struct points points_corner;

   yac_init_points(&points_corner, grid, CORNER,
                   global_coordinates_x + x_offset,
                   global_coordinates_y + y_offset);

   struct coupling_field * field;
   struct points * points_per_subdomain = &points_corner;
   struct points ** points_set = &points_per_subdomain;

   struct component * component = NULL;
   unsigned local_comp_id = yac_pointer_to_unique_id((void*)component);

   field = yac_coupling_field_new("field 1", local_comp_id, &domain, &points_set, NULL, 1);

   { // test interpolation using weights from file and avg

      struct interpolation_method * method_file, * method_avg;

      enum yac_location locations[1] = {CORNER};

      method_file =
        yac_new_interpolation_method_file(search, &domain, "file/avg stack",
                                          locations, 1, "weight_file_in.nc", 1);
      method_avg =
        yac_new_interpolation_method_avg_mask(search, &domain, "file/avg stack",
                                              &points_set, NULL, 1, 0, 0, 0, 0);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_file,
                                                                    method_avg,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src(&points_set, 1, "field 1", search,
                                            "field 2", interpolation_method_stack,
                                            filename_out, 1, mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation =
        yac_get_field_interpolation(interpolation_gen);

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[NUM_CELLS_Y_2+1+1][NUM_CELLS_X_2+1+1];

      for (unsigned i = 0; i < NUM_CELLS_Y_2+1+1; ++i)
        for (unsigned j = 0; j < NUM_CELLS_X_2+1+1; ++j)
          source_field_point_data[i][j] =
            (corner_is_on_rank[i][j] == -1)?((double)global_corner_ids[i][j]):
                                            ((double)-1);

      double * source_field_subdomain_data = &source_field_point_data[0][0];
      double ** source_field_subdomain_data_ = &source_field_subdomain_data;
      double *** source_data = &source_field_subdomain_data_;

      yac_put_field(interpolation, &source_data, 1);

      //------------------------------
      // check the written weight file
      //------------------------------

      // ensure that the weight file has been written
      MPI_Barrier(MPI_COMM_WORLD);

      if (my_rank == 0) {
        enum yac_location ref_src_locations[1] = {CORNER};
        enum yac_location ref_tgt_location = CELL;
        check_weight_file(filename_out, &ref_src_address[0][0][0],
                          &ref_tgt_address[0][0][0], &ref_weights[0][0][0],
                          ref_num_links, ref_src_locations, 1, (int*)&ref_num_links,
                          ref_tgt_address_fixed, ref_fixed_values,
                          ref_num_tgt_per_fixed_value, ref_num_fixed_values,
                          ref_tgt_location, "domain 1", "domain 2");
      }

      //---------
      // clean up
      //---------

      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_field_interpolation(interpolation);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_file);

      yac_reset_interpolation_method();
   }

   //---------
   // clean up
   //---------

   yac_coupling_field_delete(field);
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   yac_free_points(&points_corner);
   yac_delete_grid(grid);

   // delete weight file
   if (my_rank == 0) {
     unlink(filename_in);
     unlink(filename_out);
   }
}

static void
generate_input_weights() {

  for (unsigned i = 0; i < NUM_CELLS_Y_2; ++i) {
    for (unsigned j = 0; j < NUM_CELLS_X_2; ++j) {
      src_address_file[i][j][0] =
        0 + (i + NUM_CELLS_Y_2 / 2) * (NUM_CELLS_X + 1) + (j + NUM_CELLS_X_2 / 2);
      src_address_file[i][j][1] =
        1 + (i + NUM_CELLS_Y_2 / 2) * (NUM_CELLS_X + 1) + (j + NUM_CELLS_X_2 / 2);
      src_address_file[i][j][2] =
        0 + (i + 1 + NUM_CELLS_Y_2 / 2) * (NUM_CELLS_X + 1) + (j + NUM_CELLS_X_2 / 2);
      src_address_file[i][j][3] =
        1 + (i + 1 + NUM_CELLS_Y_2 / 2) * (NUM_CELLS_X + 1) + (j + NUM_CELLS_X_2 / 2);
      for (unsigned k = 0; k < 4; ++k)
        tgt_address_file[i][j][k] =
          (i + NUM_CELLS_Y_2 / 2) * NUM_CELLS_X + (j + NUM_CELLS_X_2 / 2);
      for (unsigned k = 0; k < 4; ++k)
        weights_file[i][j][k] = (double)(k + 1) * 0.1;
    }
  }
}

static void
generate_ref_weights() {

  for (unsigned i = 0, idx = 0; i < NUM_CELLS_Y; ++i) {
    for (unsigned j = 0; j < NUM_CELLS_X; ++j, ++idx) {
      ref_src_address[i][j][0] = 0 + i * (NUM_CELLS_X + 1) + j;
      ref_src_address[i][j][1] = 1 + i * (NUM_CELLS_X + 1) + j;
      ref_src_address[i][j][2] = 0 + (i + 1) * (NUM_CELLS_X + 1) + j;
      ref_src_address[i][j][3] = 1 + (i + 1) * (NUM_CELLS_X + 1) + j;
      for (unsigned k = 0; k < 4; ++k) {
        ref_tgt_address[i][j][k] = idx;
        ref_weights[i][j][k] = 0.25;
      }
    }
  }
  for (unsigned i = 0; i < NUM_CELLS_Y_2; ++i)
    for (unsigned j = 0; j < NUM_CELLS_X_2; ++j)
      for (unsigned k = 0; k < 4; ++k)
        ref_weights
          [tgt_address_file[i][j][k]/NUM_CELLS_X]
          [tgt_address_file[i][j][k]%NUM_CELLS_X][k] = weights_file[i][j][k];
}

