/**
 * @file test_grid_bnd_circle.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "tests.h"
#include "geometry.h"
#include "grid.h"
#include "grid_reg2d.h"
#include "grid_unstruct.h"
#include "test_common.h"

// tests whether all corners of grid are withing the bouding circle
static void test_circle(struct grid * grid, struct bounding_circle circle);
// tests whether a point is within the bounding circle
static unsigned point_in_circle(
  double lon, double lat, struct bounding_circle circle);

int main (void) {

   { // test simple case (2x2 regular grid)
      //---------------
      // setup
      //---------------
      double coord_x[3] = {0,1,2};
      double coord_y[3] = {0,1,2};
      unsigned num_cells[2] = {2,2};
      unsigned cyclic[2] = {0,0};
      struct grid * grid;

      for (unsigned i = 0; i < 3; ++i) coord_x[i] *= YAC_RAD;
      for (unsigned i = 0; i < 3; ++i) coord_y[i] *= YAC_RAD;

      grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      //---------------
      // testing
      //---------------
      struct bounding_circle circle;

      yac_get_grid_bounding_circle(grid, &circle);

      test_circle(grid, circle);

      //---------------
      // cleanup
      //---------------
      yac_delete_grid(grid);
   }

   { // test pole example (2x2 unstructured grid that contains the north pole)
      //---------------
      // setup
      //---------------
      double coord_x[9] = {-45,0,45, 270,0,90, 225,180,135};
      double coord_y[9] = {83,85,83, 85,90,85, 83,85,83};
      unsigned num_corners = 9;
      size_t num_cells = 4;
      size_t num_corners_per_cell[4] = {4,4,4,4};
      unsigned cell_to_corner_dep[16] = {0,1,4,3, 1,2,5,4, 3,4,7,6, 4,5,8,7};
      size_t * p_num_corners_per_cell;
      unsigned * p_cell_to_corner_dep;
      struct grid * grid;

      for (unsigned i = 0; i < 9; ++i) coord_x[i] *= YAC_RAD;
      for (unsigned i = 0; i < 9; ++i) coord_y[i] *= YAC_RAD;

      p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      //---------------
      // testing
      //---------------
      struct bounding_circle circle;

      yac_get_grid_bounding_circle(grid, &circle);

      test_circle(grid, circle);

      //---------------
      // cleanup
      //---------------
      yac_delete_grid(grid);
      yac_dep_list_delete(cell_to_vertex);
   }

   { // test ring around the globe on the equator (1x8 unstructured grid around the equator)
      //---------------
      // setup
      //---------------
      double coord_x[16] = {0,45,90,135,180,225,270,315,0,45,90,135,180,225,270,315};
      double coord_y[16] = {-5,-5,-5,-5,-5,-5,-5,-5,5,5,5,5,5,5,5,5};
      unsigned num_corners = 16;
      size_t num_cells = 8;
      size_t num_corners_per_cell[8] = {4,4,4,4,4,4,4,4};
      unsigned cell_to_corner_dep[4*8] = {0,1,9,8, 1,2,10,9, 2,3,11,10, 3,4,12,11,
                                          4,5,13,12, 5,6,14,13, 6,7,15,14, 7,0,8,15};
      size_t * p_num_corners_per_cell;
      unsigned * p_cell_to_corner_dep;
      struct grid * grid;

      for (unsigned i = 0; i < 16; ++i) coord_x[i] *= YAC_RAD;
      for (unsigned i = 0; i < 16; ++i) coord_y[i] *= YAC_RAD;

      p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      p_cell_to_corner_dep = TO_POINTER(cell_to_corner_dep);
      struct dep_list * cell_to_vertex =
        yac_dep_list_new(
          num_cells, p_num_corners_per_cell, p_cell_to_corner_dep);

      grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      //---------------
      // testing
      //---------------
      struct bounding_circle circle;

      yac_get_grid_bounding_circle(grid, &circle);

      test_circle(grid, circle);

      //---------------
      // cleanup
      //---------------
      yac_delete_grid(grid);
      yac_dep_list_delete(cell_to_vertex);
   }

   return TEST_EXIT_CODE;
}

static void test_circle(struct grid * grid, struct bounding_circle circle) {

   unsigned i, num_corners;

   num_corners = yac_get_num_grid_corners(grid);

   for (i = 0; i < num_corners; ++i) {

      double x_coord = yac_get_corner_x_coord(grid, i);
      double y_coord = yac_get_corner_y_coord(grid, i);
      if (!point_in_circle(x_coord, y_coord, circle))
         PUT_ERR("point is not in bounding circle\n");
   }
}

static unsigned point_in_circle(
  double lon, double lat, struct bounding_circle circle) {

   double xyz[3];

   struct sin_cos_angle inc_angle =
      sum_angles_no_check(circle.inc_angle, SIN_COS_TOL);

   if (compare_angles(inc_angle, SIN_COS_M_PI) >= 0) return 1 == 1;

   LLtoXYZ(lon, lat, xyz);

   struct sin_cos_angle angle = get_vector_angle_2(xyz, circle.base_vector);

   return compare_angles(angle, inc_angle) <= 0;
}

