/**
 * @file test_interpolation_method_avg.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdlib.h>
#include <stdio.h>

#include "tests.h"
#include "test_common.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_local.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_fixed.h"
#include "interpolation_method_avg.h"
#include "utils.h"
#include "geometry.h"
#include "test_interpolation_method_common.h"

static double compute_distance_weighted_ref_value(
  double tgt_coord_x, double tgt_coord_y,
  double * src_coords_x, double * src_coords_y,
  double * src_data_values, unsigned num_src_points);

int main (void) {

   { // Test 1a
     // linear point interpolation (fixed is backup) with two source processes and a
     // single target process the source fields have two struct points per subdomain
     // (the test check how the interpolation methods handle NULL target coordinates )

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       | 
      //       08  03  10  04  12  05  13
      //       |       |       |       | 
      //       04--07--05--09--06--11--07
      //       |       |       |       | 
      //       01  00  03  01  05  02  06
      //       |       |       |       | 
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src,
                          * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[5] = {0.5,1.5,2.5};
         double coordinates_y_tgt[3] = {0.5,1.5,2.5};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt", comm_p1_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt", comm_p2_src,
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_p1_tgt,
                                           0,1, "component_tgt", "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points corner_points_src[2];
      struct points cell_points_src[2];

      yac_init_points(corner_points_src+0, grid_p1_src, CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(corner_points_src+1, grid_p2_src, CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      double cell_coordinates_x_src[2][2] = {{0.5,1.5},{1.5,2.5}};
      double cell_coordinates_y_src[2][2] = {{0.5,1.5},{0.5,1.5}};

      yac_init_points(cell_points_src+0, grid_p1_src, CELL,
                      cell_coordinates_x_src[0], cell_coordinates_y_src[0]);
      yac_init_points(cell_points_src+1, grid_p2_src, CELL,
                      cell_coordinates_x_src[1], cell_coordinates_y_src[1]);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[2][2] = {{corner_points_src+0, cell_points_src+0},
                                                    {corner_points_src+1, cell_points_src+1}};
      struct points ** point_sets[2][1] = {{points_per_subdomain[0]},
                                           {points_per_subdomain[1]}};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_avg[2];
      struct interpolation_method * method_fixed[2];

      method_avg[0] = yac_new_interpolation_method_avg(search[0], &l_domain_p1_src, "avg stack", point_sets[0], NULL, 2, 0, 0, 0);
      method_avg[1] = yac_new_interpolation_method_avg(search[1], &l_domain_p2_src, "avg stack", point_sets[1], NULL, 2, 0, 0, 0);
      method_fixed[0] = yac_new_interpolation_method_fixed(1337, search[0]);
      method_fixed[1] = yac_new_interpolation_method_fixed(1337, search[1]);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][3] = {{method_avg[0],
                                                           method_fixed[0],
                                                           NULL},
                                                          {method_avg[1],
                                                           method_fixed[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][3] = {{NULL, NULL, NULL},
                                                                    {NULL, NULL, NULL}};
      int global_ids[2][9] = {{0,1,2,3,4,5,6,7,8}, {0,1,2,3,4,5,6,7,8}};
      unsigned num_points = 9;
      unsigned num_interpolated_points[2][2];

      yac_do_search_interpolation_method(
         method_stack[0], interpolation_plan_element_stack[0], global_ids[0],
         NULL, CORNER, num_points, num_interpolated_points[0], 2);
      yac_do_search_interpolation_method(
         method_stack[1], interpolation_plan_element_stack[1], global_ids[1],
         NULL, CORNER, num_points, num_interpolated_points[1], 2);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,2,4,5,6,8,9,10}},
                                     {{1,5,9,10,-1,-1,-1,-1,-1},
                                      {1,2,3,5,6,7,9,10,11}}};
      double * subdomain_in_data[2][2] = {{in_data_raw[0][1], NULL}, {in_data_raw[1][1], NULL}};
      double * halo_in_data[2][2] = {{in_data_raw[0][0], NULL}, {in_data_raw[1][0], NULL}};
      double ** in_data[2][2] = {{halo_in_data[0], subdomain_in_data[0]},
                                 {halo_in_data[1], subdomain_in_data[1]}};
      double send_buffer[2][9];

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2][2];
      struct interpolation_method_data_request * data_request[2] = {data_request_[0],
                                                                    data_request_[1]};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request + 0, 2);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request + 1, 2);

      struct interpolation_method_data_request ref_data_request[2][2] =
         {{{0, NULL, NULL}, {0, NULL, NULL}},
          {{0, NULL, NULL}, {0, NULL, NULL}}};

      check_data_request(data_request[0], ref_data_request[0], 2);
      check_data_request(data_request[1], ref_data_request[1], 2);

      unsigned ref_num_interpolated_points[2][2] = {{0,0},{0,num_points}};
      unsigned ref_num_interpolated_points_per_process[2] =
         {ref_num_interpolated_points[0][0] + ref_num_interpolated_points[0][1],
          ref_num_interpolated_points[1][0] + ref_num_interpolated_points[1][1]};
      int ref_global_ids[2][9] = {{0},{0,1,2,3,4,5,6,7,8}};
      double ref_out_data[2][9] = {{0},{1337,1337,1337,1337,1337,1337,1337,1337,1337}};

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < 2; ++j)
            if (ref_num_interpolated_points[i][j] != num_interpolated_points[i][j])
               PUT_ERR("wrong number of interpolated points\n");

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < ref_num_interpolated_points_per_process[i]; ++j) {

            if (ref_global_ids[i][j] != global_ids[i][j])
               PUT_ERR("wrong global id order\n");

            if (double_are_unequal(send_buffer[i][j], ref_out_data[i][j]))
               PUT_ERR("wrong result buffer\n");
         }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 2);
      }

      yac_free_points(corner_points_src+0);
      yac_free_points(corner_points_src+1);
      yac_free_points(cell_points_src+0);
      yac_free_points(cell_points_src+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();
   }

   { // Test 1b (as 1a, but with distance weighting
     // linear point interpolation (fixed is backup) with two source processes and a
     // single target process the source fields have two struct points per subdomain
     // (the test check how the interpolation methods handle NULL target coordinates )

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src,
                          * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[5] = {0.5,1.5,2.5};
         double coordinates_y_tgt[3] = {0.5,1.5,2.5};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt", comm_p1_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt", comm_p2_src,
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_p1_tgt,
                                           0,1, "component_tgt", "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points corner_points_src[2];
      struct points cell_points_src[2];

      yac_init_points(corner_points_src+0, grid_p1_src, CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(corner_points_src+1, grid_p2_src, CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      double cell_coordinates_x_src[2][2] = {{0.5,1.5},{1.5,2.5}};
      double cell_coordinates_y_src[2][2] = {{0.5,1.5},{0.5,1.5}};

      yac_init_points(cell_points_src+0, grid_p1_src, CELL,
                      cell_coordinates_x_src[0], cell_coordinates_y_src[0]);
      yac_init_points(cell_points_src+1, grid_p2_src, CELL,
                      cell_coordinates_x_src[1], cell_coordinates_y_src[1]);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[2][2] = {{corner_points_src+0, cell_points_src+0},
                                                    {corner_points_src+1, cell_points_src+1}};
      struct points ** point_sets[2][1] = {{points_per_subdomain[0]},
                                           {points_per_subdomain[1]}};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_avg[2];
      struct interpolation_method * method_fixed[2];

      method_avg[0] = yac_new_interpolation_method_avg(search[0], &l_domain_p1_src, "avg stack", point_sets[0], NULL, 2, 0, 1, 0);
      method_avg[1] = yac_new_interpolation_method_avg(search[1], &l_domain_p2_src, "avg stack", point_sets[1], NULL, 2, 0, 1, 0);
      method_fixed[0] = yac_new_interpolation_method_fixed(1337, search[0]);
      method_fixed[1] = yac_new_interpolation_method_fixed(1337, search[1]);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][3] = {{method_avg[0],
                                                           method_fixed[0],
                                                           NULL},
                                                          {method_avg[1],
                                                           method_fixed[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][3] = {{NULL, NULL, NULL},
                                                                    {NULL, NULL, NULL}};
      int global_ids[2][9] = {{0,1,2,3,4,5,6,7,8}, {0,1,2,3,4,5,6,7,8}};
      unsigned num_points = 9;
      unsigned num_interpolated_points[2][2];

      yac_do_search_interpolation_method(
         method_stack[0], interpolation_plan_element_stack[0], global_ids[0],
         NULL, CORNER, num_points, num_interpolated_points[0], 2);
      yac_do_search_interpolation_method(
         method_stack[1], interpolation_plan_element_stack[1], global_ids[1],
         NULL, CORNER, num_points, num_interpolated_points[1], 2);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,2,4,5,6,8,9,10}},
                                     {{1,5,9,10,-1,-1,-1,-1,-1},
                                      {1,2,3,5,6,7,9,10,11}}};
      double * subdomain_in_data[2][2] = {{in_data_raw[0][1], NULL}, {in_data_raw[1][1], NULL}};
      double * halo_in_data[2][2] = {{in_data_raw[0][0], NULL}, {in_data_raw[1][0], NULL}};
      double ** in_data[2][2] = {{halo_in_data[0], subdomain_in_data[0]},
                                 {halo_in_data[1], subdomain_in_data[1]}};
      double send_buffer[2][9];

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2][2];
      struct interpolation_method_data_request * data_request[2] = {data_request_[0],
                                                                    data_request_[1]};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request + 0, 2);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request + 1, 2);

      struct interpolation_method_data_request ref_data_request[2][2] =
         {{{0, NULL, NULL}, {0, NULL, NULL}},
          {{0, NULL, NULL}, {0, NULL, NULL}}};

      check_data_request(data_request[0], ref_data_request[0], 2);
      check_data_request(data_request[1], ref_data_request[1], 2);

      unsigned ref_num_interpolated_points[2][2] = {{0,0},{0,num_points}};
      unsigned ref_num_interpolated_points_per_process[2] =
         {ref_num_interpolated_points[0][0] + ref_num_interpolated_points[0][1],
          ref_num_interpolated_points[1][0] + ref_num_interpolated_points[1][1]};
      int ref_global_ids[2][9] = {{0},{0,1,2,3,4,5,6,7,8}};
      double ref_out_data[2][9] = {{0},{1337,1337,1337,1337,1337,1337,1337,1337,1337}};

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < 2; ++j)
            if (ref_num_interpolated_points[i][j] != num_interpolated_points[i][j])
               PUT_ERR("wrong number of interpolated points\n");

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < ref_num_interpolated_points_per_process[i]; ++j) {

            if (double_are_unequal(ref_global_ids[i][j], global_ids[i][j]))
               PUT_ERR("wrong global id order\n");

            if (double_are_unequal(send_buffer[i][j], ref_out_data[i][j]))
               PUT_ERR("wrong result buffer\n");
         }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 2);
      }

      yac_free_points(corner_points_src+0);
      yac_free_points(corner_points_src+1);
      yac_free_points(cell_points_src+0);
      yac_free_points(cell_points_src+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();
   }

   { // Test 2a
     // linear point interpolation with two source processes and a single target process
     // (test check how the interpolation method handles the case when a source
     //  process has only masked target points)

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //  target mask (X == masked)
      //
      //          0-----X-----X
      //          |     |     |
      //          |     |     |
      //          |     |     |
      //          0-----X-----X
      //          |     |     |
      //          |     |     |
      //          |     |     |
      //          0-----X-----X
      //
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src, * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         for (unsigned i = 0; i < 2; ++i ) {
            for (unsigned j = 0; j < 3; ++j ) {
               coordinates_x_src[i][j] *= YAC_RAD;
               coordinates_y_src[i][j] *= YAC_RAD;
            }
         }

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[3] = {0.5*YAC_RAD,1.5*YAC_RAD,2.5*YAC_RAD};
         double coordinates_y_tgt[3] = {0.5*YAC_RAD,1.5*YAC_RAD,2.5*YAC_RAD};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt", comm_p1_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt", comm_p2_src,
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_p1_tgt,
                                           0,1, "component_tgt", "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points points_src[2];

      yac_init_points(points_src+0, grid_p1_src, CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(points_src+1, grid_p2_src, CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
      struct points ** point_sets[2][1] = {{points_per_subdomain[0]},
                                           {points_per_subdomain[1]}};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_avg[2];

      method_avg[0] = yac_new_interpolation_method_avg(search[0], &l_domain_p1_src, "avg stack", point_sets[0], NULL, 1, 0, 0, 1);
      method_avg[1] = yac_new_interpolation_method_avg(search[1], &l_domain_p2_src, "avg stack", point_sets[1], NULL, 1, 0, 0, 1);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_avg[0],
                                                           NULL},
                                                          {method_avg[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] = {{NULL, NULL},
                                                                    {NULL, NULL}};
      int global_ids[3] = {0,3,6};
      unsigned num_interpolated_points[2];
      double coordinates_x_tgt_cells[9] = {0.5*YAC_RAD, 0.5*YAC_RAD, 0.5*YAC_RAD};
      double coordinates_y_tgt_cells[9] = {0.5*YAC_RAD, 1.5*YAC_RAD, 2.5*YAC_RAD};
      double coordinates_xyz_tgt_cells[9][3];

      for (size_t i = 0; i < 9; ++i)
        LLtoXYZ(coordinates_x_tgt_cells[i],
                coordinates_y_tgt_cells[i],
                coordinates_xyz_tgt_cells[i]);

      yac_do_search_interpolation_method(
         method_stack[0], interpolation_plan_element_stack[0],
         global_ids, coordinates_xyz_tgt_cells, CORNER, 3,
         num_interpolated_points+0, 1);
      yac_do_search_interpolation_method(
         method_stack[1], interpolation_plan_element_stack[1], NULL, NULL,
         CORNER, 0, num_interpolated_points+1, 1);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,2,4,5,6,8,9,10}},
                                     {{1,5,9,10,-1,-1,-1,-1,-1},
                                      {1,2,3,5,6,7,9,10,11}}};
      double * subdomain_in_data[2][1] = {{in_data_raw[0][1]}, {in_data_raw[1][1]}};
      double * halo_data[2][1] = {{in_data_raw[0][0]}, {in_data_raw[1][0]}};
      double ** in_data[2][2] = {{halo_data[0], subdomain_in_data[0]},
                                 {halo_data[1], subdomain_in_data[1]}};
      double send_buffer[2][3] = {{-1,-1,-1},{-1,-1,-1}};

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2];
      struct interpolation_method_data_request * data_request[2][1] = {{data_request_+0},
                                                                       {data_request_+1}};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request[0], 1);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request[1], 1);

      struct interpolation_method_data_request ref_data_request = {0, NULL, NULL};

      check_data_request(data_request[0][0], &ref_data_request, 1);
      check_data_request(data_request[1][0], &ref_data_request, 1);

      unsigned ref_num_interpolated_points[2] = {2,0};
      double ref_out_data[2][2] = {{2.5,6.5},{-1}};

      int ref_global_ids[2][2] = {{0,3},{-1}};

      for (unsigned i = 0; i < 2; ++i)
         if (ref_num_interpolated_points[i] != num_interpolated_points[i])
            PUT_ERR("wrong number of interpolated points\n");

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < ref_num_interpolated_points[i]; ++j) {

            if (double_are_unequal(ref_global_ids[i][j], global_ids[j]))
               PUT_ERR("wrong global id order\n");

            if (double_are_unequal(send_buffer[i][j], ref_out_data[i][j]))
               PUT_ERR("wrong result buffer d\n");
         }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j][0], 1);
      }

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();
   }

   { // Test 2b, as 2a but with distance weighting
     // linear point interpolation with two source processes and a single target process
     // (test check how the interpolation method handles the case when a source
     //  process has only masked target points)

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       | 
      //       08  03  10  04  12  05  13
      //       |       |       |       | 
      //       04--07--05--09--06--11--07
      //       |       |       |       | 
      //       01  00  03  01  05  02  06
      //       |       |       |       | 
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //  target mask (X == masked)
      //
      //          0-----X-----X
      //          |     |     |
      //          |     |     |
      //          |     |     |
      //          0-----X-----X
      //          |     |     |
      //          |     |     |
      //          |     |     |
      //          0-----X-----X
      //
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src, * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         for (unsigned i = 0; i < 2; ++i ) {
            for (unsigned j = 0; j < 3; ++j ) {
               coordinates_x_src[i][j] *= YAC_RAD;
               coordinates_y_src[i][j] *= YAC_RAD;
            }
         }

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[3] = {0.5*YAC_RAD,1.5*YAC_RAD,2.5*YAC_RAD};
         double coordinates_y_tgt[3] = {0.5*YAC_RAD,1.5*YAC_RAD,2.5*YAC_RAD};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt", comm_p1_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt", comm_p2_src,
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_p1_tgt,
                                           0,1, "component_tgt", "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points points_src[2];

      yac_init_points(points_src+0, grid_p1_src, CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(points_src+1, grid_p2_src, CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
      struct points ** point_sets[2][1] = {{points_per_subdomain[0]},
                                           {points_per_subdomain[1]}};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_avg[2];

      method_avg[0] = yac_new_interpolation_method_avg(search[0], &l_domain_p1_src, "avg stack", point_sets[0], NULL, 1, 0, 1, 1);
      method_avg[1] = yac_new_interpolation_method_avg(search[1], &l_domain_p2_src, "avg stack", point_sets[1], NULL, 1, 0, 1, 1);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_avg[0],
                                                           NULL},
                                                          {method_avg[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] = {{NULL, NULL},
                                                                    {NULL, NULL}};
      int global_ids[3] = {0,3,6};
      unsigned num_interpolated_points[2];
      double coordinates_x_tgt_cells[9] = {0.5*YAC_RAD, 0.5*YAC_RAD, 0.5*YAC_RAD};
      double coordinates_y_tgt_cells[9] = {0.5*YAC_RAD, 1.5*YAC_RAD, 2.5*YAC_RAD};
      double coordinates_xyz_tgt_cells[9][3];

      for (size_t i = 0; i < 9; ++i)
        LLtoXYZ(coordinates_x_tgt_cells[i],
                coordinates_y_tgt_cells[i],
                coordinates_xyz_tgt_cells[i]);

      yac_do_search_interpolation_method(
         method_stack[0], interpolation_plan_element_stack[0], global_ids,
         coordinates_xyz_tgt_cells, CORNER, 3, num_interpolated_points+0, 1);
      yac_do_search_interpolation_method(
         method_stack[1], interpolation_plan_element_stack[1], NULL, NULL,
         CORNER, 0, num_interpolated_points+1, 1);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,2,4,5,6,8,9,10}},
                                     {{1,5,9,10,-1,-1,-1,-1,-1},
                                      {1,2,3,5,6,7,9,10,11}}};
      double * subdomain_in_data[2][1] = {{in_data_raw[0][1]}, {in_data_raw[1][1]}};
      double * halo_data[2][1] = {{in_data_raw[0][0]}, {in_data_raw[1][0]}};
      double ** in_data[2][2] = {{halo_data[0], subdomain_in_data[0]},
                                 {halo_data[1], subdomain_in_data[1]}};
      double send_buffer[2][3] = {{-1,-1,-1},{-1,-1,-1}};

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2];
      struct interpolation_method_data_request * data_request[2][1] = {{data_request_+0},
                                                                       {data_request_+1}};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request[0], 1);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request[1], 1);

      struct interpolation_method_data_request ref_data_request = {0, NULL, NULL};

      check_data_request(data_request[0][0], &ref_data_request, 1);
      check_data_request(data_request[1][0], &ref_data_request, 1);

      unsigned ref_num_interpolated_points[2] = {2,0};
      double ref_out_data[2][2] = {{-1},{-1}};
      ref_out_data[0][0] =
        compute_distance_weighted_ref_value(
          0.5*YAC_RAD, 0.5*YAC_RAD,
          (double[]){0.0*YAC_RAD, 1.0*YAC_RAD, 1.0*YAC_RAD, 0.0*YAC_RAD},
          (double[]){0.0*YAC_RAD, 0.0*YAC_RAD, 1.0*YAC_RAD, 1.0*YAC_RAD},
          (double[]){0.0, 1.0, 5.0, 4.0}, 4);
      ref_out_data[0][1] =
        compute_distance_weighted_ref_value(
          0.5*YAC_RAD, 1.5*YAC_RAD,
          (double[]){0.0*YAC_RAD, 1.0*YAC_RAD, 1.0*YAC_RAD, 0.0*YAC_RAD},
          (double[]){1.0*YAC_RAD, 1.0*YAC_RAD, 2.0*YAC_RAD, 2.0*YAC_RAD},
          (double[]){4.0, 5.0, 9.0, 8.0}, 4);

      int ref_global_ids[2][2] = {{0,3},{-1}};

      for (unsigned i = 0; i < 2; ++i)
         if (ref_num_interpolated_points[i] != num_interpolated_points[i])
            PUT_ERR("wrong number of interpolated points\n");

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < ref_num_interpolated_points[i]; ++j) {

            if (ref_global_ids[i][j] != global_ids[j])
               PUT_ERR("wrong global id order\n");

            double const tol = 1.0e-6;

            if ( fabs(send_buffer[i][j]-ref_out_data[i][j]) > tol )
               PUT_ERR("wrong result buffer d\n");
         }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j][0], 1);
      }

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();
   }

   return TEST_EXIT_CODE;
}

static double compute_distance_weighted_ref_value(
  double tgt_coord_x, double tgt_coord_y,
  double * src_coords_x, double * src_coords_y,
  double * src_data_values, unsigned num_src_points) {

  double ref_value = 0.0;

  double tgt_coord_xyz[3];
  LLtoXYZ(tgt_coord_x, tgt_coord_y, tgt_coord_xyz);

  double inv_dist_sum = 0.0;

  for (unsigned i = 0; i < num_src_points; ++i) {

    double src_coord_xyz[3];
    LLtoXYZ(src_coords_x[i], src_coords_y[i], src_coord_xyz);

    double inv_dist = 1.0 / get_vector_angle(tgt_coord_xyz, src_coord_xyz);

    ref_value += src_data_values[i] * inv_dist;
    inv_dist_sum += inv_dist;
  }

  return ref_value / inv_dist_sum;
}

