/**
 * @file test_interpolation_method_conserv.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_local.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_conserv.h"
#include "utils.h"
#include "test_interpolation_method_common.h"

int main (void) {

   static double tol = 1.0e-10;

   { // conservative remapping with one source and one target process

      // the global source domain is a 2x2 grid:
      //       06--10--07--11--08
      //       |       |       |
      //       06  02  08  03  09
      //       |       |       |
      //       03--05--04--07--05
      //       |       |       |
      //       01  00  03  01  04
      //       |       |       |
      //       00--00--01--02--02
      //
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_src, * comm_tgt;

      comm_src = yac_local_comm_new();
      comm_tgt = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_src, * comm_named_tgt;

      comm_named_src = yac_named_local_comm_new(comm_src, "src");
      comm_named_tgt = yac_named_local_comm_new(comm_tgt, "tgt");

       // define domains on source

       struct grid * grid_src;
       unsigned num_cells_src[2] = {2,2};
       unsigned cyclic_src[2] = {0,0};
       double coordinates_x_src[3] = {-1,0,1};
       double coordinates_y_src[3] = {-1,0,1};

       for ( unsigned i = 0; i < 3; ++i ) {
          coordinates_x_src[i] *= YAC_RAD;
          coordinates_y_src[i] *= YAC_RAD;
       }

       grid_src = yac_reg2d_grid_new(coordinates_x_src, coordinates_y_src,
                                     num_cells_src, cyclic_src);

       struct subdomain subdomain_src;

       int global_corner_ids_src[9] = {0,1,2, 3,4,5, 6,7,8};
       unsigned corner_is_on_rank_src[9] = {-1,-1,-1, -1,-1,-1, -1,-1,-1};
       int global_cell_ids_src[4] = {0,1,2,3};
       unsigned cell_is_on_rank_src[4] = {-1,-1,-1,-1};

       yac_init_subdomain_unstruct(&subdomain_src, grid_src, comm_named_src,
                                   global_corner_ids_src, corner_is_on_rank_src,
                                   global_cell_ids_src, cell_is_on_rank_src,
                                   NULL, NULL);

       struct domain l_domain_src;

       struct subdomain * blocks_src = &subdomain_src;

       yac_init_domain(&l_domain_src, comm_named_src, &blocks_src, 1, "domain_src");

       // define a domain on the target

       struct grid * grid_tgt;
       unsigned num_cells_tgt[2] = {1,1};
       unsigned cyclic_tgt[2] = {0,0};
       double coordinates_x_tgt[2] = {-0.5,0.5};
       double coordinates_y_tgt[2] = {-0.5,0.5};

       for ( unsigned i = 0; i < 2; ++i ) {
          coordinates_x_tgt[i] *= YAC_RAD;
          coordinates_y_tgt[i] *= YAC_RAD;
       }

       grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                     num_cells_tgt, cyclic_tgt);

       struct subdomain subdomain_tgt;

       int global_corner_ids_tgt[4] = {0,1,2,3};
       unsigned corner_is_on_rank_tgt[4] = {-1,-1,-1,-1};
       int global_cell_ids_tgt[1] = {0};
       unsigned cell_is_on_rank_tgt[1] = {-1};

       yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                   global_corner_ids_tgt, corner_is_on_rank_tgt,
                                   global_cell_ids_tgt, cell_is_on_rank_tgt,
                                   NULL, NULL);

       struct domain l_domain_tgt;

       struct subdomain * blocks_tgt = {&subdomain_tgt};

       yac_init_domain(&l_domain_tgt, comm_named_tgt, &blocks_tgt, 1, "domain_tgt");

       // finish domain definition
       yac_enddef_domain(&l_domain_src);
       yac_enddef_domain(&l_domain_tgt);

       //-------------------------------------
       // init global search on both processes
       //-------------------------------------

       struct global_search * search[2];

       search[0] = yac_global_search_new(&l_domain_src, "domain_tgt", comm_src,
                                         1,0, "component_src", "component_tgt");
       search[1] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_tgt,
                                         0,1, "component_tgt", "component_src");

       //---------------------------------------
       // finish initialisation of global search
       //---------------------------------------

       yac_enddef_global_search(search[0]);
       yac_enddef_global_search(search[1]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points cell_points_src;
      yac_init_points(&cell_points_src, grid_src, CELL, NULL, NULL);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain = &cell_points_src;
      struct points ** point_sets = &points_per_subdomain;

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_conserv;

      method_conserv =
        yac_new_interpolation_method_conserv(search[0], &l_domain_src,
                                         "conserv stack", &point_sets, NULL, 1,
                                         0, 1, 0, 0, CONSERV_DESTAREA);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2] = {method_conserv, NULL};
      struct interpolation_plan_element * interpolation_plan_element_stack[2] = {NULL, NULL};
      int global_ids[1] = {0};
      unsigned num_points = 1;
      unsigned num_interpolated_points[1];

      yac_do_search_interpolation_method(method_stack, interpolation_plan_element_stack,
                                     global_ids, NULL, CELL, num_points,
                                     num_interpolated_points, 1);
      yac_enddef_interpolation_method(method_stack);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[4] = {1,2,3,4};
      double * subdomain_in_data[1] = {in_data_raw};
      double ** in_data[2] = {NULL, subdomain_in_data};
      double send_buffer[1];

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack,
                                             in_data, send_buffer);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_;
      struct interpolation_method_data_request * data_request = &data_request_;

      yac_get_source_exchange_data_interpolation_method(method_stack, &data_request, 1);

      struct interpolation_method_data_request ref_data_request = {0, NULL, NULL};

      check_data_request(data_request, &ref_data_request, 1);

      unsigned ref_num_interpolated_points = 1;
      int ref_global_ids[1] = {0};
      double ref_out_data[1] = {2.5};

      if (ref_num_interpolated_points != num_interpolated_points[0])
         PUT_ERR("wrong number of interpolated points\n");

      if (ref_global_ids[0] != global_ids[0])
         PUT_ERR("wrong global id order\n");

      if ( fabs(send_buffer[0]-ref_out_data[0]) > tol )
         PUT_ERR("wrong result buffer\n");

      //---------------
      // cleanup
      //---------------

      for (size_t i = 0; i < sizeof(method_stack) /
                      sizeof(method_stack[0]); ++i)
         if (method_stack[i] != NULL)
            yac_delete_interpolation_method(method_stack[i]);
      for (size_t i = 0; i < sizeof(interpolation_plan_element_stack) /
                      sizeof(interpolation_plan_element_stack[0]); ++i) {
         yac_free_interpolation_plan_element(interpolation_plan_element_stack[i]);
         free(interpolation_plan_element_stack[i]);
      }
      delete_interpolation_method_data_request_stack(data_request, 1);

      yac_free_points(&cell_points_src);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   { // conservative remapping with two source and one target processes

      // the global source domain is a 2x2 grid:
      //
      //       06--10--07--11--08
      //       |       |       |
      //       06  02  08  03  09
      //       |       |       |
      //       03--05--04--07--05
      //       |       |       |
      //       01  00  03  01  04
      //       |       |       |
      //       00--00--01--02--02
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_src[2], * comm_tgt;

      comm_src[0] = yac_local_comm_new();
      comm_src[1] = yac_local_comm_new();
      comm_tgt = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_src[2], * comm_named_tgt;

      comm_named_src[0] = yac_named_local_comm_new(comm_src[0], "src");
      comm_named_src[1] = yac_named_local_comm_new(comm_src[1], "src");
      comm_named_tgt = yac_named_local_comm_new(comm_tgt, "tgt");

      // define domains on source

      struct grid * grid_src[2];
      unsigned num_cells_src[2] = {2,2};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[3] = {-1,0,1};
      double coordinates_y_src[3] = {-1,0,1};

      for ( unsigned i = 0; i < 3; ++i ) {
         coordinates_x_src[i] *= YAC_RAD;
         coordinates_y_src[i] *= YAC_RAD;
      }

      grid_src[0] = yac_reg2d_grid_new(coordinates_x_src, coordinates_y_src,
                                       num_cells_src, cyclic_src);

      grid_src[1] = yac_reg2d_grid_new(coordinates_x_src, coordinates_y_src,
                                       num_cells_src, cyclic_src);

      struct subdomain subdomain_src[2];

      int global_corner_ids_src[9] = {0,1,2, 3,4,5, 6,7,8};
      unsigned corner_is_on_rank_src[2][9] = {{-1,-1, 1, -1,-1, 1, -1,-1, 1},
                                              { 0,-1,-1,  0,-1,-1,  0,-1,-1}};
      int global_cell_ids_src[4] = {0,1,2,3};
      unsigned cell_is_on_rank_src[2][4] = {{-1, 1,-1, 1},
                                            { 0,-1, 0,-1}};

      yac_init_subdomain_unstruct(&subdomain_src[0], grid_src[0], comm_named_src[0],
                                  global_corner_ids_src, corner_is_on_rank_src[0],
                                  global_cell_ids_src, cell_is_on_rank_src[0],
                                  NULL, NULL);

      yac_init_subdomain_unstruct(&subdomain_src[1], grid_src[1], comm_named_src[1],
                                  global_corner_ids_src, corner_is_on_rank_src[1],
                                  global_cell_ids_src, cell_is_on_rank_src[1],
                                  NULL, NULL);

      struct domain l_domain_src[2];

      struct subdomain * blocks_src[2] = {&subdomain_src[0],&subdomain_src[1]};

      yac_init_domain(&l_domain_src[0], comm_named_src[0], &blocks_src[0], 1, "domain_src");
      yac_init_domain(&l_domain_src[1], comm_named_src[1], &blocks_src[1], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt;
      unsigned num_cells_tgt[2] = {1,1};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[5] = {-0.5,0.5};
      double coordinates_y_tgt[3] = {-0.5,0.5};

      for ( unsigned i = 0; i < 2; ++i ) {
         coordinates_x_tgt[i] *= YAC_RAD;
         coordinates_y_tgt[i] *= YAC_RAD;
      }

      grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                    num_cells_tgt, cyclic_tgt);

      struct subdomain subdomain_tgt;

      int global_corner_ids_tgt[4] = {0,1,2,3};
      unsigned corner_is_on_rank_tgt[4] = {-1,-1,-1,-1};
      int global_cell_ids_tgt[1] = {0};
      unsigned cell_is_on_rank_tgt[1] = {-1};

      yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                  global_corner_ids_tgt, corner_is_on_rank_tgt,
                                  global_cell_ids_tgt, cell_is_on_rank_tgt,
                                  NULL, NULL);

      struct domain l_domain_tgt;

      struct subdomain * blocks_tgt = {&subdomain_tgt};

      yac_init_domain(&l_domain_tgt, comm_named_tgt, &blocks_tgt, 1, "domain_tgt");

      // finish domain definition
      yac_enddef_domain(&l_domain_src[0]);
      yac_enddef_domain(&l_domain_src[1]);
      yac_enddef_domain(&l_domain_tgt);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[3];

      search[0] = yac_global_search_new(&l_domain_src[0], "domain_tgt", comm_src[0],
                                        1,0, "component_src", "component_tgt");
      search[1] = yac_global_search_new(&l_domain_src[1], "domain_tgt", comm_src[1],
                                        1,0, "component_src", "component_tgt");
      search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_tgt,
                                        0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      yac_enddef_global_search(search[0]);
      yac_enddef_global_search(search[1]);
      yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points cell_points_src[2];
      yac_init_points(&cell_points_src[0], grid_src[0], CELL, NULL, NULL);
      yac_init_points(&cell_points_src[1], grid_src[1], CELL, NULL, NULL);

      //-------------------------------------
      // set up point set
      //-------------------------------------

      struct points * points_per_subdomain[2] = {&cell_points_src[0],&cell_points_src[1]};
      struct points ** point_sets[2] = {&points_per_subdomain[0],&points_per_subdomain[1]};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_conserv[2];

      method_conserv[0] =
        yac_new_interpolation_method_conserv(search[0], &l_domain_src[0],
                                         "conserv stack", &point_sets[0], NULL,
                                         1, 0, 1, 0, 0, CONSERV_DESTAREA);
      method_conserv[1] =
        yac_new_interpolation_method_conserv(search[1], &l_domain_src[1],
                                         "conserv stack", &point_sets[1], NULL,
                                         1, 0, 1, 0, 0, CONSERV_DESTAREA);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_conserv[0], NULL},{method_conserv[1], NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] = {{NULL, NULL},{NULL, NULL}};
      int global_ids[2][1] = {{0},{0}};
      unsigned num_points = 1;
      unsigned num_interpolated_points[2][1];

      yac_do_search_interpolation_method(method_stack[0], interpolation_plan_element_stack[0],
                                     global_ids[0], NULL, CELL, num_points,
                                     num_interpolated_points[0], 1);

      yac_do_search_interpolation_method(method_stack[1], interpolation_plan_element_stack[1],
                                     global_ids[1], NULL, CELL, num_points,
                                     num_interpolated_points[1], 1);

      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][4] = {{ 1,-1, 3,-1},
                                  { 2, 4}};
      double * remote_data[1] = {in_data_raw[1]};
      double * subdomain_in_data[1] = {in_data_raw[0]};
      double ** in_data[2][2] = {{remote_data, subdomain_in_data},{NULL, NULL}};
      double send_buffer[2][1] = {{-1},{-1}};

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0],
                                             in_data[0], send_buffer[0]);

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1],
                                             in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2];
      struct interpolation_method_data_request * data_request[2] =
         {&data_request_[0],&data_request_[1]};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], &data_request[0], 1);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], &data_request[1], 1);

      int ref_requested_global_ids[2] = {1,3};
      int ref_requested_ranks[2] = {1,1};

      struct interpolation_method_data_request ref_data_request[2] =
         {{2, ref_requested_global_ids,  ref_requested_ranks}, {0,NULL,NULL}};

      unsigned ref_num_interpolated_points[2] = {1,0};
      int ref_global_ids[1] = {0};
      double ref_out_data[2] = {2.5,-1};

      for (unsigned j = 0; j < 2; ++j ) {

         check_data_request(data_request[j], &ref_data_request[j], 1);

         if (ref_num_interpolated_points[j] != num_interpolated_points[j][0])
            PUT_ERR("wrong number of interpolated points\n");

         if (ref_global_ids[0] != global_ids[j][0])
            PUT_ERR("wrong global id order\n");

         if ( fabs(send_buffer[j][0] - ref_out_data[j] ) > tol )
            PUT_ERR("wrong result buffer\n");
      }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j ) {
         for (size_t i = 0; i < sizeof(method_stack[j])/sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (size_t i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
              sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 1);
      }

      yac_free_points(&cell_points_src[0]);
      yac_free_points(&cell_points_src[1]);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_src[0]);
      yac_free_domain(&l_domain_src[1]);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_src[0]);
      yac_delete_grid(grid_src[1]);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src[0]);
      yac_free_communicator(comm_src[1]);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src[1]);
      yac_free_communicator(comm_named_src[0]);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   { // conservative remapping with three source and two target processes

      // the global source domain is a 3x3 grid:
      //
      //       12--21--13--22--14--23--15
      //       |       |       |       |
      //       15  06  17  07  19  08  20
      //       |       |       |       |
      //       08--14--09--16--10--18--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // the global target domain is a 2x2 grid:
      //
      //       06--10--07--11--08
      //       |       |       |
      //       06  02  08  03  09
      //       |       |       |
      //       03--05--04--07--05
      //       |       |       |
      //       01  00  03  01  04
      //       |       |       |
      //       00--00--01--02--02
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

#undef NUM_SRC
#undef NUM_TGT
#define NUM_SRC 3
#define NUM_TGT 2
      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{2,3}, {3,3}, {2,3}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[NUM_SRC][4] = {{-1.5,-0.5,0.5},
                                              {-1.5,-0.5,0.5,1.5},
                                              {-0.5,0.5,1.5}};
      double coordinates_y_src[4] = {-1.5,-0.5,0.5,1.5};

      for ( unsigned j = 0; j < NUM_SRC; ++j )
         for ( unsigned i = 0; i < 4; ++i )
            coordinates_x_src[j][i] *= YAC_RAD;

      for ( unsigned i = 0; i < 4; ++i )
         coordinates_y_src[i] *= YAC_RAD;

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src[i], coordinates_y_src,
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][16] = {{0,1,2, 4,5,6, 8,9,10, 12,13,14},
                                                {0,1,2,3, 4,5,6,7, 8,9,10,11, 12,13,14,15},
                                                {1,2,3, 5,6,7, 9,10,11, 13,14,15}};
      unsigned corner_is_on_rank_src[NUM_SRC][16] = {{-1,-1,1, -1,-1,1, -1,-1,1, -1,-1,1},
                                                     {0,-1,-1,2, 0,-1,-1,2, 0,-1,-1,2, 0,-1,-1,2},
                                                     {1,-1,-1, 1,-1,-1, 1,-1,-1, 1,-1,-1}};
      int global_cell_ids_src[NUM_SRC][9] = {{0,1,3,4,6,7},
                                             {0,1,2,3,4,5,6,7,8},
                                             {1,2,4,5,7,8}};
      unsigned cell_is_on_rank_src[NUM_SRC][9] = {{-1,1,-1,1,-1,1},
                                                  {0,-1,2,0,-1,2,0,-1,2},
                                                  {1,-1,1,-1,1,-1}};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_subdomain_unstruct(&subdomain_src[i], grid_src[i], comm_named_src[i],
                                     global_corner_ids_src[i], corner_is_on_rank_src[i],
                                     global_cell_ids_src[i], cell_is_on_rank_src[i],
                                     NULL, NULL);

      struct domain l_domain_src[NUM_SRC];

      struct subdomain * blocks_src[NUM_SRC] = {&subdomain_src[0],
                                                &subdomain_src[1],
                                                &subdomain_src[2]};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[2] = {2,2};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[3] = {-1,0,1};
      double coordinates_y_tgt[3] = {-1,0,1};

      for ( unsigned i = 0; i < 3; ++i ) {
         coordinates_x_tgt[i] *= YAC_RAD;
         coordinates_y_tgt[i] *= YAC_RAD;
      }

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                          num_cells_tgt, cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[9] = {0,1,2,3,4,5,6,7,8};
      unsigned corner_is_on_rank_tgt[NUM_TGT][9] = {{-1,-1,-1, -1,-1,-1, 1,1,1},
                                                    {0,0,0, -1,-1,-1, -1,-1,-1}};
      int global_cell_ids_tgt[4] = {0,1,2,3};
      unsigned cell_is_on_rank_tgt[NUM_TGT][4] = {{-1,-1,1,1},{0,0,-1,-1}};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                     global_corner_ids_tgt, corner_is_on_rank_tgt[i],
                                     global_cell_ids_tgt, cell_is_on_rank_tgt[i],
                                     NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];

      struct subdomain * blocks_tgt[NUM_TGT] = {&subdomain_tgt[0],
                                                &subdomain_tgt[1]};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         search[i] =
           yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                 1,0, "component_src", "component_tgt");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
           yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                 0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points cell_points_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_points(&cell_points_src[i], grid_src[i], CELL, NULL, NULL);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[NUM_SRC] = {&cell_points_src[0],
                                                       &cell_points_src[1],
                                                       &cell_points_src[2]};
      struct points ** point_sets[NUM_SRC] = {&points_per_subdomain[0],
                                              &points_per_subdomain[1],
                                              &points_per_subdomain[2]};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_conserv[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         method_conserv[i] =
          yac_new_interpolation_method_conserv(search[i], &l_domain_src[i],
                                           "conserv stack", &point_sets[i],
                                           NULL, 1, 0, 1, 0, 0, CONSERV_DESTAREA);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[NUM_SRC][2] = {{method_conserv[0], NULL},
                                                                {method_conserv[1], NULL},
                                                                {method_conserv[2], NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[NUM_SRC][2] =
         {{NULL, NULL},{NULL, NULL},{NULL, NULL}};
      int global_ids[NUM_SRC][4] = {{0,1,2,3},{0,1,2,3},{0,1,2,3}};
      unsigned num_points = 4;
      unsigned num_interpolated_points[NUM_SRC][1];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_do_search_interpolation_method(method_stack[i], interpolation_plan_element_stack[i],
                                        global_ids[i], NULL, CELL, num_points,
                                        num_interpolated_points[i], 1);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_interpolation_method(method_stack[i]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[NUM_SRC][2][9] = {{{2,5,8},{1,-1,4,-1,7,-1}},
                                           {{3,6,9},{-1,2,-1,-1,5,-1,-1,8,-1}},
                                           {{-1},{-1,3,-1,6,-1,9}}};
      double * remote_data[NUM_SRC][1] = {{in_data_raw[0][0]},
                                          {in_data_raw[1][0]},
                                          {in_data_raw[2][0]}};
      double * subdomain_in_data[NUM_SRC][1] = {{in_data_raw[0][1]},
                                                {in_data_raw[1][1]},
                                                {in_data_raw[2][1]}};
      double ** in_data[NUM_SRC][2] = {{remote_data[0], subdomain_in_data[0]},
                                       {remote_data[1], subdomain_in_data[1]},
                                       {remote_data[2], subdomain_in_data[2]}};
      double send_buffer[NUM_SRC][4] = {{-1,-1,-1,-1},{-1,-1,-1,-1},{-1,-1,-1,-1}};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_execute_interpolation_plan_element(interpolation_plan_element_stack[i],
                                                in_data[i], send_buffer[i]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[NUM_SRC];
      struct interpolation_method_data_request * data_request[NUM_SRC] =
         {&data_request_[0],&data_request_[1],&data_request_[2]};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_get_source_exchange_data_interpolation_method(method_stack[i], &data_request[i], 1);

      int ref_requested_global_ids[2][3] = {{1,4,7},{2,5,8}};
      int ref_requested_ranks[2][3] =      {{1,1,1},{2,2,2}};

      struct interpolation_method_data_request ref_data_request[NUM_SRC] =
         {{3, ref_requested_global_ids[0],  ref_requested_ranks[0]},
          {3, ref_requested_global_ids[1],  ref_requested_ranks[1]},
          {0,NULL,NULL}};

      unsigned ref_num_interpolated_points[NUM_SRC] = {2,2,0};
      int ref_global_ids[NUM_SRC][2] = {{0,2},{1,3},{-1}};
      double ref_out_data[NUM_SRC][2] = {{3,6},{4,7},{-1,-1}};

      for (unsigned j = 0; j < NUM_SRC; ++j ) {

         check_data_request(data_request[j], &ref_data_request[j], 1);

         if (ref_num_interpolated_points[j] != num_interpolated_points[j][0])
            PUT_ERR("wrong number of interpolated points\n");

         for (unsigned i = 0; i < ref_num_interpolated_points[j]; ++i) {
            if (ref_global_ids[j][i] != global_ids[j][i])
               PUT_ERR("wrong global id order\n");
            // since the cells are on the sphere, we do not have perfect
            // rectangular cells and hence the reference data is very inaccurate...
            if ( fabs(send_buffer[j][i] - ref_out_data[j][i] ) > 1.0e-4 )
               PUT_ERR("wrong result buffer\n");
         }
      }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < NUM_SRC; ++j ) {
         for (size_t i = 0; i < sizeof(method_stack[j])/sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (size_t i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
              sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 1);
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_points(&cell_points_src[i]);

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);

      yac_reset_interpolation_method();
   }

   { // conservative remapping with three source and two target processes
     // this test checks all possible configurations for the conservative
     // remapping

      // the global source domain is a 4x4 grid:
      //
      //       20--36--21--37--22--38--23--39--24
      //       |       |       |       |       |
      //       28  12  30  13  32  14  34  15  35
      //       |       |       |       |       |
      //       15--27--16--29--17--31--18--33--19
      //       |       |       |       |       |
      //       19  08  21  09  23  10  25  11  26
      //       |       |       |       |       |
      //       10--18--11--20--12--22--13--24--14
      //       |       |       |       |       |
      //       10  04  12  05  14  06  16  07  17
      //       |       |       |       |       |
      //       05--09--06--11--07--13--08--15--09
      //       |       |       |       |       |
      //       01  00  03  01  05  02  07  03  08
      //       |       |       |       |       |
      //       00--00--01--02--02--04--03--06--04
      //
      // mask for the source grid (cells with "x" are masked)
      //
      //       +---+---+---+---+
      //       | x | x | x |   |
      //       +---+---+---+---+
      //       | x | x |   |   |
      //       +---+---+---+---+
      //       | x |   |   |   |
      //       +---+---+---+---+
      //       |   |   |   |   |
      //       +---+---+---+---+
      //
      // the global target domain is a 3x3 grid:
      //
      //       12--21--13--22--14--23--15
      //       |       |       |       |
      //       15  06  17  07  19  08  20
      //       |       |       |       |
      //       08--14--09--16--10--18--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

#undef NUM_SRC
#undef NUM_TGT
#define NUM_SRC 3
#define NUM_TGT 2
      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{4,2}, {4,4}, {4,2}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[5] = {0.0,1.0,2.0,3.0,4.0};
      double coordinates_y_src[NUM_SRC][5] = {{0.0,1.0,2.0},
                                              {0.0,1.0,2.0,3.0,4.0},
                                              {2.0,3.0,4.0}};

      for ( unsigned i = 0; i < 5; ++i )
         coordinates_x_src[i] *= YAC_RAD;

      for ( unsigned j = 0; j < NUM_SRC; ++j )
         for ( unsigned i = 0; i < 4; ++i )
         coordinates_y_src[j][i] *= YAC_RAD;

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src, coordinates_y_src[i],
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][25] = {{0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14},
                                                {0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24},
                                                {10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24}};
      unsigned corner_is_on_rank_src[NUM_SRC][25] = {{-1,-1,-1,-1, -1,-1,-1,-1,-1, 1,1,1,1,1},
                                                     {0,0,0,0,0, 0,0,0,0,0, -1,-1,-1,-1,-1, 2,2,2,2,2, 2,2,2,2,2},
                                                     {1,1,1,1,1, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1}};
      int global_cell_ids_src[NUM_SRC][16] = {{0,1,2,3, 4,5,6,7},
                                              {0,1,2,3, 4,5,6,7, 8,9,10,11, 12,13,14,15},
                                              {8,9,10,11, 12,13,14,15}};
      unsigned cell_is_on_rank_src[NUM_SRC][16] = {{-1,-1,-1,-1, 1,1,1,1},
                                                   {0,0,0,0, -1,-1,-1,-1, -1,-1,-1,-1, 2,2,2,2},
                                                   {1,1,1,1, -1,-1,-1,-1}};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_subdomain_unstruct(&subdomain_src[i], grid_src[i], comm_named_src[i],
                                     global_corner_ids_src[i], corner_is_on_rank_src[i],
                                     global_cell_ids_src[i], cell_is_on_rank_src[i],
                                     NULL, NULL);

      struct domain l_domain_src[NUM_SRC];

      struct subdomain * blocks_src[NUM_SRC] = {&subdomain_src[0],
                                                &subdomain_src[1],
                                                &subdomain_src[2]};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[2] = {3,2};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[4] = {0.5,1.5,2.5,3.5};
      double coordinates_y_tgt[2][3] = {{0.5,1.5,2.5}, {1.5,2.5,3.5}};

      for ( unsigned i = 0; i < 4; ++i )
         coordinates_x_tgt[i] *= YAC_RAD;

      for ( unsigned i = 0; i < 3; ++i )
         for ( unsigned j = 0; j < NUM_TGT; ++j )
            coordinates_y_tgt[j][i] *= YAC_RAD;

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt,
                                          coordinates_y_tgt[i],
                                          num_cells_tgt, cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[NUM_TGT][12] = {{0,1,2,3,4,5,6,7,8,9,10,11},
                                                {4,5,6,7,8,9,10,11,12,13,14,15}};
      unsigned corner_is_on_rank_tgt[NUM_TGT][12] = {{-1,-1,-1,-1, -1,-1,-1,-1, 1,1,1,1},
                                                     {0,0,0,0, -1,-1,-1,-1, -1,-1,-1,-1}};
      int global_cell_ids_tgt[NUM_TGT][6] = {{0,1,2,3,4,5},
                                             {3,4,5,6,7,8}};
      unsigned cell_is_on_rank_tgt[NUM_TGT][6] = {{-1,-1,-1, -1,1,1},
                                                  {0,-1,-1, -1,-1,-1}};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                     global_corner_ids_tgt[i], corner_is_on_rank_tgt[i],
                                     global_cell_ids_tgt[i], cell_is_on_rank_tgt[i],
                                     NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];

      struct subdomain * blocks_tgt[NUM_TGT] = {&subdomain_tgt[0],
                                                &subdomain_tgt[1]};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         search[i] =
           yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                 1,0, "component_src", "component_tgt");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
           yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                 0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points cell_points_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_points(&cell_points_src[i], grid_src[i], CELL, NULL, NULL);

      //-------------------------------------
      // initialise masks on source side
      //-------------------------------------

      struct mask masks[NUM_SRC];
      unsigned mask_values[NUM_SRC][16] = {{0,0,0,0, 1,0,0,0},
                                           {0,0,0,0, 1,0,0,0, 1,1,0,0, 1,1,1,0},
                                           {                  1,1,0,0, 1,1,1,0}};

      for (unsigned i = 0; i < NUM_SRC; ++i)
         yac_init_mask(masks+i, mask_values[i], cell_points_src+i);

      //-------------------------------------
      // set up point sets and masks
      //-------------------------------------

      struct points * points_per_subdomain[NUM_SRC] = {&cell_points_src[0],
                                                       &cell_points_src[1],
                                                       &cell_points_src[2]};
      struct points ** point_sets[NUM_SRC] = {&points_per_subdomain[0],
                                              &points_per_subdomain[1],
                                              &points_per_subdomain[2]};
      struct mask * masks_per_subdomain[NUM_SRC][1] = {{masks+0},
                                                       {masks+1},
                                                       {masks+2}};
      struct mask ** mask_sets[NUM_SRC] = {masks_per_subdomain[0],
                                           masks_per_subdomain[1],
                                           masks_per_subdomain[2]};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_conserv[NUM_SRC][4];
      enum yac_interpolation_method_conserv_normalisation 
         normalisation[2] = {CONSERV_DESTAREA, CONSERV_FRACAREA};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            method_conserv[i][j] =
               yac_new_interpolation_method_conserv(search[i], &l_domain_src[i],
                                                "conserv stack", &point_sets[i],
                                                &mask_sets[i], 1, 0, 1, 0, j&1,
                                                normalisation[j>>1]);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[NUM_SRC][4][2];
      struct interpolation_plan_element * interpolation_plan_element_stack[NUM_SRC][4][2];
      int global_ids[NUM_SRC][4][9] = {{{0,1,2,3,4,5}, {0,1,2,3,4,5},
                                        {0,1,2,3,4,5}, {0,1,2,3,4,5}},
                                       {{0,1,2,3,4,5,6,7,8}, {0,1,2,3,4,5,6,7,8},
                                        {0,1,2,3,4,5,6,7,8}, {0,1,2,3,4,5,6,7,8}},
                                       {{3,4,5,6,7,8}, {3,4,5,6,7,8},
                                        {3,4,5,6,7,8}, {3,4,5,6,7,8}},};
      unsigned num_points[NUM_SRC] = {6,9,6};
      unsigned num_interpolated_points[NUM_SRC][4][1];

      for (unsigned int i = 0; i < NUM_SRC; ++i) {
         for (unsigned int j = 0; j < 4; ++j) {
            method_stack[i][j][0] = method_conserv[i][j];
            method_stack[i][j][1] = NULL;
            interpolation_plan_element_stack[i][j][0] = NULL;
            interpolation_plan_element_stack[i][j][1] = NULL;
         }
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            yac_do_search_interpolation_method(method_stack[i][j],
                                           interpolation_plan_element_stack[i][j],
                                           global_ids[i][j], NULL, CELL,
                                           num_points[i],
                                           num_interpolated_points[i][j], 1);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            yac_enddef_interpolation_method(method_stack[i][j]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[NUM_SRC][2][16] = {{{5,6,7},{0,1,2,3, -1,-1,-1,-1}},
                                            {{15},{-1,-1,-1,-1, 4,5,6,7,
                                                   8,9,10,11, -1,-1,-1,-1}},
                                            {{-1},{-1,-1,-1,-1, 12,13,14,15}}};
      double * remote_data[NUM_SRC][1] = {{in_data_raw[0][0]},
                                          {in_data_raw[1][0]},
                                          {in_data_raw[2][0]}};
      double * subdomain_in_data[NUM_SRC][1] = {{in_data_raw[0][1]},
                                                {in_data_raw[1][1]},
                                                {in_data_raw[2][1]}};
      double ** in_data[NUM_SRC][2] = {{remote_data[0], subdomain_in_data[0]},
                                       {remote_data[1], subdomain_in_data[1]},
                                       {remote_data[2], subdomain_in_data[2]}};
      double send_buffer[NUM_SRC][4][6];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            for (unsigned int k = 0; k < 3; ++k)
               send_buffer[i][j][k] = -1;

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            yac_execute_interpolation_plan_element(
               interpolation_plan_element_stack[i][j], in_data[i],
               send_buffer[i][j]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[NUM_SRC][4];
      struct interpolation_method_data_request * data_request[NUM_SRC][4];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            data_request[i][j] = &data_request_[i][j];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            yac_get_source_exchange_data_interpolation_method(
               method_stack[i][j], &data_request[i][j], 1);

      int ref_num_requested_points[NUM_SRC][2] = {{3,3}, {0,1}, {0,0}};
      int ref_requested_global_ids[NUM_SRC][2][5] = {{{5,6,7}, {5,6,7}},
                                                     {{-1}, {15}},
                                                     {{-1}, {-1}}};
      int ref_requested_ranks[NUM_SRC][2][5] = {{{1,1,1},{1,1,1}},
                                                {{-1},{2}},
                                                {{-1},{-1}}};

      struct interpolation_method_data_request ref_data_request[NUM_SRC][2];


      for (unsigned int i = 0; i < NUM_SRC; ++i) {
         for (unsigned int j = 0; j < 2; ++j) {
            ref_data_request[i][j].num_values =
               ref_num_requested_points[i][j];
            if (ref_num_requested_points[i][j]) {
               ref_data_request[i][j].global_ids =
                  ref_requested_global_ids[i][j];
               ref_data_request[i][j].ranks =
                  ref_requested_ranks[i][j];
            } else {
               ref_data_request[i][j].global_ids = NULL;
               ref_data_request[i][j].ranks = NULL;
            }
         }
      }

      unsigned ref_num_interpolated_points[NUM_SRC][2] = {{2,3},{1,5},{0,0}};
      int ref_global_ids[NUM_SRC][2][5] = {{{1,2},{0,1,2}},
                                           {{5},{3,4,5,7,8}},
                                           {{-1},{-1}}};
      double ref_out_data[NUM_SRC][4][5] = {{{3.5,4.5},{1.5,3.5,4.5},
                                             {3.5,4.5},{2.0,3.5,4.5}},
                                            {{8.5},{1.25,5.25,8.5,2.5,9.0},
                                             {8.5},{5.0,7.0,8.5,10.0,12.0}},
                                            {{-1},{-1},{-1},{-1}}};

      for (unsigned j = 0; j < NUM_SRC; ++j ) {
         for (unsigned int k = 0; k < 4; ++k) {

            check_data_request(data_request[j][k], &ref_data_request[j][k&1], 1);

            if (ref_num_interpolated_points[j][k&1] != num_interpolated_points[j][k][0])
               PUT_ERR("wrong number of interpolated points\n");

            for (unsigned i = 0; i < ref_num_interpolated_points[j][k&1]; ++i) {
               if (ref_global_ids[j][k&1][i] != global_ids[j][k][i])
                  PUT_ERR("wrong global id order\n");
               // since the cells are on the sphere, we do not have perfect
               // rectangular cells and hence the reference data is very inaccurate...
               if ( fabs(send_buffer[j][k][i] - ref_out_data[j][k][i] ) > 1.0e-2 )
                  PUT_ERR("wrong result buffer\n");
            }
         }
      }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < NUM_SRC; ++j ) {
         for (unsigned int k = 0; k < 4; ++k) {
            for (size_t i = 0; i < sizeof(method_stack[j][k])/sizeof(method_stack[j][k][0]); ++i)
               if (method_stack[j][k][i] != NULL)
                  yac_delete_interpolation_method(method_stack[j][k][i]);
            for (size_t i = 0; i < sizeof(interpolation_plan_element_stack[j][k]) /
                 sizeof(interpolation_plan_element_stack[j][k][0]); ++i) {
               yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][k][i]);
               free(interpolation_plan_element_stack[j][k][i]);
            }
            delete_interpolation_method_data_request_stack(data_request[j][k], 1);
         }
      }

      for (unsigned i = 0; i < NUM_SRC; ++i)
         yac_free_mask(&masks[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_points(&cell_points_src[i]);

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);

      yac_reset_interpolation_method();
   }

   { // conservative remapping with three source and two target processes
     // this test checks how the interpolation method handles the case in which
     // it does not receive any target points because they are masked

      // the global source domain is a 4x4 grid:
      //
      //       20--36--21--37--22--38--23--39--24
      //       |       |       |       |       |
      //       28  12  30  13  32  14  34  15  35
      //       |       |       |       |       |
      //       15--27--16--29--17--31--18--33--19
      //       |       |       |       |       |
      //       19  08  21  09  23  10  25  11  26
      //       |       |       |       |       |
      //       10--18--11--20--12--22--13--24--14
      //       |       |       |       |       |
      //       10  04  12  05  14  06  16  07  17
      //       |       |       |       |       |
      //       05--09--06--11--07--13--08--15--09
      //       |       |       |       |       |
      //       01  00  03  01  05  02  07  03  08
      //       |       |       |       |       |
      //       00--00--01--02--02--04--03--06--04
      //
      // mask for the source grid (cells with "x" are masked)
      //
      //       +---+---+---+---+
      //       | x | x | x |   |
      //       +---+---+---+---+
      //       | x | x |   |   |
      //       +---+---+---+---+
      //       | x |   |   |   |
      //       +---+---+---+---+
      //       |   |   |   |   |
      //       +---+---+---+---+
      //
      // the global target domain is a 3x3 grid:
      //
      //       12--21--13--22--14--23--15
      //       |       |       |       |
      //       15  06  17  07  19  08  20
      //       |       |       |       |
      //       08--14--09--16--10--18--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // target mask:
      //
      //      +---+---+---+
      //      | x | x | x |
      //      +---+---+---+
      //      | x | x | x |
      //      +---+---+---+
      //      |   |   |   |
      //      +---+---+---+
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

#undef NUM_SRC
#undef NUM_TGT
#define NUM_SRC 3
#define NUM_TGT 2
      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{4,2}, {4,4}, {4,2}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[5] = {0.0,1.0,2.0,3.0,4.0};
      double coordinates_y_src[NUM_SRC][5] = {{0.0,1.0,2.0},
                                              {0.0,1.0,2.0,3.0,4.0},
                                              {2.0,3.0,4.0}};

      for ( unsigned i = 0; i < 5; ++i )
         coordinates_x_src[i] *= YAC_RAD;

      for ( unsigned j = 0; j < NUM_SRC; ++j )
         for ( unsigned i = 0; i < 4; ++i )
         coordinates_y_src[j][i] *= YAC_RAD;

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src, coordinates_y_src[i],
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][25] = {{0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14},
                                                {0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24},
                                                {10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24}};
      unsigned corner_is_on_rank_src[NUM_SRC][25] = {{-1,-1,-1,-1, -1,-1,-1,-1,-1, 1,1,1,1,1},
                                                     {0,0,0,0,0, 0,0,0,0,0, -1,-1,-1,-1,-1, 2,2,2,2,2, 2,2,2,2,2},
                                                     {1,1,1,1,1, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1}};
      int global_cell_ids_src[NUM_SRC][16] = {{0,1,2,3, 4,5,6,7},
                                              {0,1,2,3, 4,5,6,7, 8,9,10,11, 12,13,14,15},
                                              {8,9,10,11, 12,13,14,15}};
      unsigned cell_is_on_rank_src[NUM_SRC][16] = {{-1,-1,-1,-1, 1,1,1,1},
                                                   {0,0,0,0, -1,-1,-1,-1, -1,-1,-1,-1, 2,2,2,2},
                                                   {1,1,1,1, -1,-1,-1,-1}};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_subdomain_unstruct(&subdomain_src[i], grid_src[i], comm_named_src[i],
                                     global_corner_ids_src[i], corner_is_on_rank_src[i],
                                     global_cell_ids_src[i], cell_is_on_rank_src[i],
                                     NULL, NULL);

      struct domain l_domain_src[NUM_SRC];

      struct subdomain * blocks_src[NUM_SRC] = {&subdomain_src[0],
                                                &subdomain_src[1],
                                                &subdomain_src[2]};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[2] = {3,2};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[4] = {0.5,1.5,2.5,3.5};
      double coordinates_y_tgt[2][3] = {{0.5,1.5,2.5}, {1.5,2.5,3.5}};

      for ( unsigned i = 0; i < 4; ++i )
         coordinates_x_tgt[i] *= YAC_RAD;

      for ( unsigned i = 0; i < 3; ++i )
         for ( unsigned j = 0; j < NUM_TGT; ++j )
            coordinates_y_tgt[j][i] *= YAC_RAD;

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt,
                                          coordinates_y_tgt[i],
                                          num_cells_tgt, cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[NUM_TGT][12] = {{0,1,2,3,4,5,6,7,8,9,10,11},
                                                {4,5,6,7,8,9,10,11,12,13,14,15}};
      unsigned corner_is_on_rank_tgt[NUM_TGT][12] = {{-1,-1,-1,-1, -1,-1,-1,-1, 1,1,1,1},
                                                     {0,0,0,0, -1,-1,-1,-1, -1,-1,-1,-1}};
      int global_cell_ids_tgt[NUM_TGT][6] = {{0,1,2,3,4,5},
                                             {3,4,5,6,7,8}};
      unsigned cell_is_on_rank_tgt[NUM_TGT][6] = {{-1,-1,-1, -1,1,1},
                                                  {0,-1,-1, -1,-1,-1}};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                     global_corner_ids_tgt[i], corner_is_on_rank_tgt[i],
                                     global_cell_ids_tgt[i], cell_is_on_rank_tgt[i],
                                     NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];

      struct subdomain * blocks_tgt[NUM_TGT] = {&subdomain_tgt[0],
                                                &subdomain_tgt[1]};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         search[i] =
           yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                 1,0, "component_src", "component_tgt");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
           yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                 0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points cell_points_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_points(&cell_points_src[i], grid_src[i], CELL, NULL, NULL);

      //-------------------------------------
      //initialise masks on source side
      //-------------------------------------

      struct mask masks[NUM_SRC];
      unsigned mask_values[NUM_SRC][16] = {{0,0,0,0, 1,0,0,0},
                                           {0,0,0,0, 1,0,0,0, 1,1,0,0, 1,1,1,0},
                                           {                  1,1,0,0, 1,1,1,0}};

      for (unsigned i = 0; i < NUM_SRC; ++i)
         yac_init_mask(masks+i, mask_values[i], cell_points_src+i);

      //-------------------------------------
      // set up point sets and masks
      //-------------------------------------

      struct points * points_per_subdomain[NUM_SRC] = {&cell_points_src[0],
                                                       &cell_points_src[1],
                                                       &cell_points_src[2]};
      struct points ** point_sets[NUM_SRC] = {&points_per_subdomain[0],
                                              &points_per_subdomain[1],
                                              &points_per_subdomain[2]};
      struct mask * masks_per_subdomain[NUM_SRC][1] = {{masks+0},
                                                       {masks+1},
                                                       {masks+2}};
      struct mask ** mask_sets[NUM_SRC] = {masks_per_subdomain[0],
                                           masks_per_subdomain[1],
                                           masks_per_subdomain[2]};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_conserv[NUM_SRC][4];
      enum yac_interpolation_method_conserv_normalisation 
         normalisation[2] = {CONSERV_DESTAREA, CONSERV_FRACAREA};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            method_conserv[i][j] =
               yac_new_interpolation_method_conserv(search[i], &l_domain_src[i],
                                                "conserv stack", &point_sets[i],
                                                &mask_sets[i], 1, 0, 1, 0, j&1,
                                                normalisation[j>>1]);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[NUM_SRC][4][2];
      struct interpolation_plan_element * interpolation_plan_element_stack[NUM_SRC][4][2];
      int global_ids[NUM_SRC][4][3] = {{{0,1,2}, {0,1,2},
                                        {0,1,2}, {0,1,2}},
                                       {{0,1,2}, {0,1,2},
                                        {0,1,2}, {0,1,2}},
                                       {{-1}, {-1},
                                        {-1}, {-1}},};
      unsigned num_points[NUM_SRC] = {3,3,0};
      unsigned num_interpolated_points[NUM_SRC][4][1];

      for (unsigned int i = 0; i < NUM_SRC; ++i) {
         for (unsigned int j = 0; j < 4; ++j) {
            method_stack[i][j][0] = method_conserv[i][j];
            method_stack[i][j][1] = NULL;
            interpolation_plan_element_stack[i][j][0] = NULL;
            interpolation_plan_element_stack[i][j][1] = NULL;
         }
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            yac_do_search_interpolation_method(method_stack[i][j],
                                           interpolation_plan_element_stack[i][j],
                                           global_ids[i][j], NULL, CELL,
                                           num_points[i],
                                           num_interpolated_points[i][j], 1);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            yac_enddef_interpolation_method(method_stack[i][j]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[NUM_SRC][2][16] = {{{5,6,7},{0,1,2,3, -1,-1,-1,-1}},
                                            {{-1},{-1,-1,-1,-1, 4,5,6,7,
                                                   8,9,10,11, -1,-1,-1,-1}},
                                            {{-1},{-1,-1,-1,-1, 12,13,14,15}}};
      double * remote_data[NUM_SRC][1] = {{in_data_raw[0][0]},
                                          {in_data_raw[1][0]},
                                          {in_data_raw[2][0]}};
      double * subdomain_in_data[NUM_SRC][1] = {{in_data_raw[0][1]},
                                                {in_data_raw[1][1]},
                                                {in_data_raw[2][1]}};
      double ** in_data[NUM_SRC][2] = {{remote_data[0], subdomain_in_data[0]},
                                       {remote_data[1], subdomain_in_data[1]},
                                       {remote_data[2], subdomain_in_data[2]}};
      double send_buffer[NUM_SRC][4][3];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            for (unsigned int k = 0; k < 3; ++k)
               send_buffer[i][j][k] = -1;

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            yac_execute_interpolation_plan_element(
               interpolation_plan_element_stack[i][j], in_data[i],
               send_buffer[i][j]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[NUM_SRC][4];
      struct interpolation_method_data_request * data_request[NUM_SRC][4];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            data_request[i][j] = &data_request_[i][j];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         for (unsigned int j = 0; j < 4; ++j)
            yac_get_source_exchange_data_interpolation_method(
               method_stack[i][j], &data_request[i][j], 1);

      int ref_num_requested_points[NUM_SRC][2] = {{3,3}, {0,0}, {0,0}};
      int ref_requested_global_ids[NUM_SRC][2][3] = {{{5,6,7}, {5,6,7}},
                                                     {{-1}, {-1}},
                                                     {{-1}, {-1}}};
      int ref_requested_ranks[NUM_SRC][2][3] = {{{1,1,1},{1,1,1}},
                                                {{-1},{-1}},
                                                {{-1},{-1}}};

      struct interpolation_method_data_request ref_data_request[NUM_SRC][2];


      for (unsigned int i = 0; i < NUM_SRC; ++i) {
         for (unsigned int j = 0; j < 2; ++j) {
            ref_data_request[i][j].num_values =
               ref_num_requested_points[i][j];
            if (ref_num_requested_points[i][j]) {
               ref_data_request[i][j].global_ids =
                  ref_requested_global_ids[i][j];
               ref_data_request[i][j].ranks =
                  ref_requested_ranks[i][j];
            } else {
               ref_data_request[i][j].global_ids = NULL;
               ref_data_request[i][j].ranks = NULL;
            }
         }
      }

      unsigned ref_num_interpolated_points[NUM_SRC][2] = {{2,3},{0,0},{0,0}};
      int ref_global_ids[NUM_SRC][2][4] = {{{1,2},{0,1,2}},
                                           {{-1},{-1}},
                                           {{-1},{-1}}};
      double ref_out_data[NUM_SRC][4][6] = {{{3.5,4.5},{1.5,3.5,4.5,1.25},
                                             {3.5,4.5},{2.0,3.5,4.5,5.0}},
                                            {{-1},{-1},{-1},{-1}},
                                            {{-1},{-1},{-1},{-1}}};

      for (unsigned j = 0; j < NUM_SRC; ++j ) {
         for (unsigned int k = 0; k < 4; ++k) {

            check_data_request(data_request[j][k], &ref_data_request[j][k&1], 1);

            if (ref_num_interpolated_points[j][k&1] != num_interpolated_points[j][k][0])
               PUT_ERR("wrong number of interpolated points\n");

            for (unsigned i = 0; i < ref_num_interpolated_points[j][k&1]; ++i) {
               if (ref_global_ids[j][k&1][i] != global_ids[j][k][i])
                  PUT_ERR("wrong global id order\n");
               // since the cells are on the sphere, we do not have perfect
               // rectangular cells and hence the reference data is very inaccurate...
               if ( fabs(send_buffer[j][k][i] - ref_out_data[j][k][i] ) > 1.0e-2 )
                  PUT_ERR("wrong result buffer\n");
            }
         }
      }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < NUM_SRC; ++j ) {
         for (unsigned int k = 0; k < 4; ++k) {
            for (size_t i = 0; i < sizeof(method_stack[j][k])/sizeof(method_stack[j][k][0]); ++i)
               if (method_stack[j][k][i] != NULL)
                  yac_delete_interpolation_method(method_stack[j][k][i]);
            for (size_t i = 0; i < sizeof(interpolation_plan_element_stack[j][k]) /
                 sizeof(interpolation_plan_element_stack[j][k][0]); ++i) {
               yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][k][i]);
               free(interpolation_plan_element_stack[j][k][i]);
            }
            delete_interpolation_method_data_request_stack(data_request[j][k], 1);
         }
      }

      for (unsigned i = 0; i < NUM_SRC; ++i)
         yac_free_mask(&masks[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_points(&cell_points_src[i]);

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);

      yac_reset_interpolation_method();
   }

   { // 2nd order conservative remapping with one source and one target processes

      // the global source domain is a 3x3 grid:
      //
      //       12--21--13--22--14--23--15
      //       |       |       |       |
      //       15  06  17  07  19  08  20
      //       |       |       |       |
      //       08--14--09--16--10--18--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // the global target domain is a 2x2 grid:
      //
      //       06--10--07--11--08
      //       |       |       |
      //       06  02  08  03  09
      //       |       |       |
      //       03--05--04--07--05
      //       |       |       |
      //       01  00  03  01  04
      //       |       |       |
      //       00--00--01--02--02
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

#undef NUM_SRC
#undef NUM_TGT
#define NUM_SRC 1
#define NUM_TGT 1
      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{3,3}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[NUM_SRC][4] = {{-1.5,-0.5,0.5,1.5}};
      double coordinates_y_src[NUM_SRC][4] = {{-1.5,-0.5,0.5,1.5}};

      for ( unsigned j = 0; j < NUM_SRC; ++j )
         for ( unsigned i = 0; i < 4; ++i )
            coordinates_x_src[j][i] *= YAC_RAD;

      for ( unsigned j = 0; j < NUM_SRC; ++j )
        for ( unsigned i = 0; i < 4; ++i )
           coordinates_y_src[j][i] *= YAC_RAD;

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src[i],
                                          coordinates_y_src[i],
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][16] =
        {{0,1,2,3, 4,5,6,7, 8,9,10,11, 12,13,14,15}};
      unsigned corner_is_on_rank_src[NUM_SRC][16] =
        {{-1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1}};
      int global_cell_ids_src[NUM_SRC][9] =
        {{0,1,2,3,4,5,6,7,8}};
      unsigned cell_is_on_rank_src[NUM_SRC][9] =
        {{-1,-1,-1, -1,-1,-1, -1,-1,-1}};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
        yac_init_subdomain_unstruct(
          &subdomain_src[i], grid_src[i], comm_named_src[i],
          global_corner_ids_src[i], corner_is_on_rank_src[i],
          global_cell_ids_src[i], cell_is_on_rank_src[i], NULL, NULL);

      struct domain l_domain_src[NUM_SRC];
      struct subdomain * blocks_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
        blocks_src[i] = &subdomain_src[i];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[2] = {2,2};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[3] = {-0.5,0,0.5};
      double coordinates_y_tgt[3] = {-0.5,0,0.5};

      for ( unsigned i = 0; i < 3; ++i ) {
         coordinates_x_tgt[i] *= YAC_RAD;
         coordinates_y_tgt[i] *= YAC_RAD;
      }

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                          num_cells_tgt, cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[9] = {0,1,2,3,4,5,6,7,8};
      unsigned corner_is_on_rank_tgt[NUM_TGT][9] = {{-1,-1,-1, -1,-1,-1, -1,-1,-1}};
      int global_cell_ids_tgt[4] = {0,1,2,3};
      unsigned cell_is_on_rank_tgt[NUM_TGT][4] = {{-1,-1,-1,-1}};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                     global_corner_ids_tgt, corner_is_on_rank_tgt[i],
                                     global_cell_ids_tgt, cell_is_on_rank_tgt[i],
                                     NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];
      struct subdomain * blocks_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_TGT; ++i)
        blocks_tgt[i] = &subdomain_tgt[i];

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         search[i] =
           yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                 1,0, "component_src", "component_tgt");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
           yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                 0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points cell_points_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_points(&cell_points_src[i], grid_src[i], CELL, NULL, NULL);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[NUM_SRC];
      struct points ** point_sets[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        points_per_subdomain[i] = &cell_points_src[i];
        point_sets[i] = &points_per_subdomain[i];
      }

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_conserv[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         method_conserv[i] =
          yac_new_interpolation_method_conserv(search[i], &l_domain_src[i],
                                           "conserv stack", &point_sets[i],
                                           NULL, 1, 0, 2, 0, 0, CONSERV_DESTAREA);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[NUM_SRC][2];
      struct interpolation_plan_element * interpolation_plan_element_stack[NUM_SRC][2];
      int global_ids[NUM_SRC][4] = {{0,1,2,3}};
      unsigned num_points = 4;
      unsigned num_interpolated_points[NUM_SRC][1];

      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        method_stack[i][0] = method_conserv[i];
        method_stack[i][1] = NULL;
        interpolation_plan_element_stack[i][0] = NULL;
        interpolation_plan_element_stack[i][1] = NULL;
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_do_search_interpolation_method(method_stack[i], interpolation_plan_element_stack[i],
                                        global_ids[i], NULL, CELL, num_points,
                                        num_interpolated_points[i], 1);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_interpolation_method(method_stack[i]);


      //----------------------------------------------
      // testing data exchange and interpolated points
      //----------------------------------------------

      struct interpolation_method_data_request data_request_[NUM_SRC];
      struct interpolation_method_data_request * data_request[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i) data_request[i] = &data_request_[i];
      for (unsigned int i = 0; i < NUM_SRC; ++i)
        yac_get_source_exchange_data_interpolation_method(
          method_stack[i], &data_request[i], 1);

      unsigned ref_num_interpolated_points[NUM_SRC] = {4};
      int ref_global_ids[NUM_SRC][4] = {{0,1,2,3}};

      struct interpolation_method_data_request ref_data_request[NUM_SRC];
      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        ref_data_request[i].num_values = 0;
        ref_data_request[i].global_ids = NULL;
        ref_data_request[i].ranks = NULL;
      }

      for (unsigned j = 0; j < NUM_SRC; ++j ) {

         check_data_request(data_request[j], &ref_data_request[j], 1);

         if (ref_num_interpolated_points[j] != num_interpolated_points[j][0])
            PUT_ERR("wrong number of interpolated points\n");

         for (unsigned i = 0; i < ref_num_interpolated_points[j]; ++i)
            if (ref_global_ids[j][i] != global_ids[j][i])
               PUT_ERR("wrong global id order\n");
      }

#define NUM_TESTS (4)

      for (size_t t = 0; t < NUM_TESTS; ++t) {

        //---------------------
        // do the interpolation
        //---------------------

        double in_data_raw[NUM_TESTS][NUM_SRC][2][9] =
        {{{{-1}, {1,1,1,1,1,1,1,1,1}}},
         {{{-1}, {0,0,0,1,1,1,2,2,2}}},
         {{{-1}, {0,1,2,0,1,2,0,1,2}}},
         {{{-1}, {0,1,2,1,2,3,2,3,4}}}};
        double * remote_data[NUM_SRC][1];
        double * subdomain_in_data[NUM_SRC][1];
        double ** in_data[NUM_SRC][2];
        double send_buffer[NUM_SRC][4];

        for (unsigned int i = 0; i < NUM_SRC; ++i) {
          remote_data[i][0] = in_data_raw[t][i][0];
          subdomain_in_data[i][0] = in_data_raw[t][i][1];
          in_data[i][0] = remote_data[i];
          in_data[i][1] = subdomain_in_data[i];
          for (unsigned int j = 0; j < 4; ++j) send_buffer[i][j] = -1;
        }

        //---------------------------------
        // testing results of interpolation
        //---------------------------------

        for (unsigned int i = 0; i < NUM_SRC; ++i)
          yac_execute_interpolation_plan_element(
            interpolation_plan_element_stack[i], in_data[i], send_buffer[i]);

        double ref_out_data[NUM_TESTS][NUM_SRC][4] =
          {{{1.0,1.0,1.0,1.0}},
           {{0.75,0.75,1.25,1.25}},
           {{0.75,1.25,0.75,1.25}},
           {{1.5,2.0,2.0,2.5}}};

        for (unsigned j = 0; j < NUM_SRC; ++j ) {
           double ref_out_data_sum = 0.0, out_data_sum = 0.0;
           for (unsigned i = 0; i < ref_num_interpolated_points[j]; ++i) {
              if ( fabs(send_buffer[j][i] - ref_out_data[t][j][i] ) > 1.0e-3 )
                 PUT_ERR("wrong result buffer\n");
             out_data_sum += send_buffer[j][i];
             ref_out_data_sum += ref_out_data[t][j][i];
           }
           if ( fabs(out_data_sum - ref_out_data_sum ) > 1.0e-3 )
              PUT_ERR("wrong result buffer\n");
        }
      }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < NUM_SRC; ++j ) {
         for (size_t i = 0; i < sizeof(method_stack[j])/sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (size_t i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
              sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 1);
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_points(&cell_points_src[i]);

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);

      yac_reset_interpolation_method();
   }

   { // 2nd order conservative remapping with three source and one target processes

      // the global source domain is a 3x3 grid:
      //
      //       12--21--13--22--14--23--15
      //       |       |       |       |
      //       15  06  17  07  19  08  20
      //       |       |       |       |
      //       08--14--09--16--10--18--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // the global target domain is a 2x2 grid:
      //
      //       06--10--07--11--08
      //       |       |       |
      //       06  02  08  03  09
      //       |       |       |
      //       03--05--04--07--05
      //       |       |       |
      //       01  00  03  01  04
      //       |       |       |
      //       00--00--01--02--02
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

#undef NUM_SRC
#undef NUM_TGT
#define NUM_SRC 3
#define NUM_TGT 1
      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{2,3}, {3,3}, {2,3}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[NUM_SRC][4] = {{-1.5,-0.5,0.5},
                                              {-1.5,-0.5,0.5,1.5},
                                              {-0.5,0.5,1.5}};
      double coordinates_y_src[NUM_SRC][4] = {{-1.5,-0.5,0.5,1.5},
                                              {-1.5,-0.5,0.5,1.5},
                                              {-1.5,-0.5,0.5,1.5}};

      for ( unsigned j = 0; j < NUM_SRC; ++j )
         for ( unsigned i = 0; i < 4; ++i )
            coordinates_x_src[j][i] *= YAC_RAD;

      for ( unsigned j = 0; j < NUM_SRC; ++j )
        for ( unsigned i = 0; i < 4; ++i )
           coordinates_y_src[j][i] *= YAC_RAD;

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src[i],
                                          coordinates_y_src[i],
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][16] = {{0,1,2, 4,5,6, 8,9,10, 12,13,14},
                                                {0,1,2,3, 4,5,6,7, 8,9,10,11, 12,13,14,15},
                                                {1,2,3, 5,6,7, 9,10,11, 13,14,15}};
      unsigned corner_is_on_rank_src[NUM_SRC][16] = {{-1,-1,1, -1,-1,1, -1,-1,1, -1,-1,1},
                                                     {0,-1,-1,2, 0,-1,-1,2, 0,-1,-1,2, 0,-1,-1,2},
                                                     {1,-1,-1, 1,-1,-1, 1,-1,-1, 1,-1,-1}};
      int global_cell_ids_src[NUM_SRC][9] = {{0,1,3,4,6,7},
                                             {0,1,2,3,4,5,6,7,8},
                                             {1,2,4,5,7,8}};
      unsigned cell_is_on_rank_src[NUM_SRC][9] = {{-1,1,-1,1,-1,1},
                                                  {0,-1,2,0,-1,2,0,-1,2},
                                                  {1,-1,1,-1,1,-1}};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
        yac_init_subdomain_unstruct(
          &subdomain_src[i], grid_src[i], comm_named_src[i],
          global_corner_ids_src[i], corner_is_on_rank_src[i],
          global_cell_ids_src[i], cell_is_on_rank_src[i], NULL, NULL);

      struct domain l_domain_src[NUM_SRC];
      struct subdomain * blocks_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
        blocks_src[i] = &subdomain_src[i];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[2] = {2,2};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[3] = {-0.5,0,0.5};
      double coordinates_y_tgt[3] = {-0.5,0,0.5};

      for ( unsigned i = 0; i < 3; ++i ) {
         coordinates_x_tgt[i] *= YAC_RAD;
         coordinates_y_tgt[i] *= YAC_RAD;
      }

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                          num_cells_tgt, cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[9] = {0,1,2,3,4,5,6,7,8};
      unsigned corner_is_on_rank_tgt[NUM_TGT][9] = {{-1,-1,-1, -1,-1,-1, -1,-1,-1}};
      int global_cell_ids_tgt[4] = {0,1,2,3};
      unsigned cell_is_on_rank_tgt[NUM_TGT][4] = {{-1,-1,-1,-1}};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                     global_corner_ids_tgt, corner_is_on_rank_tgt[i],
                                     global_cell_ids_tgt, cell_is_on_rank_tgt[i],
                                     NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];
      struct subdomain * blocks_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_TGT; ++i)
        blocks_tgt[i] = &subdomain_tgt[i];

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         search[i] =
           yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                 1,0, "component_src", "component_tgt");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
           yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                 0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points cell_points_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_points(&cell_points_src[i], grid_src[i], CELL, NULL, NULL);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[NUM_SRC];
      struct points ** point_sets[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        points_per_subdomain[i] = &cell_points_src[i];
        point_sets[i] = &points_per_subdomain[i];
      }

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_conserv[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         method_conserv[i] =
          yac_new_interpolation_method_conserv(search[i], &l_domain_src[i],
                                           "conserv stack", &point_sets[i],
                                           NULL, 1, 0, 2, 0, 0, CONSERV_DESTAREA);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[NUM_SRC][2];
      struct interpolation_plan_element * interpolation_plan_element_stack[NUM_SRC][2];
      int global_ids[NUM_SRC][4] = {{0,1,2,3},{0,1,2,3},{0,1,2,3}};
      unsigned num_points = 4;
      unsigned num_interpolated_points[NUM_SRC][1];

      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        method_stack[i][0] = method_conserv[i];
        method_stack[i][1] = NULL;
        interpolation_plan_element_stack[i][0] = NULL;
        interpolation_plan_element_stack[i][1] = NULL;
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_do_search_interpolation_method(method_stack[i], interpolation_plan_element_stack[i],
                                        global_ids[i], NULL, CELL, num_points,
                                        num_interpolated_points[i], 1);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_interpolation_method(method_stack[i]);


      //----------------------------------------------
      // testing data exchange and interpolated points
      //----------------------------------------------

      struct interpolation_method_data_request data_request_[NUM_SRC];
      struct interpolation_method_data_request * data_request[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i) data_request[i] = &data_request_[i];
      for (unsigned int i = 0; i < NUM_SRC; ++i)
        yac_get_source_exchange_data_interpolation_method(
          method_stack[i], &data_request[i], 1);

      unsigned ref_num_interpolated_points[NUM_SRC] = {0,4,0};
      int ref_global_ids[NUM_SRC][4] = {{-1},{0,1,2,3},{-1}};

      struct interpolation_method_data_request ref_data_request[NUM_SRC];
      unsigned ref_num_requested_points[NUM_SRC] = {0,2,0};
      int ref_requested_global_ids[NUM_SRC][2] = {{-1},{3,5},{-1}};
      int ref_requested_ranks[NUM_SRC][2] = {{-1},{0,2},{-1}};
      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        if (ref_num_requested_points[i] == 0) {
          ref_data_request[i].num_values = 0;
          ref_data_request[i].global_ids = NULL;
          ref_data_request[i].ranks = NULL;
        } else {
          ref_data_request[i].num_values = ref_num_requested_points[i];
          ref_data_request[i].global_ids = ref_requested_global_ids[i];
          ref_data_request[i].ranks = ref_requested_ranks[i];
        }
      }

      for (unsigned j = 0; j < NUM_SRC; ++j ) {

         check_data_request(data_request[j], &ref_data_request[j], 1);

         if (ref_num_interpolated_points[j] != num_interpolated_points[j][0])
            PUT_ERR("wrong number of interpolated points\n");

         for (unsigned i = 0; i < ref_num_interpolated_points[j]; ++i)
            if (ref_global_ids[j][i] != global_ids[j][i])
               PUT_ERR("wrong global id order\n");
      }

#define NUM_TESTS (4)

      for (size_t t = 0; t < NUM_TESTS; ++t) {

        //---------------------
        // do the interpolation
        //---------------------

        double field_data[NUM_TESTS][9] =
          {{1,1,1,1,1,1,1,1,1},
           {0,0,0,1,1,1,2,2,2},
           {0,1,2,0,1,2,0,1,2},
           {0,1,2,1,2,3,2,3,4}};
        double in_data_raw[NUM_SRC][2][9];
        double * remote_data[NUM_SRC][1];
        double * subdomain_in_data[NUM_SRC][1];
        double ** in_data[NUM_SRC][2];
        double send_buffer[NUM_SRC][4];

        for (unsigned int i = 0; i < NUM_SRC; ++i) {
          for (unsigned j = 0; j < ref_num_requested_points[i]; ++j)
            in_data_raw[i][0][j] = field_data[t][ref_requested_global_ids[i][j]];
          for (unsigned j = 0; j < num_cells_src[i][0] * num_cells_src[i][1]; ++j) {
            if (cell_is_on_rank_src[i][j] == (unsigned)-1)
              in_data_raw[i][1][j] = field_data[t][global_cell_ids_src[i][j]];
            else
              in_data_raw[i][1][j] = -1.0;
          }
          remote_data[i][0] = in_data_raw[i][0];
          subdomain_in_data[i][0] = in_data_raw[i][1];
          in_data[i][0] = remote_data[i];
          in_data[i][1] = subdomain_in_data[i];
          for (unsigned int j = 0; j < 4; ++j) send_buffer[i][j] = -1;
        }

        //---------------------------------
        // testing results of interpolation
        //---------------------------------

        for (unsigned int i = 0; i < NUM_SRC; ++i)
          yac_execute_interpolation_plan_element(
            interpolation_plan_element_stack[i], in_data[i], send_buffer[i]);

        double ref_out_data[NUM_TESTS][4] =
          {{1.0,1.0,1.0,1.0},
           {0.75,0.75,1.25,1.25},
           {0.75,1.25,0.75,1.25},
           {1.5,2.0,2.0,2.5}};

        for (unsigned j = 0; j < NUM_SRC; ++j ) {
           for (unsigned i = 0; i < ref_num_interpolated_points[j]; ++i) {
              int point_idx = global_ids[j][i];
              if ( fabs(send_buffer[j][i] -
                        ref_out_data[t][point_idx] ) > 1.0e-3 )
                 PUT_ERR("wrong result buffer\n");
           }
        }
      }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < NUM_SRC; ++j ) {
         for (size_t i = 0; i < sizeof(method_stack[j])/sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (size_t i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
              sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 1);
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_points(&cell_points_src[i]);

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);

      yac_reset_interpolation_method();
   }

   { // 2nd order conservative remapping with three source and one target processes

      // the global source domain is a 5x5 grid:
      //
      //       30--55--31--56--32--57--33-58--34-59--35
      //       |       |       |       |      |      |
      //       45  20  47  21  49  22  51 23  53 24  54
      //       |       |       |       |      |      |
      //       24--44--25--46--26--48--27-50--28-52--29
      //       |       |       |       |      |      |
      //       34  15  36  16  38  17  40 18  42 19  43
      //       |       |       |       |      |      |
      //       18--33--19--35--20--37--21-39--22-41--23
      //       |       |       |       |      |      |
      //       23  10  25  11  27  12  29 13  31 14  32
      //       |       |       |       |      |      |
      //       12--22--13--24--14--26--15-28--16-30--17
      //       |       |       |       |      |      |
      //       12  05  14  06  16  07  18 08  20 09  21
      //       |       |       |       |      |      |
      //       06--11--07--13--08--15--09-17--10-19--11
      //       |       |       |       |      |      |
      //       01  00  03  01  05  02  07 03  09 04  10
      //       |       |       |       |      |      |
      //       00--00--01--02--02--04--03-06--04-08--05
      //
      // the global target domain is a 5x5 grid and is slightly shifted to the
      // source domain
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

#undef NUM_SRC
#undef NUM_TGT
#define NUM_SRC 3
#define NUM_TGT 1
      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{4,3}, {3,5}, {4,4}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[NUM_SRC][6] = {{-2.5,-1.5,-0.5,0.5,1.5},
                                              {-0.5,0.5,1.5,2.5},
                                              {-2.5,-1.5,-0.5,0.5,1.5}};
      double coordinates_y_src[NUM_SRC][6] = {{-2.5,-1.5,-0.5,0.5},
                                              {-2.5,-1.5,-0.5,0.5,1.5,2.5},
                                              {-1.5,-0.5,0.5,1.5,2.5}};

      for ( unsigned j = 0; j < NUM_SRC; ++j )
         for ( unsigned i = 0; i < num_cells_src[j][0]+1; ++i )
            coordinates_x_src[j][i] *= YAC_RAD;

      for ( unsigned j = 0; j < NUM_SRC; ++j )
        for ( unsigned i = 0; i < num_cells_src[j][1]+1; ++i )
           coordinates_y_src[j][i] *= YAC_RAD;

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src[i],
                                          coordinates_y_src[i],
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][25] =
        {{0,1,2,3,4, 6,7,8,9,10, 12,13,14,15,16, 18,19,20,21,22},
         {2,3,4,5, 8,9,10,11, 14,15,16,17, 20,21,22,23, 26,27,28,29, 32,33,34,35},
         {6,7,8,9,10, 12,13,14,15,16, 18,19,20,21,22, 24,25,26,27,28, 30,31,32,33,34}};
      unsigned corner_is_on_rank_src[NUM_SRC][25] =
        {{-1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, 2,2,2,1,1},
         {0,0,-1,-1, 0,0,-1,-1, 0,0,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1},
         {0,0,0,0,1, 0,0,0,0,1, -1,-1,-1,1,1, -1,-1,-1,1,1, -1,-1,-1,1,1}};
      int global_cell_ids_src[NUM_SRC][16] =
        {{0,1,2,3, 5,6,7,8, 10,11,12,13},
         {2,3,4, 7,8,9, 12,13,14, 17,18,19, 22,23,24},
         {5,6,7,8, 10,11,12,13, 15,16,17,18, 20,21,22,23}};
      unsigned cell_is_on_rank_src[NUM_SRC][16] =
        {{-1,-1,-1,1, -1,-1,-1,1, 2,2,2,1},
         {0,-1,-1, 0,-1,-1, 2,-1,-1, 2,-1,-1, 2,-1,-1},
         {0,0,0,1, -1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1}};

      for (unsigned int i = 0; i < NUM_SRC; ++i)
        yac_init_subdomain_unstruct(
          &subdomain_src[i], grid_src[i], comm_named_src[i],
          global_corner_ids_src[i], corner_is_on_rank_src[i],
          global_cell_ids_src[i], cell_is_on_rank_src[i], NULL, NULL);

      struct domain l_domain_src[NUM_SRC];
      struct subdomain * blocks_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
        blocks_src[i] = &subdomain_src[i];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[2] = {5,5};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[6] = {-2,-1,0,1,2,3};
      double coordinates_y_tgt[6] = {-3,-2,-1,0,1,2};

      for ( unsigned i = 0; i < 6; ++i ) {
         coordinates_x_tgt[i] *= YAC_RAD;
         coordinates_y_tgt[i] *= YAC_RAD;
      }

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                          num_cells_tgt, cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[36] =
        {0,1,2,3,4,5, 6,7,8,9,10,11, 12,13,14,15,16,17,
         18,19,20,21,22,23, 24,25,26,27,28,29, 30,31,32,33,34,35};
      unsigned corner_is_on_rank_tgt[NUM_TGT][36] =
        {{-1,-1,-1,-1,-1,-1, -1,-1,-1,-1,-1,-1, -1,-1,-1,-1,-1,-1,
          -1,-1,-1,-1,-1,-1, -1,-1,-1,-1,-1,-1, -1,-1,-1,-1,-1,-1}};
      int global_cell_ids_tgt[25] =
        {0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24};
      unsigned cell_is_on_rank_tgt[NUM_TGT][25] =
        {{-1,-1,-1,-1,-1, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1,
          -1,-1,-1,-1,-1, -1,-1,-1,-1,-1}};

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                     global_corner_ids_tgt, corner_is_on_rank_tgt[i],
                                     global_cell_ids_tgt, cell_is_on_rank_tgt[i],
                                     NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];
      struct subdomain * blocks_tgt[NUM_TGT];

      for (unsigned int i = 0; i < NUM_TGT; ++i)
        blocks_tgt[i] = &subdomain_tgt[i];

      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         search[i] =
           yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                 1,0, "component_src", "component_tgt");
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
           yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                 0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points cell_points_src[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_init_points(&cell_points_src[i], grid_src[i], CELL, NULL, NULL);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[NUM_SRC];
      struct points ** point_sets[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        points_per_subdomain[i] = &cell_points_src[i];
        point_sets[i] = &points_per_subdomain[i];
      }

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_conserv[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         method_conserv[i] =
          yac_new_interpolation_method_conserv(search[i], &l_domain_src[i],
                                           "conserv stack", &point_sets[i],
                                           NULL, 1, 0, 2, 0, 0, CONSERV_DESTAREA);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[NUM_SRC][2];
      struct interpolation_plan_element * interpolation_plan_element_stack[NUM_SRC][2];
      int global_ids[NUM_SRC][20] =
        {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18},
         {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19, 21,22,23,24},
         {5,6,7,8, 10,11,12,13, 15,16,17,18, 20,21,22,23}};
      unsigned num_points[NUM_SRC] = {16,20,16};
      unsigned num_interpolated_points[NUM_SRC][1];

      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        method_stack[i][0] = method_conserv[i];
        method_stack[i][1] = NULL;
        interpolation_plan_element_stack[i][0] = NULL;
        interpolation_plan_element_stack[i][1] = NULL;
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_do_search_interpolation_method(
            method_stack[i], interpolation_plan_element_stack[i], global_ids[i],
            NULL, CELL, num_points[i], num_interpolated_points[i], 1);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_enddef_interpolation_method(method_stack[i]);


      //----------------------------------------------
      // testing data exchange and interpolated points
      //----------------------------------------------

      struct interpolation_method_data_request data_request_[NUM_SRC];
      struct interpolation_method_data_request * data_request[NUM_SRC];

      for (unsigned int i = 0; i < NUM_SRC; ++i) data_request[i] = &data_request_[i];
      for (unsigned int i = 0; i < NUM_SRC; ++i)
        yac_get_source_exchange_data_interpolation_method(
          method_stack[i], &data_request[i], 1);

      unsigned ref_num_interpolated_points[NUM_SRC] = {1,2,1};
      int ref_global_ids[NUM_SRC][2] = {{11},{12,17},{16}};

      struct interpolation_method_data_request ref_data_request[NUM_SRC];
      unsigned ref_num_requested_points[NUM_SRC] = {7,8,4};
      int ref_requested_global_ids[NUM_SRC][8] =
        {{8,10,11,12,13,16,17},{2,6,7,11,12,16,17,22},{6,7,13,18}};
      int ref_requested_ranks[NUM_SRC][8] =
        {{1,2,2,2,1,2,2},{0,0,0,2,2,2,2,2},{0,0,1,1}};
      for (unsigned int i = 0; i < NUM_SRC; ++i) {
        if (ref_num_requested_points[i] == 0) {
          ref_data_request[i].num_values = 0;
          ref_data_request[i].global_ids = NULL;
          ref_data_request[i].ranks = NULL;
        } else {
          ref_data_request[i].num_values = ref_num_requested_points[i];
          ref_data_request[i].global_ids = ref_requested_global_ids[i];
          ref_data_request[i].ranks = ref_requested_ranks[i];
        }
      }

      for (unsigned j = 0; j < NUM_SRC; ++j ) {

         check_data_request(data_request[j], &ref_data_request[j], 1);

         if (ref_num_interpolated_points[j] != num_interpolated_points[j][0])
            PUT_ERR("wrong number of interpolated points\n");

         for (unsigned i = 0; i < ref_num_interpolated_points[j]; ++i)
            if (ref_global_ids[j][i] != global_ids[j][i])
               PUT_ERR("wrong global id order\n");
      }

#undef NUM_TESTS
#define NUM_TESTS (4)

      for (size_t t = 0; t < NUM_TESTS; ++t) {

        //---------------------
        // do the interpolation
        //---------------------

        double field_data[NUM_TESTS][25] =
          {{1,1,1,1,1,
            1,1,1,1,1,
            1,1,1,1,1,
            1,1,1,1,1,
            1,1,1,1,1},
           {0.00,0.25,0.50,0.75,1.00,
            0.00,0.25,0.50,0.75,1.00,
            0.00,0.25,0.50,0.75,1.00,
            0.00,0.25,0.50,0.75,1.00,
            0.00,0.25,0.50,0.75,1.00},
           {0.00,0.00,0.00,0.00,0.00,
            0.25,0.25,0.25,0.25,0.25,
            0.50,0.50,0.50,0.50,0.50,
            0.75,0.75,0.75,0.75,0.75,
            1.00,1.00,1.00,1.00,1.00},
           {0.00,0.25,0.50,0.75,1.00,
            0.25,0.50,0.75,1.00,1.25,
            0.50,0.75,1.00,1.25,1.50,
            0.75,1.00,1.25,1.50,1.75,
            1.00,1.25,1.50,1.75,2.00}};
        double in_data_raw[NUM_SRC][2][20];
        double * remote_data[NUM_SRC][1];
        double * subdomain_in_data[NUM_SRC][1];
        double ** in_data[NUM_SRC][2];
        double send_buffer[NUM_SRC][4];

        for (unsigned int i = 0; i < NUM_SRC; ++i) {
          for (unsigned j = 0; j < ref_num_requested_points[i]; ++j)
            in_data_raw[i][0][j] = field_data[t][ref_requested_global_ids[i][j]];
          for (unsigned j = 0; j < num_cells_src[i][0] * num_cells_src[i][1]; ++j) {
            if (cell_is_on_rank_src[i][j] == (unsigned)-1)
              in_data_raw[i][1][j] = field_data[t][global_cell_ids_src[i][j]];
            else
              in_data_raw[i][1][j] = -1.0;
          }
          remote_data[i][0] = in_data_raw[i][0];
          subdomain_in_data[i][0] = in_data_raw[i][1];
          in_data[i][0] = remote_data[i];
          in_data[i][1] = subdomain_in_data[i];
          for (unsigned int j = 0; j < 4; ++j) send_buffer[i][j] = -1;
        }

        //---------------------------------
        // testing results of interpolation
        //---------------------------------

        for (unsigned int i = 0; i < NUM_SRC; ++i)
          yac_execute_interpolation_plan_element(
            interpolation_plan_element_stack[i], in_data[i], send_buffer[i]);

        double ref_out_data[NUM_TESTS][25] =
        {{-1,-1,-1,-1,-1,
          -1,-1,-1,-1,-1,
          -1, 1, 1,-1,-1,
          -1, 1, 1,-1,-1,
          -1,-1,-1,-1,-1},
         {-1.000,-1.000,-1.000,-1.000,-1.000,
          -1.000,-1.000,-1.000,-1.000,-1.000,
          -1.000, 0.375, 0.625,-1.000,-1.000,
          -1.000, 0.375, 0.625,-1.000,-1.000,
          -1.000,-1.000,-1.000,-1.000,-1.000},
         {-1.000,-1.000,-1.000,-1.000,-1.000,
          -1.000,-1.000,-1.000,-1.000,-1.000,
          -1.000, 0.375, 0.375,-1.000,-1.000,
          -1.000, 0.625, 0.625,-1.000,-1.000,
          -1.000,-1.000,-1.000,-1.000,-1.000},
         {-1.000,-1.000,-1.000,-1.000,-1.000,
          -1.000,-1.000,-1.000,-1.000,-1.000,
          -1.000, 0.750, 1.000,-1.000,-1.000,
          -1.000, 1.000, 1.250,-1.000,-1.000,
          -1.000,-1.000,-1.000,-1.000,-1.000}};

        double out_data[25];

        for (size_t i = 0; i < sizeof(out_data) / sizeof(out_data[0]); ++i)
          out_data[i] = -1;
        for (unsigned j = 0; j < NUM_SRC; ++j )
           for (unsigned i = 0; i < ref_num_interpolated_points[j]; ++i)
              out_data[global_ids[j][i]] = send_buffer[j][i];

        for (size_t i = 0; i < sizeof(out_data) / sizeof(out_data[0]); ++i)
          if (fabs(out_data[i] - ref_out_data[t][i] ) > 1.0e-3)
            PUT_ERR("wrong result buffer\n");
      }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < NUM_SRC; ++j ) {
         for (size_t i = 0; i < sizeof(method_stack[j])/sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (size_t i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
              sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 1);
      }

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_points(&cell_points_src[i]);

      for (unsigned int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (unsigned int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (unsigned int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);

      yac_reset_interpolation_method();
   }

   yac_comm_finalize();

   return TEST_EXIT_CODE;
}


