/**
 * @file test_interpolation_method_conserv_parallel.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <mpi.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_conserv.h"
#include "utils.h"
#include "test_interpolation_method_common.h"

static void target_main(struct communicator * global_comm,
                        struct communicator * target_comm);

static void source_main(struct communicator * global_comm,
                        struct communicator * source_comm);

int main (void) {

   struct communicator * global_comm;

   global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) == 8)
   { 
      int my_global_rank = yac_get_comm_rank(global_comm);

      // split processes into source and target

      unsigned target = my_global_rank < 4;

      struct communicator * comm;

      comm = yac_comm_split(global_comm, target, 0);

      if (target)
         target_main(global_comm, comm);
      else
         source_main(global_comm, comm);

      yac_free_communicator(comm);
   } else {
      PUT_ERR("ERROR: wrong number of processes\n");
   }

   yac_free_communicator(global_comm);

   MPI_Finalize();

   return TEST_EXIT_CODE;
}

static void source_main(struct communicator * global_comm,
                        struct communicator * source_comm) {

   // corner and cell ids for a 6 x 5 grid
   //
   // 35-----36-----37-----38-----39-----40-----41
   //  |      |      |      |      |      |      |
   //  |  24  |  25  |  26  |  27  |  28  |  29  |
   //  |      |      |      |      |      |      |
   // 28-----29-----30-----31-----32-----33-----34
   //  |      |      |      |      |      |      |
   //  |  18  |  19  |  20  |  21  |  22  |  23  |
   //  |      |      |      |      |      |      |
   // 21-----22-----23-----24-----25-----26-----27
   //  |      |      |      |      |      |      |
   //  |  12  |  13  |  14  |  15  |  16  |  17  |
   //  |      |      |      |      |      |      |
   // 14-----15-----16-----17-----18-----19-----20
   //  |      |      |      |      |      |      |
   //  |  06  |  07  |  08  |  09  |  10  |  11  |
   //  |      |      |      |      |      |      |
   // 07-----08-----09-----10-----11-----12-----13
   //  |      |      |      |      |      |      |
   //  |  00  |  01  |  02  |  03  |  04  |  05  |
   //  |      |      |      |      |      |      |
   // 00-----01-----02-----03-----04-----05-----06
   //
   // the domain is distributed among the processes as follows:
   // (index == process)
   //
   // +---+---+---+---+---+---+
   // | 2 | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+---+
   // | 2 | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+---+
   // | 2 | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 | 1 |
   // +---+---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 | 1 |
   // +---+---+---+---+---+---+
   //
   //--------------------
   // set up global search
   //--------------------

   int my_source_rank = yac_get_comm_rank(source_comm);

   // define domains on source

   struct grid * grid;
   unsigned num_cells[4][2] = {{4,3}, {4,3}, {4,4}, {4,4}};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[4][5] = {{-3,-2,-1,0,1},{-1,0,1,2,3},
                                 {-3,-2,-1,0,1},{-1,0,1,2,3}};
   double coordinates_y[4][5] = {{-3,-2,-1,0},{-3,-2,-1,0},
                                 {-2,-1,0,1,2},{-2,-1,0,1,2}};

   for ( unsigned i = 0; i < 4; ++i ) {
      for ( unsigned j = 0; j < 5; ++j ) {
         coordinates_x[i][j] *= YAC_RAD;
         coordinates_y[i][j] *= YAC_RAD;
      }
   }

   grid = yac_reg2d_grid_new(coordinates_x[my_source_rank],
                             coordinates_y[my_source_rank],
                             num_cells[my_source_rank], cyclic);

   struct subdomain subdomain;

   int global_corner_ids[4][25] = {{0,1,2,3,4, 7,8,9,10,11, 14,15,16,17,18, 21,22,23,24,25},
                                   {2,3,4,5,6, 9,10,11,12,13, 16,17,18,19,20, 23,24,25,26,27},
                                   {7,8,9,10,11, 14,15,16,17,18, 21,22,23,24,25, 28,29,30,31,32, 35,36,37,38,39},
                                   {9,10,11,12,13, 16,17,18,19,20, 23,24,25,26,27, 30,31,32,33,34, 37,38,39,40,41}};
   unsigned corner_is_on_rank[4][25] = {{-1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, 2,2,2,2,3},
                                        {0,-1,-1,-1,-1, 0,-1,-1,-1,-1, 0,-1,-1,-1,-1, 2,3,3,3,3},
                                        {0,0,0,0,1, -1,-1,-1,-1,3, -1,-1,-1,-1,3, -1,-1,-1,-1,3, -1,-1,-1,-1,3},
                                        {0,1,1,1,1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1}};
   int global_cell_ids[4][16] = {{0,1,2,3, 6,7,8,9, 12,13,14,15},
                                 {2,3,4,5, 8,9,10,11, 14,15,16,17},
                                 {6,7,8,9, 12,13,14,15, 18,19,20,21, 24,25,26,27},
                                 {8,9,10,11, 14,15,16,17, 20,21,22,23, 26,27,28,29}};
   unsigned cell_is_on_rank[4][16] = {{-1,-1,-1,1, -1,-1,-1,1, 2,2,2,3},
                                      {0,-1,-1,-1, 0,-1,-1,-1, 2,3,3,3},
                                      {0,0,0,1, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3},
                                      {0,1,1,1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1}};

   yac_init_subdomain_unstruct(&subdomain, grid, source_comm,
                               global_corner_ids[my_source_rank],
                               corner_is_on_rank[my_source_rank],
                               global_cell_ids[my_source_rank],
                               cell_is_on_rank[my_source_rank],
                               NULL, NULL);

   struct domain l_domain;
   struct subdomain * blocks = &subdomain;

   yac_init_domain(&l_domain, source_comm, &blocks, 1, "domain_src");

   yac_enddef_domain(&l_domain);

   //-------------------------------------
   // do the global search
   //-------------------------------------

   struct global_search * search;

   search = yac_global_search_new(&l_domain, "domain_tgt", global_comm,
                                  1,0, "component_src", "component_tgt");
   yac_enddef_global_search(search);

   //---------------------------------
   // initialise points on source side
   //---------------------------------

   struct points cell_points;

   yac_init_points(&cell_points, grid, CELL, NULL, NULL);

   //-------------------------------------
   // set up point sets
   //-------------------------------------
   struct points * points_per_subdomain = &cell_points;
   struct points ** point_sets = &points_per_subdomain;

   //----------------------------------------
   // test generation of interpolation method
   //----------------------------------------

   struct interpolation_method * method_conserv;
   method_conserv =
    yac_new_interpolation_method_conserv(search, &l_domain, "conserv stack",
                                     &point_sets, NULL, 1, 0, 1, 0, 0,
                                     CONSERV_DESTAREA);

   //----------
   // do search
   //----------

   struct interpolation_method * method_stack[2] = {method_conserv, NULL};
   struct interpolation_plan_element * interpolation_plan_element_stack[2] = {NULL, NULL};
   int global_ids[4][20] = {{0,1,2,3,5,6,7,8,10,11,12,13},
                            {1,2,3,4,6,7,8,9,11,12,13,14},
                            {0,1,2,3,5,6,7,8,10,11,12,13,15,16,17,18,20,21,22,23},
                            {1,2,3,4,6,7,8,9,11,12,13,14,16,17,18,19,21,22,23,24}};
   unsigned num_points[4] = {12,12,20,20};
   unsigned num_interpolated_points[1];

   yac_do_search_interpolation_method(method_stack, interpolation_plan_element_stack,
                                  global_ids[my_source_rank], NULL, CELL,
                                  num_points[my_source_rank], num_interpolated_points, 1);

   yac_enddef_interpolation_method(method_stack);

   //---------------------
   // do the interpolation
   //---------------------

   double in_data_raw[4][2][16] = {{{4,10,13,14,15,16},{1,2,3,-1, 7,8,9,-1, -1,-1,-1,-1}},
                                   {{16,17,18},{-1,4,5,6, -1,10,11,12, -1,-1,-1,-1}},
                                   {{16,22,28},{-1,-1,-1,-1, 13,14,15,-1, 19,20,21,-1, 25,26,27,-1}},
                                   {{-1},{-1,-1,-1,-1, -1,16,17,18, -1,22,23,24, -1,28,29,30}}};
   double * remote_data = in_data_raw[my_source_rank][0];
   double * subdomain_in_data = in_data_raw[my_source_rank][1];
   double ** in_data[2] = {&remote_data, &subdomain_in_data};
   double send_buffer[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

   yac_execute_interpolation_plan_element(interpolation_plan_element_stack,
                                          in_data, send_buffer);

   //---------------------------------
   // testing results of interpolation
   //---------------------------------

   struct interpolation_method_data_request data_request_;
   struct interpolation_method_data_request * data_request = &data_request_;

   yac_get_source_exchange_data_interpolation_method(method_stack, &data_request, 1);

   int ref_requested_global_ids[4][6] = {{3,9,12,13,14,15},
                                         {15,16,17},
                                         {15,21,27},
                                         {-1}};
   int ref_requested_ranks[4][6] =      {{1,1,2,2,2,3},
                                         {3,3,3},
                                         {3,3,3},
                                         {-1}};

   struct interpolation_method_data_request ref_data_request[4] =
      {{6, ref_requested_global_ids[0], ref_requested_ranks[0]},
       {3, ref_requested_global_ids[1], ref_requested_ranks[1]},
       {3, ref_requested_global_ids[2], ref_requested_ranks[2]},
       {0, NULL, NULL}};

   unsigned ref_num_interpolated_points[4] = {6,4,6,4};
   unsigned ref_global_ids[4][6] = {{0,1,2,5,6,7},
                                    {3,4,8,9},
                                    {10,11,12,15,16,17},
                                    {13,14,18,19}};
   double ref_out_data[4][6] = {{4.5,5.5,6.5,10.5,11.5,12.5},
                                {7.5,8.5,13.5,14.5},
                                {16.5,17.5,18.5,22.5,23.5,24.5},
                                {19.5,20.5,25.5,26.5}};

   check_data_request(data_request, &ref_data_request[my_source_rank], 1);

   if (ref_num_interpolated_points[my_source_rank] != num_interpolated_points[0])
      PUT_ERR("wrong number of interpolated points\n");

   for (unsigned i = 0; i < ref_num_interpolated_points[my_source_rank]; ++i) {
      if (ref_global_ids[my_source_rank][i] != global_ids[my_source_rank][i])
         PUT_ERR("wrong global id order\n");
      // since the cells are on the sphere, we do not have perfect
      // rectangular cells and hence the reference data is very inaccurate...
      if ( fabs(send_buffer[i] - ref_out_data[my_source_rank][i] ) > 1.0e-3 )
         PUT_ERR("wrong result buffer\n");
   }

   //---------------
   // cleanup
   //---------------

   for (unsigned i = 0; i < sizeof(method_stack)/sizeof(method_stack[0]); ++i)
      if (method_stack[i] != NULL)
         yac_delete_interpolation_method(method_stack[i]);
   for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack) /
        sizeof(interpolation_plan_element_stack[0]); ++i) {
      yac_free_interpolation_plan_element(interpolation_plan_element_stack[i]);
      free(interpolation_plan_element_stack[i]);
   }
   delete_interpolation_method_data_request_stack(data_request, 1);

   yac_free_points(&cell_points);
   yac_global_search_delete(search);
   yac_free_domain(&l_domain);
   yac_delete_grid(grid);

   yac_reset_interpolation_method();
}

static void target_main(struct communicator * global_comm,
                        struct communicator * target_comm) {


   // corner and cell ids for a 5 x 6 grid
   //
   // 36-----37-----38-----39-----40-----41
   //  |      |      |      |      |      |
   //  |  25  |  26  |  27  |  28  |  29  |
   //  |      |      |      |      |      |
   // 30-----31-----32-----33-----34-----35
   //  |      |      |      |      |      |
   //  |  20  |  21  |  22  |  23  |  24  |
   //  |      |      |      |      |      |
   // 24-----25-----26-----27-----28-----29
   //  |      |      |      |      |      |
   //  |  15  |  16  |  17  |  18  |  19  |
   //  |      |      |      |      |      |
   // 18-----19-----20-----21-----22-----23
   //  |      |      |      |      |      |
   //  |  10  |  11  |  12  |  13  |  14  |
   //  |      |      |      |      |      |
   // 12-----13-----14-----15-----16-----17
   //  |      |      |      |      |      |
   //  |  05  |  06  |  07  |  08  |  09  |
   //  |      |      |      |      |      |
   // 06-----07-----08-----09-----10-----11
   //  |      |      |      |      |      |
   //  |  00  |  01  |  02  |  03  |  04  |
   //  |      |      |      |      |      |
   // 00-----01-----02-----03-----04-----05
   //
   // the domain is distributed among the processes as follows:
   // (index == process)
   //
   // +---+---+---+---+---+
   // | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+
   // | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+
   // | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 |
   // +---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 |
   // +---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 |
   // +---+---+---+---+---+
   //
   //--------------------
   // set up global search
   //--------------------

   int my_target_rank = yac_get_comm_rank(target_comm);

   // define a domain on the target

   struct grid * grid;
   unsigned num_cells[4][2] = {{4,4},{3,4},{3,4},{4,4}};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[4][5] = {{-2.5,-1.5,-0.5,0.5,1.5},
                                 {-0.5,0.5,1.5,2.5},
                                 {-2.5,-1.5,-0.5,0.5},
                                 {-1.5,-0.5,0.5,1.5,2.5}};
   double coordinates_y[4][5] = {{-2.5,-1.5,-0.5,0.5,1.5},
                                 {-2.5,-1.5,-0.5,0.5,1.5},
                                 {-0.5,0.5,1.5,2.5,3.5},
                                 {-0.5,0.5,1.5,2.5,3.5}};

   for ( unsigned i = 0; i < 4; ++i ) {
      for ( unsigned j = 0; j < 5; ++j ) {
         coordinates_x[i][j] *= YAC_RAD;
         coordinates_y[i][j] *= YAC_RAD;
      }
   }

   grid = yac_reg2d_grid_new(coordinates_x[my_target_rank], coordinates_y[my_target_rank],
                             num_cells[my_target_rank], cyclic);

   struct subdomain subdomain;

   int global_corner_ids[4][25] = {{0,1,2,3,4, 6,7,8,9,10, 12,13,14,15,16, 18,19,20,21,22, 24,25,26,27,28},
                                   {2,3,4,5, 8,9,10,11, 14,15,16,17, 20,21,22,23, 26,27,28,29},
                                   {12,13,14,15, 18,19,20,21, 24,25,26,27, 30,31,32,33, 36,37,38,39},
                                   {13,14,15,16,17, 19,20,21,22,23, 15,26,27,28,29, 31,32,33,34,35, 37,38,39,40,41}};
   unsigned corner_is_on_rank[4][25] = {{-1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, 2,2,2,3,3},
                                        {0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 3,3,3,3},
                                        {0,0,0,0, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3},
                                        {0,0,0,1,1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1}};
   int global_cell_ids[4][16] = {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18},
                                 {2,3,4, 7,8,9, 12,13,14, 17,18,19},
                                 {10,11,12, 15,16,17, 20,21,22, 25,26,27},
                                 {11,12,13,14, 16,17,18,19, 21,22,23,24, 26,27,28,29}};
   unsigned cell_is_on_rank[4][16] = {{-1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, 2,2,3,3},
                                      {0,-1,-1, 0,-1,-1, 0,-1,-1, 3,3,3},
                                      {0,0,0, -1,-1,3, -1,-1,3, -1,-1,3},
                                      {0,0,1,1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1}};

   yac_init_subdomain_unstruct(&subdomain, grid, target_comm,
                               global_corner_ids[my_target_rank],
                               corner_is_on_rank[my_target_rank],
                               global_cell_ids[my_target_rank],
                               cell_is_on_rank[my_target_rank],
                               NULL, NULL);

   struct domain l_domain;

   struct subdomain * blocks = &subdomain;

   yac_init_domain(&l_domain, target_comm, &blocks, 1, "domain_tgt");

   // finish domain definition
   yac_enddef_domain(&l_domain);

   //-------------------------------------
   // init global search on both processes
   //-------------------------------------

   struct global_search * search;

   search = yac_global_search_new(&l_domain, "domain_src", global_comm,
                                  0,1, "component_tgt", "component_src");
   yac_enddef_global_search(search);

   //---------------
   // cleanup
   //---------------

   yac_global_search_delete(search);
   yac_free_domain(&l_domain);
   yac_delete_grid(grid);

   yac_reset_interpolation_method();
}

