/**
 * @file test_interpolation_method_parallel.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>

#include <mpi.h>

#include "tests.h"
#include "test_common.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_fixed.h"
#include "interpolation_method_avg.h"
#include "fields.h"
#include "utils.h"
#include "test_interpolation_method_common.h"

static void target_main(struct communicator * global_comm,
                        struct communicator * target_comm);

static void source_main(struct communicator * global_comm,
                        struct communicator * source_comm);

int main (void) {

   struct communicator * global_comm;

   global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) == 4)
   {

      int my_global_rank;

      my_global_rank = yac_get_comm_rank(global_comm);

      // split processes into source an target

      unsigned target = my_global_rank < 2;

      struct communicator * comm;

      comm = yac_comm_split(global_comm, target, 0);

      if (target)
         target_main(global_comm, comm);
      else
         source_main(global_comm, comm);

      yac_free_communicator(comm);
   }

   yac_free_communicator(global_comm);

   MPI_Finalize();

   return TEST_EXIT_CODE;
}

/*
 * The source domain is distributed among 2 processes, of which on process has
 * one and the other two subdomains.
 *
 * The global source domain is a 5x4 grid:
 *
 *    24--44--25--45--26--46--27--47--28--48--29
 *    |       |       |       |       |       |
 *    34  15  36  16  38  17  40  18  42  19  43
 *    |       |       |       |       |       |
 *    18--33--19--35--20--37--21--39--22--41--23
 *    |       |       |       |       |       |
 *    23  10  25  11  27  12  29  13  31  14  32
 *    |       |       |       |       |       |
 *    12--22--13--24--14--26--15--28--16--30--17
 *    |       |       |       |       |       |
 *    12  05  14  06  16  07  18  08  20  09  21
 *    |       |       |       |       |       |
 *    06--11--07--13--08--15--09--17--10--19--11
 *    |       |       |       |       |       |
 *    01  00  03  01  05  02  07  03  09  04  10
 *    |       |       |       |       |       |
 *    00--01--01--02--02--04--03--06--04--08--05
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    01--01--01--01--01--01--01--01--01--01--01
 *    |       |       |       |       |       |
 *    01  01  01  01  01  01  01  01  01  01  01
 *    |       |       |       |       |       |
 *    01--01--01--01--01--01--01--01--01--01--10
 *    |       |       |       |       |       |
 *    01  01  01  01  01  10  10  10  10  10  10
 *    |       |       |       |       |       |
 *    01--01--00--01--01--10--10--10--10--10--10
 *    |       |       |       |       |       |
 *    00  00  00  00  00  10  10  10  10  10  10
 *    |       |       |       |       |       |
 *    00--00--00--00--00--10--10--10--10--10--10
 *    |       |       |       |       |       |
 *    00  00  00  00  10  10  10  10  10  10  10
 *    |       |       |       |       |       |
 *    00--00--00--00--10--10--10--10--10--10--10
 */
static void source_main(struct communicator * global_comm,
                        struct communicator * source_comm) {

   unsigned my_rank, i;

   my_rank = yac_get_comm_rank(source_comm);

   unsigned num_subdomains;

   if (my_rank == 0)
      num_subdomains = 2;
   else
      num_subdomains = 1;

   struct grid * grid[2];
   unsigned num_cells_p0[2][2] = {{2,2},{5,2}},
            num_cells_p1[2] =  {3,3};
   unsigned cyclic[2] = {0,0};
   double coordinates_x_p0[2][6] = {{0*YAC_RAD,1*YAC_RAD,2*YAC_RAD},
                                    {0*YAC_RAD,1*YAC_RAD,2*YAC_RAD,
                                     3*YAC_RAD,4*YAC_RAD,5*YAC_RAD}},
          coordinates_x_p1[4] = {2*YAC_RAD,3*YAC_RAD,4*YAC_RAD,5*YAC_RAD};
   double coordinates_y_p0[2][3] = {{0*YAC_RAD,1*YAC_RAD,2*YAC_RAD},
                                    {2*YAC_RAD,3*YAC_RAD,4*YAC_RAD}};
   double coordinates_y_p1[4] = {0*YAC_RAD,1*YAC_RAD,2*YAC_RAD,3*YAC_RAD};

   if (my_rank == 0)
      for (i = 0; i < 2; ++i)
         grid[i] = yac_reg2d_grid_new(coordinates_x_p0[i], coordinates_y_p0[i],
                                      num_cells_p0[i], cyclic);
   else
      grid[0] = yac_reg2d_grid_new(coordinates_x_p1, coordinates_y_p1,
                                   num_cells_p1, cyclic);

   struct subdomain subdomain[2];

   int global_corner_ids_p0[2][18] =
      {{0,1,2,6,7,8,12,13,14},
       {12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29}};
   unsigned corner_is_on_rank_p0[2][18] =
      {{-1,-1,1,-1,-1,-1,0,-1,0},
       {-1, 0,-1, 1, 1, 1,-1,-1,-1,-1,-1, 1,-1,-1,-1,-1,-1,-1}};
   int global_cell_ids_p0[2][10] =
      {{0,1,2,5,6,7},
       {10,11,12,13,14,15,16,17,18,19}};
   unsigned cell_is_on_rank_p0[2][10] =
      {{-1,-1,-1,-1,-1,-1},
       {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};

   int global_corner_ids_p1[16] =
      { 2, 3, 4, 5,  8, 9,10,11, 14,15,16,17, 20,21,22,23};
   unsigned corner_is_on_rank_p1[16] =
      {-1,-1,-1,-1,  0,-1,-1,-1,  0,-1,-1,-1,  0, 0, 0,-1};
   int global_cell_ids_p1[9] =
      {2,3,4,7,8,9,12,13,14};
   unsigned cell_is_on_rank_p1[9] =
      {-1,-1,-1,-1,-1,-1,-1,-1,-1};

   if (my_rank == 0)
      for (i = 0; i < 2; ++i)
         yac_init_subdomain_unstruct(subdomain+i, grid[i], source_comm,
                                     global_corner_ids_p0[i], corner_is_on_rank_p0[i],
                                     global_cell_ids_p0[i], cell_is_on_rank_p0[i], NULL, NULL);
   else
      yac_init_subdomain_unstruct(subdomain+0, grid[0], source_comm,
                                  global_corner_ids_p1, corner_is_on_rank_p1,
                                  global_cell_ids_p1, cell_is_on_rank_p1, NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[2] = {subdomain+0, subdomain+1};

   yac_init_domain(&domain, source_comm, subdomains, num_subdomains, "source_domain");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "target_domain", global_comm, 1, 0,
                                  "source_component", "target_component");
   yac_enddef_global_search(search);

   struct points points[2];

   if (my_rank == 0)
      for (i = 0; i < 2; ++i)
   yac_init_points(points+i, grid[i], CORNER,
                   coordinates_x_p0[i], coordinates_y_p0[i]);
   else
      yac_init_points(points+0, grid[0], CORNER,
                      coordinates_x_p1, coordinates_y_p1);

   struct points * points_per_subdomain[2][1] = {{points+0},{points+1}};
   struct points ** points_set[2] = {points_per_subdomain[0],points_per_subdomain[1]};

   struct interpolation_method * method_fixed, * method_avg;
   double fixed_value = 1337;

   method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
   method_avg = yac_new_interpolation_method_avg(search, &domain, "avg stack",
                                             points_set, NULL, 1, 0, 0, 0);

   struct interpolation_method * method_stack[] = {method_avg, method_fixed, NULL};
   struct interpolation_plan_element * plan_stack[] = {NULL, NULL, NULL};

   // for now we search on every source process for all target corners
   int global_ids[24] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23};
   double tgt_coordinates_x[24] = {0.5,1.5,2.5,3.5,4.5,5.5, 0.5,1.5,2.5,3.5,4.5,5.5,
                                   0.5,1.5,2.5,3.5,4.5,5.5, 0.5,1.5,2.5,3.5,4.5,5.5};
   double tgt_coordinates_y[24] = {0.5,0.5,0.5,0.5,0.5,0.5, 1.5,1.5,1.5,1.5,1.5,1.5,
                                   2.5,2.5,2.5,2.5,2.5,2.5, 3.5,3.5,3.5,3.5,3.5,3.5};
   double tgt_coordinates_xyz[24][3];
   for ( i = 0; i < 24; ++i )
      LLtoXYZ_deg(tgt_coordinates_x[i],
                  tgt_coordinates_y[i],
                  tgt_coordinates_xyz[i]);

   unsigned num_points = sizeof(global_ids) / sizeof(global_ids[0]);
   unsigned num_interpolated_points[2];

   // do the interpolation search

   yac_do_search_interpolation_method(method_stack, plan_stack, global_ids,
                                  tgt_coordinates_xyz, CELL, num_points,
                                  num_interpolated_points, 1);
   yac_enddef_interpolation_method(method_stack);

   // do the interpolation

   double in_data_raw_p0[3][18] = {{2,23},{0,1,-1,6,7,8,-1,13,-1},
                                   {12,-1,14,-1,-1,-1, 18,19,20,21,22,-1, 24,25,26,27,28,29}};
   double in_data_raw_p1[2][16] = {{8,14,20,21,22},{2,3,4,5, -1,9,10,11, -1,15,16,17, -1,-1,-1,23}};

   double * subdomain_in_data_p0[3][1] = {{in_data_raw_p0[0]}, {in_data_raw_p0[1]}, {in_data_raw_p0[2]}};
   double * subdomain_in_data_p1[2][1] = {{in_data_raw_p1[0]}, {in_data_raw_p1[1]}};
   double ** in_data[3];

   if (my_rank == 0) {
      in_data[0] = subdomain_in_data_p0[0];
      in_data[1] = subdomain_in_data_p0[1];
      in_data[2] = subdomain_in_data_p0[2];
   } else {
      in_data[0] = subdomain_in_data_p1[0];
      in_data[1] = subdomain_in_data_p1[1];
   }

   double out_data[num_interpolated_points[0] + num_interpolated_points[1]];

   yac_execute_interpolation_plan_element(plan_stack, in_data, out_data);

   // test results

   struct interpolation_method_data_request data_request_;
   struct interpolation_method_data_request * data_request = &data_request_;

   yac_get_source_exchange_data_interpolation_method(method_stack, &data_request, 1);

   struct interpolation_method_data_request ref_data_request[2] =
      {{2, (int[]){2,23}, (int[]){1,1}}, {5, (int[]){8,14,20,21,22}, (int[]){0,0,0,0,0}}};

   check_data_request(data_request, ref_data_request + my_rank, 1);

   unsigned ref_num_interpolated_points[2][2] = {{11,0},{9,4}};

   for (i = 0; i < 2; ++i)
      if (ref_num_interpolated_points[my_rank][i] != num_interpolated_points[i])
         PUT_ERR("wrong number of interpolated points\n");

   int ref_global_ids[2][13] = {{0,1,6,7,12,13,18,19,20,21,22},
                                {2,3,4,8,9,10,14,15,16,5,11,17,23}};

   for (i = 0; i < num_interpolated_points[0] + num_interpolated_points[1]; ++i)
      if (global_ids[i] != ref_global_ids[my_rank][i])
         PUT_ERR("wrong global id\n");

   double ref_out_data[2][13] = {{3.5,4.5,9.5,10.5,15.5,16.5,21.5,22.5,23.5,24.5,25.5},
                                 {5.5,6.5,7.5,11.5,12.5,13.5,17.5,18.5,19.5,
                                  fixed_value, fixed_value, fixed_value, fixed_value}};

   for (i = 0; i < num_interpolated_points[0] + num_interpolated_points[1]; ++i)
      if (double_are_unequal(out_data[i], ref_out_data[my_rank][i]))
         PUT_ERR("wrong result buffer\n");

   // cleanup

   for (i = 0; i < 2; ++i)
      if (plan_stack[i] != NULL)
         yac_free_interpolation_plan_element(plan_stack[i]);
   delete_interpolation_method_data_request_stack(data_request, 1);
   yac_delete_interpolation_method(method_avg);
   yac_delete_interpolation_method(method_fixed);
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   for (i = 0; i < num_subdomains; ++i)
      yac_delete_grid(grid[i]);

   yac_reset_interpolation_method();
}

/*
 * The target domain is distributed among 2 processes with one subdomain each
 *
 * The global target domain is a 5x3 grid:
 *
 *    18--33--19--34--20--35--21--36--22--37--23
 *    |       |       |       |       |       |
 *    23  10  25  11  27  12  29  13  31  14  32
 *    |       |       |       |       |       |
 *    12--22--13--24--14--26--15--28--16--30--17
 *    |       |       |       |       |       |
 *    12  05  14  06  16  07  18  08  20  09  21
 *    |       |       |       |       |       |
 *    06--11--07--13--08--15--09--17--10--19--11
 *    |       |       |       |       |       |
 *    01  00  03  01  05  02  07  03  09  04  10
 *    |       |       |       |       |       |
 *    00--01--01--02--02--04--03--06--04--08--05
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    00--00--00--00--00--00--00--10--10--10--10
 *    |       |       |       |       |       |
 *    00  00  00  00  00  00  00  10  10  10  10
 *    |       |       |       |       |       |
 *    00--00--00--00--00--00--00--10--10--10--10
 *    |       |       |       |       |       |
 *    00  00  00  00  00  00  00  10  10  10  10
 *    |       |       |       |       |       |
 *    00--00--00--00--00--00--00--10--10--10--10
 *    |       |       |       |       |       |
 *    00  00  00  00  00  00  00  10  10  10  10
 *    |       |       |       |       |       |
 *    00--00--00--00--00--00--00--10--10--10--10
 */
static void target_main(struct communicator * global_comm,
                        struct communicator * target_comm) {

   int my_rank;

   my_rank = yac_get_comm_rank(target_comm);

   struct grid * grid;
   unsigned num_cells[2][2] = {{3,3},{2,3}};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[2][4] = {{0.5,1.5,2.5,3.5},{3.5,4.5,5.5}};
   double coordinates_y[4] = {0.5,1.5,2.5,3.5};

   for ( unsigned j = 0; j < 4; ++j ) {
      coordinates_y[j] *= YAC_RAD;
      for ( unsigned i = 0; i < 2; ++i ) {
         coordinates_x[i][j] *= YAC_RAD;
      }
   }

   grid = yac_reg2d_grid_new(coordinates_x[my_rank], coordinates_y,
                             num_cells[my_rank], cyclic);

   struct subdomain subdomain;

   int global_corner_ids[2][16] =
      {{0,1,2,3,6,7,8,9,12,13,14,15,18,19,20,21},
       {3,4,5,9,10,11,15,16,17,21,22,23}};
   unsigned corner_is_on_rank[2][16] =
      {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
       {0,-1,-1,0,-1,-1,0,-1,-1,0,-1,-1}};
   int global_cell_ids[2][9] = {{0,1,2,5,6,7,10,11,12}, {3,4,8,9,13,14}};
   unsigned cell_is_on_rank[2][9] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1},{-1,-1,-1,-1,-1,-1}};

   yac_init_subdomain_unstruct(&subdomain, grid, target_comm,
                               global_corner_ids[my_rank], corner_is_on_rank[my_rank],
                               global_cell_ids[my_rank], cell_is_on_rank[my_rank], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[1] = {&subdomain};

   yac_init_domain(&domain, target_comm, subdomains, 1, "target_domain");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "source_domain", global_comm, 0, 1,
                                  "target_component", "source_component");
   yac_enddef_global_search(search);

   yac_global_search_delete(search);
   yac_free_domain(&domain);
   yac_delete_grid(grid);

   yac_reset_interpolation_method();
}

