/**
 * @file test_points.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "tests.h"
#include "test_common.h"
#include "grid.h"
#include "grid_reg2d.h"
#include "points.h"

int main (void) {

   unsigned i;

   { // testing points that are defined on the corners of the base grid
      //---------------
      // setup
      //---------------
      double grid_coord_x[3] = {0,1,2};
      double grid_coord_y[3] = {0,1,2};
      double points_coord_x[3] = {0.1,1.1,2.1};
      double points_coord_y[3] = {0.1,1.1,2.1};
      unsigned num_cells[2] = {2,2};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;
      struct points points;
      struct grid * point_grid;
      unsigned coordinate_array_sizes[2];
      unsigned ref_coordinate_array_sizes[2] = {3, 3};

      reg_grid = yac_reg2d_grid_new(grid_coord_x, grid_coord_y, num_cells, cyclic);

      //---------------
      // testing
      //---------------
      yac_init_points(&points, reg_grid, CORNER, points_coord_x, points_coord_y);

      point_grid = yac_get_point_grid(&points);

      if (yac_get_num_grid_corners(point_grid) != 9) PUT_ERR("error in get_num_grid_corners\n")
      if (yac_get_num_grid_cells(point_grid) != 4) PUT_ERR("error in get_num_grid_cells\n")
      if (yac_get_num_grid_edges(point_grid) != 12) PUT_ERR("error in get_num_grid_edges\n")

      yac_get_coordinate_array_sizes(reg_grid, CORNER, coordinate_array_sizes);
      if (coordinate_array_sizes[0] != ref_coordinate_array_sizes[0] ||
          coordinate_array_sizes[1] != ref_coordinate_array_sizes[1])
         PUT_ERR("error in get_coordinate_array_sizes\n")
      if (yac_get_data_size(points) != 9)
         PUT_ERR("error in get_data_size\n")

      double ref_coordinates[9][2] = {{0.1,0.1},{1.1,0.1},{2.1,0.1},
                                      {0.1,1.1},{1.1,1.1},{2.1,1.1},
                                      {0.1,2.1},{1.1,2.1},{2.1,2.1}};
      double coordinates[2];
      for (i = 0; i < 9; ++i) {
         yac_get_point_coordinates(&points, i, coordinates);
         if (double_are_unequal(coordinates[0], ref_coordinates[i][0]) ||
             double_are_unequal(coordinates[1], ref_coordinates[i][1]))
            PUT_ERR("error in get_point_coordinates");
      }

      yac_free_points(&points);
      yac_delete_grid(reg_grid);
   }

   { // testing points that are defined in the cells of the base grid
      //---------------
      // setup
      //---------------
      double grid_coord_x[3] = {0,1,2};
      double grid_coord_y[3] = {0,1,2};
      double points_coord_x[2] = {0.5,1.5};
      double points_coord_y[2] = {0.5,1.5};
      unsigned num_cells[2] = {2,2};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;
      struct points points;
      struct grid * point_grid;
      unsigned coordinate_array_sizes[2];
      unsigned ref_coordinate_array_sizes[2] = {2, 2};

      reg_grid = yac_reg2d_grid_new(grid_coord_x, grid_coord_y, num_cells, cyclic);

      //---------------
      // testing
      //---------------
      yac_init_points(&points, reg_grid, CELL, points_coord_x, points_coord_y);

      point_grid = yac_get_point_grid(&points);

      if (yac_get_num_grid_corners(point_grid) != 4) PUT_ERR("error in get_num_grid_corners\n")
      if (yac_get_num_grid_cells(point_grid) != 1) PUT_ERR("error in get_num_grid_cells\n")
      if (yac_get_num_grid_edges(point_grid) != 4) PUT_ERR("error in get_num_grid_edges\n")

      yac_get_coordinate_array_sizes(reg_grid, CELL, coordinate_array_sizes);
      if (coordinate_array_sizes[0] != ref_coordinate_array_sizes[0] ||
          coordinate_array_sizes[1] != ref_coordinate_array_sizes[1])
         PUT_ERR("error in get_coordinate_array_sizes\n")
      if (yac_get_data_size(points) != 4)
         PUT_ERR("error in get_data_size\n")

      double ref_coordinates[4][2] = {{0.5,0.5},{1.5,0.5},
                                      {0.5,1.5},{1.5,1.5}};
      double coordinates[2];
      for (i = 0; i < 4; ++i) {
         yac_get_point_coordinates(&points, i, coordinates);
         if (double_are_unequal(coordinates[0], ref_coordinates[i][0]) ||
             double_are_unequal(coordinates[1], ref_coordinates[i][1]))
            PUT_ERR("error in get_point_coordinates");
      }

      yac_free_points(&points);
      yac_delete_grid(reg_grid);
   }

   { // testing points that are defined on the edges of the base grid
      //---------------
      // setup
      //---------------
      double grid_coord_x[3] = {0,1,2};
      double grid_coord_y[3] = {0,1,2};
      double points_coord_x[12] = {0.5,0.0,1.5,1.0,2.0, 0.5,0.0,1.5,1.0,2.0, 0.5,1.5};
      double points_coord_y[12] = {0.0,0.5,0.0,0.5,0.5, 1.0,1.5,1.0,1.5,1.5, 2.0,2.0};
      unsigned num_cells[2] = {2,2};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;
      struct points points;
      unsigned coordinate_array_sizes[2];
      unsigned ref_coordinate_array_sizes[2] = {12, 12};

      reg_grid = yac_reg2d_grid_new(grid_coord_x, grid_coord_y, num_cells, cyclic);

      //---------------
      // testing
      //---------------
      yac_init_points(&points, reg_grid, EDGE, points_coord_x, points_coord_y);

      yac_get_coordinate_array_sizes(reg_grid, EDGE, coordinate_array_sizes);
      if (coordinate_array_sizes[0] != ref_coordinate_array_sizes[0] ||
          coordinate_array_sizes[1] != ref_coordinate_array_sizes[1])
         PUT_ERR("error in get_coordinate_array_sizes\n")
      if (yac_get_data_size(points) != 12)
         PUT_ERR("error in get_data_size\n")

      double coordinates[2];
      for (i = 0; i < 4; ++i) {
         yac_get_point_coordinates(&points, i, coordinates);
         if (double_are_unequal(coordinates[0], points_coord_x[i]) ||
             double_are_unequal(coordinates[1], points_coord_y[i]))
            PUT_ERR("error in get_point_coordinates");
      }

      yac_free_points(&points);
      yac_delete_grid(reg_grid);
   }

   return TEST_EXIT_CODE;
}

