/**
 * @file test_read_icon_parallel.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "tests.h"
#include "test_common.h"
#include "read_icon_grid.h"
#include "grid.h"
#include "geometry.h"
#include "grid2vtk.h"

#include <netcdf.h>

#include <mpi.h>

/* the grid file contains the following grid
 * (numbers == global cell/vertex indices, XX == masked cell)
 *
 *00----01----02----03----04   ------------------------
 * \ 00 /\ 11 /\ 12 /\ 13 /    \    /\ XX /\    /\    /
 *  \  /  \  /  \  /  \  /      \  /  \  /  \  /  \  /
 *   \/ 01 \/ 10 \/ 14 \/        \/    \/ XX \/    \/
 *   05----06----07----08         ------------------
 *    \ 02 /\ 09 /\ 15 /          \    /\ XX /\    /
 *     \  /  \  /  \  /            \  /  \  /  \  /
 *      \/ 03 \/ 08 \/              \/    \/ XX \/
 *      09----10----11               ------------
 *       \ 04 /\ 07 /                \    /\ XX /
 *        \  /  \  /                  \  /  \  /
 *         \/ 05 \/                    \/    \/
 *         12----13                     ------
 *          \ 06 /                      \    /
 *           \  /                        \  /
              \/                          \/
 *            14
 */

// global grid data
double vlon[15] = {-2,-1,0,1,2, -1.5,-0.5,0.5,1.5, -1,0,1, -0.5,0.5, 0};
double vlat[15] = {1.5,1.5,1.5,1.5,1.5, 0.5,0.5,0.5,0.5, -0.5,-0.5,-0.5,
                   -1.5,-1.5, -2.5};
double clon[16] = {1,1,0,0,-1,-1,-2,-1,0,0,1,1,1,1,1,0};
double clat[16] = {-1.5,-1,-1,-0.5,-0.5,0,0,0.5,0.5,0,0,-0.5,0.5,1,1,1.5};
int mask[16] = {0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0};

int vertex_of_cell[3][16] = {{1,2,6,7,10,11,13,11,8,7,3,2,3,4,4,8},
                             {2,6,7,10,11,13,14,12,11,8,7,3,4,5,8,9},
                             {6,7,10,11,13,14,15,14,12,11,8,7,8,9,9,12}};
int cells_of_vertex[6][15] = {{1, 1,11,13,14,1, 2, 9,14,3, 4, 8,5,6,7},
                              {0, 2,12,14, 0,2, 3,10,15,4, 5, 9,6,7,0},
                              {0,12,13,15, 0,3, 4,11,16,5, 6,16,7,8,0},
                              {0, 0, 0, 0, 0,0,10,13, 0,0, 8, 0,0,0,0},
                              {0, 0, 0, 0, 0,0,11,15, 0,0, 9, 0,0,0,0},
                                {0, 0, 0, 0, 0,0,12,16, 0,0,10, 0,0,0,0}};

static void write_test_grid_file(char const * file_name);

int main(void) {

  MPI_Init(NULL, NULL);

  int comm_size, comm_rank;

  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);

  if ((comm_size != 4) && (comm_rank == 0)) {
    fputs("wrong number of processes (has to be four)\n", stderr);
    exit(EXIT_FAILURE);
  }

  for (size_t i = 0; i < sizeof(vlon) / sizeof(vlon[0]); ++i) vlon[i] *= YAC_RAD;
  for (size_t i = 0; i < sizeof(vlat) / sizeof(vlat[0]); ++i) vlat[i] *= YAC_RAD;
  for (size_t i = 0; i < sizeof(clon) / sizeof(clon[0]); ++i) clon[i] *= YAC_RAD;
  for (size_t i = 0; i < sizeof(clat) / sizeof(clat[0]); ++i) clat[i] *= YAC_RAD;

  char const filename[] = "test_read_icon_parallel_grid.nc";

  if (comm_rank == 0) write_test_grid_file(filename);

  // ensure that the grid file exists
  MPI_Barrier(MPI_COMM_WORLD);

  for (int n = -1; n < comm_size + 2; ++n) {

    int nbr_vertices, nbr_cells;
    int * num_vertices_per_cell;
    int * cell_to_vertex;
    int * global_cell_ids;
    int * cell_owner;
    int * global_vertex_ids;
    int * vertex_owner;
    double * x_vertices, * y_vertices;
    double * x_cells, * y_cells;
    int * cell_mask;

    read_icon_grid_information_parallel(
      filename, MPI_COMM_WORLD, &nbr_vertices, &nbr_cells, &num_vertices_per_cell,
      &cell_to_vertex, &global_cell_ids, &cell_owner, &global_vertex_ids,
      &vertex_owner, &x_vertices, &y_vertices, &x_cells, &y_cells, &cell_mask,
      (unsigned)-1);

    int ref_nbr_cells[4] = {11, 9, 14, 9};
    int ref_nbr_vertices[4] = {11, 10, 13, 10};
    int ref_cell_to_vertex[14][3];
    int ref_global_cell_ids[4][14] = {{0,1,2,3,4,5,7,8,9,10,11},
                                      {4,5,6,7,2,3,8,9,15},
                                      {8,9,10,11,0,1,2,3,4,5,7,12,14,15},
                                      {12,13,14,15,7,8,9,10,11}};
    int ref_cell_owner[14];
    int ref_global_vertex_ids[4][13] = {{0,1,2,5,6,7,9,10,11,12,13},
                                        {5,6,7,8,9,10,11,12,13,14},
                                        {0,1,2,3,5,6,7,8,9,10,11,12,13},
                                        {1,2,3,4,6,7,8,10,11,13}};
    int ref_vertex_owner[13];

    {
      int all_cell_to_vertex[16][3] =
        {{0,1,5}, {1,5,6}, {5,6,9}, {6,9,10}, {9,10,12}, {10,12,13}, {12,13,14},
         {10,11,13}, {7,10,11}, {6,7,10}, {2,6,7}, {1,2,6}, {2,3,7}, {3,4,8},
         {3,7,8}, {7,8,11}};
      for (int i = 0; i < ref_nbr_cells[comm_rank]; ++i) {
        for (int j = 0; j < 3; ++j) {
          for (int k = 0; k < ref_nbr_vertices[comm_rank]; ++k) {
            if (all_cell_to_vertex[ref_global_cell_ids[comm_rank][i]][j] ==
                ref_global_vertex_ids[comm_rank][k]) {
              ref_cell_to_vertex[i][j] = k;
              break;
            }
          }
        }
      }
      int global_cell_owner[16] = {0,0,0,0, 1,1,1,1, 2,2,2,2, 3,3,3,3};
      for (int i = 0; i < ref_nbr_cells[comm_rank]; ++i) {
        ref_cell_owner[i] =
          global_cell_owner[ref_global_cell_ids[comm_rank][i]];
        if (ref_cell_owner[i] == comm_rank) ref_cell_owner[i] = -1;
      }
      int global_vertex_owner[15] = {0,2,3,3,3,0,2,3,3,1,2,3,1,1,1};
      for (int i = 0; i < ref_nbr_vertices[comm_rank]; ++i) {
        ref_vertex_owner[i] =
          global_vertex_owner[ref_global_vertex_ids[comm_rank][i]];
        if (ref_vertex_owner[i] == comm_rank) ref_vertex_owner[i] = -1;
      }
    }

    if (nbr_cells != ref_nbr_cells[comm_rank])
      PUT_ERR("wrong number of vertices\n");
    if (nbr_vertices != ref_nbr_vertices[comm_rank])
      PUT_ERR("wrong number of vertices\n");
    for (int i = 0; i < nbr_cells; ++i)
      if (num_vertices_per_cell[i] != 3)
        PUT_ERR("wrong number of vertices per cell\n");
    for (int i = 0; i < nbr_cells; ++i) {
      for (int j = 0; j < 3; ++j)
        if (cell_to_vertex[3 * i + j] != ref_cell_to_vertex[i][j])
          PUT_ERR("error in cell_to_vertex\n");
      if (global_cell_ids[i] != ref_global_cell_ids[comm_rank][i])
        PUT_ERR("wrong global cell id\n");
      if (cell_owner[i] != ref_cell_owner[i])
        PUT_ERR("wrong cell owner\n");
      if (double_are_unequal(x_cells[i],
                             clon[ref_global_cell_ids[comm_rank][i]]))
        PUT_ERR("wrong cell longitude\n");
      if (double_are_unequal(y_cells[i],
                             clat[ref_global_cell_ids[comm_rank][i]]))
        PUT_ERR("wrong cell latitude\n");
      if (cell_mask[i] != mask[ref_global_cell_ids[comm_rank][i]])
        PUT_ERR("wrong cell mask\n");
    }
    for (int i = 0; i < nbr_vertices; ++i) {
      if (global_vertex_ids[i] != ref_global_vertex_ids[comm_rank][i])
        PUT_ERR("wrong global vertex id\n");
      if (vertex_owner[i] != ref_vertex_owner[i]) PUT_ERR("wrong vertex owner\n");
      if (double_are_unequal(x_vertices[i],
                             vlon[ref_global_vertex_ids[comm_rank][i]]))
        PUT_ERR("wrong vertex longitude\n");
      if (double_are_unequal(y_vertices[i],
                            vlat[ref_global_vertex_ids[comm_rank][i]]))
        PUT_ERR("wrong vertex latitude\n");
    }

    free(num_vertices_per_cell);
    free(cell_to_vertex);
    free(cell_owner);
    free(vertex_owner);
    free(global_cell_ids);
    free(global_vertex_ids);
    free(x_vertices);
    free(y_vertices);
    free(x_cells);
    free(y_cells);
    free(cell_mask);
  }

  // ensure that all processes finished reading the file
  MPI_Barrier(MPI_COMM_WORLD);
  if (comm_rank == 0) unlink(filename);

  MPI_Finalize();

  return TEST_EXIT_CODE;
}

static void handle_error(int status);

static void write_test_grid_file(char const * file_name) {

  int ncid;

  // create file
  handle_error(nc_create(file_name, NC_CLOBBER, &ncid));

  int dim_cell_id, dim_vertex_id, dim_nv_id, dim_ne_id;

  // define dimensions
  handle_error(nc_def_dim(ncid, "cell", 16, &dim_cell_id));
  handle_error(nc_def_dim(ncid, "vertex", 15, &dim_vertex_id));
  handle_error(nc_def_dim(ncid, "nv", 3, &dim_nv_id));
  handle_error(nc_def_dim(ncid, "ne", 6, &dim_ne_id));


  int var_vlon_id, var_vlat_id, var_clon_id, var_clat_id, var_mask_id,
      var_v2c_id, var_c2v_id;

  // define variables
  handle_error(nc_def_var(ncid, "vlon", NC_DOUBLE, 1, &dim_vertex_id,
                          &var_vlon_id));
  handle_error(nc_def_var(ncid, "vlat", NC_DOUBLE, 1, &dim_vertex_id,
                          &var_vlat_id));
  handle_error(nc_def_var(ncid, "clon", NC_DOUBLE, 1, &dim_cell_id,
                          &var_clon_id));
  handle_error(nc_def_var(ncid, "clat", NC_DOUBLE, 1, &dim_cell_id,
                          &var_clat_id));
  handle_error(nc_def_var(ncid, "cell_sea_land_mask", NC_INT, 1, &dim_cell_id,
                          &var_mask_id));
  int dim_ids[2];
  dim_ids[0] = dim_nv_id, dim_ids[1] = dim_cell_id;
  handle_error(nc_def_var(ncid, "vertex_of_cell", NC_INT, 2, dim_ids,
                          &var_c2v_id));
  dim_ids[0] = dim_ne_id, dim_ids[1] = dim_vertex_id;
  handle_error(nc_def_var(ncid, "cells_of_vertex", NC_INT, 2, dim_ids,
                          &var_v2c_id));

  // end definition
  handle_error(nc_enddef(ncid));

  // write grid data
  handle_error(nc_put_var_double(ncid, var_vlon_id, vlon));
  handle_error(nc_put_var_double(ncid, var_vlat_id, vlat));
  handle_error(nc_put_var_double(ncid, var_clon_id, clon));
  handle_error(nc_put_var_double(ncid, var_clat_id, clat));
  handle_error(nc_put_var_int(ncid, var_mask_id, mask));
  handle_error(nc_put_var_int(ncid, var_c2v_id, &(vertex_of_cell[0][0])));
  handle_error(nc_put_var_int(ncid, var_v2c_id, &(cells_of_vertex[0][0])));

  handle_error(nc_close(ncid));
}

static void handle_error(int status) {
  if (status == NC_NOERR) return;
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  exit(EXIT_FAILURE);
}

