/**
 * @file test_sphere_part.c
 *
 * @copyright Copyright  (C)  2014 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>

#include "tests.h"
#include "test_common.h"
#include "grid_search.h"
#include "read_icon_grid.h"
#include "grid.h"
#include "sphere_part.h"
#include "grid_reg2d.h"

int main(void) {

  {
      if (!file_exists("icon_grid_R02B00.nc"))
         return EXIT_SKIP_TEST;

      struct grid * icon_grid = read_icon_grid("icon_grid_R02B00.nc");

      struct grid_search * search = yac_sphere_part_search_new(icon_grid);

      yac_delete_grid_search(search);

      yac_delete_grid(icon_grid);
   }

  {
      if ((!file_exists("icon_grid_R02B02.nc")) ||
          (!file_exists("icon_grid_R02B03_G.nc")))
         return EXIT_SKIP_TEST;

      struct grid * grid[2] = {read_icon_grid("icon_grid_R02B02.nc"),
                               read_icon_grid("icon_grid_R02B03_G.nc")};

      for (size_t i = 0; i < 2; ++i) {

        struct grid_search * search = yac_sphere_part_search_new(grid[i]);
        struct dep_list * tgt_to_src_cells =
          yac_do_cell_search(search, grid[i^1]);

        for (size_t j = 0; j < yac_get_num_elements(tgt_to_src_cells); ++j)
          if (yac_get_num_dependencies_of_element(tgt_to_src_cells, j) == 0)
            PUT_ERR("ERROR in yac_do_cell_search\n");

        yac_dep_list_delete(tgt_to_src_cells);

        unsigned num_cells = yac_get_num_grid_cells(grid[i^1]);
        struct grid_cell cell;
        yac_init_grid_cell(&cell);

        for (unsigned j = 0; j < num_cells; ++j) {

          struct bounding_circle bnd_circle;

          yac_get_grid_cell2(grid[i^1], i, &cell, &bnd_circle);

          struct dep_list * result_list =
            yac_do_bnd_circle_search(search, &bnd_circle, 1);
          if (yac_get_num_dependencies_of_element(result_list, 0) == 0)
            PUT_ERR("ERROR in yac_do_bnd_circle_search\n");
          yac_dep_list_delete(result_list);
        }
        yac_free_grid_cell(&cell);

        yac_delete_grid_search(search);
      }

      yac_delete_grid(grid[0]);
      yac_delete_grid(grid[1]);
   }

   {
      struct grid * reg_grid;
      unsigned num_cells[2] = {20,20};
      unsigned cyclic[2] = {1,0};
      double coordinates_x[20];
      double coordinates_y[21];

      for (unsigned i = 0; i < 20; ++i)
         coordinates_x[i] = (0.0 + (double)i * 18.0) * YAC_RAD;

      for (unsigned i = 0; i < 21; ++i)
         coordinates_y[i] = (-90.0 + (double)i * 9.0) * YAC_RAD;

      reg_grid = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                    num_cells, cyclic);

      struct grid_search * search = yac_sphere_part_search_new(reg_grid);

      struct bounding_circle extent;
      yac_get_grid_bounding_circle(reg_grid, &extent);

      struct dep_list * result_list =
        yac_do_bnd_circle_search(search, &extent, 1);

      if (yac_get_num_elements(result_list) != 1)
        PUT_ERR("wrong number of elements in result list");

      if (yac_get_num_dependencies_of_element(result_list, 0) != 20*20)
        PUT_ERR("wrong number of dependencies in result list");

      yac_dep_list_delete(result_list);
      yac_delete_grid_search(search);
      yac_delete_grid(reg_grid);
   }

   return TEST_EXIT_CODE;
}

